package nxm.audio.lib;

import java.util.*;
import java.rmi.*;
import java.rmi.registry.*;
import javax.sound.sampled.*;
import nxm.audio.net.*;

/**
 * playerdaemon - this should be 16bit linear, 1 channel input
 * Stand-Alone or CLASS instantiated AudioPlayer
 * Now that it's in the NeXtMidas baseline, we need to test the RMI portion.
 * Make sure RMI is optional throughout : LMT 09/23/2005
 *
 * @author Scott Zimmerman (Zeta)
 * @author Lisa Tursky (TMA)
 * @version  $Id: playerdaemon.java,v 1.5 2012/12/28 21:22:34 ntn Exp $
 */
public class playerdaemon implements Runnable {
  private static hwplayer hwp;
  private TreeMap<Integer,String> readerLookup = new TreeMap<Integer,String>();
  private Map<Integer,DataReader> readers      = Collections.synchronizedMap(new TreeMap<Integer,DataReader>());
  private TreeMap<Integer,String> openChannels = new TreeMap<Integer,String>();
  private static int maxChannels;
  private static String rmiHost="127.0.0.1";
  private static int rmiPort = 1099;
  private boolean DEBUG=false,run=true;
  private short[] decodeTable;
  private static Thread daemon = null;
  private static playerdaemon play = null;

  private RemotePlayerDelegate playerDelegate;

  public void setVerbose(boolean ptemp) { DEBUG=ptemp; }
  public static hwplayer getPlayer() { return hwp; }

  public void setSocketDebug(boolean ptemp) {
    try {
      for (Integer key : readers.keySet()) {
        DataReader dr = readers.get(key);
        dr.setVerbose(ptemp);
        readers.put(key, dr);
      }
    } catch (Exception e) { e.printStackTrace(); }
  }

  public void setSocketDebug(boolean ptemp,int channel) {
    try {
      Integer key = new Integer(channel);
      DataReader dr = readers.get(key);
      dr.setVerbose(ptemp);
      readers.put(key, dr);
    } catch (Exception e) { e.printStackTrace(); }
  }

  public void setPacketDebug(boolean ptemp) {
    try {
      for (Integer key : readers.keySet()) {
        DataReader dr = readers.get(key);
        dr.setPacketDebug(ptemp);
        readers.put(key, dr);
      }
    } catch (Exception e) { e.printStackTrace(); }
  }

  public void setPacketDebug(boolean ptemp,int channel) {
    try {
      Integer key = new Integer(channel);
      DataReader dr = readers.get(key);
      dr.setPacketDebug(ptemp);
      readers.put(key, dr);
    } catch (Exception e) { e.printStackTrace(); }
  }

  public void setRTP(boolean ptemp) {
    try {
      for (Integer key : readers.keySet()) {
        DataReader dr = readers.get(key);
        dr.setRTP(ptemp);
        readers.put(key, dr);
      }
    } catch (Exception e) { e.printStackTrace(); }
  }
  public void setRTP(boolean ptemp,int channel) {
    try {
      Integer key = new Integer(channel);
      DataReader dr = readers.get(key);
      dr.setRTP(ptemp);
      readers.put(key, dr);
    } catch (Exception e) { e.printStackTrace(); }
  }

  public void setTimeOut(int ptimeout) {
    try {
      for (Integer key : readers.keySet()) {
        DataReader dr = readers.get(key);
        dr.setTimeOut(ptimeout);
        readers.put(key, dr);
      }
    } catch (Exception e) { e.printStackTrace(); }
  }
  public void setTimeOut(int ptimeout,int channel) {
    try {
      Integer key = new Integer(channel);
      DataReader dr = readers.get(key);
      dr.setTimeOut(ptimeout);
      readers.put(key, dr);
    } catch (Exception e) { e.printStackTrace(); }
  }

  public void setHost(String host, int channel) {
    try {
      Integer key = new Integer(channel);
      DataReader dr = readers.get(key);
      dr.setHost(host); //only sets host at init
      readers.put(key, dr);
    } catch (Exception e) { e.printStackTrace(); }
  }

  public void setPort(int port, int channel) {
    try {
      Integer key = new Integer(channel);
      DataReader dr = readers.get(key);
      dr.setPort(port); //only sets port at init
      readers.put(key, dr);
    } catch (Exception e) { e.printStackTrace(); }
  }

  public void setHostPort(String host,int port,int channel,String audiotype,boolean on) {
    try {
      if (DEBUG) System.out.println("playerdaemon: (setHostPort) "+host+"/"+port+".type:"+audiotype+".chan:"+channel);
      Integer key = new Integer(channel);
      DataReader dr = readers.get(key);
      dr.setLine(null);
      dr.close(); //shuts down reading socket and run loop
      dr.setHost(host);
      dr.setPort(port);
      int audiotypenum=1;
      if (audiotype!=null && audiotype.compareToIgnoreCase("ULAW")==0) audiotypenum=0;
      dr.setAudioType(audiotypenum);
      readers.put(key,dr);
      if (on) setOff(false,channel);
      else setOff(true,channel);
    } catch (Exception e) { e.printStackTrace(); }
  }

  //for testing purposes, run it in its own thread
  public static void main(String []args) {
    try {
      play = new playerdaemon();
      daemon = new Thread(play);
      daemon.setPriority(Thread.currentThread().getPriority() - 1);
      daemon.start();
    } catch (Exception e) { e.printStackTrace(); }
  }

  //if we don't always want to do this, we can add another command line arg
  //as a flag.
  private void registerPlayer() {
    /*************************************************************************
  This is not  necessary right now because we are using RMI locally only.
  When you do not install a security manager, all remote classes must be
  found on the local classpath. Once we go to remote RMI with the registry
  located on the web server, we will need to install a security manager.

       if(System.getSecurityManager() == null) {
           System.setSecurityManager(new RMISecurityManager());
       }
     *************************************************************************/

    //host & port in name represents host of RMI registry
    //String name = "//137.20.69.227:1099/HWPlayer:137.20.69.227";
    //String rmiHost = System.getProperty("java.rmi.hostname");
    //String rmiPort = System.getProperty("java.rmi.port");
    StringBuffer nameBuf = new StringBuffer("//");
    nameBuf.append(rmiHost);
    nameBuf.append(":");
    nameBuf.append(rmiPort);
    nameBuf.append("/HWPlayer:");
    nameBuf.append("localhost");
    //TESTING--LMT
    //String name = nameBuf.toString();
    String name = "HWPlayer";
    if (DEBUG) System.out.println("binding player with name: " + name);

    try {
      playerDelegate = new RemotePlayerDelegate(this);
      Registry reg = getRMIRegistry();
      reg.rebind(name, playerDelegate);

    } catch(Exception e) {
      e.printStackTrace();
    }
  }

  private Registry getRMIRegistry() {
    Registry reg = null;
    boolean retry = true;
    while (retry == true) {
      try {
        //get Registry
        reg = LocateRegistry.getRegistry(rmiHost,rmiPort,new CustomRMISocketFactory());
        //test communication with Registry
        reg.list();
        //If no exceptions thrown by above two methods, stop retrying
        retry = false;

      } catch (ConnectException ce) {
        //Registry has not yet initialized
        retry = true;
      } catch (RemoteException re) {}
    }
    return reg;
  }

  /** Adds a DataReader for each channel to the array.
   *  TBD -- the return value is not useful for adding multiple channels? Modify?
   *  @return the index of the last channel added
   */
  public int addReader(String ihost, int iport, String audiotype, int pbufsize, int numChannels) {

    int index = -1;
    String val = ihost + ":" + iport;
    int currentSize = readers.size();
    if ((currentSize + numChannels) > maxChannels) {
      System.err.println("Cannot open " + ihost + ":" + iport + "...");
      System.err.println("Insufficient number of open channels remaining...");
    } else {
      int audiotypenum=1;
      if (audiotype!=null && audiotype.compareToIgnoreCase("ULAW")==0) audiotypenum=0;
      DataReader dataReader = null;
      if (DEBUG) System.out.println("addReader host: " + ihost+".port="+iport+".Type="+audiotype+". size="+pbufsize+".chans="+numChannels+". num="+audiotypenum);

      for (int i=0; i<numChannels; i++) {
        Integer key = openChannels.firstKey();

        //make sure channel is initialized
        SourceDataLine line = hwp.addChannel(key.intValue(), null, pbufsize); //null=AudioFormat to use default
        dataReader = new DataReader( key.intValue(), decodeTable, audiotypenum, line);
        dataReader.open(ihost,iport+i);
        dataReader.start();

        readers.put(key, dataReader);
        openChannels.remove(key);
        val = ihost + ":" + (iport+i);
        readerLookup.put(key, val);
        index = key.intValue();

      }
    }
    return index;
  }

  public String getHostPort(int channel) {
    DataReader dr = readers.get(Integer.valueOf(channel));
    StringBuffer hostPort = new StringBuffer(dr.getHost());
    hostPort.append(":");
    hostPort.append(dr.getPort());
    return hostPort.toString();
  }

  /** Constructor that will register player with RMI.
   *  This is how ELCC uses this--don't want to break
   */
  public playerdaemon() {
    this(true);
  }

  /** Use this constructor to set up the playerdaemon's audio type and set up
   *  DataReaders sufficient for the maximum number of channels
   *  the hwplayer can handle. Make subsequent calls to the addReader() method
   *  to add DataReaders for different hosts/ports.
   */
  public playerdaemon(boolean useRMI) {
    try {
      if (DEBUG) System.out.println("Opening hwplayer");
      hwp = new hwplayer();
      hwp.open();
      if (useRMI) registerPlayer();
      maxChannels = hwp.getMaxLines();
      readers = new TreeMap<Integer,DataReader>();

      initializeOpenChannels();
      decodeTable = build_mu_decode_table();

      if (hwp==null) {
        System.err.println("Error-Could not open audio player");
        return;
      }

    } catch(Exception e){
      System.err.println("Error creating:"+e);
      e.printStackTrace();
    }
    if (DEBUG) System.out.println("DONE Opening hwplayer");
  }

  public void removeReader(int channel) {
    if (readers==null) return;
    synchronized(readers) {
      Integer key = new Integer(channel);
      DataReader dr = readers.get(key);
      hwp.setOff(true,channel);
      dr.shutdown();
      readerLookup.remove(key);
      readers.remove(key);
      openChannels.put(key, "");
    }
  }

  private void initializeOpenChannels() {
    for (int i = 0; i < maxChannels; i++) {
      openChannels.put(new Integer(i), "");
    }
  }

  public void shutdown() {
    close();
  }

  public void close() {
    try {
      stopPlayer();
      hwp = null;
      if (readers!=null) readers.clear();
      readers = null;
      if (daemon!=null) {
        daemon.interrupt();
        // NTN: Thread.destroy() is deprecated and the method was never implemented.
        // daemon.destroy();
      }
      daemon = null;
      try {
        Registry reg = LocateRegistry.getRegistry(rmiHost,rmiPort,new CustomRMISocketFactory());
        reg.unbind("HWPlayer");
        if (DEBUG) System.out.println("HWPLAY unbound from:"+rmiHost+":"+rmiPort);
      } catch(Exception e) {
        System.err.println("Close Unbind Exception:"+e);
      }
    } catch(Exception e) {
      System.err.println("Close Exception:"+e);
    }
  }

  public String getPlayerChannelInfo() {
    StringBuffer buff = new StringBuffer();
    for (Integer key : readers.keySet()) {
      DataReader dr = readers.get(key);
      buff.append("channel: " + key + "\n");
      buff.append("host: " + dr.getHost() + "\n");
      buff.append("port: " + dr.getPort() + "\n\n");
    }
    return buff.toString();
  }

  public void stopPlayer(){
    try {
      run = false;  // Set GLOBAL run to FALSE...
      try {
        hwp.closeAll();
      } catch(Exception e) {
        System.err.println("Error stopping hwp:"+e);
        e.printStackTrace();
      }
      for (DataReader dr : readers.values()) {
        hwp.setOff(true,dr.getLineNumber());
        dr.shutdown();
      }
    } catch(Exception e) {
      System.err.println("Error stopping:"+e);
      e.printStackTrace();
    }
    if (DEBUG) System.out.println("Stopped Player");
  }

  public void run() {
    while (run) {
      try {
        Thread.sleep(200);
      } catch(InterruptedException ie) {}
    }
  }

  private short[] build_mu_decode_table() {
    short[] convbuf = new short[256];
    for (int i=0; i<256; i++) {
      int j = i-128;
      int x = j^127;
      int S = 1-(2*((x&128)/128));
      int L = (x&112)/16;
      int V = x&15;
      convbuf[i] = (short) Math.round(((S*(Math.pow(2,L)*(2*V+33)-33)*32767.0) / 8031.0));
    }
    return convbuf;
  }

  public void setOff(boolean off, int channel) {
    try {
      Integer key = new Integer(channel);
      DataReader dr = readers.get(key);
      if (dr == null) {
        System.err.println("playerdaemon: setOff: no channel exits for:"+channel);
        return;
      }
      SourceDataLine line = hwp.setOff(off,channel);
      if (!off) {
        hwp.setVolume(line,dr.getVolume());
        hwp.setPan(line,dr.getPan());
        dr.setLine(line);
        dr.open(dr.getHost(),dr.getPort());
        readers.put(key,dr);
      } else {
        dr.setLine(null);
        dr.close(); //shuts down reading socket but leaves the run loop
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  public void setPan(float pan, int channel) {
    Integer key = new Integer(channel);
    DataReader dr = readers.get(key);
    if (dr == null) return;
    dr.setPan(pan);
    readers.put(key,dr);
    hwp.setPan(dr.getLine(),pan);
  }

  public void setVolume(int volume, int channel) {
    Integer key = new Integer(channel);
    DataReader dr = readers.get(key);
    if (dr == null) return;
    dr.setVolume(volume);
    readers.put(key,dr);
    hwp.setVolume(dr.getLine(),volume);
  }
}
