package nxm.database.lib;

import  java.sql.*;

public class PostgreSQLDB extends NxmDB {
  public static final int defaultPort = 5432;
 
  public PostgreSQLDB() {
    super();
    if (myPort == 0) myPort = defaultPort;
  }
  
  public PostgreSQLDB(String db, String user, String passwd) {
    super("localhost", defaultPort, db, user, passwd);  
  }
  
  public PostgreSQLDB(String host, String db, String user, String passwd) {
    super(host, defaultPort, db, user, passwd);  
  }
  
  public PostgreSQLDB(String host, int port, String db, String user, String passwd) {
    super(host, port, db, user, passwd);
  }
  
  public void connect() throws DatabaseException {
    String conStr;
    try {
      myDriver = (Driver)Class.forName("org.postgresql.Driver").newInstance();
      conStr = "jdbc:postgresql://" + myHost + ":" + myPort + "/" + myName;
      if (verbose) {
        System.out.println("Connection: " + conStr);
	System.out.println("For User: " + myUserName);
      }
      myUserPassword = "";
      myConnection = DriverManager.getConnection(conStr, myUserName, myUserPassword);
      
      if (verbose) {
        SQLWarning warn = myConnection.getWarnings();
        while (warn != null) {
          System.out.println("SQLState = " + warn.getSQLState());
	  System.out.println("Message  = " + warn.getMessage());
	  System.out.println("Vendor   = " + warn.getErrorCode());
	  System.out.println("");
	  warn = warn.getNextWarning();
        }
      }
    } catch (Exception e) {     
      throw(new DatabaseException(e.getMessage()));
    }
    
   }
   
  public ResultSet getTables() {
    ResultSet rs;
    try {
      Statement stmt = myConnection.createStatement();
      String qstr = "select relname from pg_class where ";
      qstr += "relkind like 'r' and relname !~ '^pg_'";
      rs = stmt.executeQuery(qstr);
    } catch (SQLException e) {
      if (verbose) { System.out.println("getTables(): " + e);}
      rs = null;
    }
    return(rs);
  }
}
