package nxm.database.prim;

import nxm.database.lib.*;
import nxm.sys.lib.*;

import java.sql.*;

/**
   NeXtMidas database server

 @author Pete Perrino
 @version $Id: dbserver.java,v 1.5 2013/01/04 21:01:03 ntn Exp $
*/
public class dbserver extends Primitive {
  NxmDB   myDB;
  String  result;
  boolean bg;
  public static String outTypeList = "Full,3000";
  private static int OT_FULL=1, OT_3000=2;
  private int outType = OT_FULL;

  public int open() {
    String  host     = MA.getCS("HOST", "localhost");
    int     port     = MA.getL("PORT", -1);
    String  dbname   = MA.getCS("NAME");
    String  user     = MA.getCS("USER");
    String  password = MA.getCS("PASSWORD");
    verbose          = MA.getState("/VERBOSE", false);
    String  dpath    = MA.getS("/DRIVER");
    String  dtype    = MA.getS("/TYPE");
    String  durl     = MA.getS("/URL");
    outType = MA.getSelectionIndex("/OUTTYPE",outTypeList, outType); // Define table output type

    bg = MA.getState("/BG");
    if (M.pipeMode==Midas.POFF && !bg) {
       M.warning("DBSERVER can only be run in piped mode or in the background");
       return FINISH;
    }

    // specifying the exact url overrides everything else
    if (dtype == null || dtype == "") {
      if (dpath == null || dpath == "") {
        M.error("You must either specify a type or driver path");
      } else if (durl == null) {
        M.error("You must specify a database url");
      }
      myDB = new NxmDB(host, port, dbname, user, password);
      myDB.setDriverString(dpath);
      myDB.setURLString(durl);
    } else {
      String tmpS = "";
      Class<?> tmpDB;
      try {
        tmpS = NxmDB.class.getPackage().getName();
        tmpS += "." + dtype + "DB";
        if (verbose) M.info("Trying to find: " + tmpS);
        tmpDB = Class.forName(tmpS);
        if (verbose) M.info("Found: " + tmpDB);
        myDB = (NxmDB)tmpDB.newInstance();
      } catch (Exception e) {
        throw new MidasException("Could not find class: " + tmpS, e);
      }
    }

    myDB.setVerbose(verbose);
    myDB.setDriverString(dpath);
    myDB.setURLString(durl);
    myDB.setHost(host);
    if (port > 0) myDB.setPort(port);
    myDB.setName(dbname);
    myDB.setUserName(user);
    myDB.setUserPassword(password);

    return NORMAL;
  }

  public int process() {
    if (M.pipeMode==Midas.POFF && !bg) {
      return FINISH;
    }
    return NOOP;
  }

  public int close() {
    try {
      if (myDB != null) myDB.disConnect();
    } catch (Exception e) {
      M.warning("Caught exception while trying to close db connection: " + e);
    }
    return NORMAL;
  }

  public int processMessage(Message msg) {
    return NORMAL;
  }

  public Object getKey (String name) {
    return null;
  }

  public Object doQuery(String sql) {
    Object retVal;

    if (verbose) M.info("doing the string query:\n" + sql);

    if (sql != null && sql != "") {
      if (!ensureOpen()) return null;

      retVal = myDB.executeQuery(sql);
      if (retVal == null) {
        M.warning("SQL statement returned null value");
      } else {
        if ( retVal instanceof Integer) {
          Data tmp = new Data(retVal);
          return tmp;
        } else {
          boolean type3000 = (outType == OT_3000);
          Table tmp = ResultSetTable.fromResultSet((ResultSet)retVal, type3000);
          try {
            ((ResultSet)retVal).close();
          } catch (SQLException e) {
            M.warning("Caught SQL Exception when trying to close ResultSet: " + e);
          }
          return tmp;
        }
      }
    } else {
      M.warning("Empty SQL string, nothing to do");
    }
    return (new Table());
  }

  public int getLastInsertId(String tbl) {
    int retVal = 0;
    if (tbl == null || tbl == "") return retVal;
    if (!ensureOpen()) return retVal;

    String qry = "select LAST_INSERT_ID() from " + tbl;
    ResultSet rs = myDB.selectQuery(qry);
    try {
      rs.next();
      retVal = rs.getInt(1);
    } catch (java.sql.SQLException e) {
      M.warning("Caught exception: " + e);
    }
    return retVal;
  }

  public Object doQuery(Table sql) {
    M.info("Not implemented yet");
    return null;
  }

  public Object getTables() {
    ResultSet retVal = myDB.getTables();
    return (ResultSetTable.fromResultSet(retVal));
  }

  public Object describeTable(String t) {
    ResultSet retVal = myDB.describeTable(t);
    return (ResultSetTable.fromResultSet(retVal));
  }

  public Object newTableEntry(String t) {
    Table e = myDB.newTableEntry(t);
    return e;
  }

  public boolean ensureOpen() {
    if (isClosed()) {
      try {
        myDB.connect();
      } catch (Exception e) {
        M.warning("Caught exception while trying to open db connection: " + e);
        return false;
      }
    }
    return true;
  }

  public boolean isClosed() {
    if (myDB == null) return true;
    return (myDB.isClosed());
  }

  public void connect() {
    try {
      myDB.connect();
    } catch (Exception e) {
      M.warning("Caught exception while trying to open db connection: " + e);
    }
  }

  public void disConnect() {
    try {
      myDB.disConnect();
    } catch (Exception e) {
      M.warning("Caught exception while trying to open db connection: " + e);
    }
  }

}
