package nxm.nmpostlegacy.lib;

import java.io.*;
import java.net.*;
import org.apache.commons.net.ftp.*;
import org.apache.commons.net.io.*;
import java.io.BufferedOutputStream;
import java.io.FileOutputStream;

/**
 *  Title: FtpClient<p>
 *
 *  Description: This primitive is a SIMPLE Ftp file retrieval class. </p> <p>
 *
 *  Company: Zeta Associates</p>
 *  Created: February 25, 2005<br>
 *  Version  1.0
 * @author   Wayne G. Dengel II
 * @author   Neon Ngo - switch to Apache Commons Net 3.0 (which originated from ORO's NetComponents)
 * @version  $Id: FtpClient.java,v 1.2 2013/01/07 21:09:41 ntn Exp $
 */
public class FtpClient implements Runnable, org.apache.commons.net.io.CopyStreamListener {
    public final static  int                           REFUSED_CONNECTION     = -10;
    public final static  int                           UNKNOWN_CONNECT_ERROR  = -9;
    public final static  int                           NO_FILE_FOUND          = -8;
    public final static  int                           UNKNOWN_EXCEPTION      = -7;
    public final static  int                           XFER_COMPLETE          = 5;
           final static  int                           COPY_BUFFER_SIZE       = 8192; // since DEFAULT_COPY_BUFFER_SIZE = 1024

    private              double                        oldsize                = 0.0;             // The byte count from last Callback
    private              long                          filesize               = 0;               // The filesize of target
    private              boolean                       QUIET                  = false;           // GLOBAL Debug Flag
    private              Thread                        myThread               = null;            // The Thread that the LISTENER is running on
    private              FtpClient_cb                  callBack               = null;
    private              String                        ftpHost                = null;
    public               String                        filename               = null;
    private              String                        origfile               = null;
    public               String                        path                   = null;
    private              String                        username               = null;
    private              String                        password               = null;
    private              int                           port                   = 0;
    private              org.apache.commons.net.ftp.FTPClient ftp             = null;
    private              boolean                       mget                   = false;
    public               String                        fname                  = null;
    private              boolean                       deleteFile             = true;

    /**
     *  Constructor for the FtpClient object
     *
     *@param  quiet     Description of the Parameter
     *@param  ftpHost   Description of the Parameter
     *@param  path      Description of the Parameter
     *@param  filename  Description of the Parameter
     *@param  name      Description of the Parameter
     *@param  pass      Description of the Parameter
     *@param  callBack  Description of the Parameter
     */
    public FtpClient(boolean quiet, String ftpHost, String path, String filename,
            String name, String pass, FtpClient_cb callBack) {
        // Wayne, can we make this a one-line constructor:
        //   this(quiet, true, ftpHost, path, filename, name, pass, callBack);
        this.callBack = callBack;
        this.ftpHost = ftpHost;
        this.origfile = filename;
        this.path = path;
        this.QUIET = quiet;
        this.username = name;
        this.password = pass;
        myThread = new Thread(this);
        myThread.start();
        int  idx  = origfile.lastIndexOf("/");
        if (idx > -1) {
            this.filename = origfile.substring(idx + 1);
        }
        if (!QUIET) {
            info("Filename = " + this.origfile);
        }
    }

    /**
     *  Constructor for the FtpClient object
     *
     *@param  quiet      Description of the Parameter
     *@param  deleteFile Description of the Parameter
     *@param  ftpHost    Description of the Parameter
     *@param  path       Description of the Parameter
     *@param  filename   Description of the Parameter
     *@param  name       Description of the Parameter
     *@param  pass       Description of the Parameter
     *@param  callBack   Description of the Parameter
     */
    public FtpClient(boolean quiet, boolean deleteFile, String ftpHost, String path, String filename,
            String name, String pass, FtpClient_cb callBack) {
        this.callBack = callBack;
        this.ftpHost = ftpHost;
        this.origfile = filename;
        this.path = path;
        this.QUIET = quiet;
        this.username = name;
        this.password = pass;
        this.deleteFile = deleteFile;
        myThread = new Thread(this);
        myThread.start();
        int  idx  = origfile.lastIndexOf("/");
        if (idx > -1) {
            this.filename = origfile.substring(idx + 1);
        }
        if (!QUIET) {
            info("Filename = " + this.origfile);
        }
    }

    /**
     *  Constructor for the FtpClient object
     *
     *@param  quiet     Description of the Parameter
     *@param  ftpHost   Description of the Parameter
     *@param  port      Description of the Parameter
     *@param  path      Description of the Parameter
     *@param  filename  Description of the Parameter
     *@param  name      Description of the Parameter
     *@param  pass      Description of the Parameter
     *@param  callBack  Description of the Parameter
     *@param  mget      Description of the Parameter
     */
    public FtpClient(boolean quiet, String ftpHost, int port, String path, String filename,
            String name, String pass, FtpClient_cb callBack, boolean mget) {
        this.callBack = callBack;
        this.ftpHost = ftpHost;
        this.origfile = filename;
        this.path = path;
        this.QUIET = quiet;
        this.username = name;
        this.password = pass;
        this.mget = mget;
        this.port = port;

        myThread = new Thread(this);
        myThread.start();
        int  idx  = origfile.lastIndexOf("/");
        if (idx > -1) {
            this.filename = origfile.substring(idx + 1);
        }
        if (!QUIET) {
            info("Filename = " + this.filename);
        }
    }

    /**
     *  Main processing method for the FtpClient object
     */
    public void run() {
        ftp = new org.apache.commons.net.ftp.FTPClient();
        ftp.setCopyStreamListener(this);
        ftp.setBufferSize(COPY_BUFFER_SIZE);
        try {
            int  reply;
            if (port > 0) {
                ftp.connect(ftpHost, port);
            } else {
                ftp.connect(ftpHost);
            }
            if (!QUIET) {
                info("Connected to " + ftpHost);
            }
            if (!QUIET) {
                ftp.getReplyString();
            }
            reply = ftp.getReplyCode();
            if (!FTPReply.isPositiveCompletion(reply)) {
                ftp.disconnect();
                if (!QUIET) {
                    info("FTP server refused connection.");
                }
                if (callBack != null) {
                    callBack.FTPStatus(REFUSED_CONNECTION, this);
                }
                return;
            }
        } catch (IOException e) {
            if (ftp.isConnected()) {
                try {
                    ftp.disconnect();
                } catch (IOException f) {}
            }
            if (!QUIET) {
                info("Could not connect to server. " + e);
            }
            if (callBack != null) {
                callBack.FTPStatus(UNKNOWN_CONNECT_ERROR, this);
            }
            return;
        }

        try {
            ftp.login(username, password);
            if (!QUIET) {
                info(ftp.getReplyString());
            }

            FTPFile[]  myfiles  = null;
            if (mget) {
                info("Changing dir to " + origfile);
                ftp.changeWorkingDirectory(origfile);
                info(ftp.getReplyString());
                String  names[]  = ftp.listNames();
                info("Name = " + names[0]);
                fname = new String(names[0]);

                for (int fileNum = 0; fileNum < names.length; fileNum++) {
                    if (!QUIET) {
                        info(names[fileNum] + "");
                        info(ftp.getReplyString());
                    }
                    info("Looking for file " + names[fileNum]);
                    BufferedOutputStream  fos  = new BufferedOutputStream(new FileOutputStream(path + names[fileNum]));
                    ftp.setFileType(FTP.BINARY_FILE_TYPE);
                    ftp.retrieveFile(names[fileNum], fos);
                    if (this.deleteFile) ftp.deleteFile(names[fileNum]);
                    fos.close();
                    if (!QUIET) {
                        info(ftp.getReplyString());
                    }
                    if (!QUIET) {
                        info("Transfer Complete. " + origfile);
                    }
                }
                ftp.changeWorkingDirectory("/");
                ftp.removeDirectory(origfile);

            } else {
                myfiles = ftp.listFiles(origfile);
                if (myfiles == null) {
                    if (!QUIET) {
                        info(filename + " does not exist on server ");
                    }
                    if (callBack != null) {
                        callBack.FTPStatus(NO_FILE_FOUND, this);
                    }
                    disconnect();
                    return;
                }
                if (myfiles.length < 1) {
                    if (!QUIET) {
                        info(filename + " does not exist on server " + myfiles.length);
                    }
                    if (callBack != null) {
                        callBack.FTPStatus(NO_FILE_FOUND, this);
                    }
                    disconnect();
                    return;
                }

                filesize = myfiles[0].getSize();
                if (!QUIET) {
                    info(" " + myfiles[0].getName() + "  is " + filesize);
                    info(ftp.getReplyString());
                }
                BufferedOutputStream  fos  = new BufferedOutputStream(new FileOutputStream(path + filename));
                ftp.setFileType(FTP.BINARY_FILE_TYPE);
                ftp.retrieveFile(origfile, fos);
                ftp.deleteFile(origfile);
                fos.close();

                if (!QUIET) {
                    info(ftp.getReplyString());
                }
                if (!QUIET) {
                    info("Transfer Complete. " + origfile);
                }
            }
            disconnect();
            if (callBack != null) {
                callBack.FTPStatus(XFER_COMPLETE, this);
            }
        } catch (Exception ex) {
            if (!QUIET) {
                info("Unknown Exception " + ex);
            }
            if (callBack != null) {
                callBack.FTPStatus(UNKNOWN_EXCEPTION, this);
            }
        }
    }

    /**
     *  Description of the Method
     */
    private void disconnect() {
        try {
            ftp.logout();
            ftp.disconnect();
        } catch (IOException f) {}

    }

    /**
     *  Description of the Method
     *
     *@param  totalBytesTransferred  Description of the Parameter
     *@param  bytesTransferred       Description of the Parameter
     *@param  streamSize             Description of the Parameter
     */
    public void bytesTransferred(long totalBytesTransferred, int bytesTransferred, long streamSize) {
        double cursize  = ((double) totalBytesTransferred / (double) filesize) * 100;
        cursize = ((double) ((int) (cursize * 10)) / 10);
        if (cursize < oldsize + 5.0) {
            return;
        }
        oldsize = cursize;
        if (!QUIET) {
            info("total = " + oldsize);
        }
        if (callBack != null) {
            callBack.FTPPercent(oldsize, this);
        }
    }

    /**
     *  Description of the Method
     *@param  event  Description of the Parameter
     */
    public void bytesTransferred(CopyStreamEvent event) {
        info("Got event " + event);
    }

    /**
     *  Description of the Method
     *@param  str  Description of the Parameter
     */
    private void info(String str) {
        String tms = nxm.sys.lib.Time.toString(nxm.sys.lib.Time.current(),nxm.sys.lib.Time.VAX);
        System.out.println(tms+" "+getClass() +": "+ str);
//        System.out.println("FtpClient: " + str);
    }

}
