package nxm.nmpostlegacy.lib;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.net.Socket;
import java.util.StringTokenizer;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import nxm.sys.lib.MessageQueue;
import nxm.sys.lib.Table;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

/**
 * <p>Title:       nmpost (HTTP POST client request handler)</p>
 * <p>Description: This primitive accepts POSTS from a WEB or whatever. </p>
 * <P>             It expects the POSTS to be in the format below.     </p>
 * <P>             Format is for MESSAGES.  NMPOST turns post into MSG, and places in QUEUE </p>
 * <p>Company:     Zeta Associates</p>
 * @author         Wayne G. Dengel II
 * @version 1.0
 */
public class PostClient implements Runnable {
  public final static int RECVD_TABLE    = 1;
  public final static int RECVD_FILE     = 2;

  private int              bufSize        = 8192;
  private boolean          monitor        = true;
  private boolean          deBug          = false;  // GLOBAL Debug Flag
  private Thread           myThread       = null;   // The Thread that the LISTENER is running on
  public Socket            clientSocket   = null;
  private DataInputStream  dis            = null;
  private DataOutputStream dos            = null;
  public  boolean          complete       = false;
  private MessageQueue     MQ             = null;
  private String           userFileName   = null;
  private String           userFileStatus = null;
  private int              userFileSize   = 0;
  private final int        block          = 256000;
  private int              curPtr         = 0;
  private String           readStr        = null;
  private String           uPath          = null;
  private byte []          tmpBuf         = null;

  private BufferedInputStream  fis        = null;
  private BufferedOutputStream fos        = null;

  private File               df           = null;
  private int                fsize        = 0;
  private String             fname        = null;
  private int                percentold   = 0;
  private byte[]             walkBuf      = null;
  private int                totalFISsize = 0;
  private int                overLap      = 32;
  private int                readSize     = 8196;
  private boolean            reportMain   = false;

  private PostClient_cb      callBack     = null;   // WGD:  Added CallBack to support JAVA Only apps.  (No Macro/No Messages)

  private DocumentBuilderFactory factory;           // Mikes crap... Actually a MUCH better/Standard way....
  private DocumentBuilder        builder;           // Mikes crap...

  public PostClient(boolean deBug,Socket clientSocket,MessageQueue mq,String path,boolean reportMain,PostClient_cb callBack) {
    //info("PostClient create "+reportMain);
    this.clientSocket = clientSocket;
    this.MQ           = mq;
    this.uPath        = path;
    this.reportMain   = reportMain;
    this.deBug        = deBug;
    this.callBack     = callBack;
    tmpBuf            = new byte[bufSize];
    walkBuf           = new byte[readSize+overLap+64];

    try {
      df  = File.createTempFile("post", ".tmp");
      fos = new BufferedOutputStream(new FileOutputStream(df));
      fis = new BufferedInputStream(new FileInputStream(df));
      fis.mark(8196);
    } catch(Exception ex) {
      System.out.println("Exception in createTempFile of PostClient "+ex);
      complete=true;
      df=null;
      return;
    }
    myThread = new Thread(this);
    myThread.start();
    //info("PostClient create Complete");
  }

  public void run() {
    try {
      factory = DocumentBuilderFactory.newInstance();
      builder = factory.newDocumentBuilder();
      dis = new DataInputStream(clientSocket.getInputStream());
      dos  = new DataOutputStream(clientSocket.getOutputStream());
    } catch(Exception ex) {
      System.out.println("ERROR: could not get Streams "+ex);
      return;
    }

    boolean running = true;
    while (running) {
      try{
        int avail = waitAvail(200);
        if (avail == -1) {
          running = false;
        } else if (avail > 0) {
          if (avail > bufSize) avail = bufSize;
          readSocket(avail);
          running = false;
        }
      } catch (Exception ex) {
        System.out.println("ERROR: in readsocket-run "+ex);
      }
    }
    readStr = convString();

    //System.out.println("Got a "+readStr);
    if ( (isPost(readStr)) && (isText(readStr)) ) {
      String strfinal = "HTTP/1.1 200 OK\r\nServer: NeXtMidas\r\nConnection: close\r\nContent-Type: text/xml\r\n\r\n"
                      + "<XML><RESPONSE=\"YES\"/></XML>";
      try {
        dos.write(strfinal.getBytes());
        dos.flush();
      } catch (Exception ex) {
        // ignore?
      }
      try {
        fos.close();
        fis.close();
        dos.close();
        dis.close();
        clientSocket.close();
      } catch (Exception ex) {
        System.out.println("Exception CLOSING Streams...");
      }
      if (isOrigText(readStr)) {
        //System.out.println("processXML2Msg");
        ProcessXML2Msg(readStr);
      } else {
        //System.out.println("processXML2MsgExt");
        ProcessXML2MsgExt(readStr);
      }
      //ProcessXML2Msg(readStr);
      df.delete();
      complete = true;
      df = null;

    } else if (isMultiPart(readStr)) {
      running = true;
      //if(deBug) info("We are a MULTIPART...");
      fname = findFileName();
      fsize = findSize();

      while (readSocket() != -1) {
        //try{ Thread.sleep(5); }catch(Exception ex){}
        //System.out.println("Sleeping");
        //int myR = readSocket();
        //if(myR==-1){
        //  running = false;
        //}
      }

      String marker = findMarker();
      try {
        fos.close();
        fis.close();
        dos.close();
        dis.close();
      } catch (Exception ex) {
        System.out.println("Exception CLOSING Streams...");
      }
      processMultiPart(marker);
      df.delete();
      df = null;
      complete = true;
    } else {
      System.out.println("ERROR: parsing packets "+readStr);
    }

    try {
      clientSocket.close();
    } catch (Exception ex) {
      System.out.println("Can't close socket");
    }
  }

  private void processMultiPart(String marker) {
    int markinst = 1;
    while (true) {
      int markStart = findStart(marker,markinst);
      if (markStart == -1) {
        return;
      } else {
        int markStop = findStop(marker,markinst);
        markinst++;
        processPart(markStart+marker.length()+4,markStop);
      }
    }
  }

  private void processPart(int start, int stop) {
    try{
      BufferedInputStream fis = new BufferedInputStream(new FileInputStream(df));
      int totalSize = fis.available();

      int rsize = 2048;
      if (stop-start < 2048) {
        rsize  = stop-start;
      }
      byte[] bbuf = new byte[rsize];
      fis.skip(start);
      int got = fis.read(bbuf,0,rsize);
      fis.close();
      fis = null;

      if (isText(bbuf)) {
        //System.out.println("This is a TEXT Part...");
        BufferedInputStream fis2 = new BufferedInputStream(new FileInputStream(df));
        int    rsize2 = stop-start;
        byte[] bbuf2  = new byte[rsize2+64];
        fis2.skip(start);
        int got2 = fis2.read(bbuf2,0,rsize2);
        fis2.close();
        fis2 = null;
        String content = new String(bbuf2);
        //System.out.println("processPart:content is "+content);
        int beginxml   = content.indexOf("<XML>");
        int endxml     = content.indexOf("</XML>");
        String localXmlStr = content.substring(beginxml+5,endxml);
        int l1 = localXmlStr.indexOf("<USERIDENT=");
        int l2 = localXmlStr.indexOf("/>",l1);
        userFileName = localXmlStr.substring(l1+11,l2);
        l1      = localXmlStr.indexOf("<STATUS=");
        l2     = localXmlStr.indexOf("/>",l1);
        userFileStatus = localXmlStr.substring(l1+8,l2);
        l1      = localXmlStr.indexOf("<FILESIZE=");
        l2     = localXmlStr.indexOf("/>",l1);
        try {
          userFileSize = Integer.parseInt(localXmlStr.substring(l1+10,l2));
        } catch (Exception ex) {
          userFileSize = 0;
        }
        //if(deBug){
        //System.out.println("Uname is  "+userFileName);
        int xloc = userFileName.lastIndexOf('/');
        if (xloc > -1 ) {
          userFileName = userFileName.substring(xloc+1);
          //  System.out.println("Fixed Uname is  "+userFileName);
        }
        //System.out.println("status is "+userFileStatus);
        //System.out.println("fsize is  "+userFileSize);
        //}

        if (userFileSize == 0) {
          Table tb = new Table();
          tb.put("NAME",userFileName);
          tb.put("STAT",userFileStatus);
          System.out.println("Putting POSTFILE  "+tb);
          if (callBack!=null) {
            callBack.PostResult(RECVD_TABLE,tb);
          }
          else {
            if (MQ!=null) MQ.put("POSTFILE", -1, tb, "MAIN");
          }
        }
      } else if (isBinary(bbuf)) {
        //System.out.println("This is a BINARY Part...");
        int bstart = findString(start,"\r\n\r\n",1);
        int bstop  = findString(start,"\r\n",-1);
        //System.out.println("Start/Stop = "+(start+bstart)+"/"+(start+bstop));
        BufferedInputStream mfis = new BufferedInputStream(new FileInputStream(df));
        int size   = bstop-(bstart+4);
        mfis.skip(start+bstart+4);
        // Do Read/Write to FILE...
        byte[] move = new byte[8196];
        File file = null;

        if (userFileName.length() > 1 ) {
          if (callBack!=null) {
            uPath = callBack.getWritePath(uPath,userFileName);
          }
          System.out.println("Writting file to "+uPath+userFileName+" size="+size);
          file = new File(uPath+userFileName);
        } else {
          System.out.println("Writting file to "+uPath+"inputfile.tmp"+" size="+size);
          file = new File(uPath+"inputfile.tmp");
        }
        BufferedOutputStream mfos = new BufferedOutputStream(new FileOutputStream(file));
        while (size > 8196) {
          mfis.read(move,0,8196);
          mfos.write(move);
          size = size - 8196;
        }
        mfis.read(move,0,size);
        mfos.write(move,0,size);
        mfis.close();
        mfos.close();
        if (callBack!=null) callBack.PostResult(RECVD_FILE,userFileName);
        else{
          if (MQ!=null) MQ.put("POSTFILE", size, userFileName, "MAIN");
        }
      }
    } catch (Exception ex) {
      System.out.println("Exception in processPart "+ex);
    }
  }

  private int waitAvail(int max) {
    int waiting   = 0;
    //if(deBug) info("waitAvail("+max+")  begin");
    try {
      while (true) {
        int avail = dis.available();
        if ((max>0)&&(avail>max)) {
          //if(deBug) info("waitAvail("+max+")  returning "+avail);
          return avail;
        }
        if ( (avail>0) && (max<1) ) {
          //if(deBug) info("waitAvail("+max+")  returning "+avail);
          return avail;
        } else {
          waiting++;
          Thread.sleep(10);
        }
        if (waiting > 20) {
          waiting = 0;
          dos.write(0);
          //if(deBug) info("waitAvail("+max+")  Ony have "+avail);
          return avail;
        }
      }
    } catch (Exception ex) {
      //if(deBug) info("waitAvail("+max+")  Exception on WAIT "+ex);
      info("Exception in WAITAVAIL "+ex);
      return -1;
    }
  }

  private int readSocket(int avail) {
    try {
      int myread = dis.read(tmpBuf,0,avail);
      fos.write(tmpBuf,0,myread);
      curPtr = curPtr + myread;
      return myread;
    } catch (Exception ex) {
      return 0;
    }
  }

  private int readSocket() {
    try {
      int percent = 0;
      int myRead = dis.read(tmpBuf);
      if (myRead!=-1) {
        fos.write(tmpBuf,0,myRead);
        curPtr = curPtr + myRead;
        if (monitor) {
          percent = (int) ((((double)curPtr/(double)fsize)*100.0)/5);
          if(percentold!=percent){
            percentold=percent;
            if (deBug) info(reportMain +" FileXFer for "+fname+" is "+(percent*5)+" complete");
            if (reportMain) {
              if (callBack!=null) {
                callBack.PostPercent((percent*5),fname);
              }
              else if (MQ!=null) {
                Table tb = new Table();
                tb.put("NAME",fname);
                tb.put("STAT",(percent*5)+"");
                MQ.put("POSTSTAT",0,tb,"MAIN");
              }
            }
          }
        }
      }
      return myRead;
    } catch (Exception ex) {
      // ignore?
    }
    return 0;
  }

  private boolean isMultiPart(String inStr) {
    int loc = inStr.indexOf("Content-type: Multipart");
    if (loc > -1) {
      return true;
    }
    return false;
  }

  private String convString() {
    try{
      fos.flush();                    // Flush the Output Stream
      fis.reset();
      byte[] bf = new byte[fis.available()];
      fis.read(bf,0,fis.available());
      return new String(bf);
    } catch(Exception ex) {
      System.out.println("Exception in convString: PostClient "+ex);
      return null;
    }
  }

  private boolean isPost(String inStr) {
    int loc = inStr.indexOf("POST / HTTP");
    if (loc > -1) {
      return true;
    }
    return false;
  }

  private boolean isText(byte[] bbuf) {
    return isText(new String(bbuf));
  }

  private boolean isText (String inStr) {
    //System.out.println("istest is " + inStr);
    int loc = inStr.indexOf("Content-type: text");
    if (loc < 0)
      loc = inStr.indexOf("Content-Type: text");
    if (loc > -1)
      return true;
    return false;
  }

  private boolean isOrigText(String inStr) {
    int loc = inStr.indexOf("Content-type: text");
    if (loc > -1)
      return true;
    return false;
  }

  private boolean isBinary(byte[] bbuf) {
    return isBinary(new String(bbuf));
  }
  private boolean isBinary(String inStr) {
    int loc = inStr.indexOf("Content-type: binary");
    if (loc < 0)
      loc = inStr.indexOf("Content-Type: binary");
    if (loc > -1)
      return true;
    return false;
  }

  private void ProcessXML2MsgExt(String xmlStr) {  // totally cheated on how many levels i will process. I know... sorry
    // i wanted to do recursion or call internal subroutines on each node visit
    // but wanted to finish this reall quick... mjw
    try {
      //if(deBug) System.out.println("XMLSTR = "+xmlStr);
      int nStart = xmlStr.indexOf("<XML>");
      if (nStart < 0) return;                                 // return if didn't find xml tag
      String xmldata = xmlStr.substring(nStart);              // use only data between xml tags
      // System.out.println("xmdata = " + xmldata);
      byte bxml[] = xmldata.getBytes();
      ByteArrayInputStream bis = new ByteArrayInputStream(bxml,0,xmldata.length());// set up for parser
      InputSource is = new InputSource(bis);
      Document document = builder.parse(is);                  // call parser
      Element ele = document.getDocumentElement();
      //nl  = ele.getElementsByTagName(ele.getTagName());     // nodelist for xml
      NodeList nl = ele.getChildNodes();                      // childlist for xml
      Table tbl = new Table();
      Table[] tblSub = new Table[10];
      if (nl != null && nl.getLength()>0) {
        int nChild = nl.getLength();
        int nSubTbl = -1;
        for (int i = 0; i < nChild; i++) {
          if (nl.item(i).getNodeType() == org.w3c.dom.Node.TEXT_NODE) continue;
          if (nl.item(i).hasAttributes()) {
            nSubTbl++;                   // new sublevel table
            tblSub[nSubTbl] = new Table();
            tbl.put(nl.item(i).getNodeName(),tblSub[nSubTbl]);
            // System.out.println("tbl.put("+nl.item(i).getNodeName()+','+"tblSub["+nSubTbl+"])");
            NamedNodeMap nnp  = nl.item(i).getAttributes();
            if (nnp != null) {
              for (int nSubChild = 0;nSubChild <nnp.getLength(); nSubChild++){
                tblSub[nSubTbl].put(nnp.item(nSubChild).getNodeName(),nnp.item(nSubChild).getNodeValue());
                //System.out.println("tbl.put(tblSub["+nSubTbl+"].put("+nnp.item(nSubChild).getNodeName()+','+nnp.item(nSubChild).getNodeValue()+')');
              } // endfor subchild
            }
          } // has Attributes
          if (nl.item(i).hasChildNodes()) {
            NodeList nl2 = nl.item(i).getChildNodes();
            if (nl2 != null) {
              tbl.put(nl.item(i).getNodeName(),nl2.item(0).getNodeValue());
              //System.out.println("tbl.put("+ nl.item(i).getNodeName()+','+nl2.item(0).getNodeValue()+')');
            } //nl2 != null
          } // has Children
        } // end for loop
      } // if node list
      Object objVal,objMsg;
      String msgName,info,to,from;
      msgName = "UNKNOWN"; info="1"; to="MAIN"; from="NMPOST";
      if ((objMsg =tbl.getx("MSG")) != null) {
        //System.out.println("tbl.getx(MSG)="+objMsg);
        if ((objMsg instanceof Table) && ((objVal = ((Table)objMsg).getx("NAME")) != null))
          msgName = objVal.toString();
        if ((objMsg instanceof Table) && ((objVal = ((Table)objMsg).getx("INFO")) != null))
          info = objVal.toString();
        if ((objMsg instanceof Table) && ((objVal = ((Table)objMsg).getx("FROM")) != null))
          from = objVal.toString();
        if ((objMsg instanceof Table) && ((objVal = ((Table)objMsg).getx("TO")) != null))
          to = objVal.toString();
        tbl.remove("MSG");
      }

      if (callBack!=null) {
        callBack.PostResult(RECVD_TABLE,tbl);
      }
      else if(MQ!=null) {
        MQ.put(msgName,Integer.parseInt(info),tbl,to); // from if fid but filled in by rmif
      }

      // MQ.put(msgName,Integer.parseInt(info),tbl,to,from);
    } catch (Exception ex) {
      System.out.println("Exception in  ProcessXML2MsgExt : PostClient "+ex);
      return;
    }
  }

  private void ProcessXML2Msg(String xmlStr) {
    if (xmlStr==null) return;
    int beginError = xmlStr.indexOf("<XML>");
    int endError   = xmlStr.indexOf("</XML>");
    xmlStr = xmlStr.substring(beginError+5,endError);
    int beginData  = xmlStr.indexOf("<DATA>");
    int endData    = xmlStr.indexOf("</DATA>");
    String dataStr = xmlStr.substring(beginData+6,endData);
    String xmlStr2 = xmlStr.substring(0,beginData)+xmlStr.substring(endData+7);
    // Here I am making TERRIBLE ASSUMPTIONS ABOUT THE XML FORMAT....!!!!!!
    int l1 = xmlStr2.indexOf("<MSG NAME=\"");
    int l2 = xmlStr2.indexOf("\">",l1);
    String msgName = xmlStr2.substring(l1+11,l2);
    l1 = xmlStr2.indexOf("<INFO=\"");
    l2 = xmlStr2.indexOf("\"/>",l1);
    String info = xmlStr2.substring(l1+7,l2);
    l1 = xmlStr2.indexOf("<TO=\"");
    l2 = xmlStr2.indexOf("\"/>",l1);
    String to = xmlStr2.substring(l1+5,l2);
    l1 = xmlStr2.indexOf("<FROM=\"");
    l2 = xmlStr2.indexOf("\"/>",l1);
    String from = xmlStr2.substring(l1+7,l2);
    Table tbl = ProcessData(dataStr,null);
    tbl.put("MSGNAME",msgName);
    if (callBack!=null) {
      callBack.PostResult(RECVD_TABLE,tbl);
    }
    else {
      if (MQ!=null) MQ.put(msgName,Integer.parseInt(info),tbl,to,from);
    }
  }

  private Table ProcessData(String dataStr,String fmtStr) {
    dataStr = dataStr.trim();
    StringTokenizer st = new StringTokenizer(dataStr,",");
    Table t = new Table();
    while (st.hasMoreTokens()) {
      String token = st.nextToken();
      StringTokenizer tst = new StringTokenizer(token,"=");
      int ntokens = tst.countTokens();
      if (ntokens==2) {
        String name = tst.nextToken();
        String value = tst.nextToken();
        t.put(name,value);
      }
    }
    return t;
  }

  private String findFileName() {
    if (deBug) info("findFileName()  start");
    try{
      fos.flush();                    // Flush the Output Stream
      fis.reset();
      byte[] bf  = new byte[8192];
      int count  = fis.available();
      if (deBug) info("findFileName()  fis.available = "+count);
      if (count < 8192) {
        fis.read(bf,0,count);
      } else {
        fis.read(bf,0,8192);
      }
      String str = new String(bf);
      int l1 = str.indexOf("<USERIDENT=");           // Now, get the Filename TAG
      int l2 = str.indexOf("/>",l1);                 // Get endTAG
      if (l1 > -1) {
        return str.substring(l1+11,l2);              // Extract the FILENAME
      }
    } catch (Exception ex) {
      if(deBug) info("findFileName()  Exception "+ex);
    }
    if (deBug) info("findFileName()  Returjing NULL");
    return null;
  }

  private int findSize() {
    try {
      fos.flush();                    // Flush the Output Stream
      fis.reset();
      byte[] bf   = new byte[8192];
      int count   = fis.available();
      if (count < 8192) {
        fis.read(bf,0,count);
      } else {
        fis.read(bf,0,8192);
      }
      String str = new String(bf);
      int sizeloc  = str.indexOf("<FILESIZE=");       // Extract the FILESIZE from header
      int sizeloc2 = str.indexOf("/>",sizeloc);       // SO we can send USER a XFer Status
      String sizestr = "0";                           // Init to ZERO, incase of bad read
      if (sizeloc>-1) {                               // If TAG found, get the substring
        sizestr = str.substring(sizeloc+10,sizeloc2);
      }
      return Integer.parseInt(sizestr);               // Convert SubString into INTEGER
    } catch(Exception ex) {
      return 0;
    }
  }

  private String findMarker() {
    try {
      fos.flush();                    // Flush the Output Stream
      fis.reset();
      byte[] bf     = new byte[8192];
      int count     = fis.available();
      if (count < 8192) {
        fis.read(bf,0,count);
      } else {
        fis.read(bf,0,8192);
      }
      String str = new String(bf);
      int loc   = str.indexOf("boundary=");
      if (loc > -1) {
        int loc2 = str.indexOf("\r\n",loc);
        String marker = readStr.substring(loc+9,loc2);
        return marker;
      } else {
        return null;
      }
    } catch (Exception ex) {
      return null;
    }
  }

  private int findStart(String marker,int instance) {
    boolean lp = true;
    int totalWalked = 0;
    int startInstance = 0;
    try {
      while (lp) {
        BufferedInputStream fis = new BufferedInputStream(new FileInputStream(df));
        totalFISsize = fis.available();
        int toRead = readSize + overLap;
        int readPos  = totalWalked;
        if (readPos<0) readPos = 0;
        if (toRead > (totalFISsize-readPos)){
          toRead = totalFISsize-totalWalked;
          lp = false;
        }
        fis.skip(readPos);
        int got    = fis.read(walkBuf,0,toRead);
        String str = new String(walkBuf);
        boolean dp = true;
        int markStart = -1;
        int lookidx   = 0;
        while (dp) {
          markStart= str.indexOf("--"+marker+"\r\n",lookidx);
          if (markStart > -1) {
            lookidx = markStart+marker.length();
            startInstance++;
            if (startInstance==instance) {
              fis.close();
              return markStart+totalWalked;
            }
          } else {
            dp = false;
          }
        }
        totalWalked = totalWalked + readSize;
        fis.close();
        fis = null;
      }
    } catch (Exception ex) {
      System.out.println("Exception qer "+ex);
    }
    return -1;
  }

  private int findStop(String marker,int instance){
    boolean lp = true;
    int totalWalked = 0;
    int startInstance = 0;
    try{
      while (lp) {
        BufferedInputStream fis = new BufferedInputStream(new FileInputStream(df));
        totalFISsize = fis.available();
        int toRead = readSize + overLap;
        int readPos  = totalWalked;
        if (readPos<0) readPos = 0;
        if (toRead > (totalFISsize-readPos)) {
          toRead = totalFISsize-totalWalked;
          lp = false;
        }
        fis.skip(readPos);
        int got    = fis.read(walkBuf,0,toRead);
        String str = new String(walkBuf);
        boolean dp = true;
        int markStart = -1;
        int lookidx   = 0;
        while (dp) {
          markStart= str.indexOf("--"+marker+"--",lookidx);
          if (markStart > -1) {
            lookidx = markStart+marker.length()+4;
            startInstance++;
            if (startInstance==instance) {
              fis.close();
              return markStart+totalWalked;
            }
          } else {
            dp = false;
          }
        }
        totalWalked = totalWalked + readSize;
        fis.close();
        fis = null;
      }
    }catch(Exception ex){
      System.out.println("Exception qer "+ex);
    }
    return 0;
  }

  private int findString(int offset,String token,int instance) {
    boolean lp = true;
    int totalWalked = 0;
    int startInstance = 0;
    int lastInstance = 0;
    try {
      while (lp) {
        BufferedInputStream fis = new BufferedInputStream(new FileInputStream(df));
        fis.skip(offset);
        totalFISsize = fis.available();
        int toRead = readSize + overLap;
        int readPos  = totalWalked;
        if (readPos < 0) readPos = 0;
        if (toRead>(totalFISsize-readPos)) {
          toRead = totalFISsize-totalWalked;
          lp = false;
        }
        fis.skip(readPos);
        int got    = fis.read(walkBuf,0,toRead);
        String str = new String(walkBuf);
        boolean dp = true;
        int markStart = -1;
        int lookidx   = 0;
        while (dp) {
          markStart= str.indexOf(token,lookidx);
          if (markStart > -1) {
            lookidx = markStart+token.length()+4;
            startInstance++;
            if (instance == -1) {
              lastInstance = markStart+totalWalked;
            } else {
              if (startInstance==instance) {
                fis.close();
                return markStart+totalWalked;
              }
            }
          } else {
            dp = false;
          }
        }
        totalWalked = totalWalked + readSize;
        fis.close();
        fis = null;
      }
    } catch (Exception ex) {
      System.out.println("Exception qer "+ex);
    }
    if (instance==-1) {
      return lastInstance;
    }
    return 0;
  }

  private void info(String str){
    System.out.println(str);
  }

}
