package nxm.nmpostlegacy.lib;

import nxm.sys.lib.MessageQueue;
import nxm.sys.lib.Results;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.InetAddress;
import java.io.IOException;
import java.util.Vector;

/**
 * <p>Title:        nmpost (HTTP server)</p>
 * <p>Description:  This primitive accepts POSTS from a WEB or whatever. </p>
 * <P>              It expects the POSTS to be in the format below.     </p>
 * <P>              Format is for MESSAGES.  NMPOST turns post into MSG, and places in QUEUE </p>
 * <p>Company:      Zeta Associates</p>
 * <p>Version:      1.0</p>
 * @author        Wayne G. Dengel II
 * @version $Id: PostServer.java,v 1.3 2013/01/10 14:36:47 ntn Exp $
 */
public class PostServer implements Runnable {
  private boolean            deBug         = false; // GLOBAL Debug Flag
  private int                maxPorts      = 50;    // Maximum ports to try to find a free one
  private ServerSocket       server        = null;  // The SocketServer Itself
  public  int                listenPort    = 0;     // The PORT we are listening on
  private Thread             myThread      = null;  // The Thread that the LISTENER is running on
  private Vector<PostClient> parseThreads  = null;  // Hold POST clients
  private boolean            execute       = true;
  private Cleaner            cleaner       = null;  // Removes COMPLETED clients
  private MessageQueue       MQ            = null;  // Private Handle to MQ
  private Results            MR            = null;
  private String             uPath         = null;
  private boolean            reportMain    = false;
  public String              strListenAddr = null;
  private PostClient_cb      callBack      = null;

  public PostServer(MessageQueue mq,Results mr,boolean debug,int startPort,int maxPorts,
      String path,boolean reportMain,PostClient_cb callBack) throws Exception {
    boolean reTry   = true;
    int     curTry  = 0;
    this.maxPorts   = maxPorts;
    this.deBug      = debug;
    this.MQ         = mq;
    this.MR         = mr;
    listenPort      = startPort;
    this.uPath      = path;
    this.reportMain = reportMain;
    this.callBack   = callBack;
    parseThreads    = new Vector<PostClient>();  // The VECTOR that holds the POST Threads
    cleaner         = new Cleaner();             // Thread to CLEANOUT parseThreads
    cleaner.start();                             // Start the CLEANER
    while (reTry) {                              // Start looking for FREE Port
      try {
        server = new ServerSocket(listenPort);   // Create the SERVER
        reTry  = false;
        listenPort = server.getLocalPort();      // for when listenPort is 0 (any port)
      }
      catch (IOException ex) {
        if (curTry<maxPorts) {
          listenPort++;
          curTry++;
        } else {
          throw new Exception("No More Ports to use in portServer");
        }
      }
    }
    InetAddress inetaddr;
    inetaddr = InetAddress.getLocalHost();
    strListenAddr = inetaddr.getHostAddress();
    System.out.println("PostServer:Server created on "+strListenAddr+"  "+listenPort);
    if (MR!=null) {
      MR.put("POSTPORT",listenPort);
      MR.put("POSTADDR",strListenAddr);
    }
    myThread = new Thread(this);                 // Start OUR Server Thread
    myThread.start();                            // '' ''
  }

  public int getPort() { return listenPort; }
  public String getAddr() { return strListenAddr; }

  public void run() {
    while (execute) {
      try {
        Socket client = server.accept();
        PostClient tempClient = new PostClient(deBug,client,MQ,uPath,reportMain,callBack);
        parseThreads.add(tempClient);
      } catch (java.net.SocketException se) {
         if ("socket closed".equalsIgnoreCase(se.getMessage())) {} // Server Socket has been closed (not an error)
         else System.out.println("PostServer: SocketException in RUN: "+se);
      } catch (Exception ex) { System.out.println("PostServer: Exception in RUN: "+ex); }
    }
  }

  /** Stop server socket. This API matches HttpServer.stop(int) but does not
   *  implement the close timeout.
   *
   * @param seconds seconds to wait for server socket to close (ignored)
   * @since NeXtMidas 2.7.3
   */
  public void stop (int seconds) {
    execute = false;
    if (server!=null && !server.isClosed()) {
      myThread.interrupt();
      try {
        server.close();
      } catch (IOException ioe) {
        System.out.println("PostServer: Exception in stop(): "+ioe);
      }
    }
  }

  public class Cleaner extends Thread {
    public void run() {
      int total = 0;
      try {
        while (execute) {
          total = parseThreads.size();             // Get Vector Size NOW...!
          if (total>0) {                           // If items, then DO it, otherwise SLEEP
            for (int i=0;i<total;i++) {            // Walk the Vector
              PostClient pc = parseThreads.get(i); // Get Vector handle
              if (pc.complete) {                   // If DONE, then DELETE IT...
                pc = parseThreads.remove(i);       // REMOVE from VECTOR
                //System.out.println("Destroying thread");
                pc = null;                         // Mark as NULL, so GC can kill it
                i = total+2;                       // Force LOOP to EXIT, so we can reindex..
                break;
              }
            } // end for loop
            sleep(50);                        // SLEEP Some (before checking next item)
          } else {
            sleep(1000);                      // Sleep MORE (no items to clean)
          }
        } // end while loop
      } catch (Exception ex) {}
    } // end run()
  }

}
