package nxm.nmpostlegacy.lib;

import java.net.HttpURLConnection;
import java.net.URL;
import java.io.DataInputStream;

/**
   Title: WebRequestThread / getelcc <p>

   Description: This primitive receives Messages from a macro
   telling it what type of ELCC file it wants to receive.. And the Radio name
   we then send a message to the ELCC web-server. The web-server then tell us
   if it can service the request or not If it can, then we wait for a POST back
   to us telling us the NODE where the FILENAME is stored. This goes against
   most ways I write stuff like this.. In the case, I lumped it all into one
   LARGE primitive. Normally I would use separate primitives/classes, but in
   this case it makes sense.. :>) </p>

   Created:   October 12, 2004<br>
   Version:   1.0

   @author    Wayne G. Dengel II  (Zeta Associates)
   @version  $Id: WebRequestThread.java,v 1.2 2013/01/07 21:09:41 ntn Exp $
 */
public class WebRequestThread implements Runnable {
    public final static  int                  WEBOK          = 1;
    public final static  int                  WEBERROR       = -4;
    public final static  int                  AUDIO          = 1;
    public final static  int                  PSD            = 2;
    public final static  int                  PRED           = 3;
    public final static  int                  MBT            = 4;
    public final static  int                  TMIC           = 3;

    private              URL                  url            = null;
    private              HttpURLConnection    ucon           = null;
    private              DataInputStream      sis            = null;
    private              WebRequestThread_cb  callBack       = null;
    private              Thread               myThread       = null;

    private              boolean              DEBUG          = false;
    public               String               elccRadioName  = null;                             // The ELCC radio we want to DEARCH from
    public               String               filename       = null;                             // The target Filename
    private              String               web            = null;
    private              double               jstart         = 0.0;
    private              double               duration       = 0.0;
    private              String               type;
    private              String               postNode       = null;
    private              int                  postPort       = 0;
    public               String               request        = null;
    private              String               optParms       = null;
    public               String               errStr         = null;
    public               String               response       = null;

    public WebRequestThread(boolean debug, String radio, double jstart, double dur, int type, String filename, String optionalParms, WebRequestThread_cb callBack, String postNode, int postPort) {
        this.callBack = callBack;
        this.DEBUG = debug;
        this.elccRadioName = radio;
        this.jstart = jstart;
        this.duration = dur;
        this.filename = filename;
        this.optParms = optionalParms;                                                           //"&multitunes=yes&options=/nochk"
        this.postNode = postNode;
        this.postPort = postPort;

        switch (type) {
            case 1:
                this.type = new String("AUDIO");
                break;
            case 2:
                this.type = new String("PREPSD");
                break;
            case 3:
                this.type = new String("PRED");
                break;
            case 4:
                this.type = new String("MBT");
                break;
        }

        this.request = new String("http://geo:geo@" + GetWeb(radio) + "/elcccgi/postcutfile.pl?" + this.optParms);
        this.request = this.request + "&thread=" + radio;
        this.request = this.request + "&starttime=" + jstart;
        this.request = this.request + "&duration=" + dur;
        this.request = this.request + "&ftype=" + this.type;
        this.request = this.request + "&postaddr=" + this.postNode;
        this.request = this.request + "&postport=" + this.postPort;
        this.request = this.request + "&ufname=" + this.filename;
        this.request = this.request.toLowerCase();

        if (DEBUG) {
            info("STARTUP ");
            info(" Radio   " + this.elccRadioName);
            info(" Start   " + this.jstart);
            info(" Dur     " + this.duration);
            info(" File    " + this.filename);
            info(" parms   " + this.optParms);
            info(" pnode   " + this.postNode);
            info(" pport   " + this.postPort);
            info(" request " + this.request);
        }
        myThread = new Thread(this);
        myThread.start();
    }

    public void run() {
        if(DEBUG) info("Requester Running " + request);
        try {
            url = new URL(request);
            ucon = (HttpURLConnection) url.openConnection();
            String  encoding  = nxm.sys.lib.Convert.encodeBase64("geo:geo");
            ucon.setRequestProperty("Authorization", "Basic " + encoding);
            ucon.connect();
            sis = new DataInputStream(ucon.getInputStream());
        } catch (Exception e) {
            info("PostFile::Exception in openning  " + e);
            errStr = e.toString();
            if (callBack != null) {
                callBack.WEBStatus(WEBERROR, this);
            }
            return;
        }
        try {
            int      avail;
            boolean  gRun   = true;
            while (gRun) {
                avail = sis.available();
                if (avail > 0) {
                    byte[]  buf     = new byte[avail];
                    sis.read(buf, 0, avail);
                    String  output  = new String(buf);
                    int     eloc    = output.indexOf("<ERROR>");
                    int     eloc2   = output.indexOf("</ERROR>");
                    response = new String(output);
                    errStr = output.substring(eloc + 7, eloc2);
                    if ((eloc2 < 0) && (eloc < 0)) {
                    } else if (eloc2 - eloc > 7) {
                        if (callBack != null) {
                            callBack.WEBStatus(WEBERROR, this);
                        }
                    } else {
                        if (callBack != null) {
                            callBack.WEBStatus(WEBOK, this);
                        }
                    }
                    ucon.disconnect();
                    return;
                } else {
                    try {
                        Thread.sleep(10);
                    } catch (Exception ex) {}
                }
            }
        } catch (Exception e) {
            info("Exception-->" + e);
        }
        ucon.disconnect();
    }

    private String GetWeb(String radio) {
        return "NONE";
    }

    private void info(String str) {
        String tms = nxm.sys.lib.Time.toString(nxm.sys.lib.Time.current(),nxm.sys.lib.Time.VAX);
        System.out.println(tms+" "+getClass() +": "+ str);
    }
}
