package nxm.nmpostlegacy.prim;

import nxm.sys.lib.*;
import nxm.nmpostlegacy.lib.*;
import java.util.Vector;

/**
   Title:   getelcc<p>
   Created: February 25, 2005<br>
   Version: 1.0</p>
   @author  Wayne G. Dengel II
   @version $Id: getelcc.java,v 1.4 2013/01/07 21:55:46 ntn Exp $
 */
public class getelcc extends Primitive implements WebRequestThread_cb, PostClient_cb, FtpClient_cb {
  private  PostServer  ps           = null;
  private  String      uPath        = null;
  private  Vector<WebRequestThread> webrequests  = new Vector<WebRequestThread>();
  private  Vector<FtpClient>        ftprequests  = new Vector<FtpClient>();
  private  boolean     DEBUG        = false;

  public int open() {
    DEBUG = MA.isPresent("/DEBUG");
    int  port   = MA.getL("PORT", 9010);
    int  maxp   = MA.getL("/MAX", 50);
    uPath       = MA.getS("PATH", null);

    if (uPath == null) {
      String  aux_write  = (String) MR.get("AUX.WRITE");                // Find AUX Write
      uPath = (String) MR.get("AUX." + aux_write);                      // Get HARD Path
      String  user       = (String) MR.get("ENV.USER");                 // Get the USER name
      int     badchar    = uPath.indexOf('+');                          // IF the path has a /+/ then we
      if (badchar != -1) {                                              // need to insert the USERNAME in place of
        uPath = new String(uPath.substring(0, badchar) + user + "/");   // the +,
      }
      M.info("User didn;t specify path, using " + uPath);
    }
    try {
      ps = new PostServer(MQ, MR, DEBUG, port, maxp, uPath, false, this);
    } catch (Exception ex) {
      M.info("Could not start PostServer");
      return FINISH;
    }
    return NORMAL;
  }

  /**
   *  Description of the Method
   *@param  msg  Description of the Parameter
   *@return      Description of the Return Value
   */
  public synchronized int processMessage(Message msg) {                  // Process MESSAGEs
    if (!thisIsMe()) {
      MQ.put(msg);
      return 0;
    }                                                                    // If our message, ignore
    if (msg.name.startsWith("GETFILE")) {
      String            rad       = ((Table) msg.data).getS("RADIO");
      double            jstart    = ((Table) msg.data).getD("START", 0);
      double            duration  = ((Table) msg.data).getD("DUR", 0);
      int               ftype     = ((Table) msg.data).getL("TYPE", 1);
      String            fname     = ((Table) msg.data).getS("FNAME");
      String            options   = ((Table) msg.data).getS("OPTIONS");

      if (DEBUG) {
        M.info("Parsed rad is      " + rad);
        M.info("Parsed jstart is   " + jstart);
        M.info("Parsed duration is " + duration);
        M.info("Parsed ftype is    " + ftype);
        M.info("Parsed fname is    " + fname);
        M.info("Parsed options is  " + options);
      }
      webrequests.add(new WebRequestThread(DEBUG, rad, jstart, duration, ftype, fname, options, this, ps.strListenAddr, ps.listenPort));
    }
    return NORMAL;
  }

  /**
   *  Description of the Method
   *
   *@param  status  Description of the Parameter
   *@param  obj     Description of the Parameter
   */
  public synchronized void PostResult(int status, Object obj) {
    if (DEBUG)  M.info("PostResult() " + status + "  " + obj);
    switch (status) {
      case PostClient.RECVD_TABLE:{
        Table  tbl     = (Table) obj;
        Table  subTbl  = tbl.getTable("DEARCHREQ");
        if (subTbl != null) {
          String     file   = subTbl.getS("DIR") + subTbl.getS("FILE");
          String     host   = subTbl.getS("DHOST");
          String     error  = subTbl.getS("ERROR");
          if (error.length() > 0) {
            M.info("ERROR: file didn;t dearch..  " + error);
            MQ.put("DERROR", 0, error, "MAIN");
            return;
          }
          FtpClient  ftp    = new FtpClient(!DEBUG,  // QUIET Flag for FTP/Debug messages
              host,                                  // FTP Target Host
              uPath,                                 // Path to place the target file.
              file,                                  // The file to get.  FULL target directory, or must be in LOGIN
              "salem",                               // FTP account
              "Salem",                               // FTP account password
              this);
          ftprequests.add(ftp);
        } else {
          String  msgname  = tbl.getS("MSGNAME");
          if (msgname != null) {
            MQ.put(msgname, 0, tbl, "MAIN");
          }

        }
        break;
      }
    }
  }

  /**
   *  Description of the Method
   *@param  status  Description of the Parameter
   *@param  client  Description of the Parameter
   */
  public synchronized void PostStatus(int status, PostClient client) {
    if (DEBUG)  M.info("PostStatus() " + status + "  " + client);
  }

  /**
   *  Description of the Method
   *@param  percent   Description of the Parameter
   *@param  filename  Description of the Parameter
   */
  public synchronized void PostPercent(double percent, String filename) {
    if (DEBUG)  M.info("PostPercent() " + filename + "  " + percent);
  }

  /**
   *  Gets the writePath attribute of the getelcc object
   *@param  path      Description of the Parameter
   *@param  filename  Description of the Parameter
   *@return           The writePath value
   */
  public synchronized String getWritePath(String path, String filename) {
    return path;
  }

  /**
   *  Description of the Method
   *@param  status  Description of the Parameter
   *@param  client  Description of the Parameter
   */
  public synchronized void FTPStatus(int status, FtpClient client) {
    switch (status) {
      case FtpClient.XFER_COMPLETE:
        if (DEBUG) {
          M.info(client.path + client.filename + " COMPLETE");
        }
        MQ.put("FTPCOMPLETE", 0, client.filename, "MAIN");
        break;
      default:
    }
    ftprequests.remove(client);
  }

  /**
   *  Description of the Method
   *
   *@param  percent  Description of the Parameter
   *@param  client   Description of the Parameter
   */
  public synchronized void FTPPercent(double percent, FtpClient client) {
    if (DEBUG) {
      M.info(client.filename + " Percent is " + percent);
    }
    MQ.put("FTPPERCENT", (int) percent, client.filename, "MAIN");
  }

  /**
   *  Description of the Method
   *@param  status  Description of the Parameter
   *@param  client  Description of the Parameter
   */
  public synchronized void WEBStatus(int status, WebRequestThread client) {
    if (DEBUG) {
      M.info("WEBStatus = " + status);
    }
    switch (status) {
      case WebRequestThread.WEBERROR:
      {
        Table  tbl  = new Table();
        tbl.put("NAME", client.filename);
        tbl.put("ERROR", client.errStr);
        MQ.put("FAILEDFILE", 0, tbl, "MAIN");
        webrequests.remove(client);
        break;
      }
      case WebRequestThread.WEBOK:
      {
        if (DEBUG) {
          M.info("Ready to get the file:  " + client.response);
        }
        webrequests.remove(client);
        break;
      }
    }

  }

}
