package nxm.sys.inc;

/** Definitions of constants defining data types, modes, and representations.
    For historical reasons (and for compatibility with X-Midas), the data types
    and representations used by NeXtMidas do not mirror the standard Java types.
    However, analogues for most of the NeXtMidas data types are readily
    available in Java (though the names differ). This interface provides the
    identifiers necessary to properly work with the NeXtMidas data types from
    within Java.<br>
    <br>
    <a name="types"><b>NeXtMidas Data Types</b></a><br>
    The following table summarizes the data types used by NeXtMidas. Take
    careful note of the integer types (XLONG, LONG, INT) as they differ
    significantly from their Java counter-parts.
    <center><table border="1" width="90%" align="center">
      <tr><td align="center" rowspan="2"><b>Identifier</b></td>
          <td align="center" rowspan="2"><b>Midas Name</b></td>
          <td align="center" rowspan="2"><b>Java Representation</b></td>
          <td align="center" colspan="3"><b>NeXtMidas Type Character
                                         </b></td></tr>
      <tr><td align="center"><b>Data Type<br>
                 {@link #DATA_TYPE_LIST}</b></td>
          <td align="center"><b>Result Type<br>
                 {@link #RESULT_TYPE_LIST}</b></td></tr>
      <tr><td>{@link #ASCII}</td>
          <td>ASCII String</td>
          <td>&nbsp;</td>
          <td align="center">A</td>
          <td align="center">A</td></tr>
      <tr><td>{@link #STRING}</td>
          <td>Unicode String</td>
          <td>{@link java.lang.String}</td>
          <td align="center">&nbsp;</td>
          <td align="center">S</td></tr>
      <tr><td>{@link #OBJECT}</td>
          <td>Java Object</td>
          <td>{@link java.lang.Object}</td>
          <td align="center">&nbsp;</td>
          <td align="center">O</td></tr>
      <tr><td>{@link #TABLE}</td>
          <td>Table</td>
          <td>&nbsp;</td>
          <td align="center">&nbsp;</td>
          <td align="center">T</td></tr>
      <tr><td>{@link #BYTE}</td>
          <td>Byte</td>
          <td>{@link java.lang.Byte}</td>
          <td align="center">B</td>
          <td align="center">B</td></tr>
      <tr><td>{@link #DOUBLE}</td>
          <td>Double</td>
          <td>{@link java.lang.Double}</td>
          <td align="center">D</td>
          <td align="center">D</td></tr>
      <tr><td>{@link #FLOAT}</td>
          <td>Float</td>
          <td>{@link java.lang.Float}</td>
          <td align="center">F</td>
          <td align="center">F</td></tr>
      <tr><td>{@link #XLONG}</td>
          <td>Extra Long Integer</td>
          <td>{@link java.lang.Long}</td>
          <td align="center">X</td>
          <td align="center">X</td></tr>
      <tr><td>{@link #LONG}</td>
          <td>Long Integer</td>
          <td>{@link java.lang.Integer}</td>
          <td align="center">L</td>
          <td align="center">L</td></tr>
      <tr><td>{@link #INT}</td>
          <td>(Short) Integer</td>
          <td>{@link java.lang.Short}</td>
          <td align="center">I</td>
          <td align="center">I</td></tr>
      <tr><td>{@link #NIBBLE}</td>
          <td>Packed Nibble</td>
          <td>{@link java.lang.Byte}</td>
          <td align="center">N</td>
          <td align="center">N</td></tr>
      <tr><td>{@link #PACKED}</td>
          <td>Packed Bit</td>
          <td>{@link java.lang.Byte}</td>
          <td align="center">P</td>
          <td align="center">P</td></tr>
      <tr><td>{@link #BOOL}</td>
          <td>Boolean</td>
          <td>{@link java.lang.Boolean}</td>
          <td align="center">Z</td>
          <td align="center">Z</td></tr>
      <tr><td>{@link #ULAW}</td>
          <td>U-Law</td>
          <td><i>TBD</i></td>
          <td align="center"><i>TBD</i></td>
          <td align="center"><i>TBD</i></td></tr>
      <tr><td>{@link #ALAW}</td>
          <td>A-Law</td>
          <td><i>TBD</i></td>
          <td align="center"><i>TBD</i></td>
          <td align="center"><i>TBD</i></td></tr>
      <tr><td>{@link #NUMBER}</td>
          <td>Generic Number</td>
          <td>&nbsp;</td>
          <td align="center">&nbsp;</td>
          <td align="center">N</td></tr>
      <tr><td>{@link #XMVALUE}</td>
          <td>Generic XMValue</td>
          <td>{@link nxm.sys.lib.XMValue}</td>
          <td align="center">V</td>
          <td align="center">V</td></tr>
      <tr><td>{@link #XMVALUE_LIST}</td>
          <td>XMValue List (Collection)</td>
          <td>{@link nxm.sys.lib.XMValue}</td>
          <td align="center">C</td>
          <td align="center">C</td></tr>
      <tr><td>{@link #XMVALUE_KVLIST}</td>
          <td>XMValue Key-Value List</td>
          <td>{@link nxm.sys.lib.XMValue}</td>
          <td align="center">K</td>
          <td align="center">K</td></tr>
      <tr><td>{@link #XMVALUE_TABLE}</td>
          <td>XMValue Table (Hash table)</td>
          <td>{@link nxm.sys.lib.XMValue}</td>
          <td align="center">H</td>
          <td align="center">H</td></tr>
    </table></center><br>
    <br>
    Note: The NeXtMidas ASCII type represents a block of eight 8-bit ASCII
          characters (please see {@link #ASCII} for further details).<br>
    <br>
    <a name="modes"><b>NeXtMidas Data Modes</b></a><br>
    Along with the standard data types NeXtMidas supports the use of
    multi-value entities (or arrays). Multi-valued types with special
    names/meanings include:<br>
    <center><table border="1" width="90%">
      <tr><td align="center"><b>Identifier</b></td>
          <td align="center"><b>Common Name</b></td>
          <td align="center"><b>Number of Elements</b></td>
          <td align="center"><b>Type Character</b></td></tr>
      <tr><td>{@link #SCALAR}</td>
          <td>Scalar</td>
          <td align="center">1</td>
          <td align="center">S</td></tr>
      <tr><td>{@link #COMPLEX}</td>
          <td>Complex</td>
          <td align="center">2</td>
          <td align="center">C</td></tr>
      <tr><td>{@link #VECTOR}</td>
          <td>Vector</td>
          <td align="center">3</td>
          <td align="center">V</td></tr>
      <tr><td>{@link #QUAD}</td>
          <td>Quad</td>
          <td align="center">4</td>
          <td align="center">Q</td></tr>
      <tr><td>{@link #MATRIX}</td>
          <td>Matrix</td>
          <td align="center">9</td>
          <td align="center">M</td></tr>
      <tr><td>{@link #XMODE}</td>
          <td>10-Element Array</td>
          <td align="center">10</td>
          <td align="center">X</td></tr>
      <tr><td>{@link #TRANSMATRIX}</td>
          <td>Transformation Matrix</td>
          <td align="center">16</td>
          <td align="center">T</td></tr>
      <tr><td>{@link #NONHOMOGENEOUS}</td>
          <td>Nonhomogeneous</td>
          <td align="center">n/a</td>
          <td align="center">NH</td></tr>
    </table></center><br>
    <br>
    Additionally NeXtMidas supports generic references to multi-valued types
    which are identified by the character indicating the number of elements
    contained. Thus a 2-element array can be be either referred to as a
    <code>C</code> (complex) or as a <code>2</code> (2-Element Array). It is
    preferable that the letter identifier be used when the data matches the
    meaning of the identifier (e.g. a 9-Element Array should use <code>9</code>
    if it does not represent a 3x3 matrix and it should use <code>M</code> if
    it represents a matrix), however this rule is not always followed and some
    methods may improperly select between letter and number identifiers.<br>
    <br>
    The following summarizes the numerical identifiers.
    <center><table border="1" width="90%">
      <tr><td align="center"><b>Identifier</b></td>
          <td align="center"><b>Common Name</b></td>
          <td align="center"><b>Number of Elements</b></td>
          <td align="center"><b>Type Character</b></td></tr>
      <tr><td>&nbsp;</td>
          <td><i>N</i>-Element Array</td>
          <td align="center">1 .. 9</td>
          <td align="center">1 .. 9</td></tr>
    </table></center><br>
    <br>
    <a name="byte_ordering"><b>NeXtMidas Byte Ordering</b></a><br>
    NeXtMidas is able to work with numbers with differing binary
    representations. The following summarizes the different binary
    representations which NeXtMidas may work with.
    <center><table border="1" width="90%">
      <tr><td align="center"><b>Identifier</b></td>
          <td align="center"><b>Description</b></td>
          <td align="center"><b>NeXtMidas/X-Midas<br>
                                File Marker</b></td>
          <td align="center"><b>Typical Uses</b></td></tr>
      <tr><td>{@link #IEEE}</td>
          <td>IEEE format with Big-Endian Byte Ordering</td>
          <td align="center"><code>"IEEE"</code></td>
          <td>Typical of most systems from Sun Microsystems<br>
              Used internally by Java on Big-Endian machines</td></tr>
      <tr><td>{@link #EEEI}</td>
          <td>IEEE format with Little-Endian Byte Ordering</td>
          <td align="center"><code>"EEEI"</code></td>
          <td>Standard for Intel x86 processors<br>
              Used internally by Java on Little-Endian machines</td></tr>
      <tr><td>{@link #VAX}</td>
          <td>VAX Format with Little-Endian Byte Ordering</td>
          <td align="center"><code>"VAX "</code></td>
          <td>Used on VAX machines</td></tr>
      <tr><td>{@link #CRAY}</td>
          <td>CRAY Format with Big-Endian Byte Ordering</td>
          <td align="center"><code>"CRAY"</code></td>
          <td>Used on Cray machines</td></tr>
    </table></center>

    @see nxm.sys.lib.Data

    @author Jeff Schoen
    @version $Id: DataTypes.java,v 1.34 2012/10/08 20:05:59 ntn Exp $
 */
public interface DataTypes {

  /** This is a pipe-separated (<code>|</code>) list of the logical states that
      indicate a true value. Note that this must start and end with a pipe character
      (<code>|</code>) in order to be compatible with the existing implementation.
      <br>
      True States = <code>{@value}</code>
  */
  public static final String trueStates = "|Y|YES|T|TRUE|ON|";

  /** This is a pipe-separated (<code>|</code>) list of the logical states that
      indicate a false value. Note that this must start and end with a pipe character
      (<code>|</code>) in order to be compatible with the existing implementation.
      <br>
      False States = <code>{@value}</code>
  */
  public static final String falseStates = "|N|NO|F|FALSE|OFF|";

  /** List of valid data representations.
      Value = <code>{@value}</code>
  */
  public static final String repList = "IEEE,EEEI,VAX,CRAY";

  /** List of valid data modes. <br><br> Value = <code>{@value}</code> */
  public static final String modeList = "S,C,V,Q,M,X,T,1,2,3,4,5,6,7,8,9";

  /** List of valid numeric data types. <br><br> Value = <code>{@value}</code> */
  public static final String numericTypeList = "D,F,X,L,I,B,N,P";

  /** List of valid numeric data types. <br><br> Value = <code>{@value}</code> */
  public static final String typeList = numericTypeList+",A";

  /** List of valid data mode characters. <br><br> Value = <code>{@value}</code> */
  public static final String DATA_MODE_LIST = "SCVQMT";

  /** List of valid multiple data mode characters. <br><br> Value = <code>{@value}</code> */
  public static final String MULT_DATA_MODE_LIST = "123456789X";

  /** List of valid  data mode characters that match # in the checkFormat mode. <br><br>
      Value = <code>{@value}</code>
  */
  public static final String NUMERIC_DATA_TYPE_LIST = "DFXLIBNP";

  /** List of valid data type characters that match # in the checkFormat type. <br><br>
      Value = <code>{@value}</code>
      @see #RESULT_TYPE_LIST
      @see #CONVERT_TYPE_LIST
  */
  public static final String DATA_TYPE_LIST = NUMERIC_DATA_TYPE_LIST+"A";

  /** List of valid Result type characters that match * in the checkFormat type.
      Note that this includes all of the characters in {@link #DATA_TYPE_LIST}. <br><br>
      Value = <code>{@value}</code>
      @see #DATA_TYPE_LIST
      @see #CONVERT_TYPE_LIST
   */
  public static final String RESULT_TYPE_LIST = DATA_TYPE_LIST+"SOTZ"+"VCKH"; // "VCHK" = XMValue

  /** List of valid Conversion type characters. Note that this includes all of
      the characters in {@link #RESULT_TYPE_LIST} plus the following:
      <table>
        <tr>
          <th>Type</th>
          <th>Meaning</th>
        </tr>
        <tr>
          <td align=center><code>_</code></td>
          <td>Type not known, attempt to figure out automatically.</td>
        </tr>
        <tr>
          <td align=center><code>U</code></td>
          <td>Do not translate.</td>
        </tr>
        <tr>
          <td align=center><code>Q</code></td>
          <td>Quoted, type is specified along with value (e.g.
              <code>L:FOO</code>).</td>
        </tr>
      </table>
      @see #DATA_TYPE_LIST
      @see #RESULT_TYPE_LIST
   */
  public static final String CONVERT_TYPE_LIST = RESULT_TYPE_LIST+"_UQ";


  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Modes
  //////////////////////////////////////////////////////////////////////////////////////////////////
  /** Mode='S': Scalar or other 1-dimensional or non-dimensional value. */
  public static final byte SCALAR = (byte)'S';

  /** Mode='C': Complex or other 2-dimensional value. */
  public static final byte COMPLEX = (byte)'C';

  /** Mode='V': Vector or other 3-dimensional value. */
  public static final byte VECTOR = (byte)'V';

  /** Mode='Q': Quad or other 4-dimensional value. */
  public static final byte QUAD = (byte)'Q';

  /** Mode='M': Matrix or other 3x3-dimensional value. */
  public static final byte MATRIX = (byte)'M';

  /** Mode='X': Ten element or dimensional value. */
  public static final byte XMODE = (byte)'X';

  /** Mode='T': TransMatrix or other 4x4-dimensional value. */
  public static final byte TRANSMATRIX = (byte)'T';

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Types
  //////////////////////////////////////////////////////////////////////////////////////////////////
  /** Type 'P': N-bit values packed into 8-bit words.
      <br><br>
      Note: Packed bits are for storage and transport only. Prior to processing,
      packed data is automatically normalized to a full byte.
   */
  public static final byte PACKED = (byte)'P';

  /** Type 'N': (Packed Nibble) Signed 4-bit values packed into 8-bit words.
      <br><br>
      Note: Nibbles are used for storage and transport only. Prior to processing,
      Nibble data is automatically normalized to a full byte.
   */
  public static final byte NIBBLE = (byte)'N';

  /** Type 'u': Signed 8-bit u-Law value. <i>Not currently implemented.</i> */
  public static final byte ULAW = (byte)'u';

  /** Type 'a': Signed 8-bit a-Law value. <i>Not currently implemented.</i> */
  public static final byte ALAW = (byte)'a';

  /** Type 'B': Signed 8-bit Byte value (see {@link #BMIN}, {@link #BMAX}). */
  public static final byte BYTE = (byte)'B';

  /** Type 'I': Signed 16-bit (Short) Integer value (see {@link #IMIN}, {@link #IMAX}). */
  public static final byte INT = (byte)'I';

  /** Type 'L': Signed 32-bit Long Integer value (see {@link #LMIN}, {@link #LMAX}). */
  public static final byte LONG = (byte)'L';

  /** Type 'X': Signed 64-bit Extra-Long Integer value (see {@link #XMIN}, {@link #XMAX}). */
  public static final byte XLONG = (byte)'X';

  /** Type 'F': Signed 32-bit Floating Point value (see {@link #FMIN}, {@link #FMAX}). */
  public static final byte FLOAT = (byte)'F';

  /** Type 'D': Signed 64-bit Floating Point value (see {@link #DMIN}, {@link #DMAX}). */
  public static final byte DOUBLE = (byte)'D';

  /** Type 'Z': Boolean value. <br><br>
      <i>Note: Prior to NeXtMidas 2.5.0 the value of this constant was 'J' not 'Z'. This was
      corrected to match Midas standards as of 2.5.0.</i>
   */
  public static final byte BOOL = (byte)'Z';  // <-- prior to 2.5.0 this was 'J'

  /** Type 'H': Non-Homogeneous record format for Type 3000/5000/6000 files. This only appears in
      the main header of a BLUE file and is not used within macros or the shell.
      @see #XMVALUE_TABLE
   */
  public static final byte NONHOMOGENEOUS = (byte)'H';

  /** Type 'N': A generic number, defaults to {@link #DOUBLE} */
  public static final byte NUMBER = (byte)'N';

  /** Type 'A': Block of eight 8-bit ASCII values (not Unicode). Note that this differs
      from {@link #STRING} in two <b>extremely</b> important ways:
      <ol>
        <li>All values (i.e. characters) are 8-bit UTF-8 ASCII characters and
            are not in the Unicode format used by the Java type
            {@link java.lang.Character}. (This is done or historical reasons and
            for compatibility with X-Midas.)</li>
        <li>This type of data represents a block of <b>exactly</b> eight
            ASCII characters. If it holds a string that is less than eight
            characters in length, it will be padded with spaces out to fill the
            entire eight-character block. (This is done for historical reasons
            and for compatibility with X-Midas.)</li>
      </ol>
      @see #STRING
      @see nxm.sys.inc.AsciiMap
   */
  public static final byte ASCII = (byte)'A';

  /** Type 'S': A standard Unicode string as used in Java (see {@link java.lang.String}).
      Note that this differs significantly from the NeXtMidas ASCII type
      (please see {@link #ASCII} for further explanation).
      @see #ASCII
      @see java.lang.String
   */
  public static final byte STRING = (byte)'S';

  /** Type 'O': Generic Java Object (see {@link java.lang.Object}). */
  public static final byte OBJECT = (byte)'O';

  /** Type 'T': Generic Table Object (see {@link nxm.sys.lib.Table}). */
  public static final byte TABLE = (byte)'T';

  /** Type 'V': A generic {@link nxm.sys.lib.XMValue} for compatibility with X-Midas.
      @since NeXtMidas 2.7.1
   */
  public static final byte XMVALUE = (byte)'V';

  /** Type 'C': {@link nxm.sys.lib.XMValue#LIST} for compatibility with X-Midas.
      @since NeXtMidas 2.7.1
   */
  public static final byte XMVALUE_LIST = (byte)'C';

  /** Type 'C': {@link nxm.sys.lib.XMValue#KVLIST} for compatibility with X-Midas.
      @since NeXtMidas 2.7.1
   */
  public static final byte XMVALUE_KVLIST = (byte)'K';

  /** Type 'H': {@link nxm.sys.lib.XMValue#TABLE} for compatibility with X-Midas.
      @see nxm.sys.lib.XMValue
      @see #NONHOMOGENEOUS
      @since NeXtMidas 2.7.1
   */
  public static final byte XMVALUE_TABLE = (byte)'H';


  //////////////////////////////////////////////////////////////////////////////////////////////////
  // DataRep
  //////////////////////////////////////////////////////////////////////////////////////////////////
  /** DataRep 'I': IEEE data representation. The IEEE standard with big-endian byte ordering. */
  public static final byte IEEE = (byte)'I';

  /** DataRep 'E': EEEI data representation. The IEEE standard with little-endian byte ordering. */
  public static final byte EEEI = (byte)'E';

  /** DataRep 'V': VAX data representation. This is the byte ordering used by VAX systems. */
  public static final byte VAX = (byte)'V';

  /** DataRep 'C': CRAY data representation. This is the byte ordering used by Cray systems. */
  public static final byte CRAY = (byte)'C';


  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Min/Max Values
  //////////////////////////////////////////////////////////////////////////////////////////////////
  /** The minimum (bounded) value for the NeXtMidas type double. <br><br>Value = <code>{@value}</code> */
  public static final double DMIN=-1e37;

  /** The maximum (bounded) value for the NeXtMidas type double. <br><br>Value = <code>{@value}</code> */
  public static final double DMAX=1e37;

  /** The minimum (bounded) value for the NeXtMidas type float. <br><br>Value = <code>{@value}</code> */
  public static final float FMIN=-1e37f;

  /** The maximum (bounded) value for the NeXtMidas type float. <br><br>Value = <code>{@value}</code> */
  public static final float FMAX=1e37f;

  /** The minimum value for the NeXtMidas type extra-long integer. <br> <br>
      Value = <code>{@value} = {@link java.lang.Long#MIN_VALUE} + 1</code>
   */
  public static final long XMIN=java.lang.Long.MIN_VALUE+1;

  /** The maximum value for the NeXtMidas type extra-long integer. <br> <br>
      Value = <code>{@value} = {@link java.lang.Long#MAX_VALUE}</code>
   */
  public static final long XMAX=java.lang.Long.MAX_VALUE;

  /** The minimum value for the NeXtMidas type long integer. <br> <br>
      Value = <code>{@value} = {@link java.lang.Integer#MIN_VALUE} + 1</code>
   */
  public static final int LMIN=java.lang.Integer.MIN_VALUE+1;

  /** The maximum value for the NeXtMidas type long integer. <br> <br>
      Value = <code>{@value} = {@link java.lang.Integer#MAX_VALUE}</code>
   */
  public static final int LMAX=java.lang.Integer.MAX_VALUE;

  /** The maximum value for the NeXtMidas type short integer. <br> <br>
      Value = <code>{@value} = {@link java.lang.Short#MIN_VALUE}</code>
   */
  public static final short IMIN=java.lang.Short.MIN_VALUE;

  /** The maximum value for the NeXtMidas type (short) integer. <br> <br>
      Value = <code>{@value} = {@link java.lang.Short#MAX_VALUE}</code>
   */
  public static final short IMAX=java.lang.Short.MAX_VALUE;

  /** The minimum value for the NeXtMidas type byte. <br> <br>
      Value = <code>{@value} = {@link java.lang.Byte#MIN_VALUE}</code>
   */
  public static final byte BMIN=java.lang.Byte.MIN_VALUE;

  /** The maximum value for the NeXtMidas type byte. <br> <br>
      Value = <code>{@value} = {@link java.lang.Byte#MAX_VALUE}</code>
   */
  public static final byte BMAX=java.lang.Byte.MAX_VALUE;

  /** The maximum value for the NeXtMidas type packed nibble (4-bit signed integer), 2<sup>3</sup>-1 <br> <br>
      Value = <code>{@value}</code>
    */
  public static final byte NMAX=7;

  /** The minimum value for the NeXtMidas type packed nibble (4-bit signed integer), -2<sup>3</sup> <br> <br>
      Value = <code>{@value}</code>
    */
  public static final byte NMIN=(-NMAX-1); // == -8

}
