package nxm.sys.inc;

/** Defines the internal debug options.

    @author Jeff Schoen
    @version $Id: Debug.java,v 1.12 2011/02/03 15:21:00 ntn Exp $
 */
public interface Debug {
  /** Indicates that debugging messages should be turned on for I/O. */
  public static final int IO=0x001;

  /** Indicates that debugging messages should be turned on for macros. */
  public static final int MACRO=0x002;

  /** Indicates that debugging messages should be turned on for pipes. */
  public static final int PIPES=0x004;

  /** Indicates that debugging messages should be turned on for graphics. */
  public static final int GRAPHICS=0x008;

  /** Indicates that debugging messages should be turned on for results. */
  public static final int RESULTS=0x010;

  /** Indicates that debugging messages should be turned on for primitives. */
  public static final int PRIMITIVE=0x020;

  /** Indicates that debugging messages should be turned on for internals */
  public static final int SHELL=0x040;

  /** Indicates that debugging messages should be turned on for arguments. */
  public static final int ARGS=0x080;

  /** Indicates that stack traces should be included with error messages (where available).
      @see nxm.sys.lib.Midas#printStackTrace(CharSequence, Throwable)
   */
  public static final int TRACE=0x100;

  /** Indicates that debugging messages should be turned on for deprecated functionality.
      @see nxm.sys.lib.Midas#deprecate(CharSequence)
   */
  public static final int DEPRECATE=0x200;

  /** Indicates that when debugging messages are on for traces {@link #TRACE}, the trace should
      include source code where available. This has no effect is {@link #TRACE} is not enabled.
      @since NeXtMidas 2.7.1
      @see nxm.sys.lib.Midas#printStackTrace(CharSequence,Throwable)
   */
  public static final int SUPERTRACE=0x400;

  /** The comma-separated list of debugging options ({@value}). */
  public static final String debugList = "IO,Macro,Pipes,Graphics,Results,"
                                       + "Primitive,Shell,Args,Trace,"
                                       + "Deprecate,SuperTrace";
}
