package nxm.sys.inc;

/**
  Interface Definitions for Midas I/O Devices.

  This interface defines all of the methods that a user I/O terminal
  must implement to interact with the shell.

  @author Jeff Schoen
  @version $Id: IODevice.java,v 1.6 2007/08/21 15:07:13 ntn Exp $
*/
public interface IODevice {
  /** Print a non-terminated line to the terminal.

      @param text The text to print to the terminal.
   */
  public void print(String text);

  /** Print a terminated line to the terminal.

      @param text The text to print to the terminal.
   */
  public void println(String text);

  /** Get a line of input from the keyboard.

      @param prompt The text to prompt the user with.
      @param seed   The default value.
      @param flags  The flags indicating how the user is to be prompted.

      @todo Explain what flags can be used here.
   */
  public String getInput(String prompt, String seed, int flags);

  /** Check for user interrupt from the terminal (Control-C).

      @return The user interrupt.

      @todo Explain what return values are appropriate.
   */
  public int getInterrupt();

  /** Get the number of rows displayed in the terminal.

      @return the number of rows displayed in the terminal.
   */
  public int getRows();

  /** Get the number of columns displayed in the terminal.

      @return The number of columns displayed in the terminal.
   */
  public int getColumns();
}
