package nxm.sys.inc;

/** This is an extension of the {@link Keyable} type that indicates that in-line type casting is
    available.

    @author  J. Czechowski IV
    @since   NeXtMidas 2.1.0
    @version $Id: KeyableByType.java,v 1.3 2007/08/21 15:07:13 ntn Exp $
 */
public interface KeyableByType extends Keyable {
  //////////////////////////////////////////////////////////////////////////////////////////////////
  // NUMERICAL ACCESSORS
  //////////////////////////////////////////////////////////////////////////////////////////////////
  /** Gets the command line parameter at with the given name as an 8-bit Byte. <br>
      <br>
      This is a convenience method equal to <code>getB(name,(byte)0)</code>.

      @param name The parameter name.

      @return The command line parameter with the given parameter name as a Midas byte.

      @see #getB(String,byte)
   */
  public byte getB(String name);

  /** Gets the command line parameter at with the given name as an 8-bit Byte.

      @param name     The parameter name.
      @param defValue The default value for the parameter.

      @return The command line parameter with the given parameter name as a Midas byte or
              <code><i>defValue</i></code> if the parameter does not exist.
   */
  public byte getB(String name, byte defValue);

  /** Gets the command line parameter at with the given name as a 16-bit Midas Integer (Java short).
      <br>
      <br>
      This is a convenience method equal to <code>getI(name,(short)0)</code>.

      @param name The parameter name.

      @return The command line parameter with the given parameter name as a Midas integer.

      @see #getI(String,short)
   */
  public short getI(String name);

  /** Gets the command line parameter at with the given name as a 16-bit Midas Integer (Java short).

      @param name     The parameter name.
      @param defValue The default value for the parameter.

      @return The command line parameter with the given parameter name as a Midas (short)
              integer or <code><i>defValue</i></code> if the parameter does not exist.
   */
  public short getI(String name, short defValue);


  /** Gets the command line parameter at with the given name as a 32-bit Midas Long (Java int). <br>
      <br>
      This is a convenience method equal to <code>getL(name,(int)0)</code>.

      @param name The parameter name.

      @return The command line parameter with the given parameter name as a Midas long.

      @see #getL(String,int)
   */
  public int getL(String name);

   /** Gets the command line parameter at with the given name as a 32-bit Midas Long (Java int).

      @param name     The parameter name.
      @param defValue The default value for the parameter.

      @return The command line parameter with the given parameter name as a Midas long or
              <code><i>defValue</i></code> if the parameter does not exist.
   */
  public int getL(String name, int defValue);

  /** Gets the command line parameter at with the given name as 64-bit Midas eXtra long (Java long).
      <br>
      <br>
      This is a convenience method equal to <code>getX(name,0L)</code>.

      @param name The parameter name.

      @return The command line parameter with the given parameter name as a Midas eXtra long.

      @see #getX(String,long)
   */
  public long getX(String name);

  /** Gets the command line parameter at with the given name as 64-bit Midas eXtra long (Java long).

      @param name     The parameter name.
      @param defValue The default value for the parameter.

      @return The command line parameter with the given parameter name as a Midas eXtra long
              or <code><i>defValue</i></code> if the parameter does not exist.
   */
  public long getX(String name, long defValue);

  /** Gets the command line parameter at with the given name as a 32-bit Midas Float. <br>
      <br>
      This is a convenience method equal to <code>getF(name,0.0f)</code>.

      @param name The parameter name.

      @return The command line parameter with the given parameter name as a Midas float.

      @see #getF(String,float)
   */
  public float getF(String name);

  /** Gets the command line parameter at with the given name as a 32-bit Midas Float. <br>
      <br>
      This is a convenience method equal to <code>(float)getF(name,defValue)</code>.

      @param name     The parameter name.
      @param defValue The default value for the parameter.

      @return The command line parameter with the given parameter name as a Midas float
              or <code><i>defValue</i></code> if the parameter does not exist.
   */
  public float getF(String name, float defValue);


  /** Gets the command line parameter at with the given name as a 64-bit Midas Double. <br>
      <br>
      This is a convenience method equal to <code>getD(name,0.0d)</code>.

      @param name The parameter name.

      @return The command line parameter with the given parameter name as a Midas double.

      @see #getD(String,double)
   */
  public double getD(String name);

  /** Gets the command line parameter at with the given name as a 64-bit Midas Double.

      @param name     The parameter name.
      @param defValue The default value for the parameter.

      @return The command line parameter with the given parameter name as a Midas double or
              <code><i>defValue</i></code> if the parameter does not exist.
   */
  public double getD(String name, double defValue);

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // STRING ACCESSORS
  //////////////////////////////////////////////////////////////////////////////////////////////////
  /** Gets the command line parameter at with the given name as an uppercase string. <br>
      <br>
      This is a convenience method equal to <code>getS(name,"")</code>.

      @param name The parameter name.

      @return The command line parameter with the given parameter name as a uppercase string.

      @see #getS(String,String)
   */
  public String getS(String name);

  /** Gets the command line parameter at with the given name as an uppercase string.

      @param name     The parameter name.
      @param defValue The default value for the parameter.

      @return The command line parameter with the given parameter name as an uppercase string
              or <code>defValue</code> if the parameter does not exist.
   */
  public String getS(String name, String defValue);


  //////////////////////////////////////////////////////////////////////////////////////////////////
  // STATE ACCESSORS
  //////////////////////////////////////////////////////////////////////////////////////////////////
  /** Gets the boolean state of a specified result/parameter. <br>
      <br>
      This is a convenience method equal to <code>getState(name,false)</code>.

      @param name The parameter name.

      @return The command line parameter with the given parameter name as a boolean state.

      @see #getState(String,boolean)
   */
  public boolean getState(String name);

  /** Gets the boolean state of a specified result/parameter. For the purposes of this method the
      following define the states:
      <ul>
        <li>True States  = Anything that is true according to <code>StringUtil.isTrue(..)</code>
                           or a numerical value, <code>n</code> where <code>n>0</code>.</li>
        <li>False States = Anything that is false according to <code>StringUtil.isFalse(..)</code>
                           or a numerical value, <code>n</code> where <code>n<=0</code>.</li>
      </ul>
      Note that the argument is parsed twice for one of these conditions. Once without results
      substitution and then once with results substitution. This eliminates problems where a
      macro has one of these tokens as a results parameter.

      @param name     The parameter name.
      @param defValue The default value for the parameter.

      @return <code>true</code>     if the command line parameter matches a true state,
              <code>false</code>    if the command line parameter matches a false state, or
              <code>defValue</code> if the parameter (specified by <code>name</code>)
                                    does not exist.

      @see nxm.sys.lib.StringUtil#isTrue(String)
      @see nxm.sys.lib.StringUtil#isFalse(String)
  */
  public boolean getState(String name, boolean defValue);

  /** Gets the inverse of the boolean state of a specified result/parameter.
      <br>
      This is a convenience method equal to <code>getNoState(name,false)</code>.

      @param name The parameter name.

      @return The command line parameter with the given parameter name as a boolean state,
              inverted.

      @see #getState(String,boolean)
   */
  public boolean getNoState(String name);


  /** Gets the inverse of the boolean state of a specified result/parameter.
      <br>
      This is a convenience method equal to <code>!getState(name,!defValue)</code>.

      @param name     The parameter name.
      @param defValue The default value for the parameter.

      @return The command line parameter with the given parameter name as a boolean state,
              inverted.

      @see #getState(String,boolean)
   */
  public boolean getNoState(String name, boolean defValue);

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // OBJECT ACCESSORS
  //////////////////////////////////////////////////////////////////////////////////////////////////
  /** Gets the command line parameter at with the given name as an Object. <br>
      <br>
      This is a convenience method equal to <code>getO(name,null)</code>.

      @param name The parameter name.

      @return The command line parameter at the given index as a Midas URL or <code>null</code> if
              the parameter does not exist.

      @see #getO(String,Object)
   */
  public Object getO(String name);

  /** Gets the command line parameter at with the given name as a Java object. <br>
      <br>
      Implementations may feel free to translate the input value to an appropriate object (of their
      choosing), or may simply return the string version of the object.

      @param name     The parameter name.
      @param defValue The default value for the parameter.

      @return The command line parameter with the given parameter name as an Object of
              <code>defValue</code> if the parameter does not exist.
   */
  public Object getO(String name, Object defValue);
}
