package nxm.sys.inc;

import nxm.sys.libg.MWindow;
import nxm.sys.libg.MBox;

import java.awt.*;
import java.awt.event.*;

/** Defines a class that will act as a top-level window for NeXtMidas. While any
    class that implements this interface may act as a top-level window it does
    not need to be an actual top-level window (as is the case with the
    {@link nxm.sys.libg.MJPanel} class). <br>
    <br>
    This is an interface to permit {@link nxm.sys.libg.MWindow} to work with frames that may
    be implemented using either AWT or swing components.

    @see nxm.sys.libg.MWindow
    @author  J. Czechowski IV
    @version $Id: MidasFrame.java,v 1.3 2008/01/29 20:30:45 jph Exp $
    @since   NeXtMidas 1.9.2
 */
public interface MidasFrame extends WindowListener, ComponentListener {
  /** Moves and resizes this component. The new location of the top-left
      corner is specified by <code>pos.x</code> and <code>pos.y</code>, and the
      new size is specified by <code>pos.w</code> and <code>pos.h</code>.
   
      @param pos contains the coordinates and size of the MJFrame
   */
  public void setBounds(MBox pos);

  /** Resize the top-level panel in this MidasFrame according to the current
      values for insets, width, and height.
   */
  public void resizePanel();
  
  /** Returns the dimensions of this MidasFrame, minus any insets.
   
      @return the width and height of this MidasFrame
   */
  public Dimension getSize();

  /** Set the frame to FRONT, BACK or ICONIFIED.  Front also deiconifies
     
      @param mode The display mode ({@link MWindow#FRONT}, {@link MWindow#BACK},
                  or {@link MWindow#ICON}).
   */
  public void display(int mode);

  /** Gets the MWindow associated with this MidasFrame.
      @return the MWindow
      @since NeXtMidas 2.5.4
   */
  public MWindow getMWindow();

  /** Sets the MWindow associated with this MidasFrame.
      @param mw the MWindow
   */
  public void setMWindow(MWindow mw);

  /** Adds a component to this one. */
  public Component add(Component comp);

}
