package nxm.sys.inc;

import java.util.EventObject;

/** A "PropertyViewEvent" event is generated when a Java Bean property is viewed. You can register
    a {@link PropertyViewedListener} so as to be notified when the property is viewed. <br>
    <br>
    This class is intentionally similar to {@link java.beans.PropertyChangeListener}.

    @author  J. Czechowski IV
    @version $Id: PropertyViewedListener.java,v 1.4 2010/08/02 21:42:28 ntn Exp $
    @since   NeXtMidas 2.1.0
 */
public interface PropertyViewedListener {
  /** This method gets called just before a property is viewed.

      @param evt The event object describing which property is viewed.
   */
  public void propertyView (PropertyViewEvent evt);


  //////////////////////////////////////////////////////////////////////////////////////////////////
  // DEFINE A VIEW EVENT
  //////////////////////////////////////////////////////////////////////////////////////////////////

  /** Event sent when a property is viewed. <br>
      <br>
      If a {@link PropertyViewedListener} wishes to verify the value of the property, is should
      use {@link #getOldValue()}, is should not attempt to get the property from the source as
      dosing so will cause an infinite loop. <br>
      <br>
      This class is intentionally similar to {@link java.beans.PropertyChangeEvent}.
   */
  public static class PropertyViewEvent extends EventObject {
    /** NOTE: this should be changed if the structure/fields of this class changes. */
    private static final long serialVersionUID = 2010080300290L; // for Serializable interface
    /** The property name.  */
    private final String name;
    /** The property value. */
    private final Object value;

    /** Creates a new instance.

        @param source       The source of the event.
        @param propertyName The property name.
     */
    public PropertyViewEvent (Object source, String propertyName, Object oldValue) {
      super(source);
      name  = propertyName;
      value = oldValue;
    }

    /** Gets the name of the property viewed.

        @return The name of the property viewed.
     */
    public String getPropertyName () {
      return name;
    }

    /** Gets the current value of the property.

        @return The current value of the property.
     */
    public Object getOldValue () {
      return value;
    }
  }
}
