package nxm.sys.intr;

import nxm.sys.inc.*;
import nxm.sys.lib.*;
import nxm.sys.libg.Theme;

import java.awt.GraphicsConfiguration;
import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.awt.Rectangle;
import java.net.*;
import java.util.*;

/** Handles the manipulation and display of various environment variables.

    @author Jeff Schoen
    @author Jay Hallam
    @version $Id: Environment.java,v 1.54 2012/06/08 22:20:13 jgs Exp $
 */
public class Environment extends Intrinsic implements Keyable {

  public int process () {
    String  func  = MA.getU("FUNC");
    String  key   = MA.getU("KEY");
    String  label = MA.getU("VALUE");
    Object  value = null;
    boolean all   = MA.getState("/ALL");

    if (func.startsWith("SH")) {       // SHOW function
           if (key.length()==0)        showEnvironment(all);
      else if (key.startsWith("THR"))  listAllThreads();
      else if (key.startsWith("MEM"))  listAllMemory();
      else if (key.startsWith("PROP")) listAllProperties();
      else if (key.equals("HANDLERS")) nxm.sys.intr.Results.listMultipleResult(M,M.registry.handlers,"*",0,true);
      else if (key.equals("KEYGETS") || key.equals("KEYSETS")) {
        M.deprecate("ENVIRONMENT: Since NeXtMidas 2.7.0 "+key+" is deprecated (always blank).");
        nxm.sys.intr.Results.listMultipleResult(M, new nxm.sys.lib.Table(),"*",0,true);
      }
      else {
        nxm.sys.intr.Results.listSingleResult(M,M.results,"ENV."+key,0,all);
      }
    }
    else if (func.equals("SET")) {
      value = MA.getO("VALUE");
      setKey(key,value);
      if (M.macro!=null && M.macro.isTxt() && key.equals("BREAK")) {
        if (M.pipeMode!=Midas.POFF) M.pipeMode=Midas.PSTOP;
      }
    }
    else if (func.equals("GET")) {
      value = getKey(key);
      if (label == null || label.length() == 0) {
        M.type(key+" = "+value);
      } else {
        MR.put(label,value);
      }
    }
    else if (func.startsWith("NSLOOKUP")) {
      try {
        InetAddress addr = InetAddress.getByName(key);
        String hostaddr = addr.getHostAddress();
        if (label == null || label.length() == 0) {
          M.info("Inet = "+hostaddr);
        } else {
          MR.put(label,hostaddr);
        }
      } catch (Exception e) {
        throw new MidasException("NSLOOKUP exception: "+e, e);
      }
    }
    else if (func.equals("GC")) {
      if (verbose) { M.terminal.write("before GC: "); listAllMemory(); }
      System.runFinalization();
      System.gc();
      System.runFinalization();
      System.gc();
      if (verbose) { M.terminal.write("after  GC: "); listAllMemory(); }
    }
    else {
      M.error("Unknown environment function: "+func);
    }
    return FINISH;
  }

  private static String[] keys = { "DEFAULTTABLEFLAGS",  "DEFAULTTABLEMODE", "NUMNONVIRTSCREENS",
                                   "NUMSCREENS", "PRIV", "SERVER", "USENATIVE", "COMPLETIONAUXEXCLUDE" };
  public String[] getKeys () { return keys; }

  public Object setKey (String name, Object value) {
    if (name.equals("DEFAULTTABLEFLAGS")) {
      nxm.sys.lib.Table.setDefaultFlags(Convert.o2s(value));
    }
    else if (name.equals("DEFAULTTABLEMODE")) {
      nxm.sys.lib.Table.setDefaultMode(Convert.o2s(value));
    }
    else if (name.equals("PRIV")) {
      if (value.toString().equals("ALL")) Shell.setPriv(Privileges.PRIV_ALL);
    }
    else if (name.equals("COMPLETIONAUXEXCLUDE")) {
      Shell.getCurrent().setCompletionExcludedAuxes(value.toString());
    }
    else {
      if (name.equals("GRAPHICS" )) { Theme.setOptions(value.toString()); value=Theme.getOptions(); }
      if (name.equals("USENATIVE")) { Native.setUse(value.toString()); }
      if (name.equals("UIMANAGER")) { value = mergeUIManagerData((nxm.sys.lib.Table)value); }
      if (name.equals("IOOPTIONS" )){ M.io.setOptions(value.toString()); value=M.io.getOptions(); }
      if (name.equals("PAUSE" )){ M.pause = Convert.o2d(value); }
      MR.put("ENV."+name,value,nxm.sys.lib.Table.FORCE);
    }
    return getKey(name);
  }

  public Object getKey (String name) {
    if (name.equals("DEFAULTTABLEFLAGS"))     return nxm.sys.lib.Table.getDefaultFlags();
    if (name.equals("DEFAULTTABLEMODE"))      return nxm.sys.lib.Table.getDefaultMode();
    if (name.startsWith("NUMNONVIRTSCR"))     return getNumScreens(false); // get # non-virtual screens
    if (name.startsWith("NUMSCR"))            return getNumScreens(true);  // get # screens (any type)
    if (name.equals("PRIV"))                  return Privileges.getPrivMaskString();
    if (name.equals("SERVER"))                return getServer();
    if (name.equals("USENATIVE"))             return Native.getUse();
    if (name.equals("COMPLETIONAUXEXCLUDE"))  return Shell.getCurrent().getExcludedAuxes();
    return MR.getS("ENV."+name);
  }

  /** Get the server. Precedence is to look for the /REMOTE switch, then ENV.SERVER then
      the Midas Context.
      @deprecated Since NeXtMidas 2.7.1: This should be private.
   */
  @Deprecated public String getServer () {
    String value = MA.getQ("/REMOTE",null);          // try switch
    if (value==null) value = MR.getS("ENV.SERVER");  // try environment table
    if (value==null) value = M.getServer();          // try Midas Context
    return value;
  }

  /** Get the number of screens (including virtuals) in the GraphicsEnvironment. Returns 0 if headless.
      @see #getNumScreens(boolean)
   */
  public static int getNumScreens () {
    return getNumScreens(true);
  }

  /** Get the number of screens in the GraphicsEnvironment. Returns 0 if headless.
      @param virtuals true if you want to include virtual screens (aka those with cinerama on)
      @since NeXtMidas 2.8.0
   */
  public static int getNumScreens (boolean virtuals) {
    int numScreens = 0;
    if (!Theme.isHeadless()) {
      try {
        GraphicsDevice[] screens = GraphicsEnvironment.getLocalGraphicsEnvironment().getScreenDevices();
        if (virtuals) {
          numScreens = screens.length; // count all screens (including virtuals)
        }
        else {
          Rectangle virtualBounds;
          for (GraphicsDevice gd : screens) {
            GraphicsConfiguration[] gconfs = gd.getConfigurations();
            int maxx=0, maxy=0;
            for (GraphicsConfiguration gconf : gconfs) {
              virtualBounds = gconf.getBounds();
              if (virtualBounds.x != 0) maxx = virtualBounds.x;
              if (virtualBounds.y != 0) maxy = virtualBounds.y;
            }
            if (maxx == 0 && maxy == 0) numScreens++; // only count non-virtual screens
          }
        }
      } catch (Exception e) {
        Shell.warning("Unable to determine the number of screens"); // no stack trace required
      }
    }
    return numScreens;
  }

  private void showEnvironment (boolean all) {
    nxm.sys.intr.Results.listSingleResult (M,M.results,"ENV",0,all);

    // List the "special" keys/results
    M.type("------- Other Environment Settings ------");
    for (int i = 0; i < keys.length; i++) {
      String key = keys[i];
      Object value = getKey(key);
      if (value==null) value = "";
      M.type(StringUtil.padRight(key,20)+" = "+value);
    }
  }

  public void listAllThreads () {
    ThreadGroup root, parent;
    root = Thread.currentThread().getThreadGroup();
    parent = root.getParent();
    while (parent != null) {
      root = parent;
      parent = parent.getParent();
    }
    listThreadGroup (root,"");
  }

  public void listThreadGroup (ThreadGroup g, String indent) {
    if (g == null) return;
    int num_threads = g.activeCount();
    int num_groups = g.activeGroupCount();
    Thread[] threads = new Thread[num_threads];
    ThreadGroup[] groups = new ThreadGroup[num_groups];
    g.enumerate(threads,false);
    g.enumerate(groups,false);
    M.terminal.writeln( indent + "ThreadGroup: " + g.getName() +
                    " MaxPriority: " + g.getMaxPriority() +
                    (g.isDaemon()? "  Daemon": "") );
    for (int i=0; i<num_threads; i++) {
      if (threads[i] == null) continue;
      M.terminal.writeln( "  " + indent + "Thread: " + threads[i].getName() +
                             " Priority: " + threads[i].getPriority() +
                         (threads[i].isDaemon()? "  Daemon": "") );
    }
    for (int i=0; i<num_groups; i++)
      listThreadGroup(groups[i],indent+"  ");
  }

  public void listAllMemory () {
    Runtime rt = Runtime.getRuntime();
    long tm = rt.totalMemory();
    long fm = rt.freeMemory();
    M.terminal.writeln ("Memory: Total="+tm+" Free="+fm+" bytes");
  }

  public void listAllProperties () {
    Properties p = System.getProperties();
    for (Object elem : p.entrySet()) {
      M.terminal.writeln(elem.toString());
    }
  }

  /** Merges the existing UIManager data with the new input(s), so that the old
      settings are not lost. Also calls the update on the UIManager.
      @param tab The new inputs
   */
  private Object mergeUIManagerData (nxm.sys.lib.Table tab) {
    nxm.sys.lib.Table uim = MR.getTable("ENV.UIMANAGER");
    if (uim==null) uim = new nxm.sys.lib.Table();

    if (tab != null) {
      tab = nxm.sys.lib.Table.merge(uim, tab, true);
      tab = Theme.updateUIManager(tab);
      return tab;  // returns so that the ENV table gets updated correctly
    } else {
      return uim;
    }
  }

}
