package nxm.sys.intr;

import nxm.sys.lib.Intrinsic;
import nxm.sys.lib.Shell;
import nxm.sys.lib.Stack;

/**
  Shows a Shell command history

  @author Jeff Schoen
  @version $Id: History.java,v 1.12 2013/01/17 18:33:21 lsn Exp $
*/
public class History extends Intrinsic {
  /** maximum number of commands to display */
  private int maxCommands = 64;

  /** instance of shell */
  private Shell shell;

  /** instance of history */
  private Stack history;

  /** Perform a case-sensitve comparison? */
  private boolean caseSensitive = false;

  /** Main process routine print out a history list of the available commands */
  public int process () {
    caseSensitive = MA.getState("/CS", caseSensitive);
    String cmd = MA.getCS("CMD");
    int num = MA.getI("NUM");

    shell = (Shell) M.registry.get("SHELL");
    history = shell.getHistory();

    matchCommands(cmd, num);

    rewind();

    return FINISH;
  }

  /** Print all instances of commands that match the parameters */
  private void matchCommands (String cmdAsEntered, int matchNum) {
    Stack answer = new Stack(maxCommands);

    Object peekobj = history.peek ( -1 );
    int cmdnum = 1;
    int matches = 0;
    String cmd = cmdAsEntered;
    if (!caseSensitive) cmd = cmd.toUpperCase();
    // cycle through the command history until it ends
    while ( ( peekobj != null ) && ( cmdnum <= maxCommands ) ) {

      String strAsEntered = (String) peekobj;
      String str = strAsEntered;
      if (!caseSensitive) str=str.toUpperCase();
      // check to see if we match the string (will match if it's empty)
      if ( str.startsWith (cmd) ) {
        matches++;

        // check to see if we have a number of matches to look for
        if ( (matchNum == 0) || (matches <= matchNum) ) {
          answer.push(" "+strAsEntered);
        }
      }

      peekobj = history.peek ( -1 );
      cmdnum++;
    }

    Object obj = answer.pop();
    while (obj != null) {
      M.terminal.writeln(obj.toString());
      obj = answer.pop();
    }
  }

  /** Rewinds the stack to it's initial position so up and down arrows work */
  private void rewind () {
    Object peekobj = history.peek ( 1 );

    // cycle through the command history until it ends
    while ( peekobj != null ) {
      peekobj = history.peek ( 1 );
    }
  }
}

