package nxm.sys.intr;

import nxm.sys.lib.*;

/**
  Handles setting and restoring the HOME path for user macros.

  @author Jeff Schoen
  @version $Id: Homepath.java,v 1.22 2012/02/02 18:53:20 jwb Exp $
*/
public class Homepath extends Intrinsic {

  public int process () {

    // Get parameters
    String path = MA.getS("PATH");
    String opt  = MA.getS("OPT");

    // since 3.1.0/3.1.1, the 4 lines below added to support HOMEPATH 'sticking' to CWD (Bug 2481)
    boolean trackCWD = MA.getState("/CWD", false);
    if (trackCWD) {
      MR.put("AUX._TRACKCWD_", trackCWD, nxm.sys.lib.Table.FORCE, this);  // special entry in AUX table
      MR.put("AUX.HOMEPATH", MR.getS("AUX.CWD"));
    } else {
      MR.remove("AUX._TRACKCWD_");
    }

    if (path.equals("RESTORE") || MA.getState("/RESTORE")) {
      String oldpath = MR.getS("OLDHOMEPATH");
      if (oldpath!=null) {
        MR.put("AUX.HOMEPATH", oldpath);
        path = oldpath; // save to path so that we can validate that it exists below
      }
      else M.warning("Result OLDHOMEPATH not found");
      MR.remove("OLDHOMEPATH");
    }
    else if (path.length() > 0) {
      if (MA.getState("/STORE", true)) {
        String oldpath = MR.getS("AUX.HOMEPATH");
        MR.put("OLDHOMEPATH",oldpath);
      }
      if (opt.length() > 0) {
        path = M.io.name(path,"","",opt);
      } else if (path.equals("TEST")) {
        path = M.io.name("test","","","SYS");
      } else {
        FileName fn = MA.getFileName ("PATH");
        path = fn.getFullName();
        path = FileName.terminatePath (path);
      }
      MR.put("AUX.HOMEPATH", path);
    }
    else {
      path = MR.getS("AUX.HOMEPATH");
      verbose = true; // to force printing current HomePath
    }

    if (verbose) {
      M.terminal.writeln("HomePath = "+path);
    }

    if (!MA.getState("/NOWARN",false) && !M.io.exists(path)) {
      M.warning("Specified HomePath does not exist: "+path);
    }
    return FINISH;
  }

}
