package nxm.sys.intr;

import java.util.*;
import nxm.sys.inc.*;
import nxm.sys.lib.*;

/**
  Utility for editing strings from the macro language (or static).

  @author Jeff Schoen / Jay Hallam
  @version $Id: Sedit.java,v 1.56 2013/01/03 17:42:37 ntn Exp $

*/
public class Sedit extends Intrinsic {

  /** input string */
  private String istr;
  /** output string */
  private String ostr;
  private Object obj=null;

  /** How many deprecation warnings have been issued this session for the 3.1.2 SEDIT changes */
  private static int seditWarnCount = 0; // remove when SEDIT changes go as default

  /** How many deprecation warnings will be issues per session for the 3.1.2 SEDIT changes */
  private static int numSeditWarns = 3; // remove when SEDIT changes go as default

  public int process () {
    numSeditWarns = MA.getL("/NUMWARNS",numSeditWarns);  // NOT DOCUMENTED: change how many 3.1.2 deprecation warnings per session

    istr = MA.getCS("ISTR");
    ostr = MA.getU("OSTR");
    boolean dbg = MA.getState("/DEBUG");
    int tblflags = (MA.getState("/FORCE")? nxm.sys.lib.Table.FORCE : 0); // since 2.9.3

    obj = processCommandLine (istr, M, MA, 3, dbg);

    if (ostr.length()>0) {
      MR.put (ostr, obj, tblflags);
    } else {
      M.terminal.writeln ("Output string ='"+obj+"'");
    }
    return FINISH;
  }

  /** Process the command line arguments to this command.
      This is its own method to be called by inline SEDIT
  */
  public static Object processCommandLine (String inputString, Midas M, Args MA,
                                          int nextArgIndex, boolean dbg) {

    // Initialize output
    Object obj = null;
    String currString = inputString;
    int inputStringArgNum = 1;

    // Get parameters
    int argCount = nextArgIndex;
    boolean clean  = MA.getState("/CLEAN",true);
    boolean strip  = MA.getState("/STRIP",false);
    boolean legacy = MA.getState("/LEGACY",MA.getState("/SEDITLEGACY",true));  //// TODO: Change default to FALSE in 3.5.x?

    // get base
    int base = MA.getL("/BASE", (MA.getState("/ONEBASE") || MA.getState("/OB"))? 1 : 0 );

    while (argCount <= MA.getNArgs()) {

      String func    = MA.getU(argCount++);

      if (dbg) System.out.println("SEDIT: argc="+argCount+" str='"+currString+
               "' func="+func+" P1='"+MA.getU(argCount)+"' P2='"+MA.getU(argCount+1)+"'");

      // Place functions in alphabetical order
      if (func.startsWith("APPEND")) {
        String p1 = MA.getCS(argCount++);
        obj = currString + p1;
      }
      else if (func.startsWith("BETW") && legacy) {
        int imin = MA.getL(argCount++);
        int imax = MA.getL(argCount++);
        // deprecation warning added for Bug 1374 as we move to make SEDIT behave like X-Midas.
        if ( seditWarnCount++<numSeditWarns && ( (base==1 && imin<=0) || imin<0 || (base==1 && imax<=0) || imax<0) ) {
          M.deprecate("SEDIT: As of NeXtMidas 3.1.2, the current default behaviour of SEDIT BETWEEN when using "+
                       " negative numbers (and zero when one-based) will change in a future release. "+
                       "See the SEDIt explain file for details of new behaviour.");
        }
        ///////  THE CODE BELOW FOR THE SPECIAL CASE AND NEGATIVE NUMBERS DOES NOT WORK AND IS DEPRECATED AS OF 3.1.2  //////
        // Special case -- 0 indicates end of string for one-base
        if (imin >= 0 && imin < currString.length()) imin = imin - base;
        else if (base == 1 && imin == 0) imin = currString.length();
        if (imax >= 0 && imax <= currString.length()) imax = imax - base;
        else if (base == 1 && imax == 0) imax = currString.length();
        obj = getStringBetween (currString, imin, imax);

      }
      else if (func.startsWith("BETW")) {
        // support for negative value indexing added in 3.1.2 and will one day be default (Bug 1374)
        int i1 = MA.getL(argCount++); i1-=base;
        int i2 = MA.getL(argCount++); i2-=base;
        obj = getSubStringBetween (currString, i1, i2);
      }
      else if (func.startsWith("BSEA")) {
        String p1 = MA.getU(argCount++);
        int retVal = bsearch (currString,p1);
        if (retVal >= 0) retVal = retVal + base;
        obj = new Data( retVal );
      }
      else if (func.equals    ("CLEAN")) {
        obj = clean (currString);
      }
      else if (func.startsWith("ELEM")) {
        int lval = MA.getL(argCount++);
        String p2 = MA.getU(argCount++);
        obj = getWordByElement (currString,lval,p2);
      }
      else if (func.equals    ("EXT")) {
        obj = ext (currString);
      }
      else if (func.startsWith("GSUB")) {
        String p1 = MA.getU(argCount++);
        String p2 = MA.getU(argCount++);
        obj = subs (currString,p1,p2,true);
      }
      else if (func.equals    ("HEAD") || func.equals ("PATH")) {
        obj = head (currString);
      }
      else if (func.equals    ("JOIN")) {
        Object tempObj = MA.getO(inputStringArgNum); // Go back and pick as object
        String p1 = MA.getU(argCount++);
        obj = join((String [])tempObj,p1);
      }
      else if (func.equals    ("KEY")) {
        String p1 = MA.getU(argCount++);
        String p2 = MA.getU(argCount++);
        obj = getKeyValue (currString,p1,p2);
      }
      else if (func.startsWith("LEN") && legacy) {
        int index = MA.getL(argCount++,-1);
        if (index >= 0) index = index-base;
        obj = new Data( length (currString, index) );
      }
      else if (func.startsWith("LEN")) {
        int index = MA.getL(argCount++);
        obj = new Data( lengthElement (currString, index) );
      }
      else if (func.startsWith("LOCA")) {
        obj = locase (currString);
      }
      else if (func.startsWith("NEL")) {
        String p1 = MA.getU(argCount++);
        if (p1.equals("")) p1 = ",";
        obj = new Data (nelements(currString,p1));
      }
      else if (func.equals    ("NFORM")) {
        double dval = MA.getD(inputStringArgNum); // Go back and pick as double
        String p1 = MA.getU(argCount++);
        obj = nform (dval,p1);
      }
      else if (func.startsWith("PAD")) {
        // Get the arg as a string and int to handle n or +n syntax
        String newLenStr = MA.getCS(argCount);
        int newLen = MA.getL(argCount++);
        if (newLenStr !=null && newLenStr.length() > 1 && newLenStr.charAt(0) == '+') {
          newLen = currString.length() + Integer.parseInt(newLenStr.substring(1));
        }
        String padStr = MA.getCS(argCount++);
        obj = pad(currString, newLen, func, padStr);
      }
      else if (func.equals    ("PARSE")) {
        int lval = MA.getL(argCount++);
        String p2 = MA.getU(argCount++);
        // since delim is optional, decide if it is a new function instead
        // all delimiters are 1 char, all functions are > 1 char
        if (p2.length()>1) { p2=""; argCount--; }
        obj = parse(currString,lval,p2,clean);
      }
      else if (func.startsWith("PARSEAL")) {
        int lval = MA.getL(argCount++);
        String p2 = MA.getU(argCount++);
        obj = parseAll(currString,lval,p2,clean);
      }
      else if (func.startsWith("PARSEAR")) {
        obj = parseArgs (currString,M);
      }
      else if (func.startsWith("PARSEDI")) {
        String p1 = MA.getU(argCount++);
        int retval = parsedIndex (currString,p1,clean);
        if (retval >= 0) retval += base;
        obj = new Data( retval );
      }
      else if (func.startsWith("PREPEND")) {
        String p1 = MA.getCS(argCount++);
        obj = p1 + currString;
      }
      else if (func.equals    ("RANGE") && legacy) {
        int i1 = MA.getL(argCount++);
        int i2 = MA.getL(argCount);
        String p2 = MA.getU(argCount++);
        if (p2.equals("")) { i2 = currString.length(); }
        if ( base == 1 && (i1 == 0 || i2 == 0)) {
          if (i1 == 0) i1 = currString.length();
          if (i2 == 0) i2 = currString.length();
         // M.deprecate("Use of 0 to indicate the end of a string is deprecated, "
         //            +" is deprecated, use string.length() or leave off end index");
        }
        if (i1 >= 0 && i1 <= currString.length()) i1 = i1 - base;
        if (i2 >= 0 && i2 <  currString.length()) i2 = i2 - base;
        if (seditWarnCount++<numSeditWarns && (base==0 && (i1<0 || i2<0))) {
          M.deprecate("SEDIT: As of NeXtMidas 3.1.2, behaviour of negative numbers in RANGE function "+
                      "will change in a future release. See SEDIT explain file for details.");
        }
        obj = range (currString,i1,i2);
      }
      else if (func.equals    ("RANGE")) {
        // support for negative value indexing added in 3.1.2 and will one day be default (Bug 1374)
        int i1 = MA.getL(argCount++); i1-=base;
        int i2 = MA.getL(argCount++); i2-=base;
        obj = getStringInRange(currString, i1, i2);
      }
      else if (func.equals    ("ROOT")) {
        obj = root (currString);
      }
      else if (func.startsWith("SEAR")) {
        String p1 = MA.getU(argCount++);
        int retVal = search(currString,p1);
        if (retVal >= 0) retVal += base;
        obj = new Data( retVal );
      }
      else if (func.startsWith("SEL") && legacy) {
        String p1 = MA.getU(argCount++);
        int retVal = -1;
        if (seditWarnCount++<numSeditWarns && base!=1) M.deprecate("SEDIT: As of NeXtMidas 3.1.2, the current behaviour of SEDIT SELECT is deprecated. "+
                      "SEDIT SELECT will become 1-index based in a future release");
        retVal = select(currString,p1);
        if (retVal >= 0) retVal = retVal+base;
        obj = new Data( retVal );
      }
      else if (func.startsWith("SEL")) { // change to allow 1-based index in 3.1.2 for Bug 1374
        String p1 = MA.getU(argCount++);
        int retVal = selectElement(currString,p1);
        obj = new Data( retVal );
      }
      else if (func.startsWith("SPLIT")) {
        String p1 = MA.getU(argCount++);
        obj = split (currString,p1);
      }
      else if (func.equals    ("STRIM")) {
        obj = strim (currString);
      }
      else if (func.equals    ("STRIP")) {
        obj = strip (currString);
      }
      else if (func.startsWith("SUB")) {
        String p1 = MA.getU(argCount++);
        String p2 = MA.getU(argCount++);
        obj = subs (currString,p1,p2,false);
      }
      else if (func.equals    ("TAIL")) {
        obj = tail (currString);
      }
      else if (func.equals    ("TRIM")) {
        String p1 = MA.getU(argCount++);
        String p2 = MA.getU(argCount++);
        obj = trim (currString,p1,p2);
      }
      else if (func.startsWith("UPCA")) {
        obj = upcase (currString);
      }
      else if (func.equals    ("WORD")) {
        String p1 = MA.getU(argCount++);
        obj = getWordByString (currString, p1);
      }
      else {
        M.warning("Unsupported function: "+func);
      }

      if (strip) obj = strip(obj.toString());

      // Make sure the caller gets an empty string in the case of
      // a null pointer, or the word 'null'
      obj = convertNullsToEmpty (obj);

      currString = obj.toString();
    }

    return obj;
  }

  /** Convert nulls and the word 'null' to an empty string to keep the
      sedit interface consistent.  Ensures "" is returned to the user
      instead of the word 'null'. <br>
      This is the same as <code>StringUtil.convertNullsToEmpty(obj)</code>.
      @param obj The object to convert to an empty string iff it is a null.
      @return The results of calling <code>StringUtil.convertNullsToEmpty(obj)</code>.
      @see StringUtil#convertNullsToEmpty(Object)
   */
  public static Object convertNullsToEmpty (Object obj) {
    return StringUtil.convertNullsToEmpty(obj);
  }

  /** Get the value from a string containing TAG1=VAL1  TAG2=VAL2 pairs.
      A delimiter other than "=" may be used to separate the tag and value. <br>
      This is the same as <code>StringUtil.getKeyValue(inString,tag,delim)</code>.
      @param inString The input "TAG1=VAL TAG2=VAL2 ... " string
      @param tag      The name of the tag
      @param delim    The delimiter, if null or "" the equal sign (=) is used
      @return Everything from the matching tag plus delimeter to the first
              whitespace character or <code>null</code> if tag is not found.
      @see StringUtil#getKeyValue(String,String,String)
   */
  public static String getKeyValue (String inString, String tag, String delim) {
    return StringUtil.getKeyValue(inString,tag,delim);
  }

  /** Parses given element in string by "natural" comma or space delimiter.  <br>
      See {@link #parse(String,int,String,boolean)} for details.
      @param string   The string to parse.
      @param element  The index of the element to get.
      @param clean    Indicates that the string should be cleaned while parsing.
      @return The element within the parsed string at the specified index or
              <code>""</code> if <code>index</code> is invalid.
  */
  public static String parse (String string, int element, boolean clean) {
    return parse(string,element,null,clean);
  }
  /** Parses given element in string by given delimiter.  <br>
      This is the same as <code>StringUtil.getParsedElement(string,element,delim,clean)</code>.
      @param string   The string to parse.
      @param element  The index of the element to get.
      @param delim    The delimiter
      @param clean    Indicates that the string should be cleaned while parsing.
      @return The element within the parsed string at the specified index or
              <code>""</code> if <code>index</code> is invalid.
      @see StringUtil#getParsedElement(String,int,char,boolean)
      @see StringUtil#getParsedElement(String,int,boolean)
  */
  public static String parse (String string, int element, String delim, boolean clean) {
    if (delim == null || delim.length() == 0) {
      return StringUtil.getParsedElement(string,element,clean);
    } else {
      return StringUtil.getParsedElement(string,element,delim.charAt(0),clean);
    }
  }

  /** Parses string into ALL elements by "natural" comma or space delimiter.  <br>
      This is the same as <code>StringUtil.parseString(String,int,boolean)</code>.
      @param string   The string to parse.
      @param elements The maximum number of elements to get when parsing.
      @param clean    Indicates that the string should be cleaned while parsing.
      @return An array of the first <i><code>elements</code></i> elements in the string.
      @see StringUtil#parseString(String,boolean)
      @see StringUtil#parseString(String,int,boolean)
  */
  public static String[] parseAll (String string, int elements, boolean clean) {
    return StringUtil.parseString(string,elements,clean);
  }
  /** Parses string into ALL elements by given delimiter or "natural" comma or space delimiter. <br>
      @param string   The string to parse.
      @param elements The maximum number of elements to get when parsing.
      @param delim    The delimiter character as a String
      @param clean    Indicates that the string should be cleaned while parsing.
      @return An array of the first <i><code>elements</code></i> elements in the string.
      @see StringUtil#parseString(String,boolean)
      @see StringUtil#parseString(String,int,boolean)
  */
  public static String[] parseAll (String string, int elements, String delim, boolean clean) {
    if (delim == null || delim.length() == 0) {
      return parseAll(string,elements,clean);
    } else {
      return StringUtil.parseString(string,elements,delim.charAt(0),clean);
    }
  }

  /** Parses string into an Args object as if parsing a command line. <br>
      This is the same as <code>Args.parseCommand(M,string)</code>.
      @param string The string to parse the arguments from.
      @param M      The midas context to use.
      @return The parsed arguments.
      @see Args#parseCommand(Midas,String)
  */
  public static Args parseArgs (String string, Midas M) {
    return Args.parseCommand(M,string);
  }

  /** Parses string to find index of word (0 to #words-1 ). <br>
      This is the same as <code>StringUtil.getParsedElementIndex(string,item,clean)</code>.
      @param string   The string to parse.
      @param item     The element to find the index of.
      @param clean    Indicates that the string should be cleaned while parsing.
      @return The index of the specified element within the parsed string or
              <code>-1</code> if such does not exist.
      @see StringUtil#getParsedElementIndex(String,String,boolean)
    */
  public static int parsedIndex (String string, String item, boolean clean) {
    return StringUtil.getParsedElementIndex(string,item,clean);
  }

  /** Returns a CLEANed (all caps, and comma delimiter) version of string. <br>
      This is the same as <code>StringUtil.cleanString(string)</code>.
      @param string The string to clean.
      @return A cleaned version of the string.
      @see StringUtil#cleanString(String)
   */
  public static String clean (String string) {
    return StringUtil.cleanString(string);
  }

  /** This will substitute occurrances of a substring (<code>p1</code>) in
      a string (<code>string</code>) with a new string <code>p2</code>. <br>
      If <code>all</code> is <code>true</code>, this is the same as
      <code>StringUtil.replaceAll(string,p1,p2)</code><br>
      If <code>all</code> is <code>false</code>, this is the same as
      <code>StringUtil.replaceFirst(string,p1,p2)</code>
      @param string The to make replacements on.
      @param p1     The plain-text pattern identifying what is to be replaced.
      @param p2     The plain-text replacement for anything that matches the pattern.
      @param all    If this is <code>true</code> ALL occurrances of <code>p1</code> are
                    replaced with <code>p2</code>. If this is <code>false</code> only the
                    FIRST occurrance of <code>p1</code> are replaced with <code>p2</code>.
      @return A new version of <code>string</code> where occurrances of
              <code><i>p1</i></code> have been replaced with <code><i>p2</i></code>.
      @see StringUtil#replaceAll(String,String,String)
      @see StringUtil#replaceFirst(String,String,String)
   */
  public static String subs (String string, String p1, String p2, boolean all) {
    if (all) {
      return StringUtil.replaceAll(string,p1,p2);
    } else {
      return StringUtil.replaceFirst(string,p1,p2);
    }
  }

  /** Calculates the length of string through the Nth naturally delimited entry.
      Zero-based index.
      @param string Input string
      @param index  Entry number (0..n)
      @return The length.
   */
  public static int length (String string, int index) {
    if (index<0) return string.length();
    Parser p = new Parser(string);
    int ls = p.lengthd(0);
    for (int i=1; i<index+1; i++) { ls++;  ls += p.lengthd(ls); }
    return ls;
  }

  /** Calculates the length of string through the Nth naturally delimited entry.
      One-based index.
      @param string Input string
      @param index  Entry number (0..n)
      @return The length.
   */
  public static int lengthElement (String string, int index) {
    if (index<=0) return string.length();
    Parser p = new Parser(string);
    int ls = p.lengthd(0);
    for (int i=1; i<index; i++) { ls++;  ls += p.lengthd(ls); }
    return ls;
  }

  /** Extracts the "inclusive" substring from <code>inString</code> between <code>imin</code> and
      <code>imax</code>.  Indices are zero-based.  Negative numbers represent an index
      relative to the last character in the input string.
      @param inString Input string
      @param imin     Start index (0 to len-1) of substring (INCLUSIVE)
      @param imax     End index (0 to len-1) of substring (INCLUSIVE)
      @return The substring, if in range, or "" if imin>imax or any part of the min or max is beyond the range of the input String
      @since NeXtMidas 3.1.2
   */
  public static String getStringInRange (String inString, int imin, int imax) {
    if (imin<0) { imin = inString.length() + imin; }
    if (imax<0) { imax = inString.length() + imax; }
    imax++; // make it inclusive
    if (imin<0 || imax>inString.length() || imin>imax || imax<0 || imin>inString.length()) return "";
    return inString.substring(imin,imax);

  }

  /** Extracts the "inclusive" substring from <code>string</code> between <code>p1</code> and
      <code>p2</code>.  Indices are zero-based.  Negative numbers represent relative indices.
      @param string Input string
      @param p1     Start index (0 to len-1) of substring (INCLUSIVE)
      @param p2     End index (0 to len-1) of substring (INCLUSIVE)
      @return The substring.
   */
  public static String range (String string, int p1, int p2) {
    if (p2<0) p2 = string.length()+p2;
    else p2 = p2 + 1;
    if (p1<0) p1 = string.length()+p1-1;
    if ((p2<p1) || (p1>string.length())) return null;
    p2 = Math.max(p1,Math.min(string.length(),p2));
    return string.substring(p1,p2);
  }

  /** Returns a string array which is the result of splitting the string using
      the specified delimiter. The return value will be an array of strings.
      If either the value or the delimiter is null, then the method will
      simply return null. <br>
      This is the same as <code>StringUtil.split(val,delim)</code>.
      @param val The string to split
      @param delim The string to split on
      @return A new array of strings which is the result of the split, or
              <code>null</code> if either parameter is <code>null</code>.
      @see StringUtil#split(String,String)
   */
  public static String[] split (String val, String delim) {
    return StringUtil.split(val,delim);
  }

  /** Returns a string which is the result of joining the values using the
      specified delimiter. The return value will be a single string which is
      basically vals[0] + delim + vals[1] + delim + ... + vals[n - 1]. <br>
      This is the same as <code>StringUtils.join(vals,delim)</code>.
      @param vals The list of values that are to be joined.
      @param delim The string that will go between each value.
      @return The result of performing the join on the string, or
              <code>null</code> if <code>list</code> is <code>null</code> and
              <code>""</code> if <code>list</code> has a length of zero.
      @see StringUtil#join(Object[],String)
   */
  public static String join (Object[] vals, String delim) {
    return StringUtil.join(vals, delim);
  }


  /** Returns a string which is the result of joining the values using the
      specified delimiter. The return value will be a single string which is
      basically vals[0] + delim + vals[1] + delim + ... + vals[n - 1]. <br>
      This is the same as <code>StringUtil.join(list.toArray(),delim)</code>.
      @param vals The list of values that are to be joined.
      @param delim The string that will go between each value.
      @return The result of performing the join on the string, or
              <code>null</code> if <code>list</code> is <code>null</code> and
              <code>""</code> if <code>list</code> has a length of zero.
      @see StringUtil#join(Object[],String)
   */
  public static String join (Vector<?> vals, String delim) {
    return StringUtil.join(vals.toArray(),delim);
  }

  /** Removes all the leading and trailing whitespace from a string. <br>
      This is the same as <code>string.trim()</code>.
      @param string The string to trim.
      @return The trimmed string.
      @see String#trim()
   */
  public static String strim (String string) {
    return string.trim();
  }

  /** Trims off string before p1 and after p2. <br>
      This is the same as <code>StringUtil.trim(string,p1,p2)</code>.
      @param string The string to trim.
      @param p1     The beginning of the portion of the string to keep.
      @param p2     The ending of the portion of the string to keep.
      @return The string with the ends trimmed off.
      @see StringUtil#trim(String,String,String)
   */
  public static String trim (String string, String p1, String p2) {
    return StringUtil.trim(string,p1,p2);
  }

  /** Format a number.  Wrapper to  {@link nxm.sys.lib.Format} class
    @param dval Number to be formation
    @param pattern Format string pattern
    @return Formatted number as a string
    @see nxm.sys.lib.Format
  */
  public static String nform (double dval, String pattern) {
    Data D = new Data(dval);
    return D.toString(pattern);
  }

  /** Wrapper for padding LEFT Right or BOTH (center)
    @param string Input string
    @param newLen Width of output string
    @param dir Direction to pad PADLeft or LEFT, PADRight or RIGHT, PADBoth or BOTH
    @param padStr First character of this string is used to pad
    @return Padded string or original string if longer than desired input length
    @see nxm.sys.lib.StringUtil#padLeft
    @see nxm.sys.lib.StringUtil#padRight
    @see nxm.sys.lib.StringUtil#padBoth
  */
  public static String pad (String string, int newLen, String dir, String padStr) {
    char pad = ' ';
    if (padStr != null && padStr.length() != 0) {
      pad = padStr.charAt(0);
    }
    String outString = null;
    if        (dir.startsWith("PADL") || dir.equals("LEFT")) {
      outString = StringUtil.padLeft(string, newLen, pad);

    } else if (dir.startsWith("PADB") || dir.equals("BOTH") || dir.equals("CENTER")) {
     // Pad both sides, add one to left if odd
      outString = StringUtil.padBoth(string, newLen, pad);

    } else {
      outString = StringUtil.padRight(string, newLen, pad);

    }
    return outString;
  }

  /** returns the path of a filename
   @param string Full file name
   @return Head or path of file name
  */
  public static String head (String string) {
    FileName fn = new FileName(string);
    return fn.getPath();
  }

  /** returns the filename and extension of a full file name
   @param string Full file name
   @return File name root and extension
  */
  public static String tail (String string) {
    FileName fn = new FileName(string);
    return fn.getRoot()+"."+fn.getExt();
  }

  /** returns the filename without extension of a full file name
   @param string Full file name
   @return File name without path or extension
  */
  public static String root (String string) {
    FileName fn = new FileName(string);
    return fn.getRoot();
  }

  /** returns the filename extension
   @param string Full file name
   @return File name extension
  */
  public static String ext (String string) {
    FileName fn = new FileName(string);
    return fn.getExt();
  }

  /** Returns upper-case version of same string. <br>
      This is the same as <code>string.toUpperCase()</code>
      @param string The string to convert to upper-case.
      @return The upper-case string.
      @see String#toUpperCase()
   */
  public static String upcase (String string) {
    return string.toUpperCase();
  }

  /** Returns lower-case version of same string. <br>
      This is the same as <code>string.toLowerCase()</code>
      @param string The string to convert to lower-case.
      @return The lower-case string.
      @see String#toLowerCase()
   */
  public static String locase (String string) {
    return string.toLowerCase();
  }

  /** Returns number of elements in a delimited string. <br>
      This is the same as <code>nelements(string,delim.charAt(0))</code>.
      @param string  Input string
      @param delim   Delimiter, single character string
      @return The number of elements.
      @see #nelements(String,char)
  */
  public static int nelements (String string, String delim) {
    char cdelim = delim.charAt(0);
    return nelements(string, cdelim);
  }

  /** Returns number of elements in a comma delimited string. <br>
      This is the same as <code>nelements(string,',')</code>.
      @param string - Input string
      @return The number of elements (tokens) in the string.
      @see #nelements(String,char)
  */
  public static int nelements (String string) {
    return nelements(string, ',');
  }

  /** Returns number of elements in a delimited string. <br>
      Examples:
      <pre>
        nelements(" ", ',')     --&gt; 0
        nelements(",", ',')     --&gt; 2
        nelements("a,b,c", ',') --&gt; 3
      </pre>
    @param string - Input string
    @param cdelim - delimiter
  */
  public static int nelements (String string, char cdelim) {
    int retval = 0;
    if (string == null || string.length() == 0 ||
        string.equalsIgnoreCase("NULL")) {
       // No elements
    } else {
      char[] charArray = new char[string.length()];
      string.getChars(0, string.length(), charArray, 0);
      int numDelims = 0;
      for (int i=0; i <= charArray.length-1; i++) {
        if (charArray[i] == cdelim) numDelims++;
      }
      retval = numDelims+1;
    }
    return retval;
  }

  /** Case sensitive search of string for substring. <br>
      This is the same as <code>string.indexOf(p1)</code>.
      @param string The string.
      @param p1     The substring to search for.
      @return The index of the FIRST occurrance of substring within the string (zero-based)
              or <code>-1</code> if the substring is not found within the string.
      @see #bsearch(String,String)
      @see String#indexOf(String)
   */
  public static int search (String string, String p1) {
    return string.indexOf(p1);
  }

  /** Case sensitive search backwards of string for substring. <br>
      This is the same as <code>string.lastIndexOf(p1)</code>.
      @param string The string.
      @param p1     The substring to search for.
      @return The index of the LAST occurrance of substring within the string (zero-based)
              or <code>-1</code> if the substring is not found within the string.
      @see #search(String,String)
      @see String#lastIndexOf(String)
   */
  public static int bsearch (String string, String p1) {
    return string.lastIndexOf(p1);
  }

  /** Find the nth instance of a substring in an input string.
      @param inString  Input string
      @param subString Sub string to find in input string
      @return The index of the substring or <code>-1</code> if not found.
   */
  public static int selectElement (String inString, String subString) {
    return Parser.find(inString,subString,-1);
  }

  /** Find the nth instance of a substring in an input string.
      @param inString  Input string
      @param subString Sub string to find in input string
      @return The index of the substring or <code>-1</code> if not found.
   */
  public static int select (String inString, String subString) {
    int retVal = Parser.find(inString,subString,-1);
    if (retVal > 0) retVal = retVal-1;
    return retVal;
  }

  /** Get a word from a string delimitted by given character.
      @param inString    Input string
      @param elemNum     Element number (1..n)
      @param delimString The delimiter DEFAULT=comma
      @return The word.
  */
  public static String getWordByElement (String inString, int elemNum, String delimString) {
     String outString = null; // The DEFAULT
     if (delimString == null || delimString.length() == 0) {
       delimString = ",";
     }
     String[] strArray = inString.split(StringUtil.textToRegex(delimString));
     if (elemNum > 0 && (elemNum-1 < strArray.length) ) {
       outString = strArray[elemNum-1];
     }
     return outString;
  }

  /** Get the "inclusive" substring between imin and imax. A negative index means that many
      characters from the end of the String.<br>
      This is the same as <code>StringUtil.between(inString,imin,imax)</code>.
      @param inString  Input string
      @param imin      Index (0-n) of start of string to extract (inclusive)
      @param imax      Index (0-n) of end of string to extract (inclusive)
      @return The appropriate substring.
      @see StringUtil#between(String,int,int)
      @since NeXtMidas 3.1.2
   */
  public static String getSubStringBetween (String inString, int imin, int imax) {
    return StringUtil.between(inString, imin, imax);
  }

  /** Get the substring between imin and imax of string. A negative index may be
      used to indicate a relative index. <br>
      This is the same as <code>StringUtil.substring(inString,imin,imax)</code>.
      @param inString  Input string
      @param imin      Index (0-n) of start of string to extract
      @param imax      Index (0-n) of end of string to extract
      @return The appropriate substring.
      @see StringUtil#substring(String,int,int)
   */
  public static String getStringBetween (String inString, int imin, int imax) {
    return StringUtil.substring(inString,imin,imax);
  }

  /** Get a word from a string containing a substring. <br>
      For example, getWordByString("This is a string","ri") returns "string"
   @param inString  Input string
   @param subString The substring for which to search
   @return Word containing sub string
  */
  public static String getWordByString (String inString, String subString) {
    Parser p = new Parser (inString);
    String outString = null;
    String word;

    for (int i = 1; i <= p.elements(); i++) {
      word = p.get(i);
      if (word.indexOf (subString) >= 0) {
        outString = word;
      }
    }
    return outString;
  }

  /** Strip the quotes from a string. <br>
      This is the same as <code>StringUtil.stripQuotes(string)<code>.
      @param string The string to remove the quotes from.
      @return The string after removing the quotes.
      @see StringUtil#stripQuotes(String)
   */
  public static String strip (String string) {
    return StringUtil.stripQuotes(string);
  }
}
