package nxm.sys.intr;

import nxm.sys.inc.*;
import nxm.sys.lib.*;
import java.text.ParseException;

/**
  Handle time functions for the macro language.

  @author Jeff Schoen
  @version $Id: Timex.java,v 1.15 2011/02/22 21:42:39 ntn Exp $

*/
public class Timex extends Intrinsic {
  public static final double SECS_IN_YEAR = 86400.0;

  public int process () {
    DataFile hcb;

    // get format switches
    int     dp        = MA.getL("/PREC", -1);   // -1 uses default precision
    String  inFmt     = MA.getS("/INFMT",null);
    String  outFmt    = MA.getS("/FMT",  null);
    MFormat inFormat  = (inFmt  == null)? null : MFormat.getDateFormatFor(inFmt);
    MFormat outFormat = (outFmt == null)? null : MFormat.getDateFormatFor(outFmt,dp);

    // default to current CPU time
    Time time = Time.currentTime();
    boolean hasOutput = false;

    // loop through args and perform functions
    for (int n=1; n<=MA.numberOf(); n++) {
      String key   = MA.getKey(n);
      String label = MA.getU(n);

      boolean handled = true;
      if (key.equals("TIME") || key.equals("P"+n)) {
             if (label.equals("NOW")); // already done
        else if (label.equals("DAY"))        { time.floor(SECS_IN_YEAR);  }
        else if (label.equals("SSM"))        { time.modulo(SECS_IN_YEAR); }
        else if (label.equals("TERM"))       { M.terminal.writeln("Time = "+time); hasOutput=true; }
        else if (label.equals("TERMINAL"))   { M.terminal.writeln("Time = "+time); hasOutput=true; }
        else if ((n==1) && (inFormat==null)) { time = MA.getTime(n); }
        else if (n==1)                       { time = parseTime(inFormat, MA.getS(n)); }
        else if (outFormat == null)          { MR.put(label,time); hasOutput=true; }
        else                                 { MR.put(label,outFormat.format(time)); hasOutput=true; }
      }
      else if (key.equals("INFILE")) {
        hcb = MA.getDataFile("INFILE");
        hcb.open(); time.fromTimeCode(hcb); hcb.close();
      }
      else if (key.equals("INACQ")) {
        hcb = MA.getDataFile("INACQ");
        hcb.open(); time.fromAcqDate(hcb); hcb.close();
      }
      else if (key.equals("INQUAD")) {
        hcb = MA.getDataFile("INQUAD");
        hcb.open(); time.fromQuadwords(hcb); hcb.close();
      }
      else if (key.equals("ROUND")) {
        double sec = MA.getD("ROUND");
        time.round(sec);
      }
      else {
        handled = false;
      }
      if (handled) continue;
      hasOutput=true;
      if (n==1) {
         String extra = ("FILE,ACQ,QUAD".indexOf(key)>=0) ? "- Try IN"+key : "";
         M.error("Need an input parameter in 1st slot "+extra);
      }
      else if (key.equals("FILE")) {
        hcb = MA.getDataFile("FILE");
        hcb.open(hcb.INOUT); time.toTimeCode(hcb); hcb.close();
      }
      else if (key.equals("ACQ")) {
        hcb = MA.getDataFile("ACQ");
        hcb.open(hcb.INOUT); time.toAcqDate(hcb); hcb.close();
      }
      else if (key.equals("QUAD")) {
        hcb = MA.getDataFile("QUAD");
        hcb.open(hcb.INOUT); time.toQuadwords(hcb); hcb.close();
      }
      else {
             if (key.equals("SEC")) key = "MJS";
        else if (key.equals("JUL")) key = "JD";
        else if (key.equals("Y"))   key = "YEAR";
        else if (key.equals("FN"))  key = "TOFILENAME";
        Object obj = KeyObject.getKey(time,key);
        if (obj==null) M.error("Illegal time key: "+key+"="+label);
        MR.put(label,obj);
      }
    }

    // default to terminal output
    if (!hasOutput) {
      if (outFormat == null) M.terminal.writeln("Time = "+time.toString(dp));
      else                   M.terminal.writeln("Time = "+outFormat.format(time));
    }

    return FINISH;
  }

  private Time parseTime (MFormat fmt, String str) {
    try {
      Time t = Time.toTime(fmt.parseObject(str));

      if (t == null) {
        throw new MidasException("Unable to parse '"+str+"' using "+fmt+".");
      }
      return t;

    } catch (ParseException e) {
      throw new MidasException("Unable to parse '"+str+"': "+e.getMessage(), e);
    }
  }
}
