package nxm.sys.lib;

import java.applet.*;
import java.net.*;

/** A class to interface with a browser.

    @author Jeff Schoen
    @version $Id: BrowserIF.java,v 1.9 2011/06/24 17:35:33 ntn Exp $
*/
public class BrowserIF extends Object {
  /** The vendor of the browser the applet is running in. */
  public String vendor;

  /** The version of the browser the applet is running in. */
  public String version;

  /** The codebase the applet is using. */
  public String codebase;

  /** The applet context to use. */
  public AppletContext context;

  /** The midas context to use. */
  Midas M;

  /** This opens the Browser Interface setting the browser's parameters.

      @param midas  The midas context to use.
      @param applet The applet to use.
   */
  public void open (Midas midas, Applet applet) {
    M = midas;
    context = applet.getAppletContext();
    codebase = applet.getCodeBase().toString();
    if (codebase.startsWith("file:")) {
      if (codebase.charAt(7)=='|') {
        codebase = codebase.charAt(6)+":"+codebase.substring(8);
      } else {
        codebase = codebase.substring(5);
      }
    }
    try {
      vendor = System.getProperty("browser");
      version = System.getProperty("browser.version");
    }
    catch (SecurityException e) { vendor = "AppletViewer"; }
    catch (Exception e) {
      if (e.getClass().getName().equals("netscape.security.ForbiddenTargetException")) {
        System.out.println("No privilege to access local environment");
      } else {
        System.out.println("Exception "+e+" in BrowserIF");
      }
    }
    if (vendor==null) vendor = "Unknown";
    if (version==null) version = "Unknown";
  }

  /** Opens a page in the default browser window. <br>
      <br>
      This is identical to <code>showPage(name,"NeXtMidas-Documents")</code>.

      @param name   The name of the page to load.

      @see #showPage(String,String)
   */
  public void showPage (String name) {
    showPage (name, "NeXtMidas-Documents");
  }

  /** Opens a page in the named browser window. <br>
      <br>
      Calling this method is analogous to having a user click a link in a
      web page that is defined as
      <code>&lt;a&nbsp;href='<i>name</i>'&nbsp;target='<i>target</i>'&gt;</code>

      @param name   The name of the page to load (this can be a path name or a
                    URL).
      @param target The ID indicating where to load the page (see
                    {@link AppletContext#showDocument(URL,String)} for details).
   */
  public void showPage (String name, String target) {
    String fname = name;
    if (name.startsWith("//"))
      fname = "http:"+name;
    else if (name.startsWith("/"))
      fname = "file:"+name;
    else if (name.charAt(1)==':')
      fname = "file:/"+name.charAt(0)+"|"+name.substring(2);
    try {
      java.net.URL urlnew = new java.net.URL(fname);
      context.showDocument(urlnew, target);
    } catch (java.net.MalformedURLException error) {
      System.out.println("Problem displaying page: "+name);
    }
  }

  /** Indicates if the program is running as an applet.

      @return <code>true</code> if running as an applet, <code>false</code> if
              not running as an applet.
   */
  public boolean isEnabled() {
    return (context!=null);
  }

  /** Indicates if this is an applet running in a Netscape Browser.

      @return <code>true</code> if running as an applet running in Netscape,<br>
              <code>false</code> if not running as an applet in Netscape.

      @see #isEnabled()
      @see #isIE()
      @see #isAppletViewer()
      @see #isViewer()
      @see #isPlugin()
   */
  public boolean isNetscape() {
    if (context==null) return false;
    return vendor.startsWith("Netscape");
  }

  /** Indicates if this is an applet running in Microsoft Internet Explorer
      (IE).

      @return <code>true</code> if running as an applet running in IE,<br>
              <code>false</code> if not running as an applet in IE.

      @see #isEnabled()
      @see #isNetscape()
      @see #isAppletViewer()
      @see #isViewer()
      @see #isPlugin()
   */
  /** is this an Internet Explorer Browser */
  public boolean isIE() {
    if (context==null) return false;
    return vendor.startsWith("Microsoft");
  }

  /** Indicates if this is an applet running in the Java Applet Viewer.

      @return <code>true</code> if running as an applet running in the applet
              viewer,<br>
              <code>false</code> if not running as an applet in the applet
              viewer.

      @see #isEnabled()
      @see #isNetscape()
      @see #isIE()
      @see #isViewer()
      @see #isPlugin()
   */
  /** is this an Applet Viewer */
  public boolean isAppletViewer() {
    if (context==null) return false;
    return vendor.startsWith("AppletViewer");
  }

  /** Indicates if this is an applet that is the default viewer.

      @return <code>true</code> if running as an applet and acting as the
              default viewer,<br>
              <code>false</code> otherwise.

      @see #isEnabled()
      @see #isNetscape()
      @see #isIE()
      @see #isAppletViewer()
      @see #isPlugin()
   */
  public boolean isViewer () {
    if (context==null) return false;
    return "HTML".equals(M.results.getString("ENV.VIEW"));
  }
  /** Indicates if this is an applet running in a Sun plugin.

      @return <code>true</code> if running as an applet running in a plugin,<br>
              <code>false</code> if not running as an applet in a plugin.

      @see #isEnabled()
      @see #isNetscape()
      @see #isIE()
      @see #isAppletViewer()
      @see #isViewer()
   */
  public boolean isPlugin () {
    if (context==null) return false;
    return vendor.startsWith("sun.plugin");
  }
}
