package nxm.sys.lib;

import nxm.sys.inc.*;
import nxm.sys.libg.MWindow;
import nxm.sys.libg.Theme;

import java.awt.Container;

/** 
  Base class for graphical primitives.  This is a container for methods that are 
  specific to graphical primitives.  

  They normally execute in a sub-thread of the current shell.

  @author Jeff Schoen
  @version $Id: GPrimitive.java,v 1.13 2011/12/20 14:39:23 jwb Exp $

*/
public class GPrimitive	extends Command
			implements Chainable, Constants {

  /** Handle to the default graphical window. */
  public MWindow MW;

  public Object getNextLink() { return MW; }
  
  public Object getPrevLink() { return null; }

  public boolean isPipable() { return true; }
  
  static { Theme.initialize(); }  // added in 3.1.0 to make sure Graphical primitives pick up theme (Bug 2529)

  public void checkout() {
    super.checkout();
    
    if(MW != null) { MW.close(); }
    MW = null;
  }
  
  /** Gets the primary container used when displaying this graphical primitive. Subclasses may
      override this as necessary.
      @return The graphics panel or null if none is set.
      @since NeXtMidas 2.5.0
   */
  public Container getGraphicsContainer() {
    return (MW == null)? null: MW.panel;
  }
}
