package nxm.sys.lib;

import nxm.sys.inc.*;
import nxm.sys.libg.Layer;
import nxm.sys.libg.LayerKML;

import java.util.Vector;

/** Defines KML functions useful throughout NeXtMidas and NeXtMidas applications.

    @author  J. Bauer
    @version $Id: KmlFile.java,v
    @since   NeXtMidas 2.7.0
 */
public final class KmlFile extends XmlFile implements PlotFileNH, ListFile {
  public final String propertyList = super.propertyList+
                     ",XStart,XDelta,XUnits,XFrame,YStart,YDelta,YUnits,YFrame";

  private double xstart   = 0;
  private double xdelta   = 1;
  private double ystart   = 0;
  private double ydelta   = 1;
  private double timecode = 0;

  private int xunits      = 0;
  private int xframe      = 1;
  private int yunits      = 0;
  private int yframe      = 1;

  private String format        = "UNK";
  private Vector<String> lines = null;

  public KmlFile () { }

  public KmlFile (String filename) {
    init(null,filename);
  }

  public KmlFile (FileName fn) {
    init (null,fn);
  }

  public KmlFile (Object ref, Object filename) {
    init(ref,filename);
  }

  public String getMimeType () {
    // TBD: kmz files use "application/vnd.google-earth.kmz"
    return "application/vnd.google-earth.kml+xml";
  }

  public void connect(int mode) {}

  public String listHeader() {
    String list =
      "KmlFile    :  "+getURL()+"\n"
    + "Version    :  "+"KML2.2 ???\n" /* ??? */
    + "Size       :  "+getSize()+"\n";
/*    + "Format      :  "+getFormat()+"\n"
    + "X units     :  "+DataFile.getUnitsName(getXUnits())+"\n"
    + "X start     :  "+getXStart()+"\n"
    + "X delta     :  "+getXDelta()+"\n"
    + "X frame     :  "+getXFrame()+"\n"
    + "Y units     :  "+DataFile.getUnitsName(getYUnits())+"\n"
    + "Y start     :  "+getYStart()+"\n"
    + "Y delta     :  "+getYDelta()+"\n"
    + "Y frame     :  "+getYFrame()+"\n"
    + "Comment     :  "+getComment()+"\n";
*/
    return list;
  }

  public String getFormat () { return format; }
  public double getStart ()  { return ystart; }
  public double getDelta ()  { return ydelta; }
  public int    getUnits ()  { return yunits; }
  public double getXStart () { return xstart; }
  public double getXDelta () { return xdelta; }
  public int    getXUnits () { return xunits; }
  public int    getXFrame () { return xframe; }
  public double getYStart () { return ystart; }
  public double getYDelta () { return ydelta; }
  public int    getYUnits () { return yunits; }
  public int    getYFrame () { return yframe; }
  public double getTimeAt (double offset) { return timecode; }

  public void setFormat (String value) { format = value; }
  public void setXStart (double value) { xstart = value; }
  public void setXDelta (double value) { xdelta = value; }
  public void setXUnits (int    value) { xunits = value; }
  public void setXFrame (int    value) { xframe = value; }
  public void setYStart (double value) { ystart = value; }
  public void setYDelta (double value) { ydelta = value; }
  public void setYUnits (int    value) { yunits = value; }
  public void setYFrame (int    value) { yframe = value; }
  public void setTimeCode (double value) { timecode = value; }
  public Layer getLayer() { return new LayerKML();  }

  public int getMode() { return 1; }

  public int getType() { return 0; }

  public Data getDataBuffer(int size) { return null; } // TODO

  public boolean open () {
/*    // NTN: begin test case for Bug 1812
    Keywords keywords = new Keywords();
    M.info("keywords.add   (TEST)"); keywords.add("TEST", "Test Value");
    M.info("keywords.get   (TEST): "+keywords.get("TEST"));
    M.info("keywords.delete(TEST): "+keywords.delete("TEST"));
    M.info("keywords.toString():   "+keywords);
*/    // NTN: end of test case for Bug 1812

    return super.open();
  }

  public int read(Data data, int elements) { return -1; }

  private void readFile() {
    lines = new Vector<String>(1024);
    // TODO:
  }

  public final void removeData (double offset) {
    removeData(offset, 1);
  }

  public void removeData (double offset, long count) {
    if (lines == null) { readFile(); }
    Util.removeListElements(lines, (int)offset, (int)count);
  }

  public double getNumberOfRows () {
    return (lines==null)? 0 : lines.size(); // NTN: need this otherwise LIST2 throws a bunch of exceptions (still does not list anything though)
  }

  public void insertData (double offset, Object obj) {
    // TODO:
  }

  public double avail () {
    return 1;
  }

  public int setData (double offset, Table table) {
    setData(offset, "TEXT", table.getS("TEXT", ""));
    return 1;
  }

  public void setData (double element, String fieldName, Object obj) {
    if (lines == null) { readFile(); }
    if (fieldName.equals("TEXT")) {
      lines.set((int)offset, obj.toString());
    }
  }

  public void setRecordDefs (Table defs) {
    if ((defs.size() == 1) && defs.getKeys()[0].equals("TEXT")) {
      // All is good... nothing to do.
    } else {
      throw new MidasException("KmlFile: Can not alter record definitions in a TextFile.");
    }
  }

  public Table getRecordDefs () {
    Table tbl = new Table();
    tbl.setKey("TEXT", getRecordDef(0));
    return tbl;
  }

  public Table getRecordDef (int i) {
    Table tbl = new Table();
    tbl.setKey("NAME", "TEXT");
    return tbl;
  }

  public Table getDataTable (double offset) {
    if (offset < 0) return null;
    if (lines == null) { readFile(); }
    Table tbl = null;

    if ((offset < lines.size()) && (offset >= 0.0)) {
      tbl = new Table();
      tbl.setKey("TEXT", lines.get((int)offset));
    }

    return tbl;
  }

  public boolean getProtected () { return true; }

  public int getRecordDefCount () { return 1; }


  public int readDataTable(Table tbl) {
    // TODO Auto-generated method stub
    return 0;
  }

}
