package nxm.sys.lib;

import nxm.sys.inc.*;

/**
  Class to hold system logging capabilities.

  @author Jeff Schoen
  @version $Id: Logger.java,v 1.5 2010/08/02 15:42:00 ntn Exp $
*/
public class Logger {

  private Midas M;
  private DataFile lf;
  private Data lfd;
  private int size;

  private static boolean autoSave=false;
  public static void setAutoSave (boolean state) { autoSave = state; }

  public Logger (Midas midas) {
    M = midas;
  }

  /** open new log file, if exists saveOff() old file first */
  public void open (Object filename) {
    lf = new DataFile();
    lf.init(M,filename,"3000","NH",0);
    lf.setExt("mlog");
    lf.setSubRecords("TIME|3A,TYPE|1A,TEXT|XA,FROM|2A");
    lf.setTimeCode (Time.current());
    if (autoSave && lf.find(-1)) saveOff();
    lf.open(BaseFile.OUTPUT);
    lfd = lf.getDataBuffer(1);
    size=0;
  }

  /** flush log file */
  public void flush () {
    lf.flush();
  }

  /** close current log file */
  public void close () {
    lf.close();
  }

  /** save off named file with _timeTag */
  public void saveOff () {
    DataFile lfs = new DataFile();
    lfs.init(M,lf.getURL(),"3000","NH",0);
    lfs.open(BaseFile.INPUT);
    double time = lfs.getTimeCode();
    lfs.close();
    FileName fn = new FileName( lfs.getURL() );
    String tag = Time.format(time,Time.FILENAME,0);
    fn.setRoot( lfs.fn.getRoot()+"_"+tag );
    CoreIO.rename( lfs.getURL(), fn.getFullName() );
  }

  /** process log message */
  public void processMessage (Message msg) {
    String from = "UNKNOWN";
    if (msg.from instanceof IDable) from = ((IDable)msg.from).getID();
    String type = "UNKNOWN";
    if (msg.info==Midas.INFO) type = "INFO";
    else if (msg.info==Midas.WARNING) type = "WARNING";
    else if (msg.info==Midas.ERROR) type = "ERROR";
    processMessage(type, msg.data.toString(), from);
  }

  /** process log string to log file.
      @param type The type of message (only first 8 chars are used); if null, DEF="TEXT"
      @param text The message text (only first 80 chars are used); if null, DEF=""
      @param from The from ID (only first 16 chars are used); if null, DEF=""
      @since NeXtMidas 2.7.1
   */
  public void processMessage (CharSequence type, CharSequence text, CharSequence from) {
    if (lf==null || lfd==null) return;
    String time = Time.tag();
    if (type==null) type="TEXT";
    if (text==null) text="";
    if (from==null) from="";
    lfd.packS(0,24,time);
    lfd.packS(24,8,type);
    lfd.packS(32,80,text);
    lfd.packS(112,16,from);
    if ((M.debug&Debug.IO)==0) lf.write(lfd);
    size++;
  }

  public int getSize () { return size; }
}
