package nxm.sys.lib;

/**
  Security privileges

  @author Jay Hallam
  @version $Id: Privileges.java,v 1.11 2012/01/20 18:50:07 ntn Exp $
*/
public class Privileges {

  /** Privilege mask string list*/
  public static final String privsList = "EXECUTE,QUERYOS,SOCKET,FILE,NATIVE,MULTICAST";

  /** Privilege masks: <pre>
      PRIV_EXECUTE   = execute OS commands
      PRIV_QUERYOS   = query OS and System properties (for unsigned Applets)
      PRIV_SOCKET    = enable network sockets access
      PRIV_FILE      = enable file access
      PRIV_NATIVE    = enable native code
      PRIV_MULTICAST = enable sending multicast network packets
      PRIV_NONE      = no privileges (0)
      PRIV_ALL       = enable all privileges (-1 = all bits set)
      </pre>
   */
  public static final int PRIV_EXECUTE=0x1, PRIV_QUERYOS=0x2, PRIV_SOCKET=0x4,
                          PRIV_FILE=0x8,    PRIV_NATIVE=0x10, PRIV_MULTICAST=0x20,
                          PRIV_NONE=0,      PRIV_ALL=-1;

  /** Invalid privilege specified, used as return value for deprecated getPrivMaskNS().
      @deprecated in NeXtMidas 2.9.0. This was only used for Netscape 4.x support
   */
  @Deprecated
  public static final int PRIV_INVALID = -2;

  /** Table of privileges and the mapping to Netscape privilege names */
  private Table privsTable = new Table ("{"
        +"0={MASK="+PRIV_ALL+       ",PRIV=ALL,NSPRIV=NOTAPPLICABLE}"
       +",1={MASK="+PRIV_EXECUTE+   ",PRIV=EXECUTE,NSPRIV=UniversalExecAccess}"
       +",2={MASK="+PRIV_QUERYOS+   ",PRIV=QUERYOS,NSPRIV=UniversalLinkAccess}"
       +",3={MASK="+PRIV_SOCKET+    ",PRIV=SOCKET,NSPRIV=UniversalConnect}"
       +",4={MASK="+PRIV_FILE+      ",PRIV=FILE,NSPRIV=UniversalFileAccess}"
       +",5={MASK="+PRIV_NATIVE+    ",PRIV=NATIVE,NSPRIV=NOTAPPLICABLE}"
       +",6={MASK="+PRIV_MULTICAST+ ",PRIV=MULTICAST,NSPRIV=UniversalMulticast}"
       +"}");

  /** No argument constructor */
  public Privileges () {
  }

  /** Get the privileges mask string only for the enumerated ones in {@link #privsList} */
  public static String getPrivMaskString () {
    int mask = Shell.privs; //should be Shell.getPrivs();
    String maskStr = null;
    if (mask == -1 ) maskStr = "ALL";
    else             maskStr = Parser.mask2s(Privileges.privsList, mask);
    return maskStr;
  }

  /** Get the privilege name given the privilege mask
    @param privMask Mask PRIV_EXECUTE, PRIV_QUERYOS,...
    @param netscape Return a Netscape privilege string if true
    @return 1) A privilege name OR 2) a Netscape privilege name
            3) null if not found
    @deprecated Since NeXtMidas 3.1.1, this method was only used for Netscape 4.x support
                back in the old days.
  */
  @Deprecated // TODO: remove this method in the being of the 3.5.x beta series
  public String getPrivName (int privMask, boolean netscape) {
    String status = null;
    String privtag = "PRIV";
    if (netscape) privtag = "NSPRIV";
    for (Table.Iterator ti=privsTable.iterator(); ti.getNext();) {
      Table ptab = (Table) ti.value;
      String maskString = ""+ptab.getKey("MASK");
      if (maskString.equals(""+privMask)) {
        status = ""+ptab.getKey(privtag);
        break;
      }
    }
    return status;
  }
}
