package nxm.sys.libg;

import java.awt.Color;

import nxm.sys.inc.*;
import nxm.sys.lib.Shell;

import javax.swing.JScrollPane;
import javax.swing.ScrollPaneConstants;

/**
  A class to display a block of text.

  @author Jeff Schoen
  @version $Id: GText.java,v 1.27 2011/07/25 12:59:49 ntn Exp $
*/
public class GText extends GWidget {

  private MJTextArea  mjtext;
  private JScrollPane scrollPane;

  /**
     Instantiate a new Text Area, initialize the parameters, and open
     the graphics panel.
     @param parent the parent object
     @param name the name of this MJTextArea
     @param text the text to be displayed
     @param aw the width
     @param ah the height
     @param flags the widget flags
     @param mh the MessageHandler
   */
  public GText (Object parent, String name, String text, int aw, int ah, int flags, MessageHandler mh) {
    super("GText",parent,name,flags,mh);
    open (name, aw, ah, text);
    addTo (parent);
  }

  /**
     Open the graphics window for this MJTextArea with the given name.
     This MJTextArea will be sized according to the given width and height,
     and will display the given text.
     @param name the name of this MJTextArea
     @param aw the width
     @param ah the height
     @param text the text to be displayed
   */
  public void open (String name, int aw, int ah, String text) {
    open(); // get text sizes
    super.open (name,aw*tw,ah*th+ta);
    setupTextArea(text,ah,aw);
  }

  public void setupTextArea (String text, int ah, int aw) {
    mjtext = new MJTextArea (text,ah,aw);
    mjtext.setEditable(false);
    // The widget theme is not yet set here, so I override the setWidgetTheme(..) method  to
    // affect the input text colors
    mjtext.setBackground(theme.cwbg);
    mjtext.setForeground(theme.cwfg);

    // JWB: windows and Solaris do not size the scroll pane properly so for now it is only Linux
    if (Shell.osname.equals("Linux")) {
      scrollPane = new JScrollPane(mjtext);
      scrollPane.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED);
      scrollPane.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
      panel.add(scrollPane);
    } else {
      panel.add(mjtext);
    }
    setBounds();
  }

  /** Set the widget theme for this text widget and the MJTextArea.
    @param o The theme in String, Table or Object form
  */
  public void setWidgetTheme(Object o) {
    // jph - I added this method because -- Seems that the widget theme is not yet set in the open()
    //       method below , so this MAY have no thus I override the setWidgetTheme(..) method in GWidget
    //       to affect the input text colors
    super.setWidgetTheme(o);
    mjtext.setForeground(theme.getWidgetForeground());
    mjtext.setBackground(theme.getWidgetBackground());

  }

  public void resize (int mode) {
    super.resize(mode);
    setBounds();
  }

  private void setBounds () {
    mjtext.setBounds(bw,bw+titleh,pw,ph);
    if (scrollPane != null) {
      scrollPane.setBounds(bw,bw+titleh,pw,ph); // keep the scrollpane the same size as the textArea
    }
  }

  public void setValue (String value) {
    mjtext.setText(value);
  }

  public Object getValue () {
    return mjtext.getText();
  }

  public void setTheme (Theme theme) {
    super.setTheme(theme);
    mjtext.setBackground(theme.cwbg);
    mjtext.setForeground(theme.cwfg);
  }

  public boolean isStretchable() {
    return true;
  }

  public MJTextArea getMJTextArea () {
    return mjtext;
  }

  /** Sets the color of the selected text
      @param c The color to set to
      @since NeXtMidas 2.9.0
   */
  public void setSelectedTextColor (Color c) {
    mjtext.setSelectedTextColor(c);
  }

  /** Gets the color of the selected text
      @return The color
      @since neXtMidas 2.9.0
   */
  public Color getSelectedTextColor () {
    return mjtext.getSelectedTextColor();
  }

  /** Sets the background selection color of the text area
      @param c The Color to set
      @since NeXtMidas 2.9.0
   */
  public void setSelectionColor (Color c) {
    mjtext.setSelectionColor(c);
  }

  /** Gets the background color of the text area selection
      @return The color
      @since NeXtMidas 2.9.0
   */
  public Color getSelectionColor () {
    return mjtext.getSelectionColor();
  }

  /** Sets the color of the selected text
      @param color The color to set to
      @since NeXtMidas 2.9.0
   */
  public void setSelectedTextColor (String color) {
    setSelectedTextColor(MColor.getColor(color));
  }

  /** Gets the color of the selected text
      @return The color
      @since neXtMidas 2.9.0
   */
  public String getSelectedTextColorString () {
    return MColor.toString(getSelectedTextColor());
  }

  /** Sets the background selection color of the selection
      @param c The Color to set
      @since NeXtMidas 2.9.0
   */
  public void setSelectionColor (String c) {
    setSelectionColor(MColor.getColor(c));
  }

  /** Gets the background color of the selection
      @return The color
      @since NeXtMidas 2.9.0
   */
  public String getSelectionColorString () {
    return MColor.toString(getSelectionColor());
  }
}
