package nxm.sys.libg;

import nxm.sys.inc.*;
import nxm.sys.lib.*;
import java.awt.Color;

/**
  A class to hold and plot a 1 dimensional data layer.

  @author Jeff Schoen
  @version $Id: Layer1D.java,v 1.47 2012/05/14 17:10:01 jwb Exp $
*/
public class Layer1D extends Layer {

  /** Since NeXtMidas 3.1.0 this field is now protected, prior to that it was public.
      It was deprecated in NeXtMidas 2.9.2, it should be protected.  */
  protected float[] fbuf;
  /** Since NeXtMidas 3.1.0 this field is now protected, prior to that it was public.
      It was deprecated in NeXtMidas 2.9.2, instead use {@link #getXIsTime()}. */
  protected boolean xIsTime;
  private Color color2;

  /* Constructors */
  public Layer1D () {
    super();
    name = "1D";
  }

  /** setup to attach a file */
  public boolean setFile (PlotFile pf, Table tab, Args MA) {
    setFile(pf);
    setXAxis( pf.getXStart(), pf.getXDelta() );
    if (typeClass==1) {
      int ts=(int)Math.min(pf.getSize(),10.0e6);
      int fs=MA.getL("/FS",Math.min(ts,32768));
      double size = pf.getSize();
      setFrame(fs);
      int lps=MA.getL("/LPS",1);
      setSize(lps);
      setData (pf.getDataBuffer(fs*lps));
      isPaged = !realtime && (fs*lps)<ts;
      xIsTime = (pf.getXUnits()==Units.TIME);
      setTimeAlign(MA.getState("/TIMEALIGN",false));  // added for Bug 2419 in 3.1.2
    }
    else if (typeClass==2) {
      // NTN: 2011-06-07 next three line added to support plotting multi-line traces from type 2000 files (BUG 2397)
      ncolors = MA.getL("NC",32);
      cmap = MA.getSelectionIndex("CMAP", MColor.colorMapList, MColor.MULTI);
      colors = MColor.getColorMap(cmap, ncolors);

      setFrame( pf.getXFrame() );
      int lps=MA.getL("/LPS",1);
      if (lps == -1) lps = (int) pf.getSize(); // -1 means plot all line traces

      setSize(lps);
      setData(pf.getDataBuffer(lps));
      isPaged = false;
      setYUnits(Units.NONE); // dont apply here
    }
    else return false;
    return true;
  }

  /** process the plot data for this frame */
  public void process () {
    ndata = 0; if (data==null) return;
    ndata = data.size*data.ape;
    fbuf = data2fa (data,ndata,fbuf,0,MP.mode);
  }

  /* find layer bounds */
  public void findRange() {
    if (frame==0) frame=ndata;
    x1 = xstart;
    x2 = xstart + (frame-1)*xdelta;
    if (x1>x2) { double t=x1; x1=x2; x2=t; }
      y1 = 0; y2 = 1;
    if (fbuf==null || fbuf.length == 0) return;
    int spa = (pf!=null)? pf.getMode() : 1;
    if (MP.mode==MPlot.M_RVI) {
      // BUG 70 - scale correctly for none scalar (e.g. complex) data when mode is Real vs Imaginary
      x1 = x2 = fbuf[0];
      y1 = y2 = fbuf[1];
      if (spa<2) spa = 2; // force spa=2 for RvI complex mode
      for (int ii=2; ii<(ndata*spa)-1 && ii<fbuf.length; ii+=spa) {
        if (fbuf[ii]>x2)   x2=fbuf[ii];
        if (fbuf[ii]<x1)   x1=fbuf[ii];
        if (fbuf[ii+1]>y2) y2=fbuf[ii+1];
        if (fbuf[ii+1]<y1) y1=fbuf[ii+1];
      }
    } else {
      y1 = y2 = fbuf[0];
      // for M_RNI (Real and Imaginary) we need to look at all the buffered data for non scalar data
      int elements = (MP.mode==MPlot.M_RNI) ? (ndata*spa) : ndata;
      for (int i=1; i<elements && i<fbuf.length; i++) {
        if (fbuf[i]>y2) y2=fbuf[i];
        if (fbuf[i]<y1) y1=fbuf[i];
      }
    }
    if (MP.scaleAvg>0) {
      double p = 1.0/MP.scaleAvg;
      if (a1==a2) p=1.0;
      a1 = a1*(1.0-p) + y1*p;
      a2 = a2*(1.0-p) + y2*p;
    }
  }

  public void findFullRange() { findFullRange(pf); }

  public void findFullRange (PlotFile pf) {
    fx1 = x1; fx2 = x2;
    fy1 = y1; fy2 = y2;
    fz1 = z1; fz2 = z2;
    if (pf!=null && !realtime) {
      fx1 = pf.getXStart();
      // fix for Bug 1502 - plot/all does does not scale correctly
      if (typeClass==2) fx2 = fx1 + (pf.getXFrame()-1)*pf.getXDelta(); // 2D file (plotting one frame wide)
      else              fx2 = fx1 + (pf.getSize()-1)*pf.getXDelta();   // 1D file
      if (fx1>fx2) { double t=fx1; fx1=fx2; fx2=t; }
    }
  }

  /** draw the layer on the specified plot */
  public void draw (int flag) {
    super.draw(flag);
    int i1=0,i2=ndata,i12;
    if (ndata<=0) return;
 
    // Begin Bug 2419 changes 
    double startPos = xstart;
    if (typeClass==1 && pf instanceof DataFile && pf.getUnits()==nxm.sys.inc.Units.TIME && isTimeAlign()) {
      DataFile df = (DataFile)pf;
      double startTime =  pf.getTimeAt(xstart);
      Layer lay = (Layer)MP.getLayers().get(1);  // get the base layer
      if (lay!=null && lay!=this && lay.pf instanceof DataFile && startTime>MP.rx1) { 
        startPos = ((DataFile)lay.pf).getIndex(startTime,DataFile.ISTART, DataFile.ADDR_ABSCISSA_AUTO) * df.getXDelta();
      }
    } /// End changes for Bug 2419, except for changes below to MP.drawLine(..) calls ///
    
    if (size>0) i2=Math.min(frame,ndata);
    if (MP.view < MPlot.V_XYZ && MP.mode != MPlot.M_RVI) {
      if (xstart < MP.rx1)
        i1 = Math.max(0,(int)((MP.rx1-xstart)/xdelta ));
      if (xstart+ndata*xdelta > MP.rx2)
        i2 = Math.min(i2,(int)((MP.rx2-xstart)/xdelta )+2);
    }
    i12 = i2-i1;
    if (i12<=0) return;

    Line lin = line.copy();
    int spa = data.spa;

    if (size<=1) {
      if (MP.mode==MPlot.M_RVI) MP.drawLine (fbuf,i1*spa,spa,i12,lin);
      else if (MP.mode==MPlot.M_RNI && spa>1) {
        Color color1 = lin.getColor();
        for (int s=0; s<spa; ) {
          //MP.drawLine (fbuf,i1*spa+s,spa,i12,lin,xstart,xdelta); 
          MP.drawLine (fbuf,i1*spa+s,spa,i12,lin,startPos,xdelta); // since 3.1.0: changed for Bug 2419
          lin.setColor( (++s<spa)? lin.getColor().darker() : color1 );
        }
      }
      //else MP.drawLine (fbuf,i1,1,i12,lin,xstart,xdelta);
      else MP.drawLine (fbuf,i1,1,i12,lin,startPos,xdelta);   // since 3.1.0: changed for Bug 2419
    }
    else for (int i=0; i<size; i++) {
      lin.setColor(colors[i%ncolors + 1]);
      if (MP.mode==MPlot.M_RVI) MP.drawLine (fbuf,(i1+i*frame)*2,2,i12,lin);
      //else MP.drawLine (fbuf,i1+i*frame,1,i12,lin,xstart-i*xdelta*frame,xdelta);
      else MP.drawLine (fbuf,i1+i*frame,1,i12,lin,startPos-i*xdelta*frame,xdelta); // since 3.1.0: changed for Bug 2419
    }
    if (feature!=null) feature.draw(flag);
  }

  public void setSize (int value) {
    super.setSize(value);
    if (size>1) setColorMap(-1,value);
  }

  public double getPosAtTime (double time) {
    DataFile df  = (DataFile)pf;
    double delta = df.getDelta();
    double index = df.getIndexAt(time);
    return index * delta;
  }

  /** Get the time at a given mouse position.<br><br>
      This method gets the percent the mouse is across the data portion of the plot and derives which
      element (index in the file) corresponds to that location.
      @since This method substantially re-written in NeXtMidas 2.8.1 for BUG 2063
   */
  public double getTime (int px, int py) {
    if ((MP.xvy && (py<MP.iy1 || py>MP.iy2)) || (px<MP.ix1 || px>MP.ix2)) return 0;  // outside plot bounds

    boolean invert  = (MP.view==MPlot.V_IXIY) || (MP.view==MPlot.V_IXY) || (MP.view==MPlot.V_IYIX) || (MP.view==MPlot.V_YIX);
    double  delta   = ((DataFile)pf).getDelta();             // x (or y) delta of the file
    double  xMinVal = (MP.rx1 < MP.orx1)? MP.orx1: MP.rx1;   // min x data value plotted on screen
    double  xMaxVal = (MP.rx2 > MP.orx2)? MP.orx2: MP.rx2;   // max x data value plotted on screen
    double  yMinVal = (MP.ry1 < MP.ory1)? MP.ory1: MP.ry1;   // min y data value plotted on screen
    double  yMaxVal = (MP.ry2 > MP.ory2)? MP.ory2: MP.ry2;   // max y data value plotted on screen
    int     pos;                                             // mouse position
    int     startPixel;                                      // pixel loc where the first data is drawn
    int     endPixel;                                        // pixel loc where the last data is drawn

    // which position to use (we only care about the X coord if view=YX, or the Y coord if view=XY)
    if (MP.xvy) {
      if (invert) pos = py;
      else        pos = (MP.pos.h - py);   // drawn from the bottom of the screen panel so flip it
    } else {
      if (invert) pos = (MP.pos.w - px);   // drawn from the right side of the panel so flip it
      else        pos = px;
    }

    /* JWB: We go through the somewhat onerous calculations below to get the startPixel and endPixel
       instead of simply using the plot data area boundaries (MP.ix1|ix2|iy1|iy2) because if the user
       ZOOMs across one of the borders of the plot, we plot a 'blank' area corresponding to the
       portion of the ZOOM dragbox that was outside the plot-able area. This 'blank' area
       is within the ix1|iy1...etc values but has no data so we need to disregard it.
     */
    double[] rwc = new double[2]; // used to store real-world coord for min/max mouse positions
    MPoint pix = new MPoint();    // used to store pixel location of RWC mouse position boundaries
    rwc[0] = xMinVal;
    rwc[1] = yMinVal;
    MP.viewer.rwc2pix(rwc, pix);  // start point

    /* get the pixel location of the start point */
    if (MP.xvy){
      if (invert) startPixel = pix.y;               // view= xvy and inverted so draw top->down
      else        startPixel = MP.pos.h - pix.y;    // view= xvy not inverted so draw bottom->up
    } else {
      if (invert) startPixel = MP.pos.w - pix.x;    // view= yx and inverted so draw right->left
      else        startPixel = pix.x;               // view= yx not inverted so draw left->right
    }

    rwc[0] = xMaxVal;
    rwc[1] = yMaxVal;
    MP.viewer.rwc2pix(rwc, pix);  // end point

    /* get the pixel location of the end point */
    if (MP.xvy){
      if (invert) endPixel = pix.y;               // view= xvy and inverted so draw top->down
      else        endPixel = MP.pos.h - pix.y;    // view= xvy not inverted so draw bottom->up
    } else {
      if (invert) endPixel = MP.pos.w - pix.x;    // view= yx and inverted so draw right->left
      else        endPixel = pix.x;               // view= yx not inverted so draw left->right
    }

    if (pos < startPixel || pos > endPixel) return 0;                           // outside of drawn data area in plot
    double viewableIndex = ( (xMaxVal - xMinVal) / delta ) + 1;                 // viewable bounds (# of elements in view) (+1 added to include last element)
    double startIndex    = realtime? ((DataFile)pf).seek(): (xMinVal - MP.orx1) / delta; // index in file at start point of viewable area
    double relMousePos   = (pos-(double)startPixel)/(double)(endPixel-startPixel);  // percentage of mouse loc across the plot (0.000..1 - .9999999..)
    double index         = startIndex + (relMousePos * viewableIndex);              // index in file at given mouse position

    return ((DataFile)pf).getTimeAt(index);
  }

  public double getNearestYValue (int i) {
    if (fbuf==null) return 0.0;
    return (double)fbuf[i];
  }

  public boolean clear () {
    ndata=0;
    return true;
  }

  /** @return true if X units is time, otherwise false */
  public boolean getXIsTime () { // since 2.9.2
    return xIsTime;
  }

  @Override
  public Feature getFeature () {
    // This method added in 3.1.1 for Bug 2549. Only Layer1D Features should set the
    // Feature type to DATA, since it is unsupported in other Layers and causes a warning.
    if (feature!=null) return super.getFeature();
    else {
      feature = super.getFeature();
      feature.setType(Feature.DATA);  // Set Type to DATA for Layer1D
      return feature;
    }
  }

}
