package nxm.sys.libg;

import nxm.sys.inc.*;
import nxm.sys.lib.*;
import nxm.sys.libg.*;

/**
  A class to hold and plot a grid layer. 

  @author Jeff Schoen
  @version $Id: LayerGrid.java,v 1.15 2008/02/25 15:58:13 jxc Exp $

*/
public class LayerGrid extends Layer {

  private int dim;
  private float[] fbuf;

  /* Constructors */
  public LayerGrid () { 
    super(); 
    setColorMap(MColor.GRID,-1);
    name = "GRID";
  }

  /** update View */
  public void update () {
    int n,nn,ic; double r,rmax;
    if (MP.view==MP.viewLast) return;
    ndata = 0;
    if (MP.view==MPlot.V_POLAR) {
      rmax = Math.max(Math.abs(MP.ory1),Math.abs(MP.ory2));
      fbuf = new float[1028*2];
      for (n=30; n<=90; n+=30) {
        ic=MColor.C_GRID; r = rmax*n/90;
        set2D (73,ic); set2D (-180.0,r); set2D (+180.0,r);
        for (nn=-180; nn<=180; nn+=5) set2D( (double)nn, r); 
      }
      for (n=-180; n<=180; n+=30) {
        if (n==0) ic=MColor.C_AXIS; else ic=MColor.C_GRID;
        set2D (19,ic); set2D (n,0); set2D (n,90);
        for (nn=0; nn<=90; nn+=5) set2D( (double)n, rmax*nn/90); 
      }
      dim=2;
    }
    else if (MP.view==MPlot.V_LATLON || MP.view==MPlot.V_MERCATOR || MP.view==MPlot.V_GEODETIC) {
      fbuf = new float[4184*2];
      for (n=-75; n<=75; n+=15) {
        if (n==0) ic=MColor.C_AXIS; else ic=MColor.C_GRID;
        set2D (181,ic); set2D (-180,n); set2D (180,n);
        for (nn=-180; nn<=180; nn+=2) set2D(nn,n); 
      }
      for (n=-180; n<=165; n+=15) {
        if (n==0) ic=MColor.C_AXIS; else ic=MColor.C_GRID;
        set2D (87,ic); set2D (n,-85); set2D (n,85);
        set2D (n,-85); for (nn=-84; nn<=84; nn+=2) set2D(n,nn); set2D (n,85);
      }
      dim=2;
    }
    else if (MP.view==MPlot.V_XYZ) {
      int np=5,np2=np*2; 
      fbuf = new float[3*(3+np2+1)*3];
      double xc=0.0, xd=1.0/np;
      double yc=0.0, yd=1.0/np;
      double zc=0.0, zd=1.0/np;
      ic=MColor.C_AXIS; 
      set3D (np2+1,ic,0); set3D(xc,yc,zc); set3D(xc+np*xd,yc,zc); 
      for (n=-np; n<=np; n++) set3D(xc+n*xd,yc,zc); 
      set3D (np2+1,ic,0); set3D(xc,yc,zc); set3D(xc,yc+np*yd,zc); 
      for (n=-np; n<=np; n++) set3D(xc,yc+n*yd,zc); 
      set3D (np2+1,ic,0); set3D(xc,yc,zc); set3D(xc,yc,zc+np*zd); 
      for (n=-np; n<=np; n++) set3D(xc,yc,zc+n*zd); 
      ic=MColor.C_GRID; 
      dim=3;
    }
  }

  private void set2D (int x, int y) {
    fbuf[ndata++]=(float)x; fbuf[ndata++]=(float)y; 
  }
  private void set2D (double x, double y) {
    fbuf[ndata++]=(float)x; fbuf[ndata++]=(float)y; 
  }
  private void set3D (int x, int y, int z) {
    fbuf[ndata++]=(float)x; fbuf[ndata++]=(float)y; fbuf[ndata++]=(float)z;
  }
  private void set3D (double x, double y, double z) {
    fbuf[ndata++]=(float)x; fbuf[ndata++]=(float)y; fbuf[ndata++]=(float)z;
  }
    
  /* find layer bounds */
  public void findRange() { 
    if (MP.view==MPlot.V_LATLON || MP.view==MPlot.V_MERCATOR || MP.view==MPlot.V_GEODETIC) {
      x1=-180; x2=+180; y1=-90; y2=+90;
    }
  }

  public void draw (int flag) {
    super.draw(flag);
    int j,k,npts,ncol;
    if (ndata<=0) return;
    
    Line lin = line.copy();
    for (j=0; j<ndata; j=k) {
      npts = (int)fbuf[j++];  
      ncol = (int)fbuf[j++];  
      if (dim>2) j += (dim-2);
      k = j + (2+npts)*dim;
      if (MP.view==MPlot.V_GEODETIC || MP.view==MPlot.V_XYZ || MP.view==MPlot.V_POLAR) {
	j += 2*dim;
      } else {
        if (fbuf[j++]>=MP.rx2) continue;
        if (fbuf[j++]>=MP.ry2) continue;
        if (fbuf[j++]<=MP.rx1) continue;
        if (fbuf[j++]<=MP.ry1) continue;
      }
      lin.setColor(colors[ncol]);
      if (MP.view==MPlot.V_XYZ) MP.pix.setFlag(Pix.NORM);
      MP.drawLine (fbuf,j,dim,npts,lin);
      if (MP.view==MPlot.V_XYZ) MP.pix.clrFlag(Pix.NORM);
    }
  }

}
