package nxm.sys.libg;

import nxm.sys.inc.*;
import nxm.sys.lib.*;

/**
  A class to hold a layer of line segments or filled polygons.

  @author Jeff Schoen
  @version $Id: LayerPoly.java,v 1.21 2008/02/25 15:58:13 jxc Exp $
*/
public class LayerPoly extends Layer {

  public float[] fbuf;
  public int dim=2;
  private float[] ftmp = new float[6];
  private Table ctab;
  private boolean showWarning = true; // flag to show invalid color # warning in file the first time

  /* Constructors */
  public LayerPoly() {
    super();
    name = "POLY";
  }

  public LayerPoly (int cmap) {
    this();
    setColorMap(cmap);
    if (cmap==MColor.MAPLINE) ctab = MColor.getMapColorTable();
  }

  public LayerPoly (Table ctab) {
    this();
    setColorTable(ctab);
  }

  /** set the file input - if necessary - should be over-ridden */
  public boolean setFile (PlotFile pf, Table tab, Args MA) {
    super.setFile(pf,tab,MA);
    dim = pf.getMode();
    showWarning = true;
    return true;
  }

  /** process the plot data for this frame */
  public void process() {
    ndata = 0; if (data==null) return;
    ndata = data.size * data.ape;
    fbuf = new float[ndata*dim];
    data.toArray(fbuf);
  }

  /* find layer bounds */
  public void findRange() {
    int j,k,npts,ncol; float xy;
    x1=+180; x2=-180; y1=+90; y2=-90; z1=0; z2=0;
    for (j=0; j<ndata*dim; j=k) {
      npts = (int)fbuf[j];
      ncol = (int)fbuf[j+1];
      k = j + (3+npts)*dim;
      if (ncol== -998) continue;
      j += dim;
      xy = fbuf[j++];  if (xy<x1) x1=xy;
      xy = fbuf[j++];  if (xy<y1) y1=xy;
      if (dim==3) { xy = fbuf[j++];  if (xy<z1) z1=xy; }
      xy = fbuf[j++];  if (xy>x2) x2=xy;
      xy = fbuf[j++];  if (xy>y2) y2=xy;
      if (dim==3) { xy = fbuf[j++];  if (xy>z2) z2=xy; }
    }
    a1=y1; a2=y2;
  }

  public void draw (int flag) {
    super.draw(flag);
    int i,j,k,npts,ncol;
    if (ndata<=0) return;
    
    Line lin = line.copy();
    
    for (j=0; j<ndata*dim && !MP.cancel; j=k) {
      npts = (int)fbuf[j];
      ncol = (int)fbuf[j+1];
      if (showWarning && (ncol!=fbuf[j+1])) {
        // non-integer color number (potentially bad data) - print warning 1st time
        showWarning = false;
        Shell.warning("LayerPoly: expecting integer color number, but got " + fbuf[j+1]);
      }
      k = j + (3+npts)*dim;
      j += dim;
      if (MP.is2D) {
        if (fbuf[j]>=MP.rx2) continue;
        if (fbuf[j+1]>=MP.ry2) continue;
        j += dim;
        if (fbuf[j]<=MP.rx1) continue;
        if (fbuf[j+1]<=MP.ry1) continue;
        j += dim;
      }
      else if (MP.view==MPlot.V_GEODETIC || MP.view==MPlot.V_XYZ) {
        i = MP.viewer.rwc2pix (fbuf,j,dim,2,MP.tpix);
        if (i!=0); // need to draw, else try alternate corners
        else if (dim==2) {
          ftmp[0]=fbuf[j]; ftmp[1]=fbuf[j+3];
          ftmp[2]=fbuf[j+2]; ftmp[3]=fbuf[j+1];
          i = MP.viewer.rwc2pix (ftmp,0,dim,2,MP.tpix);
        }
        else if (dim==3) {
          ftmp[0]=fbuf[j]; ftmp[1]=fbuf[j+3]; ftmp[2]=fbuf[j+5];
          ftmp[3]=fbuf[j+3]; ftmp[4]=fbuf[j+1]; ftmp[5]=fbuf[j+2];
          i = MP.viewer.rwc2pix (ftmp,0,dim,2,MP.tpix);
        }
        j += 2*dim;
        if (i==0) continue;
      }
      else {
        j += 2*dim;
      }
      if (ncol== -998) continue;
      if (ncol== -999) { k=j; continue; }
      if (ncol<1) {
        // There are "special" colors handled by X-Midas' XGEOPLOT which we need to handle here.  
        lin.setColor(MColor.getSpecialXMidasColor(ncol));
      }
      else {
        int colorIndex = ((ncol-1)%(colors.length-1)) +1; // for 1..(color.length-1)
        lin.setColor(colors[colorIndex]);
      }
      MP.drawLine (fbuf,j,dim,npts,lin);
    } // end for loop
  }

  /** return a table with the current color values */
  public Table getColorTable() {
    return ctab;
  }

  /** set current color values to this table */
  public void setColorTable (Table ctab) {
    setColorMap(MColor.getColorMap(ctab));
    this.ctab = ctab;
  }

}
