package nxm.sys.libg;

import nxm.sys.lib.*;

import java.awt.*;
import java.awt.image.*;
import java.util.regex.*;

/**
  A class to handle colormaps.

  @author Jeff Schoen
  @version $Id: MColor.java,v 1.69 2012/10/18 16:42:54 ntn Exp $
*/
public class MColor extends Object {
  /** The color to use if a named color is not found or if a color is <code>null</code>. */
  private static final Color DEFAULT_COLOR = Color.white;

  private static final Pattern WHITESPACE_REGEX = Pattern.compile("\\s"); // Regex for white space

  /** Color map number for GRAYSCALE */
  public static final int GRAYSCALE = 1;

  /** Color map number for RAMP */
  public static final int RAMP = 2;

  /** Color map number for XRAMP */
  public static final int XRAMP = 3;

  /** Color map number for WHEEL */
  public static final int WHEEL = 4;

  /** Color map number for TERRAIN */
  public static final int TERRAIN = 5;

  /** Color map number for MULTI */
  public static final int MULTI = 6;

  /** Color map number for T5 */
  public static final int T5 = 7;

  /** Color map number for MAPLINE */
  public static final int MAPLINE = 8;

  /** Color map number for GRID */
  public static final int GRID = 9;

  /** Color map number for SPECTRUM */
  public static final int SPECTRUM = 10;

  /** Color map number for XDATALIST */
  public static final int XDATALIST = 11;

  /** the maximum acceptable color map number for 2D layers */
  public static final int MAXTYPE2D = 6;

  /** the maximum acceptable color map number */
  public static final int MAXTYPE=9;

  /** Color map name list: <code>{@value}</code> */
  public static final String colorMapList =
      "GrayScale,Ramp,XRamp,Wheel,Terrain,Multi,T5Layer,MapLine,Grid,Spectrum,XDatalist";

  /** Color index for CEP in T5 color map */
  public static final int C_CEP = 1;

  /** Color index for HORIZON in T5 color map */
  public static final int C_HORIZON = 2;

  /** Color index for TDOA in T5 color map */
  public static final int C_TDOA = 3;

  /** Color index for FDOA in T5 color map */
  public static final int C_FDOA = 4;

  /** Color index for T5000 in T5 color map */
  public static final int C_T5000=5;

  /** Color index for COASTLINE in MAPLINE color map */
  public static final int C_COASTLINE = 1;

  /** Color index for POLITICAL in MAPLINE color map */
  public static final int C_POLITICAL = 2;

  /** Color index for INTERPOL in MAPLINE color map */
  public static final int C_INTERPOL = 3;

  /** Color index for RIVER in MAPLINE color map */
  public static final int C_RIVER = 4;

  /** Color index for RAIL in MAPLINE color map */
  public static final int C_RAIL = 5;

  /** Color index for ROAD in MAPLINE color map */
  public static final int C_ROAD = 6;

  /** Color index for POL_OLD in MAPLINE color map */
  public static final int C_POL_OLD = 7;

  /** Color index for INTPOL_OLD in MAPLINE color map */
  public static final int C_INTPOL_OLD=8;

  /** Color index for axis in GRID color map */
  public static final int C_AXIS = 1;

  /** Color index for grid in GRID color map */
  public static final int  C_GRID=2;

  /** JGS why is this here , not used */
  public static final int NOSHARE=0x0001;

  /** Default number of colors */
  public static final int DEFCOLORS=32;

  /** Grayscale color map */
  public static final byte[] CM_GRAY = {0,0,0,0, 100,100,100,100};
  /** NeXtMidas ramp color map */
  public static final byte[] CM_RAMP = {0,10,0,23, 18,34,0,60, 36,58,20,47,
        55,74,20,28, 72,90,43,0, 87,100,72,0, 100,100,100,76};
  /** Wheel color map */
  public static final byte[] CM_WHEEL = {0,100,100,0, 20,0,80,40, 30,0,100,100,
        50,10,10,90, 65,100,0,100, 88,100,40,0, 100,100,100,0};
  /** Terrain color map */
  public static final byte[] CM_TERRAIN = {0,0,0,40, 50,0,45,55, 50,0,55,55,
        50,0,70,20, 75,75,40,25, 100,100,100,95, 100,100,100,100};
  /** {@link LayerMulti} color map */
  public static final byte[] CM_MULTI = {0,0,75,0, 17,0,90,90, 34,0,0,90,
        49,90,0,90, 66,90,0,0, 83,90,90,0, 100,95,95,95};
  /** X-Midas ramp color map */
  public static final byte[] CM_XRAMP = {0,0,0,15, 10,0,0,50, 31,0,65,75,
        50,0,85,0, 70,75,80,0, 83,100,60,0, 100,100,0,0};
  /** {@link LayerMulti} color map */
  public static final byte[] CM_SPECTRUM = {0,0,75,0, 22,0,90,90, 37,0,0,85,
        49,90,0,85, 68,90,0,0, 80,90,90,0, 100,95,95,95};

  int ncolor=0;
  Color tcolor, tcolors[];

  /** Constructor */
  public MColor () {
    super();
  }
  /** Constructor
    @param name  JGS not used
    @param color Java color object
  */
  public MColor (String name, Color color) {
    super();
    ncolor = 0;
    tcolor = color;
  }
  /** Constructor
    @param name  JGS not used
    @param colors Array of Java color objects
  */
  public MColor (String name, Color[] colors) {
    super();
    ncolor = colors.length;
    tcolors = colors;
  }

  /** Get a single color.  This is a convenience wrapper for the Java constructor
      which has component ranges from 0 to 1.0 instead of 0-100.
    @param r Red component: 0-100
    @param g Green component: 0-100
    @param b Blue component: 0-100
    @return A Java color object
  */
  public static Color getColor (int r, int g, int b) {
    Color color = new Color ( r*0.01f, g*0.01f, b*0.01f);
    return color;
  }

  /** Get a single color by name or number string. This calls {@link #getColor(Object)} and then
      applies the given alpha component to the color.
      @param obj   The name/value of the color.
      @param alpha The alpha component to use (0-255) where 255 = Completely Opaque, 0 = Completely
                   Transparent.
      @return a Java Color object
      @since NeXtMidas 2.3.1
  */
  public static Color getColor (Object obj, int alpha) {
    Color color = getColor(obj);
    if ((color != null) && (color.getAlpha() != alpha)) {
      color = new Color(color.getRed(), color.getGreen(), color.getBlue(), alpha);
    }
    return color;
  }

  /** Get a single color by name or number string. This checks to see if the object
      passed in is an instance of {@link Color}, if so it just returns that object,
      otherwise it converts the object to a string and calls {@link #getLegacyColor(String)}.
      @param obj The name/value of the color.
      @return a Java Color object
      @since NeXtMidas 2.3.1
  */
  public static Color getColor (Object obj) {
    if (obj instanceof Color)  return (Color)obj;


    // Check to see if they gave us something that has already been converted to
    // a number. If so, use the "old" conversion to color.
    if (obj instanceof Data) {
      Data data = (Data)obj;
      if (data.type != 'A') {
        int val = Convert.o2l(data);
        return new Color(val, (val & 0xFF000000) != 0 );
      }
    }
    if (obj instanceof Number) {
      int val = ((Number)obj).intValue();
      return new Color(val, (val & 0xFF000000) != 0 );
    }

    // Normal conversions.
    Color color = null;
    String str = (obj == null)? null : obj.toString();

    //if (color == null) { color = getColorFromJavaString(str); }  // might be useful if passing String form of Java color
    if (color == null) { color = getLegacyColor(str); }
    if (color == null) { color = NamedColors.getColor(str); }
    // not used: if (color == null) { color = getIntColor(str); }
    if (color == null) { color = Color.getColor(str, DEFAULT_COLOR); }

    return color;
  }

  /** Get a color defined in a String as given by Color.toString. (e.g java.awt.Color[r=0,g=255,b=255])
      @since neXtMidas 2.9.2
   */
  private static Color getColorFromJavaString (String str) {
    if ((str.indexOf("r=")<0) || (str.indexOf("g=")<0) || (str.indexOf("b=")<0)) return null;
    int red   = getIntfromColorString('r',str);
    int green = getIntfromColorString('g',str);
    int blue  = getIntfromColorString('b',str);
    if (str.indexOf("a=")>=0) {   // has an alpha value
      int alpha = getIntfromColorString('a',str);
      return new Color(red,green,blue,alpha);
    }
    return new Color(red,green,blue);
  }

  private static int getIntfromColorString (char color, String str) {
    String tmp = str;
    int startIndex = tmp.indexOf(color+"=");
    String colorString = tmp.substring(startIndex);
    int endIndex = colorString.indexOf(",")<0 ? colorString.indexOf("]") : colorString.indexOf(",");
    int colorInt = Integer.parseInt(colorString.substring(2,endIndex));
    return colorInt;
  }

  /** Get a single color by name or number string.  The recognized color
      names are the java.awt.Color names as well as the SVG standard
      colors: <br>
        TEAL, NAVY, OLIVE, DIMGRAY, BROWN, PURPLE<br>
        The color name may be prefixed with
        DARK - To increase the intensity by one unit <br>
        DEEP - To increase the intensity by three units <br>
        LIGHT - To decrease the intensity by one unit <br>
        BRIGHT - To decrease the intensity by three units <br>
        <br>
        To get a color by number the string must be prefixed with: <br>
        INDEX: - To get a color from the color map by index <br>
        # or 0X - To get a color by HEX representation (6-digit RGB or 8-digit ARGB) <br>
        <br>
        Since 2.7.0 this method supports alpha components on HEX values
     @param s the name or number String
     @return a Java Color object
  */
  private static Color getLegacyColor (String s) {

    Color color; int intense=0;
    s = s.toUpperCase();

    if (s.startsWith("INDEX:")) {
      int index = (s.charAt(6) - '0');
      return getColorByIndex(index);
    }
    if (s.startsWith("#")) s = "0X" + s.substring(1);
    if (s.startsWith("0X")) {
      String alpha = (s.length() > 8)? s.substring(2, 4) : "";
      if ("FF".equals(alpha)) s = "0X"+s.substring(4); // strip off the FF if opaque
      int value = Convert.s2l(s);
      boolean hasAlpha = ("00".equals(alpha) || ((value & 0xFF000000)!= 0) );
      return new Color(value, hasAlpha);
    }
    else if (s.startsWith("X") || s.startsWith("FF")) {
      throw new IllegalArgumentException("Specifying color using hex values must start with '#' or '0X', given: "+s);
// TODO: remove this in 3.5.0+
//      Shell.getMidasContext().deprecate("Since NeXtMidas 2.7.0. Colors must start with '#' or '0X' in future releases");
//      int value = Convert.s2l("0"+s);
//      hasAlpha = (value & 0xFF000000) != 0;
//      return (new Color(value, hasAlpha));
    }
//    else if (s.startsWith("FF")) { // TODO: remove this in 3.5.0+
//      Shell.getMidasContext().deprecate("Since NeXtMidas 2.7.0. Colors must start with '#' or '0X' in future releases");
//      int value = Convert.s2l("0X"+s.substring(2));
//      return (new Color(value));
//    }

    while (s.startsWith("DARK"))   { intense++;  s=s.substring(4); }
    while (s.startsWith("DEEP"))   { intense+=3; s=s.substring(4); }
    while (s.startsWith("LIGHT"))  { intense--;  s=s.substring(5); }
    while (s.startsWith("BRIGHT")) { intense-=3; s=s.substring(6); }

    if      (s.equals("BLACK")) color = Color.black;
    else if (s.equals("BLUE")) color = Color.blue;
    else if (s.equals("CYAN")) color = Color.cyan;
    else if (s.equals("GRAY")) color = Color.gray;
    else if (s.equals("GREEN")) color = Color.green;
    else if (s.equals("MAGENTA")) color = Color.magenta;
    else if (s.equals("ORANGE")) color = Color.orange;
    else if (s.equals("PINK")) color = Color.pink;
    else if (s.equals("RED")) color = Color.red;
    else if (s.equals("WHITE")) color = Color.white;
    else if (s.equals("YELLOW")) color = Color.yellow;
    // Colors match SVG Standard
    else if (s.equals("TEAL")) color = new Color(0,128,128);
    else if (s.equals("NAVY")) color = new Color(0,0,128);
    else if (s.equals("OLIVE")) color = new Color(128,128,0);
    else if (s.equals("DIMGRAY")) color = new Color(105,105,105);
    else if (s.equals("BROWN")) color = new Color(165,42,42);
    else if (s.equals("PURPLE")) color = new Color(128,0,128);

//    else color = Color.getColor(s,Color.white);
    else return null;
    while (intense>0) { intense--; color = color.darker(); }
    while (intense<0) { intense++; color = color.brighter(); }
    return color;
  }

  /** Get a single color from a indexed list
    JGS--are these standard colors?
    @param index the index number in the color list
    @return a Java Color object
  */
  public static Color getColorByIndex (int index) {
    Color color;
    switch (index) {
    case 1:  color = getColor(00,75,00); break;  //darkgreen
    case 2:  color = getColor(25,30,75); break;  //blue
    case 3:  color = getColor(75,50,00); break;  //brown
    case 4:  color = getColor(60,30,50); break;  //purple
    case 5:  color = getColor(00,90,75); break;  //teal
    case 6:  color = getColor(25,85,00); break;  //lightgreen
    case 7:  color = getColor(80,00,00); break;  //red
    default: color = getColor(55,00,00); break;  //dark red
    }
    return color;
  }

  /** Get a colormap by typeString or create one via an array of color names with the specified
      alpha component for transparency. See {@link #getColorMap(String, int)} for additional info.
    @param colormapType Color name in the color map list
    @param numColors Number of colors
    @param alpha transparency value for colors in the colormap (0=transparent, 255=opaque)
    @return an array of Color objects
    @since NeXtMidas 2.5.0
  */
  public static Color[] getColorMap (String colormapType, int numColors, int alpha) {
    Color colormap[] = getColorMap(colormapType, numColors);
    for (int i=0, len=colormap.length; i<len; i++) {
      Color color = MColor.getColor(colormap[i], alpha);
      colormap[i] = color;
    }
    return colormap;
  }

  /** Get a colormap by typeString or create one via an array of color names
    @param s Color name in the color map list
    @param ncolors Number of colors
    @return an array of Color objects
  */
  public static Color[] getColorMap (String s, int ncolors) {
    //if s is a defined color map in colorMapList, find it
    if (s.indexOf(',') < 0) {
      int type = Parser.find(colorMapList,s,0);
      if (type>0) return getColorMap(type, ncolors);
    }
    // if not a defined color map, assume an array of colors needs to be parsed
    if (s.charAt(0)=='"'||s.charAt(0)=='(') s = s.substring(1,s.length()-1);
    Parser p = new Parser(s,true);  // added TRUE in 3.1.0 (Bug 2456) to trim white spaces
    if (ncolors<0) ncolors = p.elements();
    Color[] colors = getColorMap(-1, ncolors); // get an ~empty color array
    for (int i=1; i<=ncolors; i++) {
      String cs = p.get(i);
      if (cs.length()>0) colors[i] = getColor(cs);
      else colors[i] = Color.black;
    }
    return colors;
  }

  /** Get a colormap by typeIndex
    @param type Color map number in {@link #colorMapList}
    @param ncolors Number of colors
    @return a Java Color object
  */
  public static Color[] getColorMap (int type, int ncolors) {
    Color[] colors;
    switch (type) {
    case GRAYSCALE:
      if (ncolors<0) ncolors=DEFCOLORS;
      colors = new Color[ncolors+1];
      interpColors (colors, CM_GRAY);
      break;
    case RAMP:
      if (ncolors<0) ncolors=DEFCOLORS;
      colors = new Color[ncolors+1];
      interpColors (colors, CM_RAMP);
      break;
    case XRAMP:
      if (ncolors<0) ncolors=DEFCOLORS;
      colors = new Color[ncolors+1];
      interpColors (colors, CM_XRAMP);
      break;
    case WHEEL:
      if (ncolors<0) ncolors=DEFCOLORS;
      colors = new Color[ncolors+1];
      interpColors (colors, CM_WHEEL);
      break;
    case TERRAIN:
      if (ncolors<0) ncolors=DEFCOLORS;
      colors = new Color[ncolors+1];
      interpColors (colors, CM_TERRAIN);
      break;
    case MULTI:
      if (ncolors<0) ncolors=DEFCOLORS;
      colors = new Color[ncolors+1];
      interpColors (colors, CM_MULTI);
      break;
    case SPECTRUM:
      if (ncolors<0) ncolors=DEFCOLORS;
      colors = new Color[ncolors+1];
      interpColors (colors, CM_SPECTRUM);
      break;
    case XDATALIST:
      if (ncolors<0) ncolors=16; /* JC4: Changed from 17 to 16 to match XDATALIST (28NOV06).*/
      colors = new Color[ncolors+1];
      interpColors (colors, CM_SPECTRUM);
      break;
    case T5:
      if (ncolors<5) ncolors=5;  // NTN: prevent ArrayOutOfBoundException when ncolors<5
      colors = new Color[ncolors+1];
      colors[C_CEP]     = getColor(30,30,30);
      colors[C_HORIZON] = getColor(00,60,00);
      colors[C_TDOA]    = getColor(80,00,00);
      colors[C_FDOA]    = getColor(80,80,00);
//Pink colors[C_T5000]   = getColor(80,00,80);
      colors[C_T5000]   = getColor(20,100,20);
      break;
    case MAPLINE:
      if (ncolors<8) ncolors=8; // NTN: prevent ArrayOutOfBoundException when ncolors<8
      colors = new Color[ncolors+1];
      colors[C_COASTLINE] = getColor(30,40,55);
      colors[C_POLITICAL] = getColor(45,45,45);
      colors[C_INTERPOL]  = getColor(20,30,30);
      colors[C_RIVER]     = getColor(00,20,50);
      colors[C_RAIL]      = getColor(25,15,15);
      colors[C_ROAD]      = getColor(40,40,00);
      colors[C_POL_OLD] = colors[C_POLITICAL];
      colors[C_INTPOL_OLD]  = colors[C_INTERPOL];
      break;
    case GRID:
      if (ncolors<2) ncolors=2; // NTN: prevent ArrayOutOfBoundException when ncolors<2
      colors = new Color[ncolors+1];
      colors[C_AXIS] = getColor(60,60,60);
      colors[C_GRID] = getColor(25,25,25);
      break;
    default:
      if (ncolors<0) ncolors=1;
      colors = new Color[ncolors+1];
      break;
    }
    colors[0] = Color.black;
    return colors;
  }

  public static String mapLegendList = "CoastLine,Political,InterPol,River,Rail,Road";

  public static Table getMapColorTable() {
    Color[] colors = getColorMap(MAPLINE,-1);
    Table ct = new Table();
    Parser p = new Parser(mapLegendList);
    for (int i=1; i<=p.elements(); i++) {
      String name = p.get(i).toUpperCase();
//      ct.put(name,colors[i]);
      ct.put(name,MColor.toString(colors[i]));
    }
    return ct;
  }

  public static Color[] getColorMap (Table ctab) {
    Table.Iterator ti = ctab.iterator();
    Color[] colors = new Color[ti.size+1];
    colors[0] = Color.black;
    for (int i=0; i<ti.size; i++) colors[i+1] = getColor(ti.values[i].toString());
    return colors;
  }

  /** Takes a given colormap (array of colors) and simply reverse the order of the colors,
      excluding position 0 (since it is reserved for the background in colormap).
      @param cmap ColorMap to be modified. Null returns null
      @return cmap (but also modifies the input array directly)
      @since NeXtMidas 2.8.2
   */
  public static Color[] reverseColorMap (Color[] cmap) {
    return Util.reverseArray(cmap, 1);
  }

  /** Takes a given colormap (array of colors) and inverts the colors, excluding position 0,
      (since it is reserved for the background in colormap). Inverting creates a photo negative
      effect. Doing this twice on a color map will return it back to normal.
      @param cmap ColorMap to be modified. null returns null.
      @return colors (but also modifies the input array directly)
      @since NeXtMidas 2.9.0
   */
  public static Color[] invertColorsMap (Color[] cmap) {
    return invertColors(cmap, 1);
  }

  /** Takes a given colormap (array of colors) and inverts the colors, excluding position 0.
      Inverting creates a photo negative effect. Doing this twice on a color map will
      return it back to normal.
      @param colors ColorMap to be modified. Null returns null
      @param startIndex index (inclusive) into colors to begin reversing the colors.
      @return colors (but also modifies the input array directly)
      @since NeXtMidas 2.9.0
   */
  public static Color[] invertColors (Color[] colors, int startIndex) {
    if (colors==null) return null;
    int len = colors.length;
    for (int ii = startIndex; ii < len; ii++) {
      colors[ii] = invert(colors[ii]);
    }
    return colors;
  }

  /** Inverts the given Color. The alpha (transparency) component will remain the same.
      Inverting creates a photo negative effect. Doing this twice on a Color will
      return it back to normal.
      @param inColor Color to be inverted. null returns null.
      @return new Color representing the inverted color of the input color
      @since NeXtMidas 2.9.0
   */
  public static Color invert (Color inColor) {
    if (inColor==null) return null;
    int argb = inColor.getRGB();
    byte alpha = (byte) ((argb >> 24) & 0xFF);
    byte red   = (byte) ((argb >> 16) & 0xff);
    byte green = (byte) ((argb >>  8) & 0xff);
    byte blue  = (byte) ((argb >>  0) & 0xff);

    // invert the RGB color (keeping the alpha component the same)
    int invertedRGB = (alpha << 24) | ((255-red) << 16) | ((255-green) << 8) | ((255-blue) << 0);

    if (alpha == 0xff) { // solid (opaque) color
      return new Color(invertedRGB);
    } else { // have alpha component (semi-transparent color)
      return new Color(invertedRGB, true);
    }
  }

  /** Interpolate colors from X-Midas color map definitions.
    @param colors The pre-allocated color array (sized to match ncolors+1).
    @param table  The interpolation table (usually copied from X-Midas).
  */
  public static void interpColors (Color[] colors, byte[] table) {
    int i,j,ncolors = colors.length-1;
    float p,pp,pq,r,g,b;
    for (i=1; i<=ncolors; i++) {
      if (ncolors==1) { p = 0.0f; } // BUG 1344: prevent p from being NaN yet still use appropriate interpolation
      else            { p = 100*((float)(i-1))/(float)(ncolors-1); }

      for (j=4; j<table.length && table[j]<(int)p; j+=4);
      pp = (p-(float)table[j-4])/(float)(table[j]-table[j-4]);
      pp = Math.max(0.0F,Math.min(1.0F,pp)); pq = 1-pp;
      r = Math.min(1.0F,(pp*(float)table[j+1] + pq*(float)table[j+1-4])/100);
      g = Math.min(1.0F,(pp*(float)table[j+2] + pq*(float)table[j+2-4])/100);
      b = Math.min(1.0F,(pp*(float)table[j+3] + pq*(float)table[j+3-4])/100);
      colors[i]=new Color(r,g,b);
    }
  }

  /** Get a Java color model. This is used by {@link MImage} when rendering the raster image for
      {@link Layer2D}. Beginning with NeXtMidas 2.5.4, the alpha (transparency) component of the
      color is used in the returned ColorModel.
      @param colors The colors in the color map.
      @return An AWT ColorModel.
  */
  public static ColorModel getColorModel (Color[] colors) {
    int size=colors.length;
    ColorModel cm;
    int bits = 8;  // we always use a byte array for the map
    byte[] r = new byte[size];
    byte[] g = new byte[size];
    byte[] b = new byte[size];
    byte[] a = new byte[size];
    for (int i=0; i<size; i++) {
      r[i] = (byte)colors[i].getRed();
      g[i] = (byte)colors[i].getGreen();
      b[i] = (byte)colors[i].getBlue();
      a[i] = (byte)colors[i].getAlpha();
    }
    cm = new IndexColorModel(bits,size,r,g,b,a);
    return cm;
  }

  /** Get a mixed color. This is frequently used when shading.
    @param c0  The first Java Color object to be mixed.
    @param c1  The second Java Color object to be mixed.
    @param mix The percentage of c0 to use (<code>1.00-mix</code> is the percentage used for c1).
               The acceptable range of mix is <code>0.00 - 1.00</code>.  If the value is outside this
               range it is clipped.
    @return A color between c0 and c1.
  */
  public static Color getColorMix (Color c0, Color c1, double mix) {
         if (mix > 1) mix=1;
    else if (mix < 0) mix=0;

    double mixi = 1.0-mix;
    int r = (int)(c0.getRed()*mixi   + c1.getRed()*mix);
    int g = (int)(c0.getGreen()*mixi + c1.getGreen()*mix);
    int b = (int)(c0.getBlue()*mixi  + c1.getBlue()*mix);
    return new Color (r,g,b);
  }

//  /** Gets a color based on it RGB color code. It is assumed that the color is opaque unless its
//      value is larger than <code>0xFFFFFF</code> or specified using the "#aarrggbb" syntax.
//    @param color The integer string for the color.
//    @return The color matching the given string or <code>null</code> if the string is invalid.
//  */
//  private static Color getIntColor (String color) {
//    Color colorObj = null;
//    if (color.startsWith("#") || color.toUpperCase().startsWith("0X")) {
//      if (color.startsWith("#")) color = "0x"+color.substring(1);
//      colorObj = new Color(Convert.o2l(color), (color.length() > 8));
//    } else if (StringUtil.isNumber(color)) {
//      int val = Convert.o2l(color);
//      //colorObj = new Color(val, (val > 0xFFFFFF));  // Old way
//      colorObj = new Color(val, (val & 0xFF000000) != 0 );
//    }
//    return colorObj;
//  }

  /** Converts a color to it's string form.
      If the color matches a named color that name will be returned.
      otherwise the Hex form of the color will be returned. If the color is
      OPAQUE, the hex form will be six digits, otherwise it will be eight.
      <br>
      Examples (color arg shown in hex):
      <pre>
        toString(ff888888)   --&gt; "GREY"
        toString(ff123456)   --&gt; "#123456"     (OPAQUE      - no alpha)
        toString(7f123456)   --&gt; "#7f123456"   (Semi-OPAQUE - has alpha)
      </pre>
      <br>

    @param color The color.
    @return The string form of the color (or <code>"null"</code> if the color is null).
  */
  public static String toString (Color color) {
    if (color == null) { return "null"; }

    String str = NamedColors.getColorName(color);  // Check if named color

    /** If not named get the RGB value and strip out the Alpha (if OPAQUE) */
    if (str == null) {
      int    rgb = color.getRGB();
      String hex = Integer.toHexString(rgb);
      hex = StringUtil.padLeft(hex, 8, '0');
      if ((rgb & 0xFF000000) == 0xFF000000) {
        hex = hex.substring(2); // Strip off the Alpha
      }
      str = "#" + hex;
    }
    return str;
  }

  /** Converts a color map to it's string form. If the color matches a named color that
      name will be returned, otherwise the Hex form of the color will be returned.
    @param colors The colors.
    @return The string form of the color map.
  */
  public static String toString (Color[] colors) {
    StringBuffer str = new StringBuffer();

    for (int i = 0; i < colors.length; i++) {
      if (i != 0) { str.append(','); }
      str.append(toString(colors[i]));
    }

    return str.toString();
  }

  /** Adjusts the image for brightness, contrast, alpha, and greyscale.
      @param img        The image to adjust
      @param xMin       The lower x value for drawing the image
      @param yMin       The lower y value for drawing the image
      @param xMax       The upper x value for drawing the image
      @param yMax       The upper y value for drawing the image
      @param contrast   The contrast.   Range is 0..100 (50 means no change).
      @param brightness The brightness. Range is 0..100 (50 means no change).
      @param alpha      The 8-bit alpha value to change. Range is 0x00 to 0xFF (-1 means no change).
      @param greyscale  Convert to greyscale (false means no change).
      @return The adjusted image (this may be a *new* image or may just be a reference to the one
              passed in if no changes were required.
      @since NeXtMidas 2.7.0
      @see #adjustColor(int,int,int,int,boolean)
   */
  public static BufferedImage adjustImage (BufferedImage img, int xMin, int yMin, int xMax, int yMax,
                                           int contrast, int brightness, int alpha, boolean greyscale) {
    if ((contrast == 50) && (brightness == 50) && (alpha == -1) && !greyscale) {
      return img; // No changes needed.
    }
    // JC4: I tried several of the image conversion tools in AWT, but all were sloooooooooooooow
    //      so I came up with this algorithm
    int height = yMax - yMin;
    int width  = xMax - xMin;
    BufferedImage out = new BufferedImage(width,height,BufferedImage.TYPE_INT_ARGB);

    int[] buf = img.getRGB(xMin, yMin, width, height, new int[width*height], 0, width);
    float cnt = (contrast   / 50f); // Multiplier for contrast
    float brt = (brightness / 50f); // Multiplier for brightness
    int   alp = (alpha == -1)? -1 : Math.min(0xFF, Math.max(0,alpha));
    for (int i = 0; i < buf.length; i++) {
      buf[i] = adjustColor0(buf[i], cnt, brt, alp, greyscale);
    }
    out.setRGB(xMin, yMin, width, height, buf, 0, width);
    return out;
  }

  /** Adjust the brightness, contrast, and alpha of a given aRGB value.
      <pre>
      Note 1: When repeatedly adjusting contrast and brightness without refreshing the original RGB values,
              changing the brightness or contrast of a R,G, or B component above the max (255)
              or below the min (0) will effectively "lose" the color, since that component
              will stop incrementing but the other components will not stop until they
              themselves hit the max/min. See LayerShape for sample implementation.
      Note 2: Adjusting the alpha value from either 0 (none) or 255 (solid) results in a
              significant performance impact due to the graphics blending required.
      </pre>
      @param rgb        The aRGB value.
      @param contrast   The contrast.   Range is 0..100 (50 means no change).
      @param brightness The brightness. Range is 0..100 (50 means no change).
      @param alpha      The 8-bit alpha value to change. Range is 0x00 to 0xFF (-1 means no change).
      @param greyscale  Convert to greyscale (false means no change).
      @return The 32 bit aRGB value
      @since NeXtMidas 2.7.0
   */
  public static int adjustColor(int rgb, int contrast, int brightness, int alpha, boolean greyscale) {
    float cnt = (contrast   / 50f); // Multiplier for contrast
    float brt = (brightness / 50f); // Multiplier for brightness
    int   alp = (alpha == -1)? -1 : Math.min(0xFF, Math.max(0,alpha));
    if ((contrast == 50) && (brightness == 50) && (alpha == -1) && !greyscale) return rgb; // No conversion required

    return adjustColor0(rgb, cnt, brt, alp, greyscale);
  }

  /** Internal version of adjustColor but without some of the checks and converts that can come
      outside of the loop in adjustImage(..).
   */
  private static int adjustColor0 (int rgb, float cnt, float brt, int alp, boolean greyscale) {
    int alpha = (rgb >> 24) & 0xFF;
    int red   = (rgb >> 16) & 0xff;
    int green = (rgb >>  8) & 0xff;
    int blue  = (rgb >>  0) & 0xff;
    int aa    = (alp != -1)? alp : alpha; // alpha (adjusted)
    int rr;                               // red   (adjusted)
    int gg;                               // green (adjusted)
    int bb;                               // blue  (adjusted)

    if (greyscale) {
      int grey = (red + green + blue) / 3;
      int gs = (int)((((grey * brt) - 128) * cnt) + 128); // grey scale

      rr = gg = bb = Math.min(0xFF, Math.max(0,gs));
    }
    else {
      // jwb - 18 Jun 08 - changed algorithm to apply brightness changes first per advice online
      int r = (int)((((red   * brt) - 128) * cnt) + 128); // red
      int g = (int)((((green * brt) - 128) * cnt) + 128); // green
      int b = (int)((((blue  * brt) - 128) * cnt) + 128); // blue

      rr = Math.min(0xFF, Math.max(0,r));
      gg = Math.min(0xFF, Math.max(0,g));
      bb = Math.min(0xFF, Math.max(0,b));
    }
    return (aa << 24) | (rr << 16) | (gg << 8) | (bb << 0);
  }

  /** Apply the specified transparency (alpha) component to every element in the image buffer array
      containing ARGB color values.
      @param rgbBuf Image array of ARGB color (32 bit) color values to have it's alpha value changed
      @param alpha  The 8-bit alpha value to change for all colors in the input imageBuf array (this
                  takes in an int rather than a byte to avoid issues with alpha being unsigned).
      @since NeXtMidas 2.7.0
   */
  public static void applyAlpha (int[] rgbBuf, int alpha) {
    if (rgbBuf==null || rgbBuf.length==0) return;
    int a = (alpha & 0xFF) << 24;
    for (int i=0, len=rgbBuf.length; i<len; i++) {
      rgbBuf[i] = a | (rgbBuf[i] & 0xFFFFFF);
    }
  }

  /** Blend two colors.<br>
      Note: To preserve transparency, blending is NOT done on the Alpha component
      @param color1  First color to blend.
      @param color2  Second color to blend.
      @param ratio   Blend ratio. 0.5 will give even blend, 1.0 will return
                     color1, 0.0 will return color2 and so on.
      @return        Blended color.
      @since         NeXtMidas 2.7.1
   */
  public static Color blend (Color color1, Color color2, double ratio) {
    float r  = (float) ratio;
    float ir = (float) 1.0 - r;

    float rgb1[] = new float[4];
    float rgb2[] = new float[4];

    color1.getComponents(rgb1);
    color2.getComponents(rgb2);

    Color color = new Color (rgb1[0] * r + rgb2[0] * ir,
                             rgb1[1] * r + rgb2[1] * ir,
                             rgb1[2] * r + rgb2[2] * ir,
                             rgb1[3]);

    return color;
  }

  /** Make an even blend between two colors. <br>
      Note: To preserve transparency, blending is NOT done on the Alpha component.
      @param color1     First color to blend.
      @param color2     Second color to blend.
      @return       Blended color.
      @since        NeXtMidas 2.7.1
   */
  public static Color blend (Color color1, Color color2) {
    return blend (color1, color2, 0.5);
  }

  public static BufferedImage colorImage (BufferedImage img, Color color, float ratio) {
    if (color != null && img != null) {
      int h = img.getHeight();
      int w = img.getWidth();
      int[] rgb = new int[h * w];
      img.getRGB(0, 0, w, h, rgb, 0, w);
      for (int xx=0; xx< rgb.length; xx++) {
        rgb[xx] = blend(new Color(rgb[xx], true), color, ratio).getRGB();
      }
      img.setRGB(0,0,w,h,rgb,0,w);
    }
    return img;
  }

  /* This is not for public use outside the NeXtMidas framework.
     This method captures 'special'  color numbers that can be in 1998, 1999 or 5000 files.
     X-Midas XGEOPLOT keys off these special numbers to adds 5 before calling mx$draw_lines  A DR
     (g765978-1) was written in Jan 2008 to document all special XGEOPLOT color numbers. This does
     not yet do all the special handling documented in this DR to include the COLORMAP=RAW
     COLORMAP=USER and foreground and background colors where I have chosen white or black.

     See docs in XM 4.5.5. or later.
  */
  static java.awt.Color getSpecialXMidasColor (int colNum) {
    java.awt.Color color = java.awt.Color.white;
    if (colNum < -6 ) {
      color = Color.white;  // X-Midas says choose foreground color
    } else {
      switch (colNum) {
        case -6:
          color = Color.white;
          break;
        case -5:
          color = Color.black;
          break;
        case -4:
          color = Color.gray;
          break;
        case -3:
          color = Color.green;
          break;
        case -2:
          color = Color.red;
          break;
        case -1:
          color = Color.yellow;
          break;
        case 0:
          color = new Color(128,0,128); // purple
          break;
        default:
          color = Color.white;
          break;
      }
    }
    return color;
  }

  /****************************************************************************/
  /* NAMED-COLOR DEFINITION CLASS                                             */
  /****************************************************************************/

  /** This class holds the standard definitions for named colors. <br>
      <br>
      The following color sets are available:
      <pre>
        Netscape 7.1 Colors - The 133 SVG colors recognized by Netscape.
        SVG Standard Colors - All 147 SVG colors (some of which are duplicates).
      </pre>

      The list of Netscape colors was obtained from a chart published at
      <a href="http://mindprod.com/jgloss/netscapecolours.html">
      http://mindprod.com/jgloss/netscapecolours.html</a>. The additional
      14 SVG colors were manually added from various sources.
    @author J. Czechowski IV
    @since  NeXtMidas 1.9.2
  */
  private static class NamedColors {
    /** Gets the color for the given color name from the most inclusive list
        of color names. <br>
        <br>
        Currently this is the same as using {@link #getSvgColor(String)} but
        may change in the future if more colors are added. <br>
        <br>
        Note that the named colors are case- and whitespace-insensitive (e.g.
        "Slate&nbsp;Blue", "slateblue", and "SlateBlue" all represent the same
        color).
      @param name The name of the color.
      @return The color associated with the given name or <code>null</code>
              if no color matches.
    */
    public static Color getColor(String name) {
      return getSvgColor(name);
    }

    /** Gets the color name for the given color from the most inclusive list
        of color names. <br>
        <br>
        Currently this is the same as using {@link #getSvgColorName(Color)} but
        may change in the future if more colors are added.
      @param color The color.
      @return The color name associated with the given color or <code>null</code>
              if no match is found.
    */
    public static String getColorName(Color color) {
      return getSvgColorName(color);
    }

    /** Gets the color for the given SVG color name. <br>
        <br>
        This method is case insensitive and ignores any white space.
      @param name The name of the color.
      @return The color associated with the given name or <code>null</code>
                if no color matches.
    */
    public static Color getSvgColor(String name) {
      Color color = getNetscapeColor(name);

      if (color == null) {
        color = getColor(name, SVG_COLOR_NAMES);
      }

      return color;
    }

    /** Gets the color name for the given SVG color.
      @param color The color.
      @return The color name associated with the given color or <code>null</code>
              if no match is found.
    */
    public static String getSvgColorName(Color color) {
      String name = getColorName(color, NS_COLOR_NAMES);

      if (name == null) {
        name = getColorName(color, SVG_COLOR_NAMES);
      }

      return name;
    }

    /** Gets the color for the given Netscape color name. <br>
        <br>
        This method is case insensitive and ignores any white space.
      @param name The name of the color.
      @return The color associated with the given name or <code>null</code>
              if no color matches.
    */
    public static Color getNetscapeColor(String name) {
      return getColor(name, NS_COLOR_NAMES);
    }

    /** Gets the color name for the given Netscape color.
      @param color The color.
      @return The color name associated with the given color or <code>null</code>
                if no match is found.
    */
    public static String getNetscapeColorName(Color color) {
      return getColorName(color, NS_COLOR_NAMES);
    }

    /** Gets the color for the given color name. <br>
        <br>
        This method is case insensitive and ignores any white space.
      @param name     The name of the color.
      @param colorSet The set of colors to use.
      @return The color associated with the given name or <code>null</code>
                if no color matches.
    */
    protected static Color getColor(String name, String[] colorSet) {
      if (name == null) {
        return null;
      }

      // Remove all spaces and make the name lower case.
      //   JC4: Note that WHITESPACE_REGEX.matcher(str).replaceAll("") is much faster than using
      //        name.replaceAll("\\s","") because the later one recompiles the regex every time.
      name = name.toLowerCase();
      name = WHITESPACE_REGEX.matcher(name).replaceAll("");

      for (int i = 1; i < colorSet.length; i+=2) {
        if (name.equals(colorSet[i])) {
          try {
            final int rgb = Integer.parseInt(colorSet[i-1], 16);

            return new Color(rgb);

          } catch(NumberFormatException e) {
            /* This should never happen. */
            throw new RuntimeException("Unable to initialize color: " + e);
          }
        }
      }

      return null; /* Name Not Found */
    }

    /** Gets the color name for the given color.
      @param color    The color.
      @param colorSet The set of colors to use.
      @return The color name associated with the given color or <code>null</code>
              if no match is found.
    */
    protected static String getColorName(Color color, String[] colorSet) {
      if (color == null) {
        return null;
      }

      final int colorRGB = color.getRGB(); /* The RGB form of the color searching for. */

      try {
        for (int i = 0; i < colorSet.length; i+=2) {
          int rgb = Integer.parseInt(colorSet[i], 16) + ((colorSet[i].length()<=6)? 0xFF000000: 0);

          if (colorRGB == rgb) {
            return colorSet[i+1];
          }
        }
      } catch(NumberFormatException e) {
        /* This should never happen. */
        throw new RuntimeException("Unable to initialize color: " + e);
      }

      return null; /* Name Not Found */
    }

    /** The 14 SVG color names NOT recognized by Netscape&nbsp;7.1. */
    private static final String[] SVG_COLOR_NAMES = new String[] {
      "00ffff", "cyan",                 "a9a9a9", "darkgrey",
      "2f4f4f", "darkslategrey",        "696969", "dimgrey",
      "ffd700", "gold",                 "808080", "grey",
      "d3d3d3", "lightgrey",            "778899", "lightslategrey",
      "cd853f", "peru",                 "ffc0cb", "pink",
      "dda0dd", "plum",                 "708090", "slategrey",
      "fffafa", "snow",                 "d2b48c", "tan" };

    /** The 133 color names recognized by Netscape&nbsp;7.1. */
    private static final String[] NS_COLOR_NAMES = new String[] {
      "f0f8ff", "aliceblue",            "faebd7", "antiquewhite",
      "00ffff", "aqua",                 "7fffd4", "aquamarine",
      "f0ffff", "azure",                "f5f5dc", "beige",
      "ffe4c4", "bisque",               "000000", "black",
      "ffebcd", "blanchedalmond",       "0000ff", "blue",
      "8a2be2", "blueviolet",           "a52a2a", "brown",
      "deb887", "burlywood",            "5f9ea0", "cadetblue",
      "7fff00", "chartreuse",           "d2691e", "chocolate",
      "ff7f50", "coral",                "6495ed", "cornflowerblue",
      "fff8dc", "cornsilk",             "dc143c", "crimson",
      "00008b", "darkblue",             "008b8b", "darkcyan",
      "b8860b", "darkgoldenrod",        "a9a9a9", "darkgray",
      "006400", "darkgreen",            "bdb76b", "darkkhaki",
      "8b008b", "darkmagenta",          "556b2f", "darkolivegreen",
      "ff8c00", "darkorange",           "9932cc", "darkorchid",
      "8b0000", "darkred",              "e9967a", "darksalmon",
      "8fbc8f", "darkseagreen",         "483d8b", "darkslateblue",
      "2f4f4f", "darkslategray",        "00ced1", "darkturquoise",
      "9400d3", "darkviolet",           "ff1493", "deeppink",
      "00bfff", "deepskyblue",          "696969", "dimgray",
      "1e90ff", "dodgerblue",           "b22222", "firebrick",
      "fffaf0", "floralwhite",          "228b22", "forestgreen",
      "ff00ff", "fuchsia",              "dcdcdc", "gainsboro",
      "f8f8ff", "ghostwhite",           "daa520", "goldenrod",
      "808080", "gray",                 "008000", "green",
      "adff2f", "greenyellow",          "f0fff0", "honeydew",
      "ff69b4", "hotpink",              "cd5c5c", "indianred",
      "4b0082", "indigo",               "fffff0", "ivory",
      "f0e68c", "khaki",                "e6e6fa", "lavender",
      "fff0f5", "lavenderblush",        "7cfc00", "lawngreen",
      "fffacd", "lemonchiffon",         "add8e6", "lightblue",
      "f08080", "lightcoral",           "e0ffff", "lightcyan",
      "fafad2", "lightgoldenrodyellow", "90ee90", "lightgreen",
      "d3d3d3", "lightgrey",            "ffb6c1", "lightpink",
      "ffa07a", "lightsalmon",          "20b2aa", "lightseagreen",
      "87cefa", "lightskyblue",         "778899", "lightslategray",
      "b0c4de", "lightsteelblue",       "ffffe0", "lightyellow",
      "00ff00", "lime",                 "32cd32", "limegreen",
      "faf0e6", "linen",                "ff00ff", "magenta",
      "800000", "maroon",               "66cdaa", "mediumaquamarine",
      "0000cd", "mediumblue",           "ba55d3", "mediumorchid",
      "9370db", "mediumpurple",         "3cb371", "mediumseagreen",
      "7b68ee", "mediumslateblue",      "00fa9a", "mediumspringgreen",
      "48d1cc", "mediumturquoise",      "c71585", "mediumvioletred",
      "191970", "midnightblue",         "f5fffa", "mintcream",
      "ffe4e1", "mistyrose",            "ffe4b5", "moccasin",
      "ffdead", "navajowhite",          "000080", "navy",
      "fdf5e6", "oldlace",              "808000", "olive",
      "6b8e23", "olivedrab",            "ffa500", "orange",
      "ff4500", "orangered",            "da70d6", "orchid",
      "eee8aa", "palegoldenrod",        "98fb98", "palegreen",
      "afeeee", "paleturquoise",        "db7093", "palevioletred",
      "ffefd5", "papayawhip",           "ffdab9", "peachpuff",
      "b0e0e6", "powderblue",           "800080", "purple",
      "ff0000", "red",                  "bc8f8f", "rosybrown",
      "4169e1", "royalblue",            "8b4513", "saddlebrown",
      "fa8072", "salmon",               "f4a460", "sandybrown",
      "2e8b57", "seagreen",             "fff5ee", "seashell",
      "a0522d", "sienna",               "c0c0c0", "silver",
      "87ceeb", "skyblue",              "6a5acd", "slateblue",
      "708090", "slategray",            "00ff7f", "springgreen",
      "4682b4", "steelblue",            "008080", "teal",
      "d8bfd8", "thistle",              "ff6347", "tomato",
      "40e0d0", "turquoise",            "ee82ee", "violet",
      "f5deb3", "wheat",                "ffffff", "white",
      "f5f5f5", "whitesmoke",           "ffff00", "yellow",
      "9acd32", "yellowgreen" };
  }
}
