package nxm.sys.libg;

import nxm.sys.inc.*;
import nxm.sys.lib.*;
import nxm.sys.libg.*;

import java.awt.*;
import java.awt.event.*;
import java.io.IOException;
import java.net.URL;

import javax.swing.JEditorPane;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;

/** A class to handle an editable text area.  Extends JEditorPane to handle mouse clicks.
 
    @author  Jeff Schoen
    @version $Id: MJEditor.java,v 1.10 2004/10/13 21:56:51 jxc Exp $
*/
public class MJEditor extends JEditorPane implements MessageHandler, HyperlinkListener {

  /** The message handler to use. */
  private MessageHandler mh;
  private MWindow mw;
  private URL curl;

  /** Constructor which uses the default number of rows/columns.  */
  public MJEditor() {
    this (null, 20, 80);
  }

  /** Constructor which uses the default scrollbar policy.
      @param text The text to initially display.
      @param rows The number of rows to initially display.
      @param cols The number of columns to initially display.
   */
  public MJEditor (String text, int rows, int cols) { 
    super();
    open (text, rows, cols);
  }

  /** Constructor helper, sets text area attributes.
      @param text The initial text to display.
      @param rows The number of rows to initially display.
      @param cols The number of columns to initially display.
   */
  public void open (String text, int rows, int cols) {

    // Create the output text area
    super.setEditable(true);
    super.setEnabled(true);

    // Get the fontsize
    Font font = new Font ("monospaced", Font.PLAIN, 12);
    super.setFont(font);
    if (text!=null) setText(text);

    addHyperlinkListener(this);

  }

  public void setFile (String url) {
    if (url.indexOf(':')<0) url = "file:"+url;
    try { setPage( url ); }
    catch (Exception e) { System.out.println("MJEdit: "+e); }
  }
  
  public void processEvent (AWTEvent e) {
    if (e.getID()==MouseEvent.MOUSE_RELEASED) {
      MouseEvent me = (MouseEvent)e;
      String msgName = null;
      if ( (me.getModifiers()&MouseEvent.BUTTON2_MASK)!=0 ) { 
        msgName = "MENU";
      }
      else if ( (me.getModifiers()&MouseEvent.BUTTON3_MASK)!=0 ) {
        msgName = "MENU";
      }
      if (mh!=null && msgName!=null) {
        if (mw!=null) {
          mw.px = me.getPoint().x;
          mw.py = me.getPoint().y;
        }
        mh.processMessage(new Message (msgName,0,null,mh,this));
      }
    }
    else if (e.getID()==MouseEvent.MOUSE_PRESSED) {
      MouseEvent me = (MouseEvent)e;
      if ((me.getModifiers()&MouseEvent.BUTTON2_MASK) != 0) {
        /* This is to prevent the automatic middle-button paste that
           typically happens when the middle mouse button is clicked
           within X-Windows.  */
        ((MouseEvent)e).consume();
      }
    }
    super.processEvent(e);
  }

  public void close() {

  }

  public void hyperlinkUpdate (HyperlinkEvent he) {
    if (he.getEventType() == HyperlinkEvent.EventType.ACTIVATED) {
      URL nurl = he.getURL();
      if (nurl.equals(curl)) return;
      try { setPage(nurl); curl=nurl; }
      catch (Exception e) { System.out.println("MJLink: "+e); }
    }
  }

  /** Set attributes to those of an MWindow.  For example, use
      <code>setAttributes(panel)</code> to set the foreground and background
      colors of this text area to match those specified in the theme for <code>panel</code>.
      @param mw The MWindow to mimic the attributes of.
      @see #setAttributes(Container)
      @see #setAttributesGW(GWidget)
   */
  public void setAttributes (MWindow mw) {
    super.setBackground(mw.theme.cbg);
    super.setForeground(mw.theme.cfg);
  }
  
  /** Set attributes to those of a GWidget.  For example, use
      <code>setAttributes(panel)</code> to set the foreground and background
      colors of this text area to match those specified in the theme for
      <code>panel</code>.
   
      @param mw The GWidget to mimic the attributes of.
   
      @see #setAttributes(MWindow)
      @see #setAttributes(Container)
      @see #setAttributesGW(GWidget)
   */
  public void setAttributesGW (GWidget mw) {
    super.setBackground(mw.theme.cwbs);
    super.setForeground(mw.theme.cwfg);
  }
  
  /** Set attributes to those of a Container.  For example, use
      <code>setAttributes(panel)</code> to set the foreground and background
      colors of this text area to match those of the Component, <code>panel</code>.
      @param panel The Container to mimic the attributes of.
      @see #setAttributes(MWindow)
      @see #setAttributesGW(GWidget)
   */
  public void setAttributes (Container panel) {
    Color cb = panel.getBackground();
    super.setBackground(cb);
    Color cf = panel.getForeground();
    super.setForeground(cf);
  }

  /** Set the message handler for the instance.
      @param mh  The message handler to use.
      @todo (JPH ???? Why do I need this???) 
  */
  public void setMessageHandler (MessageHandler mh) {
    this.mh = mh;
  }
  public void setMW (MWindow mw) {
    this.mw = mw;
  }
  
  /** {@inheritDoc} */
  public int processMessage (Message msg) {
    return processMessage (msg,null,null);
  }

  /** Process a message from parent process or config menu.
   
      @param msg    The message to process.
      @param msgid  The message ID (or <code>null</code> if not known).
      @param parent The message's parent (or <code>null</code> if not known).
   
      @return <code>1</code> on success, <code>-1</code> on failure.
   
      @todo Remove this method since it is not useful.
   */
  public int processMessage (Message msg, Object msgid, Object parent) {

    int stat=1;	/* assume its for me */

    if (msg.name.equals("MTEXT.CFG")) {
    }
    else if (msg.name.equals("AXIS")) {
    }
    else if (msg.name.equals("MODE")) {
    }
    else if (msg.name.equals("SCALE")) {
    }
    else if (msg.name.equals("SCALERANGE")) {
    }
    else if (msg.name.equals("OPTIONS")) {
    }
    else {
      stat=-1;
    }
    return stat;
  }

}
