package nxm.sys.libg;

import nxm.sys.inc.MidasFrame;
import nxm.sys.lib.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

/** A class to run a graphical Midas Panel. This class is implemented using
    swing components while {@link MPanel} uses AWT. <br>
    <br>
    Additionally (since NeXtMidas 1.9.2) this class has the ability to act as a
    {@link MidasFrame} which is used to enable the creation of swing-based Java
    bean components (see {@link MidasBean}).

    @see MWindow
    @see MPanel
    @see MidasBean

    @author  Jeff Schoen
    @version $Id: MJPanel.java,v 1.15 2012/02/10 19:49:53 ntn Exp $
 */
public class MJPanel extends JPanel implements MidasFrame {
  /** NOTE: this should be changed if the structure/fields of this class changes. */
  private static final long serialVersionUID = 2012021080311L; // for Serializable interface

  /** the MWindow associated with this panel */
  public MWindow MW;
  private String title;
  private boolean needrefresh=false;

  /** Instantiate a new MJPanel with a null layout. */
  public MJPanel() {
    this((String)null);
  }

  /** Instantiate a new MJPanel with the given registry name.
      @param name the registry name for this panel
   */
  public MJPanel (String name) {
    super();
    setLayout(null);
    setOpaque(true);
    setTitle(name);
  }

  /** Sets the title */
  void setTitle (String t) {  title = t; }  // package private only used in libg

  /** Gets the title */
  String getTitle () { return title; }      // package private only used in libg

  /** Sets the MWindow associated with this panel
      @param mw the new MWindow
   */
  public void setMWindow (MWindow mw) {
    MW = mw;
    if (needrefresh) MW.process ("REFRESH",0,null);
    needrefresh = false;
  }

  public MWindow getMWindow () {
    return MW;
  }

  public void update (Graphics g) {
    paint(g);
  }

  public void paintComponent (Graphics g) {

    if (MW == null) paintComponentBasic(g, title);
    else            MW.process("PAINT", 0, g);
  }

  /** Paints just a basic title on the component. This is used when waiting for
      it to attach.
      @param g The Graphics
      @param text The title
   */
  void paintComponentBasic (Graphics g, String text) {
    Dimension d = getSize();
    g.setClip(0, 0, d.width, d.height);
    g.clearRect(0, 0, d.width, d.height);
    g.drawRect(4, 4, d.width-8, d.height-8);
    if (text != null) g.drawString(text, 10, 20);
    needrefresh = true;
  }

// JC4: I don't think this is ever called... and it does not work for MBox...
//  /** Sets the bounds based on an arbitrary object.
//      @param obj The object which specified the bounds.
//   */
//  public void setBounds (Object obj) {
//    if (obj instanceof MBox) {
//      MBox pos = (MBox)obj;
//    }
//    else if(obj instanceof Rectangle) {
//      super.setBounds((Rectangle)obj);
//    }
//    else {
//      System.err.println("MJPanel: Can not set bounds based on a " + obj.getClass().getName());
//    }
//  }
//
  public void setBounds (MBox pos) {
    setBounds(pos.x, pos.y, pos.w, pos.h);
  }

  public void resizePanel() {
   // Do Nothing
  }

  public void display (int mode) {
    // Do Nothing
  }

  public void componentMoved (ComponentEvent e) {
    // Do nothing
  }

  public void componentResized (ComponentEvent e) {
    // Do nothing
  }

  public void componentShown (ComponentEvent e) {
    // Do nothing
  }

  public void componentHidden (ComponentEvent e) {
    // Do nothing
  }

  public void windowOpened (WindowEvent e) {
    if (MW != null) MW.process("SHOWN", 1);
  }

  public void windowActivated (WindowEvent e) {
    // Do nothing
  }

  public void windowDeactivated (WindowEvent e) {
    // Do nothing
  }

  public void windowIconified (WindowEvent e) {
    if (MW != null) MW.process("SHOWN", -1);
  }

  public void windowDeiconified (WindowEvent e) {
    repaint();
    if (MW!=null) MW.process("SHOWN",1);
  }

  public void windowClosing (WindowEvent e) {
    if (MW != null) MW.process("EXIT",0);
  }

  public void windowClosed (WindowEvent e) {
    if (MW != null) {
      for (int ic=0; ic<MW.panel.getComponentCount(); ic++) {
        Object sub = MW.panel.getComponent(ic);
        if (sub instanceof MPanel) {
          ((MPanel)sub).MW.process("SHOWN",-1);
        }
      }
    }
  }

  public Graphics getGraphics() {
    Graphics g = super.getGraphics();
    // Workaround for a possible bug in JPanel
    if (g instanceof Graphics2D) {
      ((Graphics2D)g).setBackground(getBackground());
    }
    return g;
  }
}
