package nxm.sys.libg;

/** A class to define a Pixel Buffer for {@link MPlot}, {@link View}, and {@link Layer}. This is
    used for vector line drawing and is not a replacement a buffered image.

    @author Jeff Schoen
    @version $Id: Pix.java,v 1.6 2011/02/03 15:34:46 ntn Exp $
 */
public class Pix extends Object {

  /** Pixel conversion flag used exclusively by {@link MPlot} and subclasses of {@link View}.
      <pre>
        CHECK  - Use special handling for non-visible points in {@link ViewGeo}.
        CLIP   - Use automatic bounds clipping with {@link View} (handles wrapping
                 at -180/180 for LatLon and Mercator views).
        HIDE   - <i>Not Used</i>
        NORM   - Use special normalization for drawing grid in XYZ view.
      </pre>
   */
  public static final int CHECK=0x1, CLIP=0x2, HIDE=0x4, NORM=0x8;

  public int max,off,n,flags;
  public int[] x,y;
  public float[] z;

  // This is package-private, only used by MPlot and View
  boolean save;               // Any saved?
  int[]   sx = new int[2];    // Save values X
  int[]   sy = new int[2];    // Save values Y
  float[] sz = new float[2];  // Save values Z

  public Pix (int maxsize, int flags) {
    max = maxsize+1;
    x = new int[max];
    y = new int[max];
    z = new float[max];
    setFlags(flags);
  }

  /** return how many points to do this pass and offset for next pass */
  public int setSize (int npts) {
    if (npts>max) { n=max; off=max-1; }
    else { n=npts; off=npts; }
    return n;
  }
  public void setFlags (int mask) { flags = mask; }
  public void setFlag (int mask) { flags |= mask; }
  public void clrFlag (int mask) { flags &= (~mask); }

  public int[] getXYMinMax(int num) {
    int r[] = {0,0,0,0};
    return (getXYMinMax(r,num));
  }

  public int[] getXYMinMax(int[] ini, int num) {
    int xmin,xmax,ymin,ymax;
    xmin = ini[0];
    ymin = ini[1];
    xmax = ini[2];
    ymax = ini[3];

    for (int i = 0; i < num; i++) {
      xmax = Math.max(xmax, x[i]);
      xmin = Math.min(xmin, x[i]);
      ymax = Math.max(ymax, y[i]);
      ymin = Math.min(ymin, y[i]);
    }

    int[] retVal = new int[4];
    retVal[0] = xmin;
    retVal[1] = ymin;
    retVal[2] = xmax;
    retVal[3] = ymax;

    return (retVal);
  }

  /** Creates a string of the object. This is useful for debugging purposes. The format of the
      string returned by this method may change in future versions.

      @return A string describing the object.
   */
  public String toString() {
    final StringBuffer str = new StringBuffer();

    str.append("Pix: {");

    for(int i = 0; i < n; i++) {
      str.append(" (");
      str.append(x[n]);
      str.append(',');
      str.append(y[n]);
      str.append(',');
      str.append(z[n]);
      str.append(')');
    }
    str.append(" }");

    return str.toString();
  }
}
