package nxm.sys.libxm;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import nxm.sys.lib.*;
import java.lang.reflect.*;

public class XControl implements ActionListener, ChangeListener {
  public Component myComp;
  public String name;
  public int actionMsgID = 1;
  public int changeMsgID = 1;

  public XControl() {
     this(null);
  }

  public XControl(String cname) {
    setComponent(cname);
  }

  public void setComponent(String cname) {
    name = null;
    myComp = null;

    if (cname == null) {
      return;
    }

    Midas midas = Shell.getMidasContext();
    Object comp = midas.registry.get(cname);
    if (comp == null || !(comp instanceof Component)) {
      midas.warning("Could not find Component: " + cname);
      return;
    }
    myComp = (Component)comp;
    name = cname;
  }

  public void addActionListener (int id) {
    if (myComp == null) {
      Shell.getMidasContext().warning("No Component to listen");
      return;
    }
    try {
      Method m = myComp.getClass().getMethod("addActionListener", new Class<?>[] {ActionListener.class});
      if (m != null) {
        m.invoke(myComp, new Object[] {this});
      }
    } catch (Exception e) {
      System.out.println("We caught exception trying to add action listener: " + e);
    }
    actionMsgID = id;
  }

  public void addChangeListener (int id) {
    if (myComp == null) {
      Shell.getMidasContext().warning("No Component to listen");
      return;
    }
    try {
      Method m = myComp.getClass().getMethod("addChangeListener", new Class<?>[] {ChangeListener.class});
      if (m != null) {
        m.invoke(myComp, new Object[] {this});
      }
    } catch (Exception e) {
      System.out.println("We caught exception trying to add state changed listener: " + e);
    }
    changeMsgID = id;
  }

  public void actionPerformed(ActionEvent e) {
    // System.out.println("Got ActionEvent: " + e);
    boolean continueProcess = processActionEvent(e);
    if (!continueProcess) return;

    if (actionMsgID > 0) {
      Table data = new Table();
      Message msg = null;
      Object sender = e.getSource();
      if (sender instanceof Button || sender instanceof JButton) {
        data.put("ARGS", 2);
        data.put("TYPE", "AL");
        data.put("VAL1", name);
        data.put("VAL2", e.getModifiers());
        msg = new Message("BUTTON", 0, data);
      } else {
        data.put("ARGS", 2);
        data.put("TYPE", "AL");
        data.put("VAL1", name);
        data.put("VAL2", 0);
        msg = new Message("ACTION", 0, data);
      }
      XmShellParent.sendXmMessage(actionMsgID, msg);
    }
  }

  public void stateChanged(ChangeEvent e) {
    // System.out.println("Got ChangeEvent: " + e);
    boolean continueProcess = processChangeEvent(e);
    if (!continueProcess) return;

    if (changeMsgID > 0) {
      Table data = new Table();
      Message msg = null;
      data.put("ARGS", 2);
      data.put("TYPE", "AL");
      data.put("VAL1", name);
      data.put("VAL2", 0);
      msg = new Message("CHANGE", 0, data);
      XmShellParent.sendXmMessage(changeMsgID, msg);
    }

  }

  public boolean processActionEvent(ActionEvent e) {
    // override this method to handle ActionEvents in Java
    return (true);
  }

  public boolean processChangeEvent(ChangeEvent e) {
    // override this method to handle ChangeEvents in Java
    return (true);
  }

}
