package nxm.sys.libxm;

import java.io.FilenameFilter;
import java.util.Arrays;
import nxm.sys.lib.*;

/** Resource for interacting with a X-Midas RamFile. <font color="red">Do NOT use this class
    directly, this class is for use by the SYS option tree ONLY and is subject to change without
    notice! </font><br>
    <br>
    This is based heavily on the {@link PipeResource} class used for handling NeXtMidas pipes.
    Most methods in this class simply call their counterparts in {@link XmPipe}. This class has no
    native methods and does not access any X-Midas resources directly ({@link XmPipe} handles all
    X-Midas interaction related to X-Midas pipes).
    
    @author J. Czechowski IV
    @version $Id: XmPipeResource.java,v 1.5 2006/11/07 20:04:33 jxc Exp $
 */
public class XmPipeResource extends PipeResource {
  /** The prefix for this type of resource. */
  public static final String PREFIX = "xmpipe:";
  
  /** The AUX table entry for this type of resource. */
  public static final String AUX_NAME = "XMPIPE";
  
  //////////////////////////////////////////////////////////////////////////////////////////////////
  // NOTE: The following methods all override PipeResource. Please see the PipeResource class for
  //       documentation.
  //////////////////////////////////////////////////////////////////////////////////////////////////  
  public boolean isContainer()     { return false; }
  public boolean isStream()        { return true;  }
  public boolean isPipe()          { return true;  }
  public boolean isFile()          { return false; }
  public boolean isStreaming()     { return (pipe != null)? (outlet>=0)                : false; }
  public long    getLength()       { return (pipe != null)? (long)pipe.getInByte()     : 0L;    }
  public boolean seek(long offset) { return (pipe != null)? ((XmPipe)pipe).seek(offset): false; }
  
  public void setLength(long bytes) {
    if(pipe != null) { pipe.setSize((int)bytes); }
  }
  
  public String[] getEntryList(FilenameFilter filter) {
    return filterEntries(filter, XmPipe.getPipeNames());
  }
  
  public boolean exists() {
    String[] pipeNames = XmPipe.getPipeNames();
    
    /* JC4: The setting of name is here to match that used in PipeResource... it should be moved. */
    if(urlstr.startsWith(PREFIX)) {
      name = urlstr.substring(PREFIX.length()).toUpperCase();
    } else {
      name = urlstr.toUpperCase();
    }
    
    return Arrays.asList(pipeNames).contains(name);
  }
  
  public boolean open() { 
    pipe = new XmPipe(M,name);
    outlet = pipe.attach(dir, df);
    return (outlet >= 0);
  }
  
  public boolean close() {
    if(pipe != null) { pipe.detach(outlet, df); }
    return true;
  }
  
  public boolean delete() {
    throw new MidasException("Can not delete an X-Midas pipe at this time.");
  }
  
  public int read(byte[] buf, int boff, int bytes) { 
    return pipe.read(buf, buf.length, boff, bytes, outlet);
  }
  
  public int write(byte[] buf, int boff, int bytes) { 
    return pipe.write(buf, buf.length, boff, bytes, outlet);
  }
}
