package nxm.sys.net;

import nxm.sys.lib.Query;
import nxm.sys.lib.Convert;
import nxm.sys.libg.GWidget;
import nxm.sys.libg.GLabel;

/**
  A class to implement a graphical query into an Object.

  This widget introspects an object and makes its set and get methods
  available for status and control.  A method is considered a property
  if it has a getXXX method with no arguments.  If a setXXX method of
  the same name with one argument of the getXXX methods return type,
  that property is considered controllable and will show "=>" rather than 
  just "=" between the property name and its short string value in the display.  
  @author Jeff Schoen
  @version $Id: HQuery.java,v 1.13 2012/03/09 15:17:36 jgs Exp $

*/
public class HQuery extends HSource {

  private String list;
  private Object object;
  private Query query;
  private int maxitems;
  private HQuery next;

  public HQuery (String name, Object object, String list, int maxitems) {
    this.name = name;
    this.object = object;
    this.maxitems = maxitems;
    this.list = list;
    query = new Query(object,list,maxitems);
    query.parse();
  }

  public void handleRequest (String uri, HPage hp) {

    int i,j,k;
    Query q = query;
    Object obj = object;

    // look for a set command
    String command = null;
    int qm = uri.indexOf('?');
    if (qm>0) {
      command = uri.substring(qm+1);
      uri = uri.substring(0,qm);
    }

    // walk the chain
    for (j=1; (k=uri.indexOf("/",j)) > j; j=k) {
      k++; i = uri.indexOf("/",k); 
      if (i<0) i=uri.length();
      String key = uri.substring(k,i);
      q.parse();
      obj = q.getObject(key);
      q = new Query(obj,list,maxitems);
    }

    q.parse();

    // evaluate set command
    if (command != null) {
      int es = command.indexOf('=');
      String key = command.substring(0,es);
      String value = command.substring(es+1);
      Object o = q.getObject(key);
      if (o instanceof GWidget) ((GWidget)o).setAction(value);
    }

    q.evaluate();

    queryPage (uri,q,hp);
  }

  public static void queryPage (String uri, Query q, HPage hp) {
    boolean view = true;
    hp.open();
    hp.writeln ("<html>");
    hp.writeln ("<body>");
    hp.writeln ("<h3><a href='' onClick=\"location.reload(); return false;\">Query</a> = "+uri+"</h3>");
    hp.writeln ("<table>");
    for (int i=0; i<q.getItems(); i++) {
      String name = q.getName(i);
      Object o = q.getObject(i);
      boolean isCntrl = (o instanceof GWidget);
      boolean isLabel = (o instanceof GLabel);
      String value = lookForChoices( name, q.getValue(i).toString() );
      if (view || isLabel) {
        hp.writeln ("<tr>");
        if (isLabel) hp.writeln ("<td nowrap><a href =\""+uri+"?"+name+"=Toggle\">"+name+"</a></td>");
        else if (isCntrl) hp.writeln ("<td nowrap><b>"+name+"</b></td>");
        else hp.writeln ("<td nowrap><a href =\""+uri+"/"+name+"\">"+name+"</a></td>");
        hp.writeln ("<td>"+value+"</td>");
        hp.writeln ("</tr>");
      }
      if (isLabel) view = ((GLabel)o).getActive()>0;
    }
    hp.writeln ("</table>");
    hp.writeln ("</body>");
    hp.writeln ("</html>");
    hp.close();
  }

  private static String lookForChoices (String name, String value) {
    if (value==null) return value;
    int fb = value.indexOf('[');
    int rb = value.indexOf(']');
    if (fb<0 || rb<0) return value;
    int i,j,k;
    String selects = "";
    for (i=fb; i<rb; i=j) {
      j=value.indexOf(',',i+1);
      if (j<0) j=rb;
      String item = value.substring(i+1,j); int il=item.length()-1;
      if (il>0 && item.charAt(0)=='<' && item.charAt(il)=='>') {  
        item = item.substring(1,il);
        selects += " <a href =\"?"+name+"="+item+"\"><b>"+item+"</b></a> ";
      } else {
        selects += " <a href =\"?"+name+"="+item+"\">"+item+"</a> ";
      }
      if (j!=rb) selects += ',';
    }
    return value.substring(0,fb+1)+selects+value.substring(rb);
  }

}
