package nxm.sys.net;

import nxm.sys.lib.*;
import java.io.File;

/**
  A class to display system information in an Browser.

  Supported requests are:
   Get?resultName
   Set?resultName=value
   Top
   Log
   DiskStats/auxName

  @author Jeff Schoen
  @version $Id: HSystem.java,v 1.5 2009/05/27 14:18:28 jgs Exp $

*/
public class HSystem extends HSource {

  private Midas M;

  public HSystem (String name, Midas M) {
    this.M = M;
    this.name = name+"/";
  }

  public void handleRequest (String uri, HPage hp) {

    String option = uri.substring(name.length()+1);

    if (option.startsWith("Get?")) {
      getPage (option.substring(4),hp);
    }
    else if (option.startsWith("Set?")) {
      setPage (option.substring(4),hp);
    }
    else if (option.startsWith("Top")) {
      topPage (hp);
    }
    else if (option.startsWith("Log")) {
      logPage (hp);
    }
    else if (option.startsWith("DiskStats/")) {
      String aux = option.substring(10);
      diskPage (aux,hp);
    }
    else {
      optPage (hp);
    }
  }

  /** Display system log information if available */
  public void getPage (String key, HPage hp) {
    Object value = M.results.get(key);
    hp.open();
    hp.writeln (key+"="+value);
    hp.close();
  }

  public void setPage (String key, HPage hp) {
    int ie = key.indexOf("=");
    if (ie>0) {
      String value = key.substring(ie+1);
      key = key.substring(0,ie);
      M.results.setKey(key,value);
    }
    getPage(key,hp);
  }

  /** Display system log information if available */
  public void logPage (HPage hp) {
    hp.open();
    hp.writeln ("<html>");
    hp.writeln ("<body>");
    hp.writeln ("<h3><a href='' onClick=\"location.reload(); return false;\">System Log</a></h3>");
    hp.writeln ("<br><pre>");
    hp.writeln ("Log File not opened");
    hp.writeln ("</pre>");
    hp.writeln ("</body>");
    hp.writeln ("</html>");
    hp.close();
  }

  /** Display Top CPU/Memory Usage by process */
  public void topPage (HPage hp) {
    byte[] buf = new byte[8192];
    int bytes = OS.getStats(0,buf);
    hp.open();
    hp.writeln ("<html>");
    hp.writeln ("<body>");
    hp.writeln ("<h3><a href='' onClick=\"location.reload(); return false;\">System Top</a></h3>");
    hp.writeln ("<br><pre>");
    hp.write(buf,0,bytes);
    hp.writeln ("</pre>");
    hp.writeln ("</body>");
    hp.writeln ("</html>");
    hp.close();
  }

  /** Display disk stats for a given Auxiliary */
  public void diskPage (String aux, HPage hp) {
    hp.open();
    hp.writeln ("<html>");
    hp.writeln ("<body>");
    hp.writeln ("<h3><a href='' onClick=\"location.reload(); return false;\">Aux="+aux+" Disk Stats</a></h3>");
    String list = aux;
    if (aux.equals("ALL")) list = M.results.getS("AUX.READ");
//    String[] auxes = Util.splitToArray(list,"|");
    String[] auxes = StringUtil.cleanSplit(list, "|");
    for (int i=0; i<auxes.length; i++) {
    aux = auxes[i];
    String path = M.io.getAuxPath(aux);
    Table tab = OS.getStatFS(path);
    hp.writeln ("<ul>");
    hp.writeln ("<li>Aux   = "+aux+"</li>");
    hp.writeln ("<li>Path  = "+path+"</li>");
    if (tab!=null) {
      hp.writeln ("<li>Files = "+tab.getL("FILES")+"</li>");
      hp.writeln ("<li>Total = "+tab.getL("TMBY")+" Mby</li>");
      hp.writeln ("<li>Free  = "+tab.getL("FMBY")+" Mby</li>");
      hp.writeln ("<li>Used  = "+tab.getL("USED")+"%</li>");
    }
    hp.writeln ("</ul>");
    }
    hp.writeln ("</body>");
    hp.writeln ("</html>");
    hp.close();
  }

  /** Display menu of all available display pages */
  public void optPage (HPage hp) {
    hp.open();
    hp.writeln ("<html>");
    hp.writeln ("<head><title>HSystem Options Page</title></head>");
    hp.writeln ("<body>");
    hp.writeln ("<center><h2>NeXtMidas System Information</h2></center>");
    hp.writeln ("<ul>");
    hp.writeln ("<li><a href='/nmroot/index.html'>JavaDocs</a> - System tree documentation</li>");
    hp.writeln ("<li><a href='LogTop'>System Log</a> - Top of System Log File</li>");
    hp.writeln ("<li><a href='LogTail'>System Log</a> - Tail of System Log File</li>");
    hp.writeln ("<li><a href='LogAll'>System Log</a> - All of System Log File</li>");
    hp.writeln ("<li><a href='TopCPU'>Top CPU/Mem</a> - List of Top CPU/Memory Processes on Server</li>");
    Table taux = M.results.getTable("AUX");
    String raux = taux.getString("READ");
    Parser auxes = new Parser (raux,'|');
    for (auxes.reset();  auxes.more(); ) {
      String aux = auxes.next();
      String path = M.io.getAuxPath(aux);
      hp.writeln ("<li><a href='DiskStats/"+aux+"'>Disk Aux="+aux+"</a> Path="+path+" Disk Stats</li>");
    }
    hp.writeln ("</ul>");
    hp.writeln ("</body>");
    hp.writeln ("</html>");
    hp.close();
  }

}
