package nxm.sys.net; //not needed for non-nxm use

import java.net.*;
import java.util.zip.Deflater;
import nxm.sys.lib.Convert; //for pack
import nxm.sys.lib.Data; //for format
import nxm.sys.lib.ICEPacket; //for RTP Header of ICE Packet
import nxm.sys.lib.Shell; //for rep in IcePacket Header
import nxm.sys.lib.Time; //for current Time

/**
  dgnetsend - outputs Multicast messages. May choose multicast or udp but do not mix.

  @author Scott Zimmerman / Jay Hallam
  @version $Id: dgnetsend.java,v 1.17 2012/11/07 22:05:13 ntn Exp $
*/
public class dgnetsend {
   //settable by accessor methods
  private boolean verbose = false;
  private boolean packetDebug = false;
  private boolean multicast = true;
  private boolean keepopen=false;
  private Deflater compressor = null;
  private String host=new String("238.0.0.1"); // remote host
  private int port=18800;                      // remote port
  private int ttl=1; //ttl is unsigned 8-bit, 0 <= ttl <= 0xFF
  private int rtphdrlen=0;
  private int icehdrlen=0;
  private int ssrc=0,csrc0=0;
  private short payload=0;

  //private variables
  private static final int MAXPKTBUF=65536; //max data - TODO: remove? not used
  private InetAddress addr=null;
  private DatagramSocket dsock=null;
  private MulticastSocket msock=null;
  private String osname = (System.getProperty("os.name")).toUpperCase(); // MSZ work-around for windows
  private boolean windows = (osname.startsWith("WINDOWS"));
  private DatagramPacket dpacki = null;

  /** ICE packet handling */
  ICEPacket icePacket = new ICEPacket("SB", ICEPacket.headerLength, 0);

  /** Max 16-bit (2 byte) unsigned integer. Value = 2<sup>16</sup> = {@value}. */
  static final int  MAX_UNSIGNED_INT_2 = 65535;
  /** Max 32-bit (4 byte) unsigned integer. Value = 2<sup>32</sup> = {@value}. */
  static final long MAX_UNSIGNED_INT_4 = 4294967296L;
  /** Max 32-bit (4 byte) signed integer. Value = 2<sup>31</sup> - 1 = {@value}. */
  static final int  MAX_INT_4  = Integer.MAX_VALUE;

  public static final int MAXPKT=32768;
  private byte[] rtpbuf = new byte[MAXPKT];
  private int timestamp=0;
  int seqnum=0;
  int count=0; // TODO: count should be 'long' to cover max 32-bit unsigned integer

  public dgnetsend() { }
  public dgnetsend(String phost,int pport) {
    host=phost;
    port=pport;
  }

  /** Get the next sequence number to use
      @return next sequence number to use
      @since NeXtMidas 2.8.2
   */
  public int getNextSeqNum() {
    if (seqnum>=MAX_UNSIGNED_INT_2) return 0;
    return (seqnum+1);
  }

  /** Get the current sequence number
      @return current sequence number
      @since NeXtMidas 2.8.2
   */
  public int getSeqNum() { return seqnum; }

  /** Set the current sequence number
      @param parm the new sequence number to use
      @since NeXtMidas 2.8.2
   */
  public void setSeqNum(int parm) { seqnum=parm; }

  /** Get the time-to-live (TTL) */
  public int getTTL() { return ttl; }
  /** Set the time-to-live (TTL) */
  public void setTTL(int pittl) {
    try {
      ttl=pittl;
      if (msock==null) return;
      if (multicast) msock.setTimeToLive(ttl);
      if (verbose) System.out.println("dgnetsend: setting ttl "+ttl);
    } catch (Exception e) {
      System.err.println("dgnetsend: Invalid ttl="+pittl+". TTL Must be >=0 and <=255");
      e.printStackTrace();
    }
  }
  /** @deprecated Since NeXtMidas 2.5.5: see Bug 1367 */
  @Deprecated public int getTimeOut() { return 0; }
  /** @deprecated Since NeXtMidas 2.5.5: see Bug 1367  */
  @Deprecated public void setTimeOut(int pTimeOut) { System.out.println("dgnetsend: setTimeOut() deprecated: did nothing"); }
  public boolean getVerbose() { return verbose; }
  public void setVerbose(boolean pVerbose) {
    verbose=pVerbose;
    if (verbose) System.out.println("dgnetsend: setting Verbose "+pVerbose);
  }
  public boolean getPacketDebug() { return packetDebug; }
  public void setPacketDebug(boolean pPacketDebug) {
    packetDebug=pPacketDebug;
    if (verbose) System.out.println("dgnetsend: setting PacketDebug "+pPacketDebug);
  }
  public void setHost(String phost) {
    host=phost;
    if (verbose) System.out.println("dgnetsend: setting host "+phost);
  }
  public String getHost() { return host; }
  public void setPort(int pport) {
    port=pport;
    if (verbose) System.out.println("dgnetsend: setting port "+pport);
  }
  public int getPort() { return port; }
  public void setKeepOpen(boolean pkeepopen) {
    keepopen=pkeepopen;
    if (verbose) System.out.println("dgnetsend: setting keepopen "+pkeepopen);
  }
  public short getPayload() { return payload; }
  public void setPayload(short ppayload) {
    payload=ppayload;
    if (verbose) System.out.println("dgnetsend: setting payload "+payload);
  }
  public int getSSRC() { return ssrc; }
  public void setSSRC(int pssrc) {
    ssrc=pssrc;
    if (verbose) System.out.println("dgnetsend: setting ssrc "+ssrc);
  }
  public int getCSRC() { return csrc0; }
  public void setCSRC(int pcsrc) {
    csrc0=pcsrc;
    if (verbose) System.out.println("dgnetsend: setting csrc(0) "+csrc0);
  }
  public boolean getCompress() { if (compressor!=null) return true; else return false; }
  public void setCompress(boolean pcompress) {
    if (pcompress) {
      compressor=new Deflater();
    } else {
      compressor=null;
    }
    if (verbose) System.out.println("dgnetsend: setting compress "+pcompress);
  }
  /** Enable/Disable the local loopback of multicast datagrams.
      @see MulticastSocket#setLoopbackMode(boolean)
      @since NeXtMidas 3.0.0
   */
  public void setLoopbackMode(boolean disable) {
    if (msock==null) return;
    try {
      msock.setLoopbackMode(disable);
    } catch (Exception e) {
      e.printStackTrace();
    }
  }
  /** Is the local loopback of multicast datagrams enabled/disabled?
      @see MulticastSocket#getLoopbackMode()
      @since NeXtMidas 3.0.0
   */
  public boolean getLoopbackMode() {
    if (msock==null) return false;
    try {
      return msock.getLoopbackMode();
    } catch (Exception e) {
      e.printStackTrace();
    }
    return false;
  }
  public MulticastSocket getMulticastSocket() { if  (multicast) return msock; return null; } // since 3.0.0
  public DatagramSocket getDatagramSocket()   { if (!multicast) return dsock; return null; } // since 3.0.0

  ////////////////////////////////////////////////////////////////////////////////
  public boolean open(String phost,int pport,short pptype,int pssrc,int pcsrc0) {
    host=phost;
    port=pport;
    payload=pptype;
    ssrc=pssrc;
    csrc0=pcsrc0;
    keepopen=true;
    return open();
  }
  ////////////////////////////////////////////////////////////////////////////////
  public boolean open(String phost,int pport) {
    host=phost;
    port=pport;
    keepopen=true;
    return open();
  }
  public boolean open(short pptype,int pssrc,int pcsrc0) {
    payload=pptype;
    ssrc=pssrc;
    csrc0=pcsrc0;
    keepopen=true;
    return open();
  }
  ////////////////////////////////////////////////////////////////////////////////
  public boolean open() {
    try {
      close();
      if (!keepopen) { //!keepopen means writing messages to anywhere so open both sockets because host may be multicast or udp
        msock = new MulticastSocket();
        msock.setTimeToLive(ttl);
        dsock = new DatagramSocket();
      } else {
        if (host==null) { System.err.println("dgnetsend: No Host to open"); return false; }

        InetSocketAddress isa = new InetSocketAddress(host,port);
        addr = InetAddress.getByName(host);
        multicast = addr.isMulticastAddress(); // e.g. 224.0.0.0 through 239.255.255.255 for IPv4
        if (verbose && multicast) System.out.println("dgnetsend: Set Multicast on for "+isa);

        if (multicast) {
          // MSZ: work-around for Java Bug 4701650 on Windows (09JAN06), NeXtMidas Bug 299
          //        msock = new MulticastSocket(isa);
          if (windows) { msock = new MulticastSocket(port); }
          else         { msock = new MulticastSocket(isa);  }
          msock.setTimeToLive(ttl);
        } else dsock = new DatagramSocket(); // Bug 1969 (for unicast, on any available port); previously used isa (which is remote host/port)
        if (verbose) System.out.println("dgnetsend: Connected to Socket. Host:"+host+". Port:"+port);
      }
      if (verbose) System.out.println("dgnetsend: Done open. dsock=" +dsock+" msock="+msock);
      return true;
    } catch (Exception e) {
      System.err.println("dgnetsend: Open Error:"+e+". For IP="+host+". Port="+port);
      if (e instanceof BindException && "Cannot assign requested address".equals(e.getMessage())) { // nice info msg for Bug 1864
        System.err.println("dgnetsend: On RHEL4+ (system with dual IPv4 and IPv6 stacks), try defining\n"
                         + "           java.net.preferIPv4Stack=true system property to use IPv4 stack.");
      }
      e.printStackTrace();
    }
    return false;
  }
  ////////////////////////////////////////////////////////////////////////////////
  public void close() {
    if (verbose) System.out.println("dgnetsend: Closing");
    try {
      if (dsock!=null) { dsock.close(); dsock=null; }
      if (msock!=null) { msock.close(); msock=null; }
    } catch (Exception e) {
      System.err.println("dgnetsend: Exception: "+e);
      e.printStackTrace();
    }
  }
  //writeRTP methods - String input
  ////////////////////////////////////////////////////////////////////////////////
  public boolean writeRTP(String sbuf) { return writeRTP(sbuf,payload,ssrc); }
  ////////////////////////////////////////////////////////////////////////////////
  public boolean writeRTP(String sbuf,short ptype,int ssrcid) { //matches old calls
    try {
      byte[] buf = sbuf.getBytes();
      if (buf==null) { System.err.println("dgnetsend: Null Input"); return false; }
      return writeRTP(buf,buf.length,ptype,ssrcid,0,csrc0);
    } catch (Exception e) {
      System.err.println("dgnetsend: writeRTP Exception: "+e);
      e.printStackTrace();
    }
    return false;
  }
  //writeRTP methods - byte input
  ////////////////////////////////////////////////////////////////////////////////
  public boolean writeRTP(byte[] dbuf,int size) { return writeRTP(dbuf,size,payload,ssrc,0,csrc0); }
  ////////////////////////////////////////////////////////////////////////////////
  public boolean writeRTP(byte[] dbuf,int size,int offset) { return writeRTP(dbuf,size,payload,ssrc,offset,csrc0) ; }
  ////////////////////////////////////////////////////////////////////////////////
  public boolean writeRTP(byte[] dbuf,int size,short ptype,int ssrcid) { return writeRTP(dbuf,size,ptype,ssrcid,0,csrc0); }
  ////////////////////////////////////////////////////////////////////////////////
  public boolean writeRTP(byte[] dbuf,int size,short ptype,int ssrcid,int offset) {
    return writeRTP(dbuf,size,ptype,ssrcid,offset,csrc0);
  }
  ////////////////////////////////////////////////////////////////////////////////
  public boolean writeRTP(byte[] dbuf,int size,short ptype,int ssrcid,int offset,int csrc0id) {
    try {
      timestamp+=size;
      if (seqnum>=MAX_UNSIGNED_INT_2) seqnum=0;
      else seqnum++;
      if (count>=MAX_INT_4) count=0; // TODO: count should be a 'long' to cover max 32-bit unsigned integer
      else count++;
      rtphdrlen = setRTPHeader(rtpbuf,null,0,ptype,ssrcid,seqnum,timestamp,0,csrc0); //first 0 means no header
      if (compressor==null) System.arraycopy(dbuf,offset,rtpbuf,rtphdrlen,size); //copy data into rtpbuf
      else {
        compressor.setInput(dbuf,offset,size);
        compressor.finish();
        size=compressor.deflate(rtpbuf,rtphdrlen,size);
      }
      try {
        if (dsock==null && msock==null) {
          System.err.println("dgnetsend: Null Socket/buffer");
          if (verbose) {
            for (StackTraceElement ste: Thread.currentThread().getStackTrace()) {
              System.err.println(ste);
            }
          }
          return false;
        }
        dpacki = new DatagramPacket(rtpbuf,size+rtphdrlen,addr,port);
        if (dpacki==null) return false;
        if (multicast) msock.send(dpacki);
        else dsock.send(dpacki);
        if (packetDebug) System.out.println("dgnetsend: Packet size "+(size+rtphdrlen)+"-"+new String(rtpbuf,0,size+rtphdrlen));
      } catch (Exception e) {
        System.err.println("dgnetsend writeRTP socket: Exception: "+e);
        e.printStackTrace();
      }
      return true;
    } catch (Exception e) {
      System.err.println("dgnetsend: writeRTP Exception: "+e);
      e.printStackTrace();
    }
    return false;
  }
  //writeRTPICE methods where hdr is passed in the byte buffer
  ////////////////////////////////////////////////////////////////////////////////
  public boolean writeRTPICE(byte[] dbuf,int size,short ptype) { //contains ICE Hdr in dbuf
     return writeRTPICE(dbuf,size,ptype,ssrc,csrc0);
  }
  public boolean writeRTPICE(byte[] dbuf,int size,short ptype,int ssrc,int csrc0) { //contains ICE Hdr in dbuf
    try {
      timestamp+=size-ICEPacket.headerLength;
      if (seqnum>=MAX_UNSIGNED_INT_2) seqnum=0;
      else seqnum++;
      rtphdrlen = setRTPHeader(rtpbuf,dbuf,ICEPacket.headerLength,ptype,ssrc,seqnum,timestamp,0,csrc0);
      if (compressor==null) System.arraycopy(dbuf,ICEPacket.headerLength,rtpbuf,rtphdrlen,size-ICEPacket.headerLength); //copy data into rtpbuf
      else {
        compressor.setInput(dbuf,ICEPacket.headerLength,size-ICEPacket.headerLength);
        compressor.finish();
        size=compressor.deflate(rtpbuf,rtphdrlen,size-ICEPacket.headerLength);
      }
      try {
        if (dsock==null && msock==null) {
          System.err.println("dgnetsend: Null Socket/buffer");
          return false;
        }
        dpacki = new DatagramPacket(rtpbuf,size-ICEPacket.headerLength+rtphdrlen,addr,port);
        if (multicast) msock.send(dpacki);
        else dsock.send(dpacki);
        if (packetDebug) System.out.println("dgnetsend: Packet size "+rtpbuf.length+"-"+new String(rtpbuf)+":::"+dpacki);
      } catch (Exception e) {
        System.err.println("dgnetsend writeRTPICE socket: Exception: "+e);
        e.printStackTrace();
      }
      return true;
    } catch (Exception e) {
      System.err.println("dgnetsend: writeRTPICE Exception: "+e);
      e.printStackTrace();
    }
    return false;
  }
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  public boolean writeRTPICE(byte[] dbuf,byte icebuf[],int size,short ptype,int ssrc,int csrc0) { //contains ICE Hdr in sep buffer
    return writeRTPICE(dbuf,icebuf,size,ptype,ssrc,csrc0,0);
  }

  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  public boolean writeRTPICE(byte[] dbuf,byte icebuf[],int size,short ptype,int ssrc,int csrc0,int offset) { //contains ICE Hdr in sep buffer
    try {
      timestamp+=size;
      if (seqnum>=MAX_UNSIGNED_INT_2) seqnum=0;
      else seqnum++;
      rtphdrlen = setRTPHeader(rtpbuf,icebuf,ICEPacket.headerLength,ptype,ssrc,seqnum,timestamp,0,csrc0);
      if (compressor==null) System.arraycopy(dbuf,offset,rtpbuf,rtphdrlen,size); //copy data into rtpbuf
      else {
        compressor.setInput(dbuf,offset,size);
        compressor.finish();
        size=compressor.deflate(rtpbuf,rtphdrlen,size);
      }
      try {
        if (dsock==null && msock==null) {
          System.err.println("dgnetsend: Null Socket/buffer");
          return false;
        }
        dpacki = new DatagramPacket(rtpbuf,size+rtphdrlen,addr,port);
        if (multicast) msock.send(dpacki);
        else dsock.send(dpacki);
        if (packetDebug) System.out.println("dgnetsend: Packet size "+rtpbuf.length+"-"+new String(rtpbuf)+":::"+dpacki);
      } catch (Exception e) {
        System.err.println("dgnetsend writeRTPICE socket: Exception: "+e);
        e.printStackTrace();
      }
      return true;
    } catch (Exception e) {
      System.err.println("dgnetsend: writeRTPICE Exception: "+e);
      e.printStackTrace();
    }
    return false;
  }
  //writeRTPICE methods which create ICE header
  ////////////////////////////////////////////////////////////////////////////////
  public boolean writeRTPICE(byte[] dbuf,int size,double jtime,double p1,double p2) {
    return writeRTPICE(dbuf,size,jtime,p1,p2,payload,ssrc,0,csrc0);
  }
  ////////////////////////////////////////////////////////////////////////////////
  public boolean writeRTPICE(byte[] dbuf,int size,short ptype,int ssrcid) {
    return writeRTPICE(dbuf,size,Time.current(),0.0,0.0,ptype,ssrcid,0,csrc0);
  }
  ////////////////////////////////////////////////////////////////////////////////
  public boolean writeRTPICE(byte[] dbuf,int size,double jtime,double p1,double p2,int offset) {
    return writeRTPICE(dbuf,size,jtime,p1,p2,payload,ssrc,offset,csrc0);
  }
  ////////////////////////////////////////////////////////////////////////////////
  public boolean writeRTPICE(byte[] dbuf,int size,double jtime,double p1,double p2,short ptype,int ssrcid) {
    return writeRTPICE(dbuf,size,jtime,p1,p2,ptype,ssrcid,0,csrc0);
  }
  ////////////////////////////////////////////////////////////////////////////////
  public boolean writeRTPICE(byte[] dbuf,int size,double jtime,double p1,double p2,short ptype,int ssrcid,int offset) {
    return writeRTPICE(dbuf,size,jtime,p1,p2,ptype,ssrcid,offset,csrc0);
  }
  ////////////////////////////////////////////////////////////////////////////////
  public boolean writeRTPICE(byte[] dbuf,int size,double jtime,double p1,double p2,short ptype,int ssrcid,int offset,int csrc0id) {
    if (count>=MAX_INT_4) count=0;
    else count++;
    icehdrlen = defICE(icePacket.buf,count,size,"SB",jtime,p1,p2,0);
    return writeRTPICE(dbuf,size,icehdrlen,ptype,ssrcid,offset,csrc0id);
  }
  ////////////////////////////////////////////////////////////////////////////////
  public boolean writeRTPICE(byte[] dbuf,int size,String format,double jtime,double p1,double p2,short ptype,int ssrcid,int offset,
                             int csrc0id) {
    if (count>=MAX_INT_4) count=0;
    else count++;
    icehdrlen = defICE(icePacket.buf,count,size,format,jtime,p1,p2,0);
    return writeRTPICE(dbuf,size,icehdrlen,ptype,ssrcid,offset,csrc0id);
  }
  ////////////////////////////////////////////////////////////////////////////////
  //should not be public as it relies on icePacket.buf to be set
  private boolean writeRTPICE(byte[] dbuf,int size,int icehdrlen,short ptype,int ssrcid,int offset,int csrc0id) {
    try {
      timestamp+=size;
      if (seqnum>=MAX_UNSIGNED_INT_2) seqnum=0;
      else seqnum++;
      rtphdrlen = setRTPHeader(rtpbuf,icePacket.buf,icehdrlen,ptype,ssrc,seqnum,timestamp,offset,csrc0id);
      if (compressor==null) System.arraycopy(dbuf,offset,rtpbuf,rtphdrlen,size); //copy data into rtpbuf
      else {
        compressor.setInput(dbuf,0,size);
        compressor.finish();
        size=compressor.deflate(rtpbuf,rtphdrlen,size);
      }
      try {
        if (dsock==null && msock==null) {
          System.err.println("dgnetsend: Null Socket/buffer");
          return false;
        }
        dpacki = new DatagramPacket(rtpbuf,size+rtphdrlen,addr,port);
        if (multicast) msock.send(dpacki);
        else dsock.send(dpacki);
        if (packetDebug) System.out.println("dgnetsend: Packet size "+rtpbuf.length+"-"+new String(rtpbuf)+":::"+dpacki);
      } catch (Exception e) {
        System.err.println("dgnetsend writeRTPICE socket: Exception: "+e);
        e.printStackTrace();
      }
      return true;
    } catch (Exception e) {
      System.err.println("dgnetsend: writeRTPICE Exception: "+e);
      e.printStackTrace();
    }
    return false;
  }
  //write methods w/no added RTP header
  ////////////////////////////////////////////////////////////////////////////////
  public boolean write(String sbuf) {
    boolean ret=false;
    try {
      byte[] buf = sbuf.getBytes();
      if (buf==null) { System.err.println("dgnetsend: Null Input"); return ret; }
      ret=write(buf);
    } catch (Exception e) {
      System.err.println("dgnetsend: write Exception: "+e);
      e.printStackTrace();
    }
    return ret;
  }
  ////////////////////////////////////////////////////////////////////////////////
  public boolean write(byte[] buf) {
    if (buf==null || buf.length<1) {
       System.err.println("dgnetsend: Null data input");
       return false;
    }
    if (!keepopen) {
       System.out.println("dgnetsend: Must use write(host,port,buf) if not using keepopen");
       return false;
    }
    try {
      try {
         if (dsock==null && msock==null) {
           System.err.println("dgnetsend: Null Socket/buffer");
           return false;
         }
         dpacki = new DatagramPacket(buf,buf.length,addr,port);
         if (multicast) msock.send(dpacki);
         else dsock.send(dpacki);
         if (packetDebug) System.out.println("dgnetsend: Packet size "+buf.length+"-"+new String(buf)+":::"+dpacki);
         return true;
      } catch (SocketTimeoutException ste) {
         if (verbose) System.out.println("dgnetsend: SocketTimeOutException" + ste);
      }
    } catch (java.io.InterruptedIOException e) {
      if (verbose) System.out.println("dgnetsend: Interrupted:"+e);
    } catch (java.net.SocketException e) {
      if (verbose) System.out.println("dgnetsend: Socket Closed:"+e);
    } catch (Exception e) {//receive can throw IO, PortUnreachable and IllegalBlockingMode Exceptions
      System.err.println("dgnetsend: write Exception: "+e);
      e.printStackTrace();
    }
    return false;
  }
  ////////////////////////////////////////////////////////////////////////////////
  public boolean write(String phost,int pport,String sbuf) {
    boolean ret=false;
    try {
      byte[] buf = sbuf.getBytes();
      ret=write(phost,pport,buf);
    } catch (Exception e) {
      System.err.println("dgnetsend: write Exception: "+e);
      e.printStackTrace();
    }
    return ret;
  }
  ////////////////////////////////////////////////////////////////////////////////
  public boolean write(String phost,int pport,byte[] buf) {
    boolean ret=false;
    if (phost==null) { System.err.println("dgnetsend: No Host for write "); return ret; }
    try {
      boolean tmulticast=false;
      int iprefix = Integer.parseInt(phost.substring(0,3));
      if (iprefix>=224 && iprefix<=239) { tmulticast=true; }
      if (verbose && tmulticast) System.out.println("dgnetsend: Multicast:"+iprefix);

      try {// connect to socket ports
        dpacki = new DatagramPacket(buf,buf.length,InetAddress.getByName(phost),port);
        if (tmulticast) msock.send(dpacki);
        else dsock.send(dpacki);
        ret=true;
      } catch (SocketTimeoutException ste) {
        if (verbose) System.out.println("dgnetsend: SocketTimeOutException" + ste);
      } catch (java.io.InterruptedIOException e) {
        if (verbose) System.out.println("dgnetsend: Interrupted:"+e);
      } catch (java.net.SocketException e) {
        if (verbose) System.out.println("dgnetsend: Socket Closed:"+e);
      } catch (Exception e) {//receive can throw IO, PortUnreachable and IllegalBlockingMode Exceptions
        System.err.println("dgnetsend: write Exception: "+e);
        e.printStackTrace();
      }
    } catch (Exception e) {
       System.err.println("dgnetsend: Problem with stream address:"+phost+":"+pport+" ->"+e);
       e.printStackTrace();
    }
    return ret;
  }
  ////////////////////////////////////////////////////////////////////////////////
  //creation and handling of ice packet header for output
  ////////////////////////////////////////////////////////////////////////////////
  public int defICE(byte[] buf,int count,int size,String format,double jtime,double p1,double p2,int boffset) {
    jtime = jtime+5e-7; //copied from Time.java - fromJ1950(double sec)
    double wsec = Math.floor(jtime);
    double fsec = 1e-6*Math.floor(1e6*(jtime-wsec));
    return setICEPacket((byte)0,count,size,(short)0,(short)1,(byte)0,(byte)0,Shell.rep,format,(byte)0,(byte)0,0.0,wsec,fsec,p1,p2);
  }
  public int defICE(byte[] buf,int count,int size,double jtime,double p1,double p2,int boffset) {
    return defICE(buf,count,size,"SB",jtime,p1,p2,boffset);
  }
  //////////////////////////////////////////////////////////////////////////////// main setICEPacket
  public int setICEPacket(byte type, int count, int size, short user, short channel, byte id1, byte id2,
                            byte rep, String format, byte mode, byte status, double offset, double wsec,
                            double fsec, double p1, double p2) {
     int ilen = setICEPacket(type,count,size,user,channel,id1,id2,rep,format,mode,status,offset,wsec,fsec);
     icePacket.setUserData(p1,0);
     icePacket.setUserData(p2,1);
     return ilen;
  }
  ////////////////////////////////////////////////////////////////////////////////
  public int setICEPacket(byte type, int count, int size, short user, short channel, byte id1, byte id2,byte rep, String format,
                            byte mode, byte status, double offset, double wsec,double fsec) {
     icePacket.setUserKey(type); //user defined key
     icePacket.setCount(count); //packet number since start of xfer
     int bpa = Data.getBPA(format); //get Bytes per atom
     int nsize=size;
     if (bpa>0) nsize=nsize/bpa;
     icePacket.setSize(nsize); //data elements in packet
     icePacket.setUser(user); //user defined field
     icePacket.setChannel(channel); //port channel number
     icePacket.setID(id1, id2); //subsystem IDs
     icePacket.setRep(rep); //data rep (Vax,Ieee,Eeei) - usually Shell.rep
     icePacket.setFormat(format); // Midas format digraph, sets bytes per atom (mode*type),mode, and type (MODE=S,C,V.., TYPE=F,D,L,...)
     icePacket.setTC(mode,status,offset,wsec,fsec); // Set timecode mode, status, offset whole seconds and fractional seconds
     return ICEPacket.headerLength;
  }
  //These are not used in ELCC Streams as ELCC has user data sections as doubles only, only floats added here although other available
  ////////////////////////////////////////////////////////////////////////////////
  public int defICE(byte[] buf,int count,int size,int boffset) {
    return setICEPacket((byte)0,count,size,(short)0,(short)1,(byte)0,(byte)0,Shell.rep,"SB",(byte)0,(byte)0,0.0,0.0,0.0,0.0,1.0);
  }
  ////////////////////////////////////////////////////////////////////////////////
  public boolean writeRTPICE(byte[] dbuf,int size,double jtime,float p1,float p2,float p3,float p4,short ptype,int ssrcid,int offset,
                             int csrc0id) {
    icehdrlen = defICE(icePacket.buf,count,size,jtime,p1,p2,p3,p4,0);
    return writeRTPICE(dbuf,size,icehdrlen,ptype,ssrcid,offset,csrc0id);
  }
  ////////////////////////////////////////////////////////////////////////////////
  public int defICE(byte[] buf,int count,int size,double jtime,float p1,float p2,float p3,float p4,int boffset) {
    jtime = jtime+5e-7; //copied from Time.java - fromJ1950(double sec)
    double wsec = Math.floor(jtime);
    double fsec = 1e-6*Math.floor(1e6*(jtime-wsec));
    return setICEPacket((byte)0,count,size,(short)0,(short)1,(byte)0,(byte)0,Shell.rep,"SI",(byte)0,(byte)0,0.0,wsec,fsec,p1,p2,p3,p4);
  }
  ////////////////////////////////////////////////////////////////////////////////
  public int setICEPacket(byte type, int count, int size, short user, short channel, byte id1, byte id2,
                            byte rep, String format, byte mode, byte status, double offset, double wsec,
                            double fsec,float p1,float p2,float p3,float p4) {
     int ilen = setICEPacket(type,count,size,user,channel,id1,id2,rep,format,mode,status,offset,wsec,fsec);
     icePacket.setUserData(p1,0);
     icePacket.setUserData(p2,1);
     icePacket.setUserData(p3,2);
     icePacket.setUserData(p4,3);
     return ilen;
  }
  ////////////////////////////////////////////////////////////////////////////////
  //creation and handling of rtp packet header for output

  //  The RTP header has the following format:
  // 0                   1                   2                   3
  // 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
  //+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  //|V=2|P|X|  CC   |M|     PT      |       sequence number         |
  //+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  //|                           timestamp                           |
  //+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  //|           synchronization source (SSRC) identifier            |
  //+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+
  //|            contributing source (CSRC) identifiers             |
  //|                             ....                              |
  //+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  //The CC in RTPVPXCC tells how many CSRC's (0-15) follow the fixed header
  //If the X bit in the RTP header is one, a variable-length header
  //extension is appended to the RTP header, following the CSRC list if
  //present. The header extension contains a 16-bit length field that
  //counts the number of 32-bit words in the extension, excluding the
  //four-octet extension header (therefore zero is a valid length).
  //P is padding and not used here at all
  // 0                   1                   2                   3
  // 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
  //+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  //|      defined by profile       |           length              |
  //+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  //|                        header extension                       |
  //|                             ....                              |

   // Should be sending out RTCP with ReceiverReport (RR)

  ////////////////////////////////////////////////////////////////////////////////
  public int setRTPHeader(byte[] buf,byte[] xhdr,int xlen,short ptype,int ssrcid,int snum,int tstamp,int boffset,int csrc0id) {
    //only added 1 csrc although 16 are allowed
    int pos=12;
    if (xhdr!=null && xlen>0) buf[boffset]=(byte)0x90; //set V=2 and the X flag
    else buf[boffset]=(byte)0x80; //only set the V=2 flag
    if (csrc0id!=0) {
      buf[boffset]|=0x01; //set cc flag to 1
      pos=16;
    }
    buf[boffset+1]=(byte)ptype;
    buf[boffset+1]&=(byte)0x7F; //zero out mark and use only up to 128 of ptype
    //now do seq number
    netpackI2S(buf,boffset+2,snum);
    //now do timestamp
    netpackL(buf,boffset+4,tstamp); //should be different type
    //now do ssrc - which should be some type of ip based id
    netpackL(buf,boffset+8,ssrcid);
    if (csrc0id!=0) netpackL(buf,boffset+12,csrc0id);
    if (xhdr==null || xlen<=0 || xlen>65000) return pos; //65535 is maxpkt - but assume any extension > 65000 is invalid
    //now do x hdr >0 - xstream uses 255 as compound header - dont deal with here
    //elcc packet type is 16
    buf[boffset+pos]=0;
    buf[boffset+pos+1]=16;
    int xlen4=xlen>>2;
    netpackI2S(buf,boffset+pos+2,xlen4); //length of x hdr in words
    //put in the xhdr data now
    System.arraycopy(xhdr,0,buf,pos+4,xlen); //copy data into rtpbuf
    return pos+4+xlen;
  }
  /////////////////////////////////////////////////////////////
  //These force the output to IEEE format
  public static void netpackI2S (byte[] buf, int off, int val) {
    buf[off+1] = (byte)(val&0xFF);
    buf[off+0] = (byte)((val>>8)&0xFF); //ignores highest 2 bytes
  }
  /////////////////////////////////////////////////////////////
  public static void netpackL (byte[] buf, int off, int val) {
    buf[off+3] = (byte)(val&0xFF);
    buf[off+2] = (byte)((val>>8)&0xFF);
    buf[off+1] = (byte)((val>>16)&0xFF);
    buf[off+0] = (byte)((val>>24)&0xFF);
  }
}
