package nxm.sys.prim;

import nxm.sys.inc.*;
import nxm.sys.lib.*;
import nxm.sys.libg.GProgress;

/**
  Copies data from one file to another.

  @author Jeff Schoen
  @version $Id: noop.java,v 1.18 2012/10/04 12:44:49 jgs Exp $
*/
public class noop extends Primitive {

  public  static final String throttleList = "RealTime,Full,Block";
  private static final int REALTIME=1,FULL=2,BLOCK=3;
  private DataFile hi,ho;
  private Data data;
  private boolean wrap=false;
  private int throttle;
  private double tdelta,tstart,offs,size;
  private double rateWindow, rateWindowOffset=0;
  private static double RW_DEF=1.0, RW_MIN=0.01; 	// rate window default and minimum
  private GProgress gpw;


  public int open() {

    hi = MA.getDataFile ("IN");
    hi.open();

    size = MA.getD("/SIZE");
    offs = MA.getD("/START",0.0);

    ho = MA.getDataFile ("OUT",hi,0);
    if (size>0) ho.setSize(size);
    ho.open();

    if (MA.find("/PACKET")) {	// extract packets
      FileName pfn = MA.getFileName("/PACKET");
      ho.getPacketHandler().setFileName(ho,pfn);
    }
    if (MA.find("/GPW")) {	// graphical progress widget
      String text = "NOOP "+hi.getURL()+" to "+ho.getURL();
      gpw = new GProgress(this,"Transfer Progress","Run,Pause,Abort",text,0,0,0,this);
      gpw.setLabel("PROGRESS");
    }

    xfer = Math.max(1,(int)(bufSize/hi.dbpe));
    xfer = MA.getL("/TL",xfer);
    data = hi.getDataBuffer(xfer);
    wrap = MA.getState("/WRAP");

    tstart = 0;
    setRate( MA.getD("/RATE") );
    setRateWindow( MA.getD("/RATEWINDOW",RW_DEF) );
    throttle = MA.getState("/RT")? REALTIME : FULL;
    throttle = MA.getChoice("/THROTTLE",throttleList,throttle); // Override any /RT input

    if (offs>0) hi.seek(offs);
    if (hi.getFormatMode() != ho.getFormatMode()) {
      M.error("Mixed mode reformats are not supported: "+hi.getFormat()+" -> "+ho.getFormat());
    }
    return (NORMAL);
  }

  public int process() {

    if (throttle==REALTIME) {
      double tcur = Time.current();
      if (tstart==0 && hi.avail()<xfer) return (NOOP);
      if (tstart==0) tstart=tcur;
      if (tcur < (tstart+(ho.seek()-rateWindowOffset)*tdelta)) return (NOOP);
      if (tcur > (tstart+rateWindow)) resetWindow();
    }
    else if (throttle==BLOCK) return (NOOP);

    if (size>0 && ho.seek()>=size) return (FINISH);
    int n = hi.read(data);
    if (gpw!=null) gpw.setProgress(getProgress());
    if (size>0 && n+ho.seek()>size) n=(int)(size-ho.seek());

    if (n>0) ho.write(data,n);
    else if (n==0) return (NOOP);
    else if (wrap && hi.isFile()) hi.seek(0.0);
    else return (FINISH);
    return (NORMAL);
  }

  public int processMessage (Message msg) {
    if (msg.name.equals("PROGRESS")) {
      String text = (String)msg.data;
      // restart rate control when resuming after a pause
           if (text.equals("RUN")) { setState(RESUME); resetWindow(); }
      else if (text.equals("PAUSE")) setState(PAUSE);
      else if (text.equals("ABORT")) setState(FINISH);
      else if (text.equals("EXIT"))  setState(FINISH);
      return NORMAL;
    }
    else return super.processMessage(msg);
  }

  public int close() {
    hi.close();
    ho.close();
    if (gpw!=null) gpw.close();
    return (NORMAL);
  }

  /** get the 0->1 progress of the transfer */
  public double getProgress() {
    return hi.seek()/hi.getSize();
  }

  /** set RealTime transfer delta in sec per sample */
  public void setTDelta (double value) {
    tdelta = value;
  }
  /** set RealTime transfer rate in bytes/sec */
  public void setRate (double value) {
    if (value<=0) tdelta = hi.getDelta();
    else tdelta = hi.dbpe / value;
  }
  /** set RealTime throttle mode */
  public void setThrottle (String value) { 
    throttle = Parser.find(throttleList,value,throttle,0,Parser.E_WARNING);
  }

  /** set RealTime window (in sec) over which the rate should be controlled */
  public void setRateWindow (double value) {
    if (value < RW_MIN) {
      M.warning("Rate window reset to minimum value of "+RW_MIN+" seconds.");
      rateWindow = RW_MIN;
    } else {
      rateWindow = value;
    }
  }

  private void resetWindow() {
    tstart = Time.current();
    rateWindowOffset = ho.seek();
  }

}
