package nxm.sys.prim;

import java.applet.Applet;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;
import java.lang.reflect.*;
import java.util.*;
import javax.swing.*;
import javax.imageio.ImageIO;
import nxm.sys.inc.*;
import nxm.sys.lib.*;
import nxm.sys.libg.*;

import static java.lang.Math.max;
import static java.lang.Math.min;

/**
  Container for other plots and/or widgets.

  @author Jeff Schoen
  @version $Id: panel.java,v 1.165 2012/12/27 15:51:41 ntn Exp $
*/
public class panel extends GPrimitive implements WidgetPanel, Filter {

  private static final int sbw=4;  // Size of the border slider line (in pixels)
  private static final int bw=3;   // Border width (in pixels)
  private        final int maxLine=69;
  private        final int maxPane=65;
  private        final int maxWidget=256;

  private int exit=-1, filterMode=0;
  private int nline=0, npane=0, nwidget=0;
  private int iwz0=0, iwz1=0, iwz2=0;	// left/right widget group current indices
  private int dlh=50; 			// default logger height
  private int cfgx,cfgy,configure,options,eventFilter,dragline;
  private Line[] line = new Line[maxLine];
  private Pane[] pane = new Pane[maxPane];
  private GWidget[] widget = new GWidget[maxWidget];
  private String cfgfile,regname,titlebar;
  //themeName;  // removed in 3.3.0 since it is not needed
  private double lastPollTime;
  private int nTimers = 0;
  private double[] timer = null;
  private double[] timerinit = null;
  private Table cpos = new Table();
  private boolean inline,init=true,needLayout,headless;
  private int gnx,gny;
  private Object wmsgid;
  private GLabel curGLabel,lastCollapsableGLabel,lastOpenedGLabel;
  private MBox[] bbox = new MBox[5];
  private boolean logger;
  private boolean addNewLoggerPane;
  private int nleft,yleft,ybott,iScroll,oScroll,vScroll,cursorDir;
  private int needCompress, prevCompress, curCompress, maxCompress=10;
  private String  userCursor  = null; // The cursor set by the user, restored after line drag.
  private Table   xmCanvasTbl = null; // Table of XmCanvas elements that need to be configured.
  private String  altTitle;
  private String  currentTitle;
  private int     mobile,mpad=640;
  private Robot   robot;
  private boolean tsetup       = false;   //  /tsetup is present (with or without a Table)
  private Table   setupTable   = null;    //  setup Table
  private boolean parentResult = false;   //  the setup table is a parent result variable
  private boolean localResult  = false;   //  the setup table is a local result variable
  private boolean holdLayout   = false;   //  hold layout refreshes (Bug 2214)

  /** use {@link #getPanes()}. */
  private Table panes = new Table();
  /** use {@link #getControls()}. */
  private Table cntrls;

  private static final int EXIT_RETURN=0x1, EXIT_MENU=0x2,
          EXIT_MESSAGE=0x4, EXIT_WINDOW=0x8, EXIT_ALL=0xF;
  private static final String exitList = "RETURN,MENU,MESSAGE,WINDOW";

  private static final int ROTATETABS   = 0x01; // Option: Tabs Rotate.
  private static final int PUSHPOPX     = 0x02; // Option: Show 'X' for Push/Pop.
  private static final int COMPRESS     = 0x04; // Option: Compress widgets.
  private static final int COLLAPSE     = 0x08; // Option: Collapse roll-downs.
  private static final int LOCKSETUP    = 0x10; // Option: Lock setup (can't drag lines).
  private static final int LOCKCONTROLS = 0x20; // Option: Lock controls (can't drag lines).
  private static final int MINIDRAGZONE = 0x40; // Option: Tiny drag bars.
  private static final int ADVPUSHPOPX  = 0x80; // Option: Ctrl+click on 'X' to bring up pop menu (Since 3.1.0)

  private int optionsDefault = PUSHPOPX|COMPRESS|ADVPUSHPOPX; // The default options. ADVPUSHPOPX added for Bug 2458 in 3.1.0
  private int optionsDefaultMnM = LOCKSETUP|LOCKCONTROLS; // The default options.

  /** The list of options. */
  public static String optionsList = "RotateTabs,PushPopX,CompressControls,CollapseControls,"
                                   + "LockSetup,LockControls,MiniDragZone,AdvPushPopX";

  private static final int TABS_ROTATE = 0x1;            // Tab Option: Rotate tabs.
  private static final String tabOptionsList = "ROTATE"; // The list of tab options.

  private int tabOptions = TABS_ROTATE; // The tab options.

  /** Panel border options string list which corresponds with the BORDER Options. */
  public String bordersList = "Left,Right,Top,Bottom,TopLeft,TopRight,BottomLeft,BottomRight";
  /** Mnemonics for BORDER flag options.  Users should use names from the border list */
  private static final int BORD_LEFT=0x1, BORD_RIGHT=0x2, BORD_TOP=0x4, BORD_BOTTOM=0x8,
                           BORD_TOPLEFT=0x10, BORD_TOPRIGHT=0x20,
                           BORD_BOTTOMLEFT=0x40, BORD_BOTTOMRIGHT=0x80;

  private int borders=BORD_LEFT|BORD_RIGHT|BORD_TOP|BORD_BOTTOM;

  private static final int VLINE=1, HLINE=2, VLINEX=-1, HLINEX=-2;

  private static final int SETUP  = 1;
  private static final int TSETUP = 2;

  /** The pane/parent to place any controls on. If this is null, then MW is used.  */
  private MidasControls ctlPane = null;

  /** The top-level frame/panel being used if /JSETUP is used and the panel is a Container.  */
  private Container container = null;

  /** panel(s) of Macro that may be added to parent Macro */
  private Table  subPanels;
  /** name of the subPanel. This must match what is in the parent setup file */
  private String subPanelName;

  private boolean shareTheme;  // whether or not to use the Global theme
  private Theme theme;         // the theme to use  - since 3.3.0 Panel gets it's own theme reference

  @Override
  public int open () {
    String panelClass = MA.getCS("/JSETUP", null, "");  // null if not present, "" for state switch
    titlebar= "Panel";
    cfgfile = MA.getS("FILE");
    tsetup = MA.find("/TSETUP");
    if (tsetup)  setupTable = configTSetup();
    if (MA.find("/SETUP") ) setConfigFileName(SETUP);
    setTimers (MA.getL("/NTIMERS",3));
    regname = MA.getU("/PANES","WIN");
    String rname = MA.getU("/WMSGID");
    if (rname.length()>0) wmsgid = M.registry.getCommand(rname,5);
    if (M.macro==null) cntrls = new Table();
    else cntrls = M.macro.getControls();
    if (MA.find("/CONTROLS")) MR.put(MA.getS("/CONTROLS","GC"),cntrls);
    if (M.pipeMode==Midas.PINIT) exit=EXIT_WINDOW; else exit=EXIT_ALL;
    exit  = MA.getOptionMask("/EXIT",exitList,exit);
    if (MA.find("/TABS")) M.deprecate("The /TABS= list is now in the /OPTIONS= list");
    tabOptions = MA.getOptionMask("/TABS",tabOptionsList,tabOptions);
    if ((tabOptions&TABS_ROTATE)==0) optionsDefault &= ~ROTATETABS;
    headless = MA.getState("/HEADLESS", Shell.isHeadless());
    if (!headless && Shell.isHeadless()) // error when /HEADLESS=FALSE if the shell is headless
      M.error("PANEL: Can not set '/HEADLESS=" + MA.getS("/HEADLESS") + "' on a headless system.");
    inline = MA.getState("/INLINE",true);
    logger = MA.getState("/LOGGER");
    String themeName = MR.getS("ENV.THEME");
    Object theme = MA.getO("/THEME",themeName);
     // BEGIN BUG 2618  OPEN() CHANGES
    shareTheme = MA.getState("/SHARETHEME");  // for backwards compatibility
    this.theme = shareTheme ? Theme.global() : Theme.global().copy();
    if (theme!=null && theme instanceof Theme) this.theme = (Theme) theme;
    // END BUG 2618 OPEN() CHANGES
    boolean mnm = theme.toString().equals("MNM");
    options = MA.getOptionMask("/OPTIONS",optionsList,mnm?optionsDefaultMnM:optionsDefault);
    MW = new MWindow(titlebar,this);

    panes.put("PANE0",MW);

    if (panelClass == null) {
      if (!headless) MW.open();
      useDefaultSetup();
      addNewLoggerPane = logger;
      if (mnm) {
        Theme.setOptions("+MOBILE");
        mobile = -1;
        load("nxm.sys.mcr.mnm.mmp");
        mobile = 1;
        try { robot = new Robot(); }
        catch (Exception e) { M.printStackTrace("Unable to instantiate Robot class for event generation ",e); }
      }
      if (setupTable != null ) {
        try { setSetupTable(setupTable); }
        catch (Exception e) { M.printStackTrace("Error loading setup from table "+setupTable,e); useDefaultSetup(); }
      }
      else if (cfgfile.length()>0 && !tsetup) {
        try { load(cfgfile); }
        catch (Exception e) { M.printStackTrace("Error loading setup from mmp file "+cfgfile,e); useDefaultSetup(); }
      }
      // only add logger panel if not defined in setup file and /logger is present  (since 3.1.3 - Bug 2623)
      addNewLoggerPane = logger && addNewLoggerPane;
    }
    else {
      int width = 150;
      loadPanelClass(panelClass);
      if (ctlPane != null) {
        width = ctlPane.getWidth();
        if (!headless) MW.open();
      }
      for (int i=0; i<=4; i++) bbox[i] = new MBox();
      nline=4;
      for (int i=0; i<=4; i++) line[i]=new Line();
      setLine(0,HLINE,0,0.0F,0.0F,0,0);     // dummy
      setLine(1,VLINE,width,0.0F,1.0F,0,0); // left edge
      pane[0] = new Pane(this, "PANE0");
    }
    if (addNewLoggerPane && logger) addPane (-1, "LOGGER", "Message Logger", "LOGGER", 12, 1,2,4,0, 0);

    Table t = MA.getTable("/GRID");
    if (t!=null) autoGrid(t);

    if (headless) {
      init=false;
    } else {
      if (container == null) {                // No /JSETUP=
        MW.addTo(this);
      }
      else if (container instanceof Frame) {  // /JSETUP= with a Frame
        if (ctlPane != null) {
          MW.addTo(this); // <-- needed for the pane0 init
        }
      }
      else {                                  // /JSETUP= with a Panel
        MW.addTo(this); // <-- may change MW.parent

        if (MW.parent instanceof Container) {
          // JC4: If we don't do this, we get into situations where the JPanel
          //      specified via /JSETUP= does not show up, particularly in an
          //      Applet. This is because it is nominally given a size of [0,0]
          //      even if the preferred size of the JPanel is set. This in
          //      its self doesn't do the resizing, the .pack() call (below)
          //      does the rest -- except for applets where the size is already
          //      set by the browser.
          Container _parent = (Container)MW.parent;
          _parent.remove(container);
          _parent.setLayout(new BorderLayout());
          _parent.add(container, BorderLayout.CENTER);

          if (_parent instanceof Window) {
            // JC4: Added the frame.pack() call in Sept 2012 -- Without this, JOGL
            //      components on a /JSETUP= panel can get an exception when trying
            //      to access the "packer" in com.sun.opengl.util.j2d.TextRenderer's
            //      getBackingStore(..) following a user resizing the window. The
            //      best example of this is when using WorldWind with an
            //      AbstractTacticalSymbol visible.
            ((Window)_parent).pack();
          }
//          else if (_parent instanceof Applet) {
//            // JC4: Not sure why this is necessary... but if we don't do it we
//            //      get some of the toolbar widgets drawing funny (such as in
//            //      DEMOWIND;WORLDWIND when run as an applet).
//            container.invalidate();
//            container.revalidate();
//          }
        }
      }
      MW.setTheme(theme,inline);
      this.theme = MW.theme;  // make sure our Theme reference is correct - added in 3.3.0
      resize();
      refresh();
    }
    M.registry.put(regname,panes);

    if (logger) Shell.runCommand(M,"MSGLOG/ID=LOGGER/BANNER=PANE/WRAP="+mnm);

    return NORMAL;
  }

  @Override
  public int process () {
    try {
      if (xmCanvasTbl != null) {
        // Any X-Midas displays need to be pulled into the panel after M$SYNC has been called (i.e.
        // at the top of process()). If not running from X-Midas, xmCanvasTbl will be null.
        String keys[] = xmCanvasTbl.getKeys();
        for (int i = 0; i < keys.length; i++) {
          KeyObject.setKey(xmCanvasTbl.get(keys[i]), "XS", keys[i].substring(2));
        }
        xmCanvasTbl = null;
      }

      if (init && M.pipeMode!=Midas.PINIT) {
        if (curGLabel!=null)  maxCompress=curGLabel.pos.h/4;
        curGLabel=null;
        updateLastWidget();
        init=false;
      }

      // wait for pipe system to be running
      if (init) return NOOP;

      double time = Time.current();
      // If timers set, send timer message with timer number in the info field
      for (int i=0; i<nTimers; i++) {
        if (timer[i]!=0) {
          if (timer[i]!=0 && (time-timerinit[i])>Math.abs(timer[i])) {
            MQ.put("TIMER",i,new Double(time),"MAIN",this);
            if (timer[i]>0) timerinit[i]=time; else timer[i]=0;
          }
        }
      }

      for (int i=1; i<=nwidget; i++) widget[i].poll();

      if (needLayout) layout();
      if (needCompress!=0 || prevCompress!=0) {
        curCompress += needCompress;
        layout();
      }
      else if (is(COLLAPSE) && lastOpenedGLabel!=null) {
        lastOpenedGLabel.setActive(GLabel.aCOLLAPSABLE);
        lastOpenedGLabel = null;
      }
      if (time-lastPollTime>0.3) {
        lastPollTime = time;
        refreshTitle();
      }
    } catch (Exception e) {
      // Bug 1366 - PANEL should never crash
      M.printStackTrace(Thread.currentThread().toString(), e);
    }
    return NOOP;
  }

  @Override
  public int close () {
    int to=10; // wait for primitives writing to panes to go away
    for (Table.Iterator ti=panes.iterator(); ti.getNext(); ) {
      while (to>0 && M.registry.get(ti.key)!=null) { to--; Time.sleep(0.1); }
    }
    if (mobile!=0) Theme.setOptions("-MOBILE");
    MW.theme.setInline(false);
    MW.close();
    M.registry.remove(regname);

    if (container != null) container.setVisible(false);

    return NORMAL;
  }

  /** Loads the default setup for the panel. */
  private void useDefaultSetup () {
    for (int i=0; i<=4; i++) bbox[i]=new MBox();
    nline=4; for (int i=0; i<=4; i++) line[i]=new Line();
    setLine(0,HLINE,0,0.0F,0.0F,0,0);   // dummy
    setLine(1,VLINE,150,0.0F,0.5F,0,0); // left edge
    setLine(2,VLINE,50,1.0F,0.5F,0,0);  // right edge
    setLine(3,HLINE,50,0.5F,0.0F,1,2);  // top edge
    setLine(4,HLINE,50,0.5F,1.0F,1,2);  // bottom edge
    pane[0] = new Pane(this, "PANE0");
  }

  @Override
  public Container getGraphicsContainer () {
    // override to return the /JSETUP frame as needed
    return (container != null)? container: super.getGraphicsContainer();
  }

  @Override
  public void checkout () {
    for (int i=0; i < widget.length; i++) {
      if (widget[i] != null) widget[i].close();
    }
    super.checkout();
  }

  private boolean is (int mask) {
    return (options&mask) != 0;
  }

  public void refresh () {
    if (MQ != null) MQ.remove("REFRESH");
    if ((MW != null) && (MW.panel != null)) MW.panel.repaint();
  }

  /** A more comprehensive refresh than {@link #refresh()}. This method will call
      repaint on each widget and pane's graphics panel and then call {@link #refresh()}
      @since NeXtMidas 3.3.0
   */
  public void refreshAll () {
    if (!isHideControls()) {
      for (int i=0; i<widget.length; i++) {
        if (widget[i]!=null && widget[i].panel!=null) widget[i].panel.repaint();
        if (widget[i] instanceof GText) {
          ((GText)widget[i]).getMJTextArea().setAttributesGW((GText)widget[i]);   // make sure Gtext text area colors get updated
        }
      }
      for (int i=0; i<pane.length; i++) {
        if (pane[i]!=null && pane[i].panel!=null) pane[i].panel.repaint();
      }
    }
    refresh();
  }

  private void paint (Graphics g) {
    if (headless) return;
    MW.setGraphics(g);
    for (int i=1; i<=npane; i++) {
      Pane p = pane[i];  // just a pane refresh ?
      if (p!=null && p.pos.equals(MW.getClipBounds())) { p.paint(); return; }
    }
    MW.clear();
    for (int i=1; i<=nline; i++) {
      Line l=line[i];
      if (i<=4 && l.fp==0 && (borders&(1<<(i-1)))!=0) continue;
      MW.shadowbox(l.x1, l.y1, l.x2-l.x1, l.y2-l.y1, MWindow.SB_SEPERATOR);
      int xm = (l.x1+l.x2)/2;
      int ym = (l.y1+l.y2)/2;
      g.setColor(MW.theme.cwts);
      if (i==1 || i==2) {
        g.drawLine (l.x1+bw,l.y1+1*sbw,l.x2-bw,l.y1+2*sbw);
        g.drawLine (l.x2-bw,l.y1+2*sbw,l.x1+bw,l.y1+3*sbw);
        g.drawLine (l.x1+bw,l.y1+3*sbw,l.x2-bw,l.y1+4*sbw);
      }
      else if (i==3 || i==4) {
        g.drawLine (l.x1+1*sbw,l.y1+bw,l.x1+2*sbw,l.y2-bw);
        g.drawLine (l.x1+2*sbw,l.y2-bw,l.x1+3*sbw,l.y1+bw);
        g.drawLine (l.x1+3*sbw,l.y1+bw,l.x1+4*sbw,l.y2-bw);
      }
      else if (is(MINIDRAGZONE) && !is(LOCKSETUP)) {
        if (l.dir==HLINE) g.drawLine (xm-sbw,ym,xm+sbw,ym);
        if (l.dir==VLINE) g.drawLine (xm,ym-sbw,xm,ym+sbw);
      }
      if (configure>0) {
        MW.shadowbox(l.x1, l.y1, l.x2-l.x1, l.y2-l.y1, MWindow.SB_ALERT);
      }
    }
    for (int i=1; i<=npane; i++) {
      if (pane[i]!=null) pane[i].paint();
    }
    paintScroll(g);
  }

  private void paintScroll (Graphics g) {
    Line l = line[1];
    if (l != null) {
      vScroll = Math.max(10,Math.min(200,l.y2-l.y1-sbw*8));
      int xm = (l.x1+l.x2)/2;
      int ym = l.y2 - sbw*2 - vScroll;
      MWindow.drawVSlider(xm,ym,0,vScroll, (double)iScroll/nwidget,0, g,MW.theme);
    }
  }

  private void resize () {
    if (headless || isHideControls()) return;
    for (int i=1; i<=4; i++) fixLine(i); // borders need 2 passes of fixLine
    for (int i=1; i<=nline; i++) fixLine(i);
    for (int i=1; i<=npane; i++) fixPane(i);
    setBorderBoxes();
    layout();
  }

  /** Hold application of future layout changes (e.g. widget updates).
      Typically set to true before a large batch of gcontrol updates and then
      set back to false at the conclusion of those updates.
      @param hold True to holdLayout refreshes, false to allow them (normal operations)
      @since NeXtMidas 2.8.2
   */
  public void setHoldLayout (boolean hold) { holdLayout = hold; }

  /** Hold application of future layout changes (e.g. widget updates)? */
  public boolean getHoldLayout () { return holdLayout; }

  private void layout () {
    if (init || holdLayout) return;  // holdLayout check added in 2.8.2 for Bug 2214
    if (isHideControls()) return;
    Graphics g = MW.panel.getGraphics();
    if (g == null) return;

    nleft=yleft=0; ybott=bbox[1].y+bbox[1].h;
    if (!is(COMPRESS)) curCompress=0;
    iwz0=iwz1=iwz2=0;
    if (headless) return;
    int needCompressO = needCompress;
    int prevCompressO = prevCompress;
    lastOpenedGLabel = null;
    lastCollapsableGLabel = null;
    for (int i=1; i<=nwidget; i++) fixWidget(i,1);
    needLayout=false;
    prevCompress = needCompressO;
    needCompress = 0;
    if (needCompressO==0 && prevCompressO!=0) return;
    if (yleft>ybott && curCompress<maxCompress) needCompress=1;
    if (curCompress>0 && ybott-yleft>nleft && prevCompress<=0) needCompress=-1;
    if (is(COLLAPSE) && needCompress>0 && lastCollapsableGLabel!=null) {
      lastCollapsableGLabel.setActive(GLabel.aCOLLAPSED);
      needCompress=0;
    }
    if (!is(COMPRESS)) needCompress=0;
    if (yleft>ybott && iScroll<0) iScroll=0;
    if (yleft<ybott && iScroll==0) iScroll=-1;
    if (iScroll != oScroll) {
      paintScroll(g);
      oScroll = iScroll;
    }
  }

  @Override
  public int processMessage (Message msg) {
    String text;
//if (!"POINTER".equals(msg.name)) System.out.println("PANEL: got msg: "+msg.toString(-1)+" ["+getID()+"]"); // DEBUG

    // convert a few special sequences to specifically named messages
    if (msg.name.equals("KEYPRESS")) {
      text = msg.getS();
      if (text.equalsIgnoreCase("M")) msg.name = "MENU";
      if (text.equalsIgnoreCase("Enter") && (exit&EXIT_RETURN)!=0) msg.name = "EXIT";
    }
    if (msg.name.equals("BUTTON") && msg.info==2) {
      if (cfgClick (MW.px,MW.py)) return 1;
      msg.name = "MENU";
    }

    // all drawing is done via the PAINT message, which only the drawing thread sends.
    if (msg.name.equals("PAINT")) {
      paint ((Graphics)msg.data);
    }
    else if (msg.name.equals("RESIZE")) {
      resize();
    }
    /* if initializing queue for me */
    else if (msg.name.equals("POINTER")) {
      int i = selLine (MW.px, MW.py);
      int cdir = (i>0 && msg.info!=-9)? line[i].dir : 0;
      if (cdir!=cursorDir) setResizeCursor(cdir);
      cursorDir = cdir;
    }
    else if (msg.name.equals("DRAG")) {
      MBox box = (MBox)msg.data;
      if (dragline==0) dragline = selLine (box.x,box.y);
      if (dragline>0) movLine(dragline,box.x+box.w,box.y+box.h);
      if (dragline==-1) movScroll(box.y+box.h);
    }
    else if (msg.name.equals("DRAGBOX")) {
      MBox box = (MBox)msg.data;
      if (msg.info==1 && dragline>0) movLine(dragline,box.x+box.w,box.y+box.h);
      dragline=0;
    }
    else if (msg.name.equals("BUTTON")) {
      int i = selLine (MW.px,MW.py);
      if (i==0) {
        i = selPane (MW.px,MW.py);
        if (i!=0) {
          // Bug 2458: Check to see if CTRL key was held while clicking
          if (msg.data instanceof MouseEvent){
            MouseEvent me = (MouseEvent)msg.data;
            boolean isCtrl = (me.getModifiersEx()&InputEvent.CTRL_DOWN_MASK)!=0;
            hitPane(i,isCtrl);
          }
          else hitPane(i);
        }
      }
      else if (i==-1) movScroll(MW.py);
      else if (((i<=2 && MW.py>line[i].y1 && MW.py<line[i].y1+sbw*5) ||
          (i>2 && i<=4 && MW.px>line[i].x1 && MW.px<line[i].x1+sbw*5)) ) {
        // hit the auto collapse/expand border zones
        borders ^= (1<<(i-1)); resize(); refresh();
      }
    }
    else if (msg.name.equals("SHOW")) {
      MW.display(msg.info);
    }
    else if (msg.name.equals("SHOWN")) {
      if (MW.status!=MWindow.INIT) MW.status=msg.info;
      sendMessage (msg);
    }
    else if (msg.name.equals("REFRESH")) {
      refresh(); // doesn't actually draw
    }
    else if (msg.name.equals("RESIZE")) {
      resize();
    }
    else if (msg.name.equals("MENU")) {
      new GMenu (MW,"Panel","Configure,Controls,Borders,Options,Theme,Query,Macro,Debug,Exit",0,0,this);
    }
    else if (msg.name.equals("PANEL")) {
      text = msg.getS();
      if (text.equals("CONFIGURE")) {
      if (configure>0) text="NoModify"; else text="Modify";
        new GMenu (MW,"Panel.cfg","Setup,Clear,Load,SaveResult,Save,"+text,0,0,this);
      }
      else if (text.equals("CONTROLS")) {
        new GMenu (MW,"Controls","Display,Events",0,0,this);
      }
      else if (text.equals("BORDERS")) {
        new GMenu (MW,"Borders",bordersList,borders,GMenu.TOGGLE,this);
      }
      else if (text.equals("OPTIONS")) {
        new GMenu (MW,"Options",optionsList, options,GMenu.TOGGLE,this);
      }
      else if (text.equals("THEME")) {
        new GMenu (MW,"Theme",MWindow.themeList,0,0,this);
      }
      else if (text.equals("QUERY")) {
        new GQuery (MW,"Query Results",MR,0,this);
      }
      else if (text.equals("MACRO")) {
        new GMenu (MW,"Macro","Run,Pause,End,Registry,Results,Pipes,Commands,Remotes",0,0,this);
      }
      else if (text.equals("DEBUG")) {
        new GMenu (MW,"Debug","Results,PipeMon,MsgMon,Terminal",0,0,this);
      }
      else if (text.equals("EXIT")) {
        msg.name = "EXIT";
        if ((exit&EXIT_MENU)!=0) MQ.put(msg);
      }
    }
    else if (msg.name.equals("MACRO")) {
      text = msg.getS();
           if (M.macro==null);
      else if (text.equals("RUN")) M.macro.setPipeMode(Midas.PWAIT);
      else if (text.equals("PAUSE")) M.macro.setPipeMode(Midas.PPAUSE);
      else if (text.equals("END")) M.macro.setPipeMode(Midas.PSTOP);
      else if (text.equals("REGISTRY")) new GQuery (MW,"Query Registry",M.registry.table,0,this);
      else if (text.equals("RESULTS")) new GQuery (MW,"Query Results",MR,0,this);
      else if (text.equals("PIPES")) {
        Table t = (Table)MR.get("RAM");
        new GMenu (MW,"Pipe",t.getKeyList(),0,GMenu.STICKY,this);
      }
      else if (text.equals("COMMANDS")) {
        Table t = (Table)MR.get("REG"); filterMode=1;
        new GMenu (MW,"Command",t.getKeyList(this),0,GMenu.STICKY,this);
      }
      else if (text.equals("REMOTES")) {
        Table t = (Table)MR.get("REG"); filterMode=2;
        new GMenu (MW,"Remote",t.getKeyList(this),0,GMenu.STICKY,this);
      }
    }
    else if (msg.name.equals("DEBUG")) {
      text = msg.getS();
           if (text.equals("RESULTS")) new GQuery (MW,"Query Results",MR,0,this);
      else if (text.equals("PIPEMON")) Shell.runCommand (M,"GCONTROL,PIPEMON");
      else if (text.equals("MSGMON")) Shell.runCommand (M,"GCONTROL,MSGMON");
      else if (text.equals("TERMINAL")) Shell.runCommand (M,"SHELLGUI/ATTACH/BG/DISPLAY=FRONT");
    }
    else if (msg.name.equals("PIPE")) {
      text = msg.getS();
      new GQuery (MW,"Query "+text,M.pipes.get(text),GQuery.NOEDIT,this);
    }
    else if (msg.name.equals("COMMAND") || msg.name.equals("REMOTE")) {
      text = msg.getS();
      new GQuery (MW,"Query "+text,M.registry.get(text),0,this);
    }
    else if (msg.name.equals("PANEL.CFG")) {
      text = msg.getS();
      if (text.equals("MODIFY")) {
        if (is(LOCKSETUP)) {
          new GAlert (MW,"Setup","Setup is locked, can not modify.","OK",1,0,this,4);
        } else {
          configure=1; refresh();
        }
      }
      else if (text.equals("NOMODIFY")) {
        configure=0; refresh();
      }
      else if (text.equals("CLEAR")) {
        for (int i=npane; i>0; i--) delPane(i);
        npane=0; nline=4; refresh();
      }
      else if (text.equals("SETUP")) {
        int nChars = getPromptNumChars(80); // Desire an 80 width input prompt
        if (M.macro!=null) new GPrompt (MW,"LoadSetup","",nChars,0,this);
        else new GAlert (MW,"Setup","No macro available for setup","OK",1,0,this);
      }
      else if (text.equals("LOAD")) {
        int nChars = getPromptNumChars(cfgfile.length());
        new GPrompt (MW,"LoadFile",cfgfile,nChars,0,this);
      }
      else if (text.equals("SAVERESULT")) {
        if (cfgfile.contains(".mmp") || cfgfile.contains(".tbl")) cfgfile = "setupTable";
        int nChars = Math.max(getPromptNumChars(cfgfile.length()), 25);
        new GPrompt (MW,"SaveTableResult",cfgfile,nChars,0,this);
      }
      else if (text.equals("SAVE")) {
        if (tsetup) {
          if (!cfgfile.endsWith(".tbl")) setConfigFileName(TSETUP);
        } else setConfigFileName(SETUP);
        int nChars = Math.max(getPromptNumChars(cfgfile.length()), 25);
        new GPrompt (MW,"SaveFile",cfgfile,nChars,0,this);
      }
    }
    else if (msg.name.equals("CFG.ADD")) {
      text = msg.getS();
           if (text.equals("HLINE")) addLine (HLINE,cfgx,cfgy);
      else if (text.equals("VLINE")) addLine (VLINE,cfgx,cfgy);
      else if (text.equals("HLINEX")) addLine (HLINEX,cfgx,cfgy);
      else if (text.equals("VLINEX")) addLine (VLINEX,cfgx,cfgy);
      else if (text.equals("PANE")) addPane (0,cfgx,cfgy);
      configure=1;
    }
    else if (msg.name.equals("THEME")) {
      String themeName = msg.getS();
      MW.setTheme(themeName,inline);
      this.theme = MW.theme;
      for (int i=1; i<=nwidget; i++) {
        if (!widget[i].hasLocalTheme()) widget[i].setTheme(MW.theme);
      }
      for (int i=1; i<=npane; i++) pane[i].setTheme(MW.theme);
      resize();
      refreshAll(); // make sure everything refreshes
    }
    else if (msg.name.equals("CONTROLS")) {
      text = msg.getS();
      if (text.equals("DISPLAY")) { borders ^= BORD_LEFT; resize(); refresh(); }
      else if (text.equals("EVENTS"))
        new GMenu (MW,"Events",MWindow.eventFilterList,eventFilter,GMenu.TOGGLE,this);
    }
    else if (msg.name.equals("EVENTS")) {
      eventFilter = msg.info;
      for (int i=1; i<=nwidget; i++) widget[i].setEventFilter(eventFilter);
    }
    else if (msg.name.equals("BORDERS")) {
      //jph -- Use set method here instead, it does the same thing
      //TBD -- Use setBorders(String) preferred
      //borders = msg.info; resize(); refresh();
      setBorders(msg.info);
    }
    else if (msg.name.equals("OPTIONS")) {
      options = msg.info; refresh();
    }
    else if (msg.name.equals("LOADSETUP")) {
      String cfgfile = M.macro.getFileName()+"p"+msg.getS();
      load(cfgfile); resize(); refresh();
    }
    else if (msg.name.equals("LOADSETUPTABLE")) {
      setSetupTable(Convert.o2t(msg.data)); resize(); refresh();
    }
    else if (msg.name.equals("LOADFILE")) {
      load(msg.getS()); resize(); refresh();
    }
    else if (msg.name.equals("SAVEFILE")) {
      save(msg.getS());
    }
    else if (msg.name.equals("SAVETABLERESULT")) {
      saveResult(msg.getS());
    }
    else if (msg.name.equals("SAVETABLEFILE")) {
      saveSetupTable(msg.getS());
    }
    else if (msg.name.equals("WINDOW")) {
      // since 3.3.0: handle mouse and messages for ICON/DEICONIED (BUG 2690) on PANEL
      if ("DEICONIFIED".equals(msg.data)) {
        MW.display(MWindow.FRONT);
      } else if ("ICON".equals(msg.data)) {
        MW.display(MWindow.ICON);
      }
      // TODO: The above ~5 lines of code should move to a common place in GPrimitive (or MWindow)
      //       It also matches code in MWindow.process(String name, int info, Object data)

      if (wmsgid==null) {
        if ((exit&EXIT_WINDOW)!=0 && "CLOSING".equals(msg.data) && M.macro!=null) {
          M.macro.setPipeMode(Midas.PSTOP);
        }
      } else sendWMessage (msg);
    }
    else if (msg.name.equals("EXIT")) {
      if ((exit&EXIT_MESSAGE)!=0) return FINISH;
    }
    else if (msg.name.equals("CURSOR")) {
      setCursor(msg.data.toString());
    }

    return NORMAL;
  }

  /* Compute the number of characters for the prompt for SAVE or LOAD configuration file */
  private int getPromptNumChars (int desiredNumChars) {
    int panelWidth = MW.panel.getWidth();
    int charWidth  = MW.fm.charWidth('W'); // Find width of widest character
    int maxChars   = Math.max(10, (panelWidth/charWidth)-17);
    // This is, reserve some space for SAVEFILE or LOAD title and border
    int numChars = Math.min(desiredNumChars,maxChars);
    return numChars;
  }

  private boolean cfgClick (int x, int y) {
    int i;
    if (configure!=1) return false;
    i = selLine (x,y);
    if (i>4) {
      Line l = line[i];
      new GMenu (MW,"Cfg.Line","Cancel,Extend,Shrink,Fix,UnFix,Delete",1,0,l);
      return true;
    }
    i = selPane (x,y);
    if (i>0) {
      Pane p = pane[i];
      cfgx=x; cfgy=y;
      new GMenu (MW,"Cfg.Pane","Cancel,TitleText,TitleColor,TitleSize,Attach,Delete,AddTab",1,0,p);
      return true;
    }
    if (x<line[1].x2) return false;
    cfgx=x; cfgy=y;
    new GMenu (MW,"Cfg.Add","Cancel,Vline,HLine,VLineX,HLineX,Pane",1,0,this);
    return true;
  }

  public static String eventFilterList = MWindow.eventFilterList;

  /** set or modify event filter */
  public void setEventFilter (String mask) {
    int ef = 0;
    if (mask.equals("ENABLE")) ef=0;
    else if (mask.equals("DISABLE")) ef=-1;
    else ef = Parser.mask(eventFilterList,mask,eventFilter);
    for (int i=1; i<=nwidget; i++) widget[i].setEventFilter(ef);
    eventFilter = ef;
  }
  public String getEventFilter () {
    return Parser.mask2s(eventFilterList,eventFilter);
  }

  /** set or modify options */
  public void setOptions (String mask) {
    options = Parser.mask(optionsList,mask,options);
    layout();
    refresh();
  }

  private void sendWMessage (Message msg) {
    if (wmsgid instanceof MessageHandler) {
      if (wmsgid == msg.from) return;    // don't send it back
      msg.to = wmsgid; msg.from = this;
      ((MessageHandler)wmsgid).processMessage(msg);
    }
  }

  private int getix (float x) {
    int i1 = line[1].x2-bw;
    int i2 = line[2].x1+bw;
    return (int)(x*(i2-i1))+i1;
  }

  private int getiy (float y) {
    int i1 = line[3].y2-bw;
    int i2 = line[4].y1+bw;
    return (int)(y*(i2-i1))+i1;
  }

  private float getrx (int x) {
    int i1 = line[1].x2-bw;
    int i2 = line[2].x1+bw;
    return (float)(x-i1)/(float)Math.max(1,i2-i1);
  }

  private float getry (int y) {
    int i1 = line[3].y2-bw;
    int i2 = line[4].y1+bw;
    return (float)(y-i1)/(float)Math.max(1,i2-i1);
  }

  private void addLine (int dir, int x, int y) {
    float rx,ry; int i,i1,i2,ii1,ii2, n=nline+1;
    boolean extend = (dir<0);
    if (extend) dir = -dir;
    if (getrx(x)<=0||getrx(x)>=1||getry(y)<=0||getry(y)>=1) {
      new GAlert (MW,"AddLine","Must be inside the four fixable walls","OK",1,0,this);
      return;
    }
    line[n] = new Line();
    line[n].dir=dir;
    line[n].rx=rx=getrx(x);
    line[n].ry=ry=getry(y);
    if (dir==VLINE) {
      line[n].x1 = x-bw; line[n].x2 = x+bw;
      if (extend) { i1=3; i2=4; }
      else for (i1=3,i2=4,i=5; i<=nline; i++) {
        ii1 = line[i].i1; ii2 = line[i].i2;
        if (line[i].dir==HLINE && line[ii1].rx<rx && line[ii2].rx>rx
          && line[i].ry<ry && line[i].ry>line[i1].ry) i1=i;
        if (line[i].dir==HLINE && line[ii1].rx<rx && line[ii2].rx>rx
          && line[i].ry>ry && line[i].ry<line[i2].ry) i2=i;
      }
      line[n].i1 = i1;
      line[n].i2 = i2;
    }
    if (dir==HLINE) {
      line[n].y1 = y-bw; line[n].y2 = y+bw;
      if (extend) { i1=1; i2=2; }
      else for (i1=1,i2=2,i=5; i<=nline; i++) {
        ii1 = line[i].i1; ii2 = line[i].i2;
        if (line[i].dir==VLINE && line[ii1].ry<ry && line[ii2].ry>ry
          && line[i].rx<rx && line[i].rx>line[i1].rx) i1=i;
        if (line[i].dir==VLINE && line[ii1].ry<ry && line[ii2].ry>ry
          && line[i].rx>rx && line[i].rx<line[i2].rx) i2=i;
      }
      line[n].i1 = i1;
      line[n].i2 = i2;
    }
    fixLine(n);
    nline=n;
    refresh();
  }

  private void addPane (int ip, int x, int y) {
    Pane p = new Pane(this);
    if (ip==0) for (ip=1; ip<=npane && pane[ip]!=null; ip++);
    pane[ip]=p;
    p.ix1 = p.ix2 = p.iy1 = p.iy2 = 0;
    for (int i=1; i<=nline; i++) {
      if (line[i].dir == VLINE) {
        if (line[i].y1<y && line[i].y2>y) {
          if (line[i].x2<x && (p.ix1==0 || line[i].x2>line[p.ix1].x2)) p.ix1=i;
          if (line[i].x1>x && (p.ix2==0 || line[i].x1<line[p.ix2].x1)) p.ix2=i;
        }
      } else {
        if (line[i].x1<x && line[i].x2>x) {
          if (line[i].y2<y && (p.iy1==0 || line[i].y2>line[p.iy1].y2)) p.iy1=i;
          if (line[i].y1>y && (p.iy2==0 || line[i].y1<line[p.iy2].y1)) p.iy2=i;
        }
      }
    }
    p.name = "PANE"+ip;
    p.title = "Pane "+ip;
    p.ts = 12;
    fixPane(ip);
    panes.put(p.name,p);
    npane = Math.max(ip,npane);
    refresh();
  }

  private void delPane (Pane p) {
    for (int i=1; i<=npane; i++) if (p==pane[i]) delPane(i);
  }

  private void delPane (int i) {
    if (i==0) return;
    Pane p = pane[i];
    if (p.panel != null) MW.panel.remove(p.panel);
    for (--npane; i<=npane; i++) pane[i] = pane[i+1];
    pane[npane+1] = null;      //bug 1881
    getPanes().remove(p.name); //bug 1881
    refresh();
  }

  private void fixPane (int i) {
    Pane p = pane[i];
    if (p==null) return;
    int x1=0, x2=MW.pos.w, y1=0, y2=MW.pos.h;
    if (p.ix2==1) {           x2 = line[1].x1;
      if ((borders&BORD_TOPLEFT)!=0) y1 = line[3].y2;
      if ((borders&BORD_BOTTOMLEFT)!=0) y2 = line[4].y1;
    }
    else if (p.ix1==2) {      x1 = line[2].x2;
      if ((borders&BORD_TOPRIGHT)!=0) y1 = line[3].y2;
      if ((borders&BORD_BOTTOMRIGHT)!=0) y2 = line[4].y1;
    }
    else if (p.iy2==3) {      y2 = line[3].y1;
      if ((borders&BORD_TOPLEFT)==0) x1 = line[1].x2;
      if ((borders&BORD_TOPRIGHT)==0) x2 = line[2].x1;
    }
    else if (p.iy1==4) {      y1 = line[4].y2;
      if ((borders&BORD_BOTTOMLEFT)==0) x1 = line[1].x2;
      if ((borders&BORD_BOTTOMRIGHT)==0) x2 = line[2].x1;
    }
    else {
      x1 = line[p.ix1].x2;
      x2 = line[p.ix2].x1;
      y1 = line[p.iy1].y2;
      y2 = line[p.iy2].y1;
    }
    p.x=x1; p.y=y1+3*p.ts/2; p.w=x2-p.x; p.h=y2-p.y;
    if (i>0 && mobile!=0) { p.w=mpad; p.h=mpad; }
    p.hidden=false;
    if (p.ix2==1) p.hidden = ((borders&BORD_LEFT)==0);
    if (p.ix1==2) p.hidden = ((borders&BORD_RIGHT)==0);
    if (p.iy2==3) p.hidden = ((borders&BORD_TOP)==0);
    if (p.iy1==4) p.hidden = ((borders&BORD_BOTTOM)==0);

    for (int ip=1; ip<i; ip++) {
      Pane pp = pane[ip];
      if (pp!=null && pp.ix1==p.ix1 && pp.iy1==p.iy1) {
        if (p.getTabManager() == null) {
          if (pp.getTabManager() == null) {
            pp.setTabManager(new TabManager());
          }
          p.setTabManager(pp.getTabManager());
        }
      }
    }
    if (p.panel==null) return;
    if (p.mw!=null && p.mw.pushParent!=null) return;
    boolean vis = (!p.hidden) && ((p.getTabManager() == null)
                || p.getTabManager().isVisibleTab(p));
    p.panel.setVisible(vis);
    if (vis) p.panel.setBounds(p.x,p.y,p.w,p.h);
  }

  /** This is called when a tab is clicked, and other times.
      @param i The pane number.
   */
  private void hitPane (int i) {
    hitPane(i,false);
  }

  /** This is called when a tab is clicked, and other times.
      @param i The pane number.
      @param isCtrl Whether or not the CTRL key is held down
      @since NeXtMidas 3.1.0 (Bug 2458)
   */
  private void hitPane (int i, boolean isCtrl) {
    if (i<0) {
      Pane p = pane[-i];
      if (p!=null && p.panel!=null) p.panel.setVisible(true);
      if (p!=null && p.mw!=null && p.mw.status!=MWindow.EXIT) p.mw.pop(-1,isCtrl&&((options&ADVPUSHPOPX)!=0));
    } else {
      Pane p = pane[i];
      if (p.getTabManager() != null) {
        p.makeVisibleTab(); /* calls updateTabs() */
      } else {
        resize();
        refresh();
      }
    }
  }

  /** This calls a resize() and a refresh() to update the tabs. */
  private void updateTabs () {
    resize();
    refresh();
  }

  private void movScroll (int y) {
    int y2 = line[1].y2 - sbw*2;
    int y1 = y2 - vScroll;
    double f = (double)(y-y1)/(y2-y1);
    int i = (int)(f*nwidget);
    if (setScroll(i)) layout();
  }

  private boolean setScroll (int i) {
    if (i==iScroll) return false;
    iScroll = i;
    return true;
  }

  private void setLine (int i, int dir, int fp, float rx, float ry, int i1, int i2) {
    line[i].dir=dir;
    line[i].fp=fp;
    line[i].rx=rx;
    line[i].ry=ry;
    line[i].i1=i1;
    line[i].i2=i2;
    if (dir==VLINE) {
      line[i].y1=0; line[i].y2=MW.pos.h;
      line[i].x1=getix(rx); line[i].x2=getix(rx);
    } else {
      line[i].x1=0; line[i].x2=MW.pos.w;
      line[i].y1=getiy(ry); line[i].y2=getix(ry);
    }
  }

  private void delLine (Line l) {
    for (int i=1; i<=nline; i++) if (l==line[i]) delLine(i);
  }

  private int delLine (int i) {
    for (int j=5; j<=nline; j++) {
      if (line[j].i1>=i) line[j].i1--;
      if (line[j].i2>=i) line[j].i2--;
    }
    for (int j=i; j<nline; j++) line[j]=line[j+1];
    nline--;
    resize();
    refresh();
    return 0;
  }

  private void fixLine (int i) {
    Line l = line[i]; int i1,i2;
    if (i<=4) {
      if ((borders&(1<<(i-1)))==0) { i1=0; i2=sbw+bw+bw; }
      else if (l.fp==0) i1=i2=0;
      else { i1=l.fp-sbw-bw; i2=l.fp+bw; }
      if (i==1) {
        l.x1=i1; l.x2=i2;
        if ((borders&BORD_TOPLEFT)!=0) l.y1=line[3].y2; else l.y1=0;
        if ((borders&BORD_BOTTOMLEFT)!=0) l.y2=line[4].y1; else l.y2=MW.pos.h;
      }
      if (i==2) {
        l.x1=MW.pos.w-i2; l.x2=MW.pos.w-i1;
        if ((borders&BORD_TOPRIGHT)!=0) l.y1=line[3].y2; else l.y1=0;
        if ((borders&BORD_BOTTOMRIGHT)!=0) l.y2=line[4].y1; else l.y2=MW.pos.h;
      }
      if (i==3) {
        l.y1=i1; l.y2=i2;
        if ((borders&BORD_TOPLEFT)!=0) l.x1=0; else l.x1=line[1].x2;
        if ((borders&BORD_TOPRIGHT)!=0) l.x2=MW.pos.w; else l.x2=line[2].x1;
      }
      if (i==4) {
        l.y1=MW.pos.h-i2; l.y2=MW.pos.h-i1;
        if ((borders&BORD_BOTTOMLEFT)!=0) l.x1=0; else l.x1=line[1].x2;
        if ((borders&BORD_BOTTOMRIGHT)!=0) l.x2=MW.pos.w; else l.x2=line[2].x1;
      }
    }
    else if (l.dir==HLINE) {
      l.x1 = getix(line[l.i1].rx)+bw;
      l.x2 = getix(line[l.i2].rx)-bw;
      l.y1 = getiy(line[i].ry)-bw;
      l.y2 = getiy(line[i].ry)+bw;
    }
    else if (l.dir==VLINE) {
      l.y1 = getiy(line[l.i1].ry)+bw;
      l.y2 = getiy(line[l.i2].ry)-bw;
      l.x1 = getix(line[i].rx)-bw;
      l.x2 = getix(line[i].rx)+bw;
    }
  }

  /** Moves a line.
      @param i  The line index.
      @param x  The x coordinate (in pixels).
      @param y  The y coordinate (in pixels).
   */
  private void movLine (int i, int x, int y) {
    if (is(LOCKSETUP)) return;
    Line l = line[i];
    x = Math.max(0,Math.min(MW.pos.w,x));
    y = Math.max(0,Math.min(MW.pos.h,y));
         if (i==1) l.fp=x;
    else if (i==2) l.fp=MW.pos.w-x;
    else if (i==3) l.fp=y;
    else if (i==4) l.fp=MW.pos.h-y;
    else if (l.dir==HLINE) { l.ry = getry(y); if (l.fp!=0) l.fp=y; }
    else if (l.dir==VLINE) { l.rx = getrx(x); if (l.fp!=0) l.fp=x; }
    if (i>=1 && i<=4) borders |= (0x1<<(i-1));
    if (i==1 && curCompress!=0) {
      options &= ~COMPRESS;
      resize();
      refresh();
      resize();
      refresh();
      options |= COMPRESS;
    }
    resize();
    refresh();
  }

  /** Indicates the line that is at a given point.
      @param x The x-coord.
      @param y The y-coord.
      @return The line at (x,y).
   */
  private int selLine (int x, int y) {
    for (int i=1; i<=nline; i++) {
      Line l=line[i]; if (l==null) continue;
      if (x>=l.x1 && x<=l.x2 && y>=l.y1 && y<=l.y2) {	// on the separator
        // don't allow select of the controls scroll bar
        if (i==1 && y>l.y2-sbw*4-vScroll) return -1;
        // setup is locked, no resize, but collapse is ok
        if (configure<=0 && i>4 && is(LOCKSETUP)) continue;
        if (configure<=0 && is(MINIDRAGZONE)) {
          if ((i==1 || i==2) && y>l.y1+4*sbw) continue;
          if ((i==3 || i==4) && x>l.x1+4*sbw) continue;
          if (i>4 && l.dir==HLINE && Math.abs(x-(l.x1+l.x2)/2)>sbw*2) continue;
          if (i>4 && l.dir==VLINE && Math.abs(y-(l.y1+l.y2)/2)>sbw*2) continue;
        }
        return i;
      }
    }
    return 0;
  }

  /** Indicates the pane that is at a given point.
      @param x The x-coordinate
      @param y The y-coordinate
      @return The pane at (x,y).
   */
  private int selPane (int x, int y) {
    for (int i=1; i<=npane; i++) {
      Pane p=pane[i]; if (p==null) continue;
      int hit = p.clickHit(x,y);
      if (hit!=0) return i*hit;
    }
    return 0;
  }

  /** Arrange the widget on the panel
      @param i The index of the widget in the {@link #widget} array
      @param mode The display mode. <br>
        -1 = set bounds even if widget is not shown. Used by initial call in {@link #addWidget(GWidget)}
         1 = do not set bounds if widget is not shown. Used by calls from {@link #layout()}
   */
  private void fixWidget (int i, int mode) {
    GWidget gw=widget[i];
    GLabel gwg=gw.pGroup;
    boolean isLabel=(gw instanceof GLabel);
    boolean show=(gw.isShow())&&(gwg==null || gwg.isShow()); // A hidden group means a hidden control

    if (headless) { gw.display(-1); return; }

    // recalc rx,ry if this was moved by the user
    if (gw.rMoved && !is(LOCKCONTROLS)) {
      setPos2RXY(gw);
      if (isLabel || gwg==null || gwg.rx>0);
      else if (gw.rx>0) { gw.pDetach=true; gw.setFlag(GWidget.INLINE,false); }
      else if (gw.rx<0) { gw.pDetach=false; gw.setFlag(GWidget.INLINE,true); }
      gw.rMoved=false;
    }

    // no show by label activity ?
    if (gwg==null);
    else if (isLabel) show &= (gwg.getActive()!=GLabel.aHIDDEN);
    else if (!gw.pDetach) {
      show &= (gwg.getActive()>0);
      gw.rx=gwg.rx;
    }
    if (gw.rx<0 && gw.pos.w!=line[1].x1) {
      gw.setCompression(0);
      gw.setBounds(0,gw.pos.y,line[1].x1,gw.pos.h);
    }
    if (i<=iScroll) show = false;
    //if (!show) { if (mode>0) gw.display(-1); return; } // changed to block below for Bug 2217
    if (!show && mode>0 ) {
      gw.display(GWidget.ICON);
      return;
    }

    // no show by border activity ?
         if (gw.rx<0) show = (borders&BORD_LEFT)!=0;
    else if (gw.rx>1) show = (borders&BORD_RIGHT)!=0;
    else if (gw.ry<0) show = (borders&BORD_TOP)!=0;
    else if (gw.ry>1) show = (borders&BORD_BOTTOM)!=0;
    // if (!show) { if (mode>0) gw.display(-1); return; }  // changed to block below for Bug 2217
    if (!show && mode>0 ) {
      gw.display(GWidget.ICON);
      return;
    }

    // find the next row for an outside group entry
    if (gwg!=null && gwg!=gw && gwg.rx>0 && gwg.rx<1 && !gw.pDetach) {
      int x = widget[iwz0].pos.x;
      int y = widget[iwz0].pos.y + widget[iwz0].pos.h;
      int w = widget[iwz0].pos.w;
      int h = gw.pos.h;
      gw.fixWidth(w);
      gw.setBounds(x,y,w,h);
      iwz0=i;
    }
    else if (gw.rx<0) {    // find next row in left column
      int x = 0;
      int y = bbox[1].y;
      int w = line[1].x1;
      gw.fixWidth(w);
      // don't refresh during compression move to reduce flicker
      if (needCompress!=0) gw.eventFilter |= GWidget.NOREFRESH;
      else gw.eventFilter &= ~GWidget.NOREFRESH;
      gw.setCompression(curCompress);
      int h = gw.pos.h;
      if (iwz1>0) y = widget[iwz1].pos.y+widget[iwz1].pos.h;
      gw.setBounds (x,y,w,h);
      if ((borders&BORD_BOTTOMLEFT)!=0) show = y+h < ybott; // scrolled off-screen bottom
      else show = y+5 < ybott; // scrolled off-screen bottom - allow partial widgets
      if (show && needCompress==0 && prevCompress!=0) gw.refresh();
      yleft = y+h;
      nleft++;
      iwz1=i;
      if (isLabel && gwg.getActive()==GLabel.aOPEN) lastOpenedGLabel=gwg;
      if (isLabel && gwg.getActive()==GLabel.aCOLLAPSABLE) lastCollapsableGLabel=gwg;

    }
    else if (gw.rx>1) {    // find next row in right column
      int x = line[2].x2;
      int y = bbox[2].y;
      int w = MW.pos.w-line[2].x2;
      int h = gw.pos.h;
      if (iwz2>0) y = widget[iwz2].pos.y+widget[iwz2].pos.h;
      gw.fixWidth(w);
      gw.setBounds(x,y,w,h);
      show = y+h < bbox[2].y+bbox[2].h; // scrolled off-screen bottom
      iwz2=i;
    }
    else {     // if not in the left border, find scaled location
      setRXY2Pos(gw);
      gw.setBounds(gw.pos);
      if (isLabel) iwz0=i;
    }
    if (mode<0)    ;                // skip (show/hide)
    else if (show) gw.display(1);   // show widget
    else           gw.display(-1);  // hide widget
  }

  private void setBorderBoxes () {
    int bx1, bx2, by1, by2;
    // calc middle region
    bx1=line[1].x2;
    bx2=line[2].x1;
    by1=line[3].y2;
    by2=line[4].y1;
    bbox[0].setBounds(bx1,by1,bx2-bx1,by2-by1);
    // calc the left border
    bx1=0;
    bx2=line[1].x2;
    by1=0; if ((borders&BORD_TOPLEFT)!=0) by1=line[3].y2;
    by2=MW.pos.h; if ((borders&BORD_BOTTOMLEFT)!=0) by2=line[4].y1;
    bbox[1].setBounds(bx1,by1,bx2-bx1,by2-by1);
    // calc the right border
    bx1=line[2].x1;
    bx2=MW.pos.w;
    by1=0; if ((borders&BORD_TOPRIGHT)!=0) by1=line[3].y2;
    by2=MW.pos.h; if ((borders&BORD_BOTTOMRIGHT)!=0) by2=line[4].y1;
    bbox[2].setBounds(bx1,by1,bx2-bx1,by2-by1);
    // calc the top border
    bx1=0; if ((borders&BORD_TOPLEFT)==0) bx1=line[1].x2;
    bx2=MW.pos.w; if ((borders&BORD_TOPRIGHT)==0) bx2=line[2].x1;
    by1=0;
    by2=line[3].y2;
    bbox[3].setBounds(bx1,by1,bx2-bx1,by2-by1);
    // calc the bottom border
    bx1=0; if ((borders&BORD_BOTTOMLEFT)==0) bx1=line[1].x2;
    bx2=MW.pos.w; if ((borders&BORD_BOTTOMRIGHT)==0) bx2=line[2].x1;
    by1=line[4].y1;
    by2=MW.pos.h;
    bbox[4].setBounds(bx1,by1,bx2-bx1,by2-by1);
  }

  private void setPos2RXY (GWidget gw) {
    int x = gw.pos.x + gw.pos.w/2;
    int y = gw.pos.y + gw.pos.h/2;
    int i;
    for (i=0; i<5; i++) { if (bbox[i].within(x,y)) break; }
    if (i==5) i=0; // malformed - assume in middle region
    gw.rx = (x-bbox[i].x)/(double)bbox[i].w;
    gw.ry = (y-bbox[i].y)/(double)bbox[i].h;
    if (i==1) gw.rx-=1;
    if (i==2) gw.rx+=1;
    if (i==3) gw.ry-=1;
    if (i==4) gw.ry+=1;
    //System.out.println("SetPos2RXY "+gw+" "+i+" "+x+" "+y+" "+gw.rx+" "+gw.ry);
  }

  private void setRXY2Pos (GWidget gw) {
    int i=0; double rx=gw.rx, ry=gw.ry;
         if (rx<0) { i=1; rx+=1; }
    else if (rx>1) { i=2; rx-=1; }
    else if (ry<0) { i=3; ry+=1; }
    else if (ry>1) { i=4; ry-=1; }
    gw.pos.x = (int)(rx*bbox[i].w) + bbox[i].x - gw.pos.w/2;
    gw.pos.y = (int)(ry*bbox[i].h) + bbox[i].y - gw.pos.h/2;
  }

  /** Setup config when using /TSETUP */
  private Table configTSetup () {
    Table setupTable =  MA.getTable("/TSETUP",null,Args.NO_ERROR);   // get from /tsetup switch, if given

    if (M.results.getTable(MA.getS("/TSETUP")) != null) localResult = true;   // check for it in local results table
    if (setupTable == null) {    // check to see if parent result
      try {
        setupTable = M.results.getPrevious().getTable(MA.getS("/TSETUP") );
        if (setupTable != null) parentResult = true;
      } catch (Exception e) { e.printStackTrace(); }
    }
    String setupTableFileName = MA.getS("/TSETUP", null);
    if (setupTableFileName == null || setupTableFileName.equals("")) { setConfigFileName(TSETUP); } // default name
    else cfgfile = setupTableFileName.toLowerCase();
    if (setupTable == null) {
      setupTable =  Convert.o2t(cfgfile,M);   // look for setupTable as a file
    }
    return setupTable;
  }

  private void setConfigFileName (int setupType) {
    if (M.macro!=null) {
      cfgfile = M.macro.getFileName();
      titlebar = M.macro.name;
    } else {
      Args targs = Shell.parseCommand(M,cfgfile);
      String option = targs.option.toLowerCase();
      cfgfile = Macro.getFileName(M,cfgfile,option);
    }
    if (setupType==TSETUP) cfgfile = cfgfile.substring(0,cfgfile.lastIndexOf('.'))+".tbl";  // make a .tbl file name
    if (setupType==SETUP ) cfgfile = cfgfile+"p"+MA.getS("/SETUP").toLowerCase();           // make a .mmp file name

  }

  /** Saves an MMP or a TBL file. */
  private void save (String fname) {
    String text;
    String base = fname.toLowerCase();
    if (base.endsWith(".tbl")) { saveSetupTable(fname); return; }  // save as a TBL file
    TextFile tf = new TextFile(M,fname);
    if (!tf.open(TextFile.OUTPUT|TextFile.NOABORT)) {
      M.info("Could not open setup file: "+tf.getURL());
      return;
    }
    if (MW != null) {
      text = "{TYPE=POS,X="+MW.pos.x+",Y="+MW.pos.y+",W="+MW.pos.w+",H="+MW.pos.h+"}";
      tf.writeln(text);
    }
    if (borders!=0xF) {
      text = "{TYPE=BORDERS,MASK="+Convert.l2x(borders)+"}";
      tf.writeln(text);
    }
    for (int il=1; il<=nline; il++) {
      Line l = line[il];
      text = "{TYPE=LINE,INDEX="+il+",DIR="+l.dir+",X="+l.rx+",Y="+l.ry+
          ",I1="+l.i1+",I2="+l.i2+",FP="+l.fp+"}";
      tf.writeln(text);
    }
    for (int ip=1; ip<=npane; ip++) {
      Pane p = pane[ip];
      if (p==null) continue;
      text = "{TYPE=PANE,INDEX="+ip+",NAME="+p.name+",TITLE="+p.title+",TH="+p.ts+
          ",IX1="+p.ix1+",IX2="+p.ix2+ ",IY1="+p.iy1+ ",IY2="+p.iy2+"}";
      tf.writeln(text);
    }
    for (int iw=1; iw<=nwidget; iw++) {
      GWidget w = widget[iw];
      if (w==null) continue;
      if (w.pos.x==0) continue;
      if (w.pGroup!=null && w.pGroup!=w && w.pDetach==false) continue;
      text = "{TYPE=CPOS,NAME="+w.label+",X="+w.pos.x+",Y="+w.pos.y+
          ",W="+w.pos.w+",H="+w.pos.h+"}";
      tf.writeln(text);
    }
    M.info("Saved file to: "+tf.getURL());
    tf.close();
  }

  /** Loads an MMP file. */
  private void load (String fname) {
    String text;
    TextFile tf = new TextFile(M,fname);
    tf.open(TextFile.INPUT|TextFile.NOABORT);
    if (!tf.isOpen) { M.info("Could not open setup file: "+tf.getURL()); return; }
    while ( (text=tf.readProper()) != null) {
      Table t = new Table(text);
      loadSetupLine(t);
    }
    tf.close();
    // initial state of logger if not defined is opened
    if (logger && line[4].fp==0) line[4].fp=dlh;
  }

  /** Loads a single line from the setup file/table.
      This should only be called from load(..) and loadTableSetup(..).
   */
  private void loadSetupLine(Table t) {
      String type = t.getS("TYPE");
      String name = t.getS("NAME");
      // if the logger pane is already defined, do not add it again in open() (since 3.1.3 - Bug 2623)
      if (name!=null && name.equals("LOGGER")) addNewLoggerPane = false;
      int i = t.getL("INDEX",0);
      if (mobile>0) {
        if (type.equals("PANE")) addPane (i,t.getS("NAME"),t.getS("TITLE"),t.getS("ID"),10,1,2,3,4,0);
      }
      else if (type.equals("PANE")) {
        addPane (i,t.getS("NAME"),t.getS("TITLE"),t.getS("ID"),t.getL("TH"),
          t.getL("IX1"),t.getL("IX2"),t.getL("IY1"),t.getL("IY2"),t.getL("FLAG",0));
      }
      else if (type.equals("GRID")) {
        autoGrid(t);
      }
      else if (type.equals("LINE")) {
        addLine (i,t.getL("DIR"),t.getF("X"),t.getF("Y"),
          t.getL("I1"),t.getL("I2"),t.getL("FP"),t.getL("FLAG",0));
      }
      else if (type.equals("POS")) {
        // Check setup file Position to ensure it can be displayed on the screen. (Bug 1398)
        int xpos   = t.getL("X");
        int ypos   = t.getL("Y");
        int width  = t.getL("W");
        int height = t.getL("H");
        Rectangle screenBounds = GraphicsEnvironment.getLocalGraphicsEnvironment().
                                 getDefaultScreenDevice().getDefaultConfiguration().getBounds();
        int maxX = screenBounds.x + screenBounds.width;         // right edge of screen. adds in Xinerama values
        int maxY = screenBounds.y + screenBounds.height;        // bottom edge of screen.
        if (width  > maxX) width  = maxX;               // keep width no larger than screen
        if (height > maxY) height = maxY;               // keep height no larger than screen
        if ((xpos+width)  > maxX) xpos = maxX - width;  // if x off screen, move back on
        if ((ypos+height) > maxY) ypos = maxY - height; // if y off screen, move back on
        MW.setBounds(xpos,ypos,width,height);
      }
      else if (type.equals("CPOS")) {
        //String name = t.getS("NAME");  //moved to top of method to check for LOGGER
        MBox pos = new MBox(t.getL("X"),t.getL("Y"),t.getL("W"),t.getL("H"));
        cpos.put(name,pos);
      }
      else if (type.equals("BORDERS")) {
        borders = t.getL("MASK");
      }
      else M.warning("Unknown table type "+type+" in setup file");
  }

  /** Loads the setup from a table.
      @param t The table containing the setup.
   */
  public void setSetupTable (Table t) {
    Table.Iterator iter = t.iterator();
    while (iter.hasNext()) {
      loadSetupLine(t.getTable(iter.next().toString(),true));
    }

    // initial state of logger if not defined is opened
    if (logger && line[4].fp==0) line[4].fp=dlh;
  }

  /** Gets the current setup as a table. */
  public Table getSetupTable () {
    Table tbl = new Table();
    if (MW != null) {
      tbl.put("POS", "{TYPE=POS,X="+MW.pos.x+",Y="+MW.pos.y+",W="+MW.pos.w+",H="+MW.pos.h+"}");
    }
    if (borders!=0xF) {
      tbl.put("BORDERS", "{TYPE=BORDERS,MASK="+Convert.l2x(borders)+"}");
    }
    for (int il=1; il<=nline; il++) {
      Line l = line[il];
      tbl.put("LINE_"+il, "{TYPE=LINE,INDEX="+il+",DIR="+l.dir+",X="+l.rx+",Y="+l.ry+
                          ",I1="+l.i1+",I2="+l.i2+",FP="+l.fp+"}");
    }
    for (int ip=0; ip<=npane; ip++) {
      Pane p = pane[ip];
      if (p!=null) {
        tbl.put("PANE_"+ip, "{TYPE=PANE,INDEX="+ip+",NAME="+p.name+",TITLE="+p.title+",TH="+p.ts+
                            ",IX1="+p.ix1+",IX2="+p.ix2+ ",IY1="+p.iy1+ ",IY2="+p.iy2+"}");
      }
    }
    for (int iw=1; iw<=nwidget; iw++) {
      GWidget w = widget[iw];
      if ((w!=null) && (w.pos.x!=0) && w.pGroup==null && w.pGroup==w && w.pDetach!=false) {
        tbl.put("WIDGET_"+iw, "{TYPE=CPOS,NAME="+w.label+",X="+w.pos.x+",Y="+w.pos.y+
                              ",W="+w.pos.w+",H="+w.pos.h+"}");
      }
    }

    return tbl;
  }

  /** Save a setup table that is a result variable
      @param resName The name of the result
      @since NeXtMidas 2.9.0
   */
  public void saveResult (String resName) {
    String loc = "";
    if (parentResult || (!parentResult && !localResult)) {
      M.results.getPrevious().put(resName.toUpperCase(), getSetupTable());  // save result to Parent Midas context
      loc = "parent";
    }
    if (localResult) {
      M.results.put(resName.toUpperCase(), getSetupTable());           // save result in local Midas context
      loc = "local";
    }
    M.info("Saved result "+resName+" to "+loc+" results table.");
  }

  /** Saves a TBL file. */
  public void saveSetupTable (String fname) {
    TextFile tf = new TextFile(M,fname);
    tf.open(TextFile.OUTPUT|TextFile.FORCEABORT);
    Table.toTextFile(tf,getSetupTable(),"");
    tf.close();
    //getSetupTable().toTextFile(tf);  // changed to lines above per BUG 2189
    M.info("Saved setup table to: "+tf.getURL());
  }

  private void autoGrid (Table t) {
    int nx = t.getL("NX",4);
    int ny = t.getL("NY",4);
    String name = t.getS("NAME","GP");
    String title = t.getS("TITLE","Pane");
    int ts = t.getL("TH",10);
    String fill = t.getS("FILL","LRTB");

    nline=4; // erase other lines first
    for (int ix=1; ix<nx; ix++) addLine(-1,VLINE,(float)ix/(float)nx,0.5F,3,4,0,0);
    for (int iy=1; iy<ny; iy++) addLine(-1,HLINE,0.5F,(float)iy/(float)ny,1,2,0,0);

    int ip=0; // add panes
    if (mobile!=0);
    else if (fill.equals("TBLR")) {
      for (int ix=1; ix<=nx; ix++) {
        int ix1=1; if (ix>1) ix1=4+ix-1;
        int ix2=2; if (ix<nx) ix2=4+ix;
        for (int iy=1; iy<=ny; iy++) {
          int iy1=3; if (iy>1) iy1=4+nx-1+iy-1;
          int iy2=4; if (iy<ny) iy2=4+nx-1+iy;
          ip++; addPane (ip,name+ip,title+ip,id+ip,ts,ix1,ix2,iy1,iy2,0);
        }
      }
    }
    else if (fill.equals("LRTB")) {
      for (int iy=1; iy<=ny; iy++) {
        int iy1=3; if (iy>1) iy1=4+nx-1+iy-1;
        int iy2=4; if (iy<ny) iy2=4+nx-1+iy;
        for (int ix=1; ix<=nx; ix++) {
          int ix1=1; if (ix>1) ix1=4+ix-1;
          int ix2=2; if (ix<nx) ix2=4+ix;
          ip++; addPane (ip,name+ip,title+ip,id+ip,ts,ix1,ix2,iy1,iy2,0);
        }
      }
    }
    gnx = nx;
    gny = ny;
  }

  private void updateLastWidget() {
    for (int i=1; i<=nwidget; i++) {
      GWidget gw = widget[i];
      if (gw==null || gw.pGroup==null) continue;
      if (i==nwidget || widget[i+1] instanceof GLabel) {
        gw.pGroup.lastWidget = gw;
        gw.refresh();
      }
    }
  }

  @Override
  public Object addWidget (GWidget gw) {
    if (gw==null) return null;
    if (isHideControls()) return null;
    if (!cntrls.containsKey(gw.label)) {
      cntrls.put(gw.label,gw);
    }
    if (gw instanceof GLabel) curGLabel=(GLabel)gw;
    GLabel gwg = curGLabel;
    int i = ++nwidget;
    if (gw.group != null) {
      for (i=1; i<nwidget && (widget[i].group==null || !widget[i].group.equals(gw.group)); i++);
      gwg = (GLabel)widget[i]; // not adding to curGLabel
      for (;i<nwidget && widget[i].group.equals(gw.group); i++);
      for (int j=nwidget; j>i; j--) widget[j]=widget[j-1];
    }
    if (gwg!=null) { gw.pGroup=gwg; gw.setGroup(gwg.label); } //gwg.lastWidget=gw; }
    widget[i] = gw;
    if (!widget[i].hasLocalTheme()) gw.setTheme(theme);  // enabled this in 3.3.0
    gw.setFlag(GWidget.INLINE,inline);
    gw.setWidgetPanel(this);
    MBox pos = (MBox)cpos.get(gw.label);
    if (pos!=null) {
      gw.pos.setBounds(pos); setPos2RXY(gw);
      if (gwg!=null && gwg.rx<0) gw.pDetach=true;
    } else if (gwg!=null && gw!=gwg) {
      gw.rx=gwg.rx;
    } else {
      gw.rx=-1;
    }
    gw.parent=gw.MW=MW;
    if (headless) return gw;
    gw.resize(1);
    fixWidget(i,-1); // default placement in left border.
    MW.addOn(gw.panel);
    if (gw instanceof GPanel) panes.put(gw.label,gw);
    return gw;
  }

  @Override
  public void layoutWidget (GWidget gw) {
    needLayout = true;
  }

  @Override
  public boolean accept (Object entry) {
    boolean rmif = entry instanceof nxm.sys.prim.rmif;
    boolean remote = entry instanceof nxm.sys.net.Rmif.Remote;
    boolean command = entry instanceof nxm.sys.lib.Command;
    if (filterMode==1 && command) return true;
    if (filterMode==2 && rmif) return true;
    if (filterMode==2 && remote) return true;
    return false;
  }

  /** get a timer value by index */
  public double getTimer (int i) {
    if (i < 0 || i >= nTimers) {
      throw new MidasException("Invalid timer index = "+i);
    }
    return timer[i];
  }

  /** Set the number of timers.  Initializes (RESETS) timer arrays*/
  public void setTimers (int in_nTimers) {
    if (in_nTimers < 0 ) {
      M.error ("Invalid number of timers = "+in_nTimers);
    } else {
      nTimers = in_nTimers;
      if (nTimers == 0) {
        timer = null;
        timerinit = null;
      } else {
        timer = new double[nTimers];
        timerinit = new double[nTimers];
      }
    }
  }

  /** Sets the mouse cursor when over the panel to a given cursor type. This version will reset the
      cursor after a given number of seconds have elapsed.
      @param name The name of the cursor to set (see {@link MWindow#cursorList}), cant be "CUSTOM".
      @param sec  The number of seconds to wait.
   */
  public void setCursor (String name, double sec) {
    if (MW != null) {
      MW.setCursor(name, sec);
      if (userCursor != null) userCursor = getCursor();
    } else {
      M.warning("PANEL: Can not set cursor, MW is null");
    }
  }

  public void setAltTitle (String title) {
    altTitle = title;
  }
  private void refreshTitle () {
    if (altTitle==null) {
      if (currentTitle==null) return;
    }
    if (altTitle!=null) {
      if (currentTitle==null);
      else if (altTitle.equals(currentTitle)) return;
    }
    currentTitle = altTitle;
    ((Frame)MW.parent).setTitle(currentTitle);
  }


  /** Sets the mouse cursor when over the panel to a given cursor type.
      @param name The name of the cursor to set (see {@link MWindow#cursorList}), cant be "CUSTOM".
   */
  public void setCursor (String name) {
    if (MW != null) {
      MW.setCursor(name);
      if (userCursor != null) userCursor = getCursor();
    } else {
      M.warning("PANEL: Can not set cursor, MW is null");
    }
  }

  /** Sets the cursor when in the proper location to resize.
      @param dir The direction from lines[i].dir or -1 to reset to normal.
   */
  private void setResizeCursor (int dir) {
    if (dir > 0) {
      if (userCursor == null) userCursor = getCursor();
      MW.setCursor( (dir==VLINE)? "WResize" : "SResize");
    }
    else if (userCursor != null) {
      MW.setCursor(userCursor);
      userCursor = null;
    }
  }

  /** Gets the current mouse cursor.
      @return The current mouse cursor, may be "CUSTOM".
   */
  public String getCursor () {
    return (MW != null)? MW.getCursor() : "Default";
  }

  /** set a timer value by index */
  public void setTimer (int i, double value) {
    if (i < 0 ) {
      M.error ("Invalid timer index = "+i);
    } else {
      timer[i]=value;
      timerinit[i]=Time.current();
    }
  }

  public void addWindow (MWindow mw) {
    int n = ++npane;
    mw.MW = MW;
    if (pane[n]==null) addPane(n,""+n,""+n,""+n, 10, 1,2,3,4, 0);
    pane[n].add(mw);
  }

  /** Control borders on the panel (see {@link #bordersList}) */
  public void setBorders (String str) {
    setBorders(Parser.mask(bordersList,str,borders));
  }
  /** Get the panel borders mask as a string (see {@link #bordersList}) */
  public String getBordersString () {
    return Parser.mask2s(bordersList,borders);
  }

  /** Users should use (see {@link #getBordersString()}) */
  public double getBorders () { return borders; }
  /** Users should not use this method.  Users should use (see {@link #setBorders(String)}) */
  public void setBorders (int value) { borders=value; resize(); refresh(); }
  /** Users should not use this method.  Users can toggle borders options with
    a <code>~</code> before the borders option to toggle.  (see {@link #setBorders(String)}) */
  public void setToggleBorders (int value) { borders^=value; resize(); refresh(); }

  public Table getControls () { return cntrls; }
  public Table getPanes () { return panes; }

  public Pane[] getPanesArray() {
    Pane[] pa = new Pane[npane+1];
    for (int i=0; i<=npane; i++) pa[i]=pane[i];
    return pa;
  }

  private int lbu=0, lbm=0, ldp=-1;
  public java.io.ByteArrayOutputStream handleMobileRequest (String name) {
    //System.out.println("HandleMobileRequest: "+name);
    int iarg = name.indexOf('?');
    if (iarg>0) {
      Table t = new Table("{"+name.substring(iarg+1)+"}");
      String type=t.getS("TYPE");
      int x=t.getL("X"), y=t.getL("Y"), ip=t.getL("IP");
      Container c = (ip==0)? MW.panel : pane[ip].panel;
      while (c!=null) {
        x += c.getLocation().x;
        y += c.getLocation().y;
        c = c.getParent();
      }
      if (type.equals("MOVE")) {
        robot.mouseMove(x,y);
      }
      else if (type.equals("PRESS")) {
        robot.mouseMove(x,y);
        int bu = t.getL("B");
        int ab = Math.abs(bu);
        int bm = (ab==3)? InputEvent.BUTTON3_MASK : (ab==2)? InputEvent.BUTTON2_MASK : InputEvent.BUTTON1_MASK ;
        if (bm!=lbm && lbu>0)  robot.mouseRelease(lbm); // auto finish last button
        if (bu>0 || lbu<=0) robot.mousePress(bm);
        if (bu<0) robot.mouseRelease(bm);
        lbu = bu; lbm = bm; 	// last Button
      }
      else if (type.equals("KEY")) {
        if (ldp != ip) { // get focus
          robot.mouseMove(x,y);
          robot.mousePress(InputEvent.BUTTON3_MASK);
          robot.mouseRelease(InputEvent.BUTTON3_MASK);
          ldp = ip;
        }
        String keys = t.getS("K");
        while (keys.length()>0) {
          char ch = keys.charAt(0);
          int key = (int)ch;
          if (ch>='a' && ch<='z') key -= 32;
          robot.keyPress(key);
          robot.keyRelease(key);
          keys = keys.substring(1);
        }
        int key = KeyEvent.VK_ENTER;
        robot.keyPress(key);
        robot.keyRelease(key);
      }
      name = name.substring(0,iarg);
    }
    if (name.startsWith("DisplayPane")) {
      int ip = (name.charAt(11)-'0');
      Container p = (ip==0)? MW.panel : pane[ip].panel;
      //Container p = (ip==0)? widget[1].panel.getParent() : pane[ip].panel;
      int w=(ip==0)?320:p.getWidth(), h=p.getHeight();
      try {
        java.io.ByteArrayOutputStream baos = new java.io.ByteArrayOutputStream();
        BufferedImage image = new BufferedImage(w,h,BufferedImage.TYPE_INT_RGB);
        p.print(image.getGraphics());
        ImageIO.write(image, "png", baos);
        //System.out.println("Image pane="+ip+" size="+baos.size());
        return baos;
      }
      catch (Exception e) { M.printStackTrace("Unable to export image: ", e); }
    }
    return null;
  }

  public void addPane (int i, String name, String title, String id, int ts,
                       int ix1,int ix2, int iy1,int iy2, int flags) {
    if (i<0) i=npane+1;
    Pane p = pane[i];
    if (p==null) p = pane[i] = new Pane(this);
    p.name = name;
    p.title= title;
    // shorten names for mobile tabbed screens
    if (mobile!=0 && ix1!=0 && ix2!=0 && iy1!=0 && iy2!=0) p.tag = ""+i;
    p.ts   = ts;
    p.ix1  = ix1;
    p.ix2  = ix2;
    p.iy1  = iy1;
    p.iy2  = iy2;
    p.flags = flags;
    panes.put(p.name,p);
    npane = Math.max(npane,i);
  }

  public void addLine (int i, int dir, float rx, float ry, int i1, int i2, int fp, int flags) {
    if (i<0) i=nline+1;
    Line l = line[i];
    if (l==null) l=new Line();
    l.dir = dir;
    l.rx  = rx;
    l.ry  = ry;
    l.i1  = i1;
    l.i2  = i2;
    l.fp  = fp;
    l.flags = flags;
    line[i] = l;
    nline = Math.max(nline,i);
  }


  public class Line implements MessageHandler {
    float rx,ry;
    int i1,i2;
    int x1,x2;
    int y1,y2;
    int w,dir;
    int fp=0;
    int flags=0;

    public int processMessage (Message msg) {
      if (msg.name.equals("CFG.LINE")) {
        String text = msg.getS();
        if (text.equals("DELETE")) delLine(this);
        else if (text.equals("FIX")) { if (dir==VLINE) fp=(x1+x2)/2; else fp=(y1+y2)/2; }
        else if (text.equals("UNFIX")) fp=0;
      }
      return 1;
    }
  }

  /** A pane in a panel. */
  public class Pane implements MessageHandler {
    private final panel pan;

    public String name;
    public String title;
    public String tag;
    public MWindow MW;
    public MWindow mw;
    public Container panel;
    public int ix1=1,ix2=2,iy1=3,iy2=4;
    public int x,y,w,h;
    public int ts=0;
    public int flags=0;
    public boolean hidden=false;
    public TabManager tabManager = null;
    public MBox pos = new MBox();

    /** @deprecated Since NeXtMidas 2.5.4: Only PANEL should be creating instances of Pane. */
    @Deprecated public Pane (MWindow MW) {
      this.MW  = MW;
      this.pan = null;
    }

    /** Creates a new instance. */
    Pane (panel p, String title) {
      this(p);
      this.title = title;
    }

    /** Creates a new instance. */
    Pane (panel p) {
      this.pan = p;
      this.MW  = p.MW;
    }

    /** Sets the pane's TabManager.
        @param mgr The tab manager for this tab.
     */
    private void setTabManager (TabManager mgr) {
      tabManager = mgr;
      mgr.addTab(this);
    }

    /** Gets the pane's TabManager. */
    private TabManager getTabManager () {
      return tabManager;
    }

    /** Indicates if the pane is in a tab.
        @return Indicates if the pane is in a tab.
     */
    public boolean isInTab () {
      return getTabManager() != null;
    }

    /** Indicates if the pane is the visible tab. */
    public boolean isVisibleTab () {
      return isInTab() && getTabManager().isVisibleTab(this);
    }

    /** Makes this pane the visible tab. */
    public void makeVisibleTab () {
      if (isInTab() && !isVisibleTab()) {
        getTabManager().setVisibleTab(this);
        updateTabs();
      }
    }

    /** Makes this pane a non-visible tab (if other tabs exist). */
    public void makeNonVisibleTab () {
      Shell.warning("Pane.makeNonVisibleTab() does not work for all pane types.");
      if (isInTab() && isVisibleTab()) {
        getTabManager().setNonVisibleTab(this);
        updateTabs();
      }
    }

    /** Sets the title of the pane */
    public void setTitle (String value) { title=value; refresh(); }

    /** Sets the title size of the pane in pixels */
    public void setTS (int value) { if (mobile==0) ts=value; refresh(); }

    public void setTheme (String value) { if (mw!=null) mw.setTheme(value); }

    public void setTheme (Object value) { if (mw!=null) mw.setTheme(value); }

    public void add (MWindow mw) { this.mw=mw; add(mw.panel); }

    public void add (Container panel) {
      this.panel = panel;
      panel.setBounds(x,y,w,h);
      MW.addOn(panel,-1);
      panel.setVisible(!hidden);
    }

    public void remove (Component panel) {
      if ((MW != null) && (MW.panel != null)) MW.panel.remove(panel);
      pan.refresh();
    }

    public void refresh () {
      MW.refresh(pos);
    }

    public void paint () {
      int xi=x; // top-left X Pos (pixel)
      int wi=w; // width of the pane
      int yi=y; // top left Y Pos (pixel)
      int hi=h; // height
      int tsp=3*ts/2;   // area around text??
      if (tsp>0) { yi-=tsp; hi+=tsp; }  // add text "zone" to borders
      pos.setBounds(xi,yi,wi,hi);
      if (tabManager != null) {
        wi/=tabManager.getTabCount();
        xi+=tabManager.getTabIndex(this) * wi;
      }
      if (ts>0 && !hidden) {
        int color = ((tabManager==null) || tabManager.isVisibleTab(this))? MWindow.HILIT: MWindow.LOLIT;
        MW.shadowbox(xi,yi,wi,tsp,is(PUSHPOPX)?MWindow.SB_SHADEX:MWindow.SB_SHADE);
        MW.drawString((tag!=null)?tag:title,xi+wi/2,y-4,ts,MWindow.CENTER|color);
      }
      if (panel==null) {  // if no Graphics panel in the Pane, draw a rectangle border and the pane name in the center
        MW.g.drawRect(xi+2,y+2,wi-4,h-4);  // rectangle size reduced slightly in 3.1.1 to remove artifacts  (Bug 2511)
        MW.drawString(name,xi+wi/2,y+h/2,12,0);
      }
    }

    private int clickHit (int xp, int yp) {
      int xi=pos.x, wi=pos.w, tsp=3*ts/2;
      if (tabManager != null) {
        wi/=tabManager.getTabCount();
        xi+=tabManager.getTabIndex(this) * wi;
      }
      if (xp<xi || xp>xi+wi || yp<y-tsp || yp>y+h) return 0;
      if (is(PUSHPOPX) && xp>xi+wi-15) return -1;
      return 1;
    }

    public void setDisplay (String str) {
      if (str.equals("FRONT")) hitPane(myIndex());
      else if (str.equals("POP")) { panel.setVisible(true); mw.pop(1); }
      else if (str.equals("PUSH")) mw.pop(0);
    }

    private int myIndex () {
      for (int i=1; i<=npane; i++) if (pane[i]==this) return i;
      return 0;
    }

    @Override
    public int processMessage (Message msg) {
      if (msg.name.equals("CFG.PANE")) {
        String text = msg.getS();
             if (text.equals("TITLETEXT"))  new GPrompt (MW,"TitleText",title,title.length(),0,this);
        else if (text.equals("TITLESIZE"))  new GMenu (MW,"TitleSize","0,8,10,12,14,16,20,24",0,0,this);
        else if (text.equals("TITLECOLOR")) new GPrompt (MW,"TitleColor","ForeGround",0,this);
        else if (text.equals("ATTACH"))     new GMenu (MW,"Attach",M.registry.table.getKeyList(),0,0,this);
        else if (text.equals("ADDTAB"))     addPane(0,cfgx,cfgy);
        else if (text.equals("DELETE"))     delPane(this);
        else M.warning("Unrecognized message "+msg.name+" text "+text);
        configure=1;
      }
      else if (msg.name.equals("TITLETEXT"))  { setTitle(msg.getS()); }
      else if (msg.name.equals("TITLESIZE"))  { setTS(Convert.o2l(msg.data)); resize(); }
      else if (msg.name.equals("TITLECOLOR")) { M.warning("Setting color is TBD"); }
      else if (msg.name.equals("ATTACH"))     { attach(msg.getS()); }
      else if (msg.name.equals("SUBPANELS"))  { attachSubPanel(msg.getS()); }
      return NORMAL;
    }

    /** Attach an object (either a primitive or a Macro) to the panel. If the
        object is a Macro, the macro panel is added
        @param name The name of the object as it is in the registry
     */
    private void attach (String name) {
      Object entry = M.registry.get(name);
      if (entry instanceof Macro) getSubPanels((Macro)entry, name);
      else attach(entry,name);
    }

    /** Attaches an object to the panel
        @param entry A GPrimitive Object
        @since NeXtMidas 2.7.3
     */
    private void attach (Object entry, String name) {
      if (entry instanceof GPrimitive) {
        mw = ((GPrimitive)entry).MW;
        mw.pushParent = this;
        mw.pop(0);
        this.name = name;
      }
      else M.info ("Attach Object must be an instanceof GPrimitive: "+name);
    }

    /** Attach a panel to this panel
        @param name The name of the panel to attach
        @since NeXtMidas 2.7.3
     */
    public void attachSubPanel (String name) {
      attach(subPanels.get(name), subPanelName);
    }

    /** Get any and all panels from a Macro
        @param m The Macro
        @param name The id or name given to the sub-panel in the parent macro
        @since NeXtMidas 2.7.3
     */
    private void getSubPanels (Macro m, String name) {
      panel p = (panel) m.M.registry.get("PANEL");
      if (p != null) attach(p,name);      // if default, go ahead and attach it
      else {
        subPanels  = new Table();         // clear out the subPanels table
        subPanelName = name;              // keep the name in sync with what the parent expects
        Table t    = m.M.registry.table;  // get the list of objects
        String[] s = t.getKeys();         // get the keys of the objects
        Object tmp = null;
        for (int i=0; i<s.length; i++) {
          tmp =  t.getO(s[i]);            // get the object
          if (tmp instanceof panel) subPanels.put(s[i], tmp);  // if the obj is a panel, add it to the sub-panels table
        }
        // If there is one panel, add it. If there are many, present a menu to the user
        if (subPanels.getSize() > 0) {
          if (subPanels.getSize() == 1) {
            String id = subPanels.getKeys()[0];
            attach(subPanels.getO(id), subPanelName);
          }
          else new GMenu (MW,"SubPanels",subPanels.getKeyList(),0,0,this);
        }
      }
    }

    @Override
    public String toString() {
      return "Pane "+name+" x="+x+" y="+y+" w="+w+" h="+h;
    }
  }

  /** Loads the configuration of a panel from a class.
      @param className The fully-qualified name of the class to load the config from.
   */
  private void loadPanelClass (String className) {
    if ((className.trim().length() == 0) && (M.macro != null)) {
      className = M.macro.name.toLowerCase();
    }

    // The default package the class will be in
    String defPkg = "";
    if (M.macro != null) { defPkg = "nxm." + M.macro.option.toLowerCase() + ".mcr."; }

    String nameLC = className.toLowerCase();
    String nameTC = className.substring(0,1).toUpperCase() + nameLC.substring(1);

    // Possible class names
    String[] names = new String[] { defPkg+className, defPkg+nameLC, defPkg+nameTC, className, nameLC, nameTC };

    Class<?> configClass = null;

    // Try to figure out what the name of the class is
    for (int i = 0; (i < names.length) && (configClass == null); i++) {
      try {
        configClass = Class.forName(names[i]);
      } catch (Exception e) {
        /* ignore */
      }
    }

    if (configClass == null) {
      M.warning ("Unable to load panel class '" + className + "': Class not found.");
    }
    else {
      try {
        final MessageHandler mh          = M.macro;
        final Object         config      = configClass.newInstance();
        final Message        closeMsg    = new Message("WINDOW", 0, "CLOSING", config, this);
        final WindowListener listener    = new WindowAdapter() {
          @Override
          public void windowClosing (WindowEvent e) {
            mh.processMessage(closeMsg);
          }
        };
        //loadDisplays(configClass.getFields(), config, false);
        loadDisplays(configClass.getDeclaredFields(), config, true);

        if (config instanceof Frame) {
          Frame frame = (Frame)config;
          container = frame;
          frame.addWindowListener(listener);
          frame.setVisible(true);
          MW.parent = frame;                          // <-- PARENT = FRAME
        }
        else if (config instanceof Container) {
          container = (Container)config;
          container.setVisible(true);
          MW.panel = container;                      // <-- PANEL = CONTAINER
        }
        else {
          M.warning("Class loaded with /JSETUP is a "+Util.getClassName(config)+", "
                   +"expected a Swing container");
          if (config instanceof Component) {
            // JC4: This next line is backwards-compatible with NeXtMidas 3.1
            //      and earlier but doesn't make much sense... I suspect this
            //      was part of my initial attempt to support non-Frame
            //      components that didn't work out.
            ((Component)config).setVisible(true);
          }
        }
      } catch (Exception e) {
        M.printStackTrace("Unable to load panel class '"+className+"'", e);
      }
    }
  }

  /** Loads the fields for a class. (Used by loadPanelClass)
      @param fields  The fields to load.
      @param config  The configuration instance.
      @throws Exception If an error occurs.
   */
  private void loadDisplays (Field[] fields, Object config, boolean noPublic) throws Exception {
    AccessibleObject.setAccessible(fields, true);
    MessageHandler mh   = M.macro;
    Object         to   = null;
    Object         from = this;

    Map<Object,String> controls = new HashMap<Object,String>();
    Map<String,Object> groups   = new HashMap<String,Object>();

    // Find the controls and displays then add the displays and put the controls in the appropriate table.
    for (int i = 0; i < fields.length; i++) {
      String name  = fields[i].getName().toUpperCase();
      Object value = fields[i].get(config);

      if (name.equals("PANE0")) {                     // The pane for gcontrol
        if (value instanceof MidasControls) {
          ctlPane = (MidasControls)value;
          ctlPane.setMWindow(MW);
        } else {
          M.error("PANE0 must be an instance of MidasControls.");
        }
      }
      else if (value instanceof PushPopPanel) {
        PushPopPanel ppp = (PushPopPanel)value;
        getPanes().put(name, ppp.getMidasDisplay());
      }
      else if (value instanceof MidasDisplay) {     // A NeXtMidas display
        ((MidasDisplay)value).setMWindow(MW);
        getPanes().put(name, value);
      }
      else if ((value != null) && value.getClass().getName().equals("nxm.sys.libxm.XmDisplay")) {
        // An X-Midas display
        if (xmCanvasTbl == null) { xmCanvasTbl = new Table(); }
        if (name.toUpperCase().startsWith("XS") && StringUtil.isInteger(name.substring(2))) {
          xmCanvasTbl.put(name, value);
        } else {
          M.error("Illegal name for an XmCanvas, given '"+name+"'. Must use name in the form xsN "
                 +"where N is the value of /XS=. For example use the name xs3 for /XS=3.");
        }
      }
      else if (value instanceof ButtonGroup) {
        groups.put(name, value);
      }
      else if (value instanceof JPanel) {
        final JPanel jpanel = (JPanel)value;
        MouseAdapter mouseListener = new MouseAdapter() {
          @Override
          public void mouseClicked(MouseEvent evt) {
            if (evt.getButton() == MouseEvent.BUTTON2) {
              processMessage("BUTTON", 2, evt);
            }
          }
        };
        jpanel.addMouseListener(mouseListener);
      }
      else if (value instanceof Widget) {           // A nxm.sys.inc.Widget
        controls.put(value, name);
      }
      else if (value instanceof Component) {        // A Swing/AWT component
        controls.put(value, name);
      }
    }

    // Take care of any groups.
    for (Map.Entry<String,Object> e : groups.entrySet()) {
      String name  = e.getKey();
      Object value = e.getValue();

      if (value instanceof ButtonGroup) {
        Enumeration<?> buttons = ((ButtonGroup)value).getElements();

        while (buttons.hasMoreElements()) {
          AbstractButton button  = (AbstractButton)buttons.nextElement();
          String         btnName = controls.remove(button);

          cntrls.put(name, new WidgetWrapper(button, mh, name, btnName, to, from));
        }
      }
    }

    // Take care of the rest of the controls.
    for (Map.Entry<Object,String> e : controls.entrySet()) {
      Object value = e.getKey();
      String name  = e.getValue();

      WidgetWrapper w;
      if (value instanceof Widget) w = new WidgetWrapper((Widget)value,    mh, name, null, to, from);
      else                         w = new WidgetWrapper((Component)value, mh, name, null, to, from);

      cntrls.put(name, w);
    }
  }

  /** Gets the panel to use for the controls.
      @return Returns the panel to use for the controls or <code>null</code>
              if the default should be used.
   */
  public MidasControls getMidasControls () {
    return ctlPane;
  }

  /** Hide normal gcontrols? (/JSETUP= with no PANE0) */
  private boolean isHideControls () {
    return (container != null) && (ctlPane == null);
  }

  /** Indicates if the panel is running in headless mode.
      @return <code>true</code> if in headless mode, <code>false</code> otherwise.
      @since NeXtMidas 1.9.5
   */
  public boolean isHeadless () { return headless; }


  /** Manages a set of tabbed panes. */
  public class TabManager {
    private java.util.List<Object> panes = new LinkedList<Object>();  // The panes in this tabbed area.
    private int      visiblePane = 0;                 // The index of the visible pane.

    /** Creates a new instance.  */
    private TabManager () {
    }

    /** Adds a pane.
        @param pane The pane to add.
     */
    public synchronized void addTab (Pane pane) {
      panes.add(pane);
    }

    /** Sets the active tab.
        @param index The index of the tab to show.
     */
    public synchronized void setVisibleTab (int index) {
      if (is(ROTATETABS)) {
        Object pane = panes.get(index);
        panes.remove(index);
        panes.add(0, pane);
        visiblePane = 0;
      } else {
        visiblePane = index;
      }
    }

    /** Sets the active tab.
        @param pane The tab to show.
     */
    public synchronized void setVisibleTab (Pane pane) {
      setVisibleTab(getTabIndex(pane));
    }

    /** Sets the active tab to be a tab other than the given tab (unless it is the only tab).
        @param pane The tab to not show.
     */
    public synchronized void setNonVisibleTab (Pane pane) {
      if (isVisibleTab(pane)) {
        if (is(ROTATETABS)) {
          setVisibleTab(1);
        } else {
          int paneIndex = getTabIndex(pane) + 1;
          if (paneIndex < getTabCount()) {
            setVisibleTab(paneIndex);
          } else {
            setVisibleTab(0);
          }
        }
      }
    }

    /** Gets the active tab. */
    public synchronized Pane getVisibleTab () {
      return (Pane)panes.get(visiblePane);
    }

    /** Indicates if the given pane is visible.
        @param pane The pane to check for being visible.
        @return Indication of whether or not it is visible.
     */
    public synchronized boolean isVisibleTab (Pane pane) {
      return pane == getVisibleTab();
    }

    /** Gets the index number of a given pane. */
    public synchronized int getTabIndex (Pane pane) {
      return panes.indexOf(pane);
    }

    /** Gets the number of tabs. */
    public synchronized int getTabCount () {
      return panes.size();
    }
  }
}
