package nxm.sys.prim;

import nxm.sys.inc.Installation;
import nxm.sys.lib.*;
import nxm.sys.libg.*;

/**
  A class to run the Midas shell from a Graphical User Interface.

  @author Jeff Schoen
  @version $Id: shellgui.java,v 1.36 2013/01/17 18:33:21 lsn Exp $
*/
public class shellgui extends GPrimitive {

  /** Exit options mask for /EXIT= switch. */
  public static final String exitList = "COMMAND,MENU,MESSAGE,WINDOW";
  public static final int EXIT_COMMAND=0x1, EXIT_MENU=0x2,
          EXIT_MESSAGE=0x4, EXIT_WINDOW=0x8, EXIT_ALL=0xF;

  /** Shell driving this SHELLGUI terminal */
  public Shell shell;
  /** Terminal driving the graphical text window */
  public GTerminal terminal;

  /** The parent terminal to back this terminal with */
  private Terminal saveterm;
  /** Is the shell is "attached" to the shell it was run from. */
  private boolean  attach;
  /** User specified exit options. See {@link #exitList} for mask options. */
  private int exit;
  private boolean exiting;
  /** Text area (window) for the terminal input/output */
  private MWText mwtext;

  public int open () {
    int lines  = MA.getL("/LINES",2000); // BUG 1830: to help with scroll issues
    int rows   = MA.getL("/ROWS",   30);
    int cols   = MA.getL("/COLS",  100); // BUG 1830: 80 normally cause output to wrap via "+"
    int fontsz = MA.getL("/FONTSZ", 12);
    int flags  = MJText.TEXT | MJText.BACKL | MJText.FIFO | MJText.INPUT | MJText.WRAP | MJText.COLOR ;

    boolean addPlus = MA.getState("/ADDPLUS", true);  // BUG 1563: Give users ability to turn OFF
    if (addPlus) flags |= MJText.ADDPLUS;             //   plus sign at the end of a wrapped line

    // The text window
    MW = mwtext = new MWText("GShell",lines,rows,cols,fontsz,flags,this);
    MW.open();

    // The terminal to drive the text window
    terminal = new GTerminal(M,mwtext);
    saveterm = M.terminal;

    attach = MA.getState("/ATTACH",false);

    // The shell to drive the terminal
    shell = new Shell(M, attach);

    if (attach) {
      if (M.macro==null) saveterm.leach = terminal; // 2011-03-15 BUG 1908 (don't need to set leach when in a macro)
    }
    else shell.M.detach();

    exit = (M.pipeMode==Midas.PINIT)? 0:EXIT_ALL;
    exit = MA.getOptionMask("/EXIT",exitList,exit);

    MW.addTo(this);

    shell.M.terminal = terminal;
    if (verbose) terminal.writeln("Welcome to NeXtMidas "+Installation.NM_VERSION+" ...");

    return NORMAL;
  }

  public int process () {
    shell.process();
    if (exit(EXIT_COMMAND)) return FINISH;
    return  NORMAL;
  }

  public int processMessage (Message msg) {
    int status = NORMAL;
    if      (msg.name.equals("EXIT"))                  exit(EXIT_MESSAGE);
    else if (msg.name.equals("MENUEXIT"))              exit(EXIT_MENU);
    else if (msg.name.equals("WINDOW") && msg.info==0) exit(EXIT_WINDOW);
    else if (msg.name.equals("RESIZE"))                MW.resize(0);
    else if (msg.name.equals("CLEAR"))                 clear();
    else status = MW.processMessage(msg);
    return status;
  }

  private boolean exit (int flag) {
    if ((flag&exit)!=0) exiting=true;
    if (exiting && flag!=EXIT_COMMAND) terminal.setTextInput("EXIT");
    return exiting;
  }

  public int close () {
    if (attach) {
      M.terminal = saveterm;
      saveterm.leach = null;
    }
    MW.close();
    return NORMAL;
  }

  /** Clear text in terminal window.
      @since NeXtMidas 2.9.0
   */
  public void clear () { mwtext.mtext.removeAll(); }


}
