package nxm.sys.prim;

import nxm.sys.inc.*;
import nxm.sys.lib.*;
import nxm.sys.net.DatagramData;
import nxm.sys.net.DatagramData.IceTimeCallBack;
import java.util.Arrays;
import java.util.StringTokenizer;

/**
  sourcedg - uses DatagramData class to reads (Raw or RTP) from Multicast or unicast datagram stream

  @author Scott Zimmerman
  @version $Id: sourcedg.java,v 1.32 2012/01/23 17:50:51 ntn Exp $
*/
public class sourcedg extends Primitive implements IceTimeCallBack {
  //may be multicast or unicast and may or may not be rtp or sap - SB in only
  private TextFile logfile;
  private DataFile pout,convout;          //pout is normal output, convout is for audio streams - converted from ULAW to Lin16
  private Data outdata;                   //hold convout data
  private DatagramData dgd = null;

  private boolean abort=false, dgdclosed=true, linear=true, linearin=true, init=false;

  private short[] outbuf;
  private short[] decodeTable;
  private Table t3kstruct = null;
  private byte[] buf;
  private byte[] fbuf = new byte[1024];

  private static final int MAXPKTBUF=32768;
  private static final String defaulthost="127.0.0.1",defaultxmlmsgname="XMLMSGS";
  private static final int defaultport=18800,defaulttimeout=50,defaultintl=1000,defaulttll=128;
  private static final double defaultxstart=0.0,defaultxdelta=1.0;

  private String host=defaulthost,midasdatatype="SB";
  private int port=defaultport,timeOut=defaulttimeout;
  private int intl=defaultintl,outtl=intl*2,midasfiletype=1000,tll=defaulttll;
  private int offsetdata=0;         //offset of where to start reading data in packets
  private String msgnameforxml=defaultxmlmsgname;  //the XML main name sent to processMessage
  private Object msgid=null;        //where to send messages, from /MSGID=
  private int rtpfiltertype=-1;     //filter on specific rtp type - used to pull a "channel" out of stream based on type
  private String msgidname=null;      //msgid's name

  private int curlen=0; //curlen is last >0 length
  private int pktfrcount=0; //bytes written to current frame so far
  private int fullwrites,remain;   //used to calculate fill bytes needed per packet
  /** basically for startup condition to find a header = Current HeaDeR, but not always updated */
  private int chdr=-1;
  private int len,offset,fillnum,pktsinframe,seq,pcount,psize,i,j;
  private byte fillValue = 0;       // value used to fill missing buffers
  private int framesz=0;
  private double ystartjtime=0;
  long numrecs;

  private boolean verbose=false;      //debugging messages
  private boolean packetDebug=false;  //Datagram data prints out detailed packet info
  private boolean rtp=true;           //rtp packets
  private boolean sap=false;          //sap type packets
  private boolean alladdr=false;      //set to a port and any address
  private boolean xml3000out=true;    //converts XML input into 3000 file output - tags must match 3000 file columns
  private boolean icehdrinpipe=false; //include icehdr in data?
  private boolean icehdrframe=false;  //set framesize from ice header data
  private boolean fill=false;         //fill requires constant packet sizes, doesn't work with variable data packets
  private boolean msgonly=false;      //used to output only Tables converted from XML to processMessage loop

  private boolean doTimeCode = false; // Indicates if time-code processing is being performed.
  private boolean doXSXD     = true;  // Used with doTimeCode to enable updates of XS and XD.

  /** The time code start time and delta */
  private double xStart=defaultxstart;
  private double xDelta=defaultxdelta;
  private double dgdxs=0.0, dgdxd=0.0, dgjtime=0.0;
  private int dgdsize=0, dgdbytes=0;
  private String dgdformat=null;

  /** For debugging, converts time of the datagram to readable form. */
  private Time dgTime = new Time();

///////////////////////////////////////////////////////////////////////////////////////////////////
  public boolean getVerbose () { return verbose; }
  public void setVerbose (boolean ptemp) { verbose=ptemp; }
  public void setDebug (boolean ptemp) { verbose=ptemp; }  //same as setVerbose

  /** Get is debugging DatagramData socket settings/major events
      @since NeXtMidas 2.7.4
  */
  public boolean getSocketDebug () {
    if (dgd==null) return false;
    try {
      return dgd.getVerbose();
    } catch (Exception e) {
      e.printStackTrace();
    }
    return false;
  }
  public void setSocketDebug (boolean ptemp) {
    if (dgd==null) return;
    try {
      dgd.setVerbose(ptemp);
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  /** Get is debugging DatagramData packets, e.g. to print out detailed packet info.
      @since NeXtMidas 2.7.4
   */
  public boolean getPacketDebug() { return packetDebug; }
  public void setPacketDebug (boolean ptemp) {
    packetDebug=ptemp;
    if (dgd==null) return;
    try {
      dgd.setPacketDebug(ptemp);
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  /** Get is reading RTP type packets
      @since NeXtMidas 2.7.4
   */
  public boolean getRTP() { return rtp; }
  public void setRTP (boolean ptemp) {
    rtp=ptemp;
    if (dgd==null) return;
    try {
      dgd.setRTP(ptemp);
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  /** Get is reading SAP packets
      @since NeXtMidas 2.7.4
   */
  public boolean getSAP() { return sap; }
  public void setSAP (boolean ptemp) {
    sap=ptemp;
    if (dgd==null) return;
    try {
      dgd.setSAP(ptemp);
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  /** Get filter on specific RTP type - used to pull a "channel" out of stream based on type.
      @return -1 if no RTP filter is set, otherwise RTP type that being filtered
      @since NeXtMidas 2.7.4
   */
  public int getRTPFilter () { return rtpfiltertype; }
  public void setRTPFilter (int ptemp) {
    rtpfiltertype=ptemp;
    if (dgd==null || rtpfiltertype<0) return;
    try {
      dgd.setRTPFilter(ptemp);
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  public String getHost () { return host; }
  public void setHost (String ptemp) {
    host=ptemp;
    if (dgd==null) return;
    try {
      dgd.setHost(ptemp);
    } catch (Exception e) {
      e.printStackTrace();
    }
  }
  public int getPort () { return port; }
  public void setPort (int ptemp) {
    port=ptemp;
    if (dgd==null) return;
    try {
      dgd.setPort(ptemp);
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  /** Get socket timeout (ms) to use for Datagrams.
      @since NeXtMidas 2.7.4
   */
  public int getTimeOut() { return timeOut; }
  public void setTimeOut (int ptemp) {
    timeOut=ptemp;
    if (dgd==null) return;
    try {
      dgd.setTimeOut(ptemp);
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  /** Get is listening to a specified port and any address (host).
      @since NeXtMidas 2.7.4
   */
  public boolean getAllHosts () { return alladdr; }
  public void setAllHosts (boolean ptemp) {
    alladdr=ptemp;
    if (dgd==null) return;
    try {
      dgd.setAllHosts(ptemp);
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  public int getOffset () { return offsetdata; }
  public void setOffset (int ptemp) {
    if (ptemp<0 || ptemp>MAXPKTBUF) offsetdata=0;
    else offsetdata=ptemp;
  }
  public boolean getXML3000 () { return xml3000out; }
  public void setXML3000 (boolean ptemp) { xml3000out=ptemp; }
  public int getTimeLineLength () { return tll; }
  public void setTimeLineLength (int ptemp) {
    if (ptemp<0) tll=defaulttll;
    else tll=ptemp;
  }
  public boolean getIceHdrFrame() { return icehdrframe; }
  public void setIceHdrFrame (boolean ptemp) { icehdrframe=ptemp; }

  public boolean getIceHdrInPipe () { return icehdrinpipe; }
  public void setIceHdrInPipe (boolean ptemp) {
    icehdrinpipe=ptemp;
    if (dgd==null) return;
    try {
      dgd.setIceHdrInPipe(ptemp);
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  public byte getFillValue() { return fillValue; }
  public void setFillValue(byte filler) {
    fillValue = filler;
  }

  public boolean getFill () { return fill; }
  public void setFill (boolean ptemp) {
    fill=ptemp;
    if (fill) {
      setIceHdrInPipe(false); //why am I setting this here?
      setRTP(true);
      dgd.setIceData(true);
      setIceHdrFrame(true); //need to fix this for Fill with no ICE Hdr in Frame
      if (verbose) M.info("Fill requires TC processing and NO XS-XD restarts");
    }
    if (dgd!=null) {
      try {
        dgd.setFill(ptemp);
      } catch (Exception e) {
        e.printStackTrace();
      }
    }
  }
  public void setFillBufValue (byte val) { Arrays.fill(fbuf,val); }
  public int setFillBuf (byte[] pfbuf) {
    try {
      if (pfbuf.length>0) {
        fbuf = new byte[pfbuf.length];
        System.arraycopy(pfbuf,0,fbuf,0,fbuf.length);
      }
    } catch (Exception e) {
      e.printStackTrace();
      M.error("sourcedg: setFillBuff failed:"+e);
    }
    if (fbuf==null) return 0;
    return fbuf.length;
  }

  public String getMSGID() { return msgidname; }
  public void setMSGID(String ptemp) {
    msgidname=ptemp;
    msgid=null;
    if (msgidname!=null && msgidname.length()>0) msgid = M.registry.getCommand(msgidname,5);
    if (msgonly && (msgid==null || !(msgid instanceof MessageHandler))) {
      M.warning("MSGONLY requires valid /MSGID - msgonly turned off");
      setMsgOnly(false);
    }
  }

  public boolean getMsgOnly() { return msgonly; }
  public void setMsgOnly(boolean ptemp) { msgonly=ptemp; }
  public boolean getTimeCode() { return doTimeCode; }
  public void setTimeCode(boolean ptemp) { doTimeCode=ptemp; }
  public boolean getXSXD() { return doXSXD; }
  public void setXSXD(boolean ptemp) { doXSXD=ptemp; } //Restart pipe on XS,XD changes?
  public String getMidaDataType() { return midasdatatype; }
  public void setMidasDataType(String ptemp) {
    if (ptemp==null || ptemp.length()<2) midasdatatype="SB";
    else midasdatatype=ptemp;
  }
  public double getXStart() { return xStart; }
  public void setXStart(double ptemp) { xStart=ptemp; }
  public double getXDelta() { return xDelta; }
  public void setXDelta(double ptemp) { xDelta=ptemp; }
  public double getYStartTime() { return ystartjtime; }
  public void setYStartTime(double ptemp) { ystartjtime=ptemp; }
  public double getPacketXStart() { return dgdxs; }  // since 3.0.0
  public double getPacketXDelta() { return dgdxd; }  // since 3.0.0
  public double getPacketJtime() { return dgjtime; } // since 3.0.0
  public int getSize() { return dgdsize; }           // since 3.0.0
  public int getBytes() { return dgdbytes; }         // since 3.0.0
  public String getFormat() { return dgdformat; }    // since 3.0.0

  public String getXMLMSGName() { return msgnameforxml; }
  public void setXMLMSGName(String ptemp) {
    if (ptemp==null || msgnameforxml.length()<1) msgnameforxml=defaultxmlmsgname;
    else msgnameforxml=ptemp;
  }

  public String toString () { return "DG: Host="+host+" Port="+port+" RTP="+rtp; }
//////////////////////////////////////////////////////////////////////////////////////////////////////////
  public int open () {
    try {
      if (!init) { // skip the below initialization block on a RESTART
        init=true;
        // 2010-12-21 only init DatagramData once
        dgd = new DatagramData();
        if (dgd==null) { M.error("sourcedg: Could not open socket"); return FINISH; }
        if (MA.find("/VERBOSE")) setVerbose(true);
        if (MA.find("/TIMECODE")) setTimeCode(true);
        if (MA.find("/PACKETS")) setPacketDebug(true);
        if (MA.find("/NOXSXD")) setXSXD(false);
        if (MA.find("/NORTP")) setRTP(false);
        if (MA.find("/ULAW")) { linear=false; linearin=false; }
        if (MA.find("/SAP")) setSAP(true);
        if (MA.find("/ALLHOSTS")) setAllHosts(true);
        if (MA.find("/MIDASSI")) setMidasDataType("SI");
        if (MA.find("/ICEHDRINPIPE")) setIceHdrInPipe(true);
        if (MA.find("/ICEHDRFRAME")) setIceHdrFrame(true);
        if (MA.find("/NORMAL3000")) setXML3000(false);
        if (MA.find("/MSGONLY")) setMsgOnly(true);  //wgd used XMLTBL but MSGONLY does same thing - XMLTBL XML Message always sent to MAIN
        setFrameSize(MA.getL("/FRAMESIZE",1024),false);
        if (MA.find("/FILL")) { setFill(true); setFillValue(MA.getB("/FILLVALUE", (byte)0)); }
        setAutoSeqResetMAX(MA.getL("/AUTOFILLRESET",0)); //to automatically reset fill stuff on certain counts
        setMSGID(MA.getU("/MSGID"));
        setTimeOut(MA.getL("/TIMEOUT",defaulttimeout));
        setRTPFilter(MA.getL("/RTPFILTER",-1));
        setOffset(MA.getL("/OFFSET",0));

        midasfiletype = MA.getL("/FTYPE",1000);
        intl = MA.getL("/INTL",defaultintl);
        if (intl<=1) intl=defaultintl;

        setXStart(MA.getD("/XSTART",defaultxstart));
        setXDelta(MA.getD("/XDELTA",defaultxdelta));
        setYStartTime(MA.getD("/INITIALTIME",-1.0));    //Initial ystart time (-1=Do not set, use qualifiers)
        setTimeLineLength(MA.getL("/TLL",defaulttll));  //TimeLine Length
        setXMLMSGName(MA.getS("/MSGNAME",defaultxmlmsgname));
        // check for log file
        if (MA.find("/LOG")) {
          logfile = MA.getTextFile("/LOG");
          if (logfile.open(TextFile.OUTPUT|TextFile.NOABORT)) {
            logfile.writeln("sourcedg log started at "+Time.tag());
          } else {
            M.warning("sourcedg: Unable to open log file "+logfile.getName());
            logfile = null;
          }
        }
        String hostport=MA.getS("HOSTPORT","");
        if (verbose) M.info("HOSTPORT="+hostport);
        i=hostport.indexOf(':');
        if (i<0) {
          if (sap) { host="224.2.127.254"; port=9875; }
          else { host=defaulthost; port=defaultport; }
        } else {
          host=hostport.substring(0,i).toLowerCase();
          port=Integer.parseInt(hostport.substring(i+1));
        }
        setHost(host);
        setPort(port);

        setFillBufValue(fillValue);

        outbuf = new short[MAXPKTBUF];
        buf = new byte[MAXPKTBUF];
        decodeTable = build_mu_decode_table();
      } //end initial startup

      chdr=-1; //reset current packet header
      if (!linear) outtl=intl*2; //2 is for linear out and byte in
      else outtl=intl;

      if (verbose) {
        M.info("sourcedg: TimeOut="+timeOut+". Verbose="+verbose+". PacketDebug="+packetDebug);
        M.info("sourcedg: RTP="+rtp+". XStart="+xStart+". XDelta="+xDelta);
        M.info("sourcedg: Out Type="+midasdatatype+" .SAP="+sap+" .ALL Hosts="+alladdr);
        M.info("sourcedg: Host="+host+". Port="+port);
      }

      if (msgonly) { //no pipes for messages so can stop here
        if (verbose) M.info("sourcedg: Messaging only");
        if (!getDgState()) {
          if (MA.find("/ON")) setDgState(MA.getState("/ON"));
        }
        return NORMAL;
      }

      pout = MA.getDataFile("OUT",""+midasfiletype,midasdatatype,DataFile.OUTPUT,0,null); //name,types,formats,0=size
      if (verbose) M.info("Pout="+pout);
      if (midasfiletype==2000) {
        pout.setType(midasfiletype);
        pout.setFrameSize(framesz);
        if (ystartjtime>=0) { pout.setYStart(ystartjtime); }
      } else if (midasfiletype==3000) {
        pout.setType(midasfiletype);
        try {
          t3kstruct = MA.getTable("/STRUCT");
          String tempstr="(";
          for (Table.Iterator ti=t3kstruct.iterator(); ti.getNext(); ) {
            tempstr=tempstr+ti.key+"|"+ti.value+",";
          }
          tempstr=tempstr.substring(0,tempstr.lastIndexOf(","));
          tempstr=tempstr+")";
          if (verbose) M.info("sourcedg: Setting 3000 file output");
          pout.setSubRecords(tempstr);
        } catch (Exception e) {
          throw new MidasException("sourcedg: Type 3000 file structure not defined. Use /STRUCT switch", e);
        }
      }
      pout.setTimeLineLength(tll);
      pout.setXStart(xStart);
      pout.setXDelta(xDelta);
      if (ystartjtime>=0) { pout.setTimeAt(ystartjtime); }
      if (doTimeCode){
        if (doXSXD) { pout.setXUnits(nxm.sys.inc.Units.FREQUENCY); }
        pout.setYUnits(nxm.sys.inc.Units.TIME);
        dgd.registerTimeCB(this);
      }
      pout.open();

      if (MA.find("CONVOUT")) {
        if (verbose) M.info("sourcedg: Convout created");
        convout = MA.getDataFile("CONVOUT","1000","SI",DataFile.OUTPUT,1,null);
        convout.setXStart(xStart);
        convout.setXDelta(xDelta);
        if (doTimeCode && doXSXD) convout.setXUnits(nxm.sys.inc.Units.TIME);
        convout.open();
        outdata=convout.getDataBuffer(outtl,INT);
      }

      if (getDgState()) setDgState("ON");
      else if (MA.find("/ON")) setDgState(MA.getState("/ON"));

    } catch (Exception e) {
      throw new MidasException("sourcedg: open failed", e);
    }

    return NORMAL;
  }
//////////////////////////////////////////////////////////////////////////////////////////////////////////
  public int process () {
    if (dgdclosed) return NOOP;
    if (abort) return NORMAL;
    return (read())? NORMAL: NOOP;
  }
//////////////////////////////////////////////////////////////////////////////////////////////////////////
  public int close () {
    try {
      if (verbose) System.out.println("Closing");
      if (dgd!=null) { dgd.close(); } //doesnt clear seq numbers
      FillPipeAtEnd();
      if (pout!=null) { pout.update(); pout.close(); }
      if (convout!=null) { convout.update(); convout.close(); }
      if (logfile!=null) {
        logfile.writeln("sourcedg log stopped at "+Time.tag());
        logfile.close();
      }
    } catch (Exception e) {
      throw new MidasException("sourcedg: close failed", e);
    }
    return NORMAL;
  }
//////////////////////////////////////////////////////////////////////////////////////////////////////////
  public void FillPipeAtEnd() {
    try {
      if (fill && framesz>0) { //check for left over data on close and fill in left over
        if (verbose) M.info("sourcedg: FillPipeAtEnd and checking for leftover data.");
        if (curlen>0) {
          if (pktfrcount<framesz && pktfrcount>0) {
            fillbuf(fbuf,framesz-pktfrcount);
            if (verbose) M.info("sourcedg: FillPipeAtEnd and filling buffer with "+(framesz-pktfrcount));
            pktfrcount=framesz;
          }
        }
      }
    } catch (Exception e) {
      throw new MidasException("sourcedg: FillPipeAtEnd failed", e);
    }
  }
//////////////////////////////////////////////////////////////////////////////////////////////////////////
  public synchronized int processMessage (Message msg) {
    if (verbose) M.info("sourcedg: Processing a message: "+msg.name);
    if (msg.name.equals("OPENNET")) {
      setDgState("OFF");
      String hostport=(String)msg.data;
      i=hostport.indexOf(':');
      if (i<0) {
        host=defaulthost;
        port=defaultport;
      } else {
        host=hostport.substring(0,i).toLowerCase();
        port=Integer.parseInt(hostport.substring(i+1));
      }
      if (verbose) M.info("sourcedg: Host="+host+". Port="+port);
      setHost(host);
      setPort(port);
      setDgState("ON");
    } else if (msg.name.equals("CLOSENET")) {
      setDgState("OFF");
    } else if (msg.name.equals("EXITNET")) {
      setDgState("OFF");
      close();
      return FINISH;
    }
    return NORMAL;
  }
//////////////////////////////////////////////////////////////////////////////////////////////////////////
//  /** @deprecated Since NeXtMidas 2.5.0 - This illegally overrides {@link Command#setState(String)}.
//                  use {@link #setDgState(String)} in place of this.
//   */
//  public void setState(String pvalue) {
//    setDgState(pvalue);
//  }

  /** Turns on/off sourcedg.
      @since NeXtMidas 2.5.0
   */
  public void setDgState(String pvalue) {
    setDgState(StringUtil.isTrue(pvalue));
  }
//////////////////////////////////////////////////////////////////////////////////////////////////////////
  /** Turns on/off sourcedg.
      @since NeXtMidas 2.5.0
   */
  public void setDgState(boolean on) {
    if (verbose) M.info("sourcedg: Set DgState:"+on);
    try {
      chdr=-1;
      if (!dgdclosed) {
        dgdclosed=true;
        if (dgd!=null) dgd.leave();
        FillPipeAtEnd();
      }
      if (on && dgd!=null) dgdclosed=!dgd.open();
    } catch (Exception e) {
       e.printStackTrace();
       M.error("sourcedg: setDgState failed:"+e);
    }
    if (verbose) M.info("sourcedg: Done:"+on+" and closed set="+dgdclosed);
  }
//////////////////////////////////////////////////////////////////////////////////////////////////////////
  /** Indicates if sourcedg is on/off.
      @since NeXtMidas 2.5.0
   */
  public boolean getDgState() { return !dgdclosed; }
  //** Set the linearin flag dynamically
  public void setLinear() { linearin=true; }
  public void setULAW() { linearin=false; }
  public void setAutoSeqResetMAX (int value) {
    if (dgd==null) return;
    try {
      dgd.setAutoSeqResetMAX(value);
    } catch (Exception e) {
      e.printStackTrace();
    }
  }
  /** Get the frame size of the output data file */
  public int getFrameSize () { return pout.getFrameSize(); }
  /** Set the frame size of the output data file. Also, updates the header. */
  public void setFrameSize (int value,boolean setpipe) {
    framesz=value;
    if (setpipe && pout!=null) {
      pout.setFrameSize(value);
      pout.update();
    }
  }
  public void setFrameSize (int value) { setFrameSize(value,true); if (fill) chdr = -1; }

  /** Get the Y delta of the output data file */
  public double getYDelta () { return pout.getYDelta(); }
  /** Set the Y delta of the output data file. Also, updates the header.  */
  public void setYDelta (double value) {
    pout.setYDelta(value);
    pout.update();
  }
//////////////////////////////////////////////////////////////////////////////////////////////////////////
  /** Reads a packet, returns true if a packet was read, false otherwise. This is called by {@link #process()}. */
  public boolean read () {
    try {
       if (dgd==null) return false;
       len=dgd.read(buf,offsetdata);
       if (len<=0) return true;
       curlen=len;
       offset=0;
       if (fill) {
         fillnum=dgd.getFillNum(); //fillnum just tells us the count of missing packets - no header calculations
         if (!icehdrframe && framesz<=0) { //framesz<=0 for 1000 file with fill on
           if (verbose) System.out.println("FILL:: 1000 file fill="+fillnum);
           if (fillnum>0) fillbuf(fbuf,len,fillnum);
           return true;
         }
         if (verbose) System.out.println("FILL:: Packet:"+dgd.getCurHdrSeq()+":"+dgd.getSeq()+":"+len+",fn="+fillnum);

         if (chdr<0) { //initial startup cases
           if (!dgd.isIceHdrPacket()) {
             if (verbose) System.out.println("FILL:: Skipping startup packets until header");
             return true; //startup and need to skip until packet with a header
           }
           chdr=dgd.getCurHdrSeq();
           if (verbose) System.out.println("FILL:: Startup packet:"+chdr);
           if (icehdrframe) {
             if (dgd.getBytes()>0) setFrameSize(dgd.getBytes(), true); //currently does not check all headers - only startup -should do all the time
             if (verbose) System.out.println("FILL:: Set Frame Size="+framesz);
           }
           pktfrcount=0; //gets incremented on write
           fillnum=0; //no fill needed on startup with hdr
           dgramTime(dgd.getICETime(),dgd.getDU2(),dgd.getDU1());
         }

         if (fillnum>0) { //know we skipped packets before this one when get here
           if (pktfrcount>=framesz) { //filled last packet so either skip or reset
             if (dgd.isIceHdrPacket()) { //current packet is header case
               pktfrcount=0; //gets incremented on write
               dgramTime(dgd.getICETime(),dgd.getDU2(),dgd.getDU1());
             } else { //ignore this packet
               if (verbose) System.out.println("FILL:: Skip packet, not header and previous frame filled fn="+fillnum+
                                               " pktfrcount="+pktfrcount+" framesz="+framesz+" len="+len);
               return true;
             }
           } else { //missing frame data
             if (dgd.isIceHdrPacket()) { //need to fill buffer as current is header frame
               fillbuf(fbuf,framesz-pktfrcount);
               if (verbose) System.out.println("FILL:: Header case framesz="+framesz+" pktfrcount="+pktfrcount+",fn="+fillnum);
               pktfrcount=0;
               dgramTime(dgd.getICETime(),dgd.getDU2(),dgd.getDU1());
             } else { //not header frame
               pktsinframe=framesz/len;
               chdr=dgd.getCurHdrSeq();
               seq=dgd.getSeq();
               pcount=seq-chdr;
               if (pcount<0) {
                 if (verbose) System.out.println("FILL:: rollover or prev case seq="+seq+",hdr="+chdr);
                 if (pcount<=-100) pcount=pcount+65536; //likely case rollover, else very strange case
                 else {
                   if (verbose) System.out.println("FILL:: Out of order case seq="+seq+",hdr="+chdr);
                   return true;
                 }
               }
               if (pcount>pktsinframe) {
                 fillbuf(fbuf,framesz-pktfrcount);
                 if (verbose) System.out.println("FILL:: Skip Cross frame case pktsinframe="+pktsinframe+" fn="+fillnum+
                                                 " pcount="+pcount+" seq="+seq+" chdr="+chdr+",fill="+(framesz-pktfrcount));
                 pktfrcount=framesz;
                 return true;
               } else {
                 pktfrcount+=fillbuf(fbuf,fillnum*len);
                 if (verbose) System.out.println("FILL:: Within frame case pktsinframe="+pktsinframe+" fn="+fillnum);
               }
             }
           }
         } else { //when fillnum<0 by more than some frames, restart
           if (dgd.isIceHdrPacket()) {
             if (verbose && pktfrcount<framesz && pktfrcount!=0) System.out.println("FILL:: ERROR - frame not filled in framesz="+framesz+","+pktfrcount);
             pktfrcount=0;   //reset on packetheader
             dgramTime(dgd.getICETime(),dgd.getDU2(),dgd.getDU1());
           } else if (pktfrcount>=framesz) {
             if (verbose) System.out.println("FILL:: Skip packet, looking for header and previous frame filled");
             return true;
           }
         }
       }

       if (len>0) {
         if (msgonly) { //no pipe case
           String tmpbuf = new String(buf,0,len);
           if (verbose) System.out.println("XML In="+tmpbuf);
           Table XMLTab = XmlFile.fastXmlToTable(tmpbuf, null, null, -1, true);
           sendMessage((MessageHandler)msgid,new Message(msgnameforxml,1,XMLTab));
         } else {
//         if (pout.avail()<len) {
//           M.info("sourcedg: Pipe full - dumping data. Len="+len+". Avail="+pout.avail());
//           double p = pout.avail();
//           len=(int)p;
//         }
           if (pout==null) {
             M.info("sourcedg: No output pipe.");
             return true;
           }
           if (pout.io==null) {
             M.error("sourcedg: Pipe open failed");
             return true;
           }
           if (midasfiletype==2000 || midasfiletype==1000) {
             if (verbose) System.out.println("  Writing len="+len);
             pout.write(buf,0,len);
             if (fill) pktfrcount+=len;
             if (pktfrcount>framesz) {
               if (verbose) System.out.println("ERROR IN CALCULATION="+framesz+">pkt="+pktfrcount);
               pktfrcount=framesz;
             }
             if (convout!=null && convout.io!=null && intl>0) {
               numrecs=(long)(len/intl);
               psize=len;
               if (numrecs<1) numrecs=1;
               if (len>intl) psize=intl;
               offset=0;

               if (convout.avail()<(len+len)) M.info("sourcedg: Convpipe full: len="+len+", avail="+convout.avail());
               else {
                  if (!linear && !linearin) {
                   for (j=0;j<numrecs;j++) {
                     outbuf=outdata.castI(false);
                     for (i=0;i<psize; i++) outbuf[i]=decodeTable[buf[offset+i]+128];
                     outdata.uncast(outbuf,true);
                     convout.write(outdata,psize);
                     offset+=psize;
                   }
                 } else convout.write(buf,0,len);
               }
             }
           } else if (midasfiletype==3000) {
           //Skip the first 8 bytes
             if (xml3000out) {
               String tmpbuf = new String(buf,8,buf.length-8);
               //Only handles XML in single string
               if (verbose) System.out.println("SDG: XML '"+tmpbuf+"'");
               Table XMLTab = XmlFile.fastXmlToTable(tmpbuf, null, null, -1, true);
               Table outtab = matchTableInfo(t3kstruct,XMLTab);
               if (verbose) {
                 System.out.println("SDG: XML as a Table "+XMLTab);
                 System.out.println("SDG: OUTPUT Table "+outtab);
               }
               //XML data must match type 3000 file structure
               pout.setData(1,outtab);
             } else pout.write(buf,0,len);
           }
         }
       }
       return true;
     } catch (Exception e) {//receive can throw IO, SocketTimeOut,PortUnreachable and IllegalBlockingMode Exceptions
        e.printStackTrace();
        M.warning("sourcedg: Read failed. "+e);
     }
     return false;
  }
//////////////////////////////////////////////////////////////////////////////////////////////////////////
  public int fillbuf(byte[] fbuf,int len,int fillnum) {
    if (fillnum<=0 || len<1) return 0;
    return fillbuf(fbuf,len*fillnum);
  }
  public int fillbuf(byte[] fbuf,int totbytes) {
    if (verbose) System.out.println("  Filling totbytes="+totbytes);
    if (pout==null) return -1;
    if (totbytes<1) return 0;
    if (totbytes<=fbuf.length) {
      pout.write(fbuf,0,totbytes);
      return totbytes;
    }
    fullwrites=totbytes/fbuf.length;
    for (i=0;i<fullwrites;i++) {
      pout.write(fbuf,0,fbuf.length);
    }
    remain=totbytes-(fullwrites*fbuf.length);
    if (remain>0){
      pout.write(fbuf,0,remain);
    }
    return totbytes;
  }
//////////////////////////////////////////////////////////////////////////////////////////////////////////
  public void sendMessage (MessageHandler msgid, Message msg) {
//    if (msgid instanceof MessageHandler) { // NTN: since 3.1.1: this check is NOT needed
      if (msgid == msg.from) return;    // don't send it back
      msg.to = msgid;
      msg.from = this;
      msgid.processMessage(msg);
//    }
  }
//////////////////////////////////////////////////////////////////////////////////////////////////////////
  private Table matchTableInfo (Table structure, Table inTable) {
    String keyvals = structure.getKeyList();
    StringTokenizer keyST = new StringTokenizer(keyvals,",",false);
    Table retTab = new Table();

    while (keyST.hasMoreTokens()) {
      String structKey = keyST.nextToken();
      if (structKey.length()>4) structKey = structKey.substring(0,4);
      String fmt = structure.getS(structKey);
      String tblVal = inTable.getS(structKey);
      if (tblVal != null) {
        if (fmt.equals("SD")) retTab.put(structKey,Convert.s2d(tblVal));
        else if (fmt.equals("SL")) retTab.put(structKey,Convert.s2l(tblVal));
        //else if (fmt.equals("SB"))
        else retTab.put(structKey,tblVal);
      }
    }
    return retTab;
  }
//////////////////////////////////////////////////////////////////////////////////////////////////////////
  private short[] build_mu_decode_table () {
    short[] convbuf = new short[256];

    int i,j,x,S,L,V;
    for (i=0; i<256; i++) {
      j = i-128;
      x = j^127;
      S = 1-(2*((x&128)/128));
      L = (x&112)/16;
      V = x&15;
      convbuf[i] = (short) Math.round(((S*(Math.pow(2,L)*(2*V+33)-33)*32767.0)/8031.0));
    }
    return convbuf;
  }
//////////////////////////////////////////////////////////////////////////////////////////////////////////
  public void dgramTime (double jtime, double inxStart, double inxDelta){
    try {
//      if (packetDebug) {
      if (verbose) {
        dgTime.fromJ1950(jtime);
        System.out.println("sourcedg: dgramTime = "+dgTime+" XS="+inxStart+" XD="+inxDelta);
      }

      pout.setTimeAt(jtime);
      if (convout!=null) convout.setTimeAt(jtime);
      dgjtime=jtime;
      dgdxs=inxStart;
      dgdxd=inxDelta;
      dgdsize=dgd.getSize();
      dgdbytes=dgd.getBytes();
      dgdformat=dgd.getFormat();

      if ((icehdrframe || doXSXD) && (inxStart!=getXStart() || inxDelta!=getXDelta())) {
        if (verbose) M.info("Update message XS="+inxStart+"("+getXStart()+") XD="+inxDelta+"("+getXDelta()+")");
        setXStart(inxStart);
        setXDelta(inxDelta);
        if (icehdrframe && msgid instanceof MessageHandler) {
          Table XMLTab = new Table();
          XMLTab.put("TYPE","XSXDUPD");
          XMLTab.put("XD",inxDelta);
          XMLTab.put("XS",inxStart);
          XMLTab.put("SIZE",dgdsize);
          XMLTab.put("BYTES",dgdbytes);
          XMLTab.put("FORMAT",dgdformat);
          sendMessage((MessageHandler)msgid,new Message(msgnameforxml,1,XMLTab));
        }
        if (doXSXD) {
          if (verbose) M.info("sourcedg: Restarting with xStart="+inxStart+" xDelta="+inxDelta);
          restartSourcedg(jtime);
        }
      }
    } catch (Exception e) {
      e.printStackTrace();
      M.warning("sourcedg: checkIceParms failed. "+e);
    }
  }
//////////////////////////////////////////////////////////////////////////////////////////////////////////
  public void restartSourcedg() { setState(RESTART); }
  private void restartSourcedg(double jtime) {
    setYStartTime(jtime);
    restartSourcedg();
  }
}
