//
//  Copyright (C) 2002 Constantin Kaplinsky, Inc.  All Rights Reserved.
//
//  This is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This software is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this software; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
//  USA.
//

//
// HTTPConnectSocketFactory.java together with HTTPConnectSocket.java
// implement an alternate way to connect to VNC servers via one or two
// HTTP proxies supporting the HTTP CONNECT method.
//
// --------------------------------------------------------------------
// Modified on 16-Apr-2003 by Warren Knight
// To support Basic Authentication from the HTTP Proxy Server
// --------------------------------------------------------------------
package nxm.vnc.libvnc.tmde;
import nxm.vnc.libvnc.tightvnc.*;

import java.applet.*;
import java.net.*;
import java.io.*;

class HTTPConnectSocketAuthFactory implements SocketFactory {

  public Socket createSocket(String host, int port, Applet applet)
  throws IOException {

    return createSocket(host, port,
                        applet.getParameter("PROXYHOST"),
                        applet.getParameter("PROXYPORT"));
  }

  public Socket createSocket(String host, int port, String[] args)
  throws IOException {

    return createSocket(host, port,
                        readArg(args, "PROXYHOST"),
                        readArg(args, "PROXYPORT"));
  }

  public Socket createSocket(String host, int port,
                             String proxyHost, String proxyPortStr)
  throws IOException {

    int proxyPort = 0;
    if (proxyPortStr != null) {
      try {
        proxyPort = Integer.parseInt(proxyPortStr);
      } catch (NumberFormatException e) { }
    }

    if (proxyHost == null || proxyPort == 0) {
      System.out.println("Incomplete parameter list for HTTPConnectSocket");
      return new Socket(host, port);
    }

    System.out.println("HTTP CONNECT via proxy " + proxyHost +
                       " port " + proxyPort);

    // Attempt to connect to the proxy without authentication
    HTTPConnectSocketAuth s =
      new HTTPConnectSocketAuth(host, port, proxyHost, proxyPort);

    // If authentication was required retry with password
    AuthProxy authenticator = new AuthProxy();
    authenticator.pack();
    authenticator.setLocationRelativeTo(null);
    while (s.isClosed() && s.isAuthenticationRequired()) {
      String authscheme = s.getAuthScheme();

      System.out.println("HTTP CONNECT via proxy " + proxyHost +
                         " port " + proxyPort + " Auth " +authscheme);

//    authenticator.show();            // ORIG prior to 2.9.2
      authenticator.setVisible(true);  // since 2.9.2 to fix Deprecated warning
      // Wait for user entering a password.
      synchronized(authenticator) {
        try {
          authenticator.wait();
        } catch (InterruptedException e) {
        }
      }
//    authenticator.hide();            // ORIG prior to 2.9.2
      authenticator.setVisible(false); // since 2.9.2 to fix Deprecated warning
      if (authenticator.isCanceled()) break;
      String credentials = authenticator.username.getText() + ":" + authenticator.password.getText();
      String base64credentials = new String(Base64.encode(credentials.getBytes()));

      // Try to recreate the Socket with the new credentials
      s = new HTTPConnectSocketAuth(host, port,
                                    proxyHost, proxyPort,
                                    authscheme, base64credentials);

      if (s.isAuthenticationRequired()) {
        authenticator.setMessage("Sorry try again");
      }
    } // end while loop
    return (Socket)s;
  }

  private String readArg(String[] args, String name) {

    for (int i = 0; i < args.length; i += 2) {
      if (args[i].equalsIgnoreCase(name)) {
        try {
          return args[i+1];
        } catch (Exception e) {
          return null;
        }
      }
    }
    return null;
  }
}

