c************
c  Description:	Structure Def for Midas Blue Header Control Block
c
c  Author:	JGS 4/2/90
c
c  $Id: headers.incp,v 1.7 2009/08/11 00:25:53 ntn Exp $
c  $Revision: 1.7 $
c  $Modtime:   20 Mar 1996 14:09:06  $
c
c************
c

! The following structure definition is the same as the M_PTR
! definition, but was created separately because I was concerned that
! if TYPES.INC was included in HEADERS.INC it would cause errors in
! existing primitives that were already including TYPES.INC.
	structure /H_PTR/
!if _64BIT
!
!else
	  integer*4 p
	  integer*4 FILL
!endif
	end structure

	integer*4  L_quadwords,L_components,L_subrecords,L_outlets
	integer*4  L_keywords,L_header_size

	parameter (L_quadwords=12)	! max quadwords in 5000 file
	parameter (L_components=14)	! max components in 5000 file
	parameter (L_subrecords=26)	! max subrecords in 3000 file
	parameter (L_outlets=8)		! max outlets for a pipe
	parameter (L_keywords=92)	! max internal keywords characters
	parameter (L_header_size=1536)	! size of header structure

	structure /SUBRECSTRUCT/
	  character*4	name			! 4 char subr name
	  character*2	format			! subr format code
	  integer*2	offset			! byte offset into record
	end structure

	structure /COMPSTRUCT/
	  character*4	name			! 4 char comp name
	  character*2	format			! comp format code
	  byte		type			! comp type (see tables)
	  byte		units			! comp units (see tables)
	end structure

! Added for type 6000 support
        integer*4 L_t6subr_str_len      ! length of T6 subr. def. string
        parameter (L_t6subr_str_len = 96)
        integer*4 L_t6subr_len          ! size of T6 subr. def. table
        parameter (L_t6subr_len = 64)

! Define offset parameter constants for the fields in the Type 6000 subrecord
! definition structure, since it is held in dynamic memory buffers and we
! sometimes use M$MOVE to get/put info

        integer*4 T6O_name
        parameter (T6O_name = 0)
        integer*4 T6O_minval
        parameter (T6O_minval = 24)
        integer*4 T6O_maxval
        parameter (T6O_maxval = 32)
        integer*4 T6O_offset
        parameter (T6O_offset = 40)
        integer*4 T6O_num_elts
        parameter (T6O_num_elts = 44)
        integer*4 T6O_units
        parameter (T6O_units = 48)
        integer*4 T6O_uprefix
        parameter (T6O_uprefix = 52)
        integer*4 T6O_format
        parameter (T6O_format = 56)

        structure /T6SUBR_STR_STRUCT/   ! Type 6000 subrecord definition
        union                   ! in Extended Header
          map
          character*24 name             ! subrecord name
          character*24 minval           ! minimum value of subrecord
          character*24 maxval           ! maximum value of subrecord
          character*8 offset            ! byte offset into record
          character*4 num_elts          ! # elements in subrecord
          character*4 units             ! subrecord units code
          character*2 format            ! subr format code
          character*3 uprefix           ! units Greek Prefix code
          character*3 reserved          ! pad to 8-byte boundry
          end map
          map
            character*(L_t6subr_str_len) definition ! whole subr. def. string
          end map
        end union
        end structure

        structure /T6SUBR_STRUCT/       ! Type 6000 binary subrecord
        union           ! definition in memory
          map
          character*24 name     ! subrecord name
          real*8 minval         ! minimum value of subrecord
          real*8 maxval         ! maximum value of subrecord
          integer*4 offset              ! byte offset into record
          integer*4 num_elts    ! # elements in subrecord
          integer*4 units               ! subrecord units code
          integer*4 uprefix             ! units Greek Prefix code
          character*2 format    ! subrecord format code
          character*6 reserved  ! pad to 8-byte boundry
          end map

          map
            byte definition(L_t6subr_len)       ! whole subrecord def. table
          end map
        end union
        end structure


	structure /HEADER/

	union					! file basic header
	 map
	  character*512	main
	 end map

	 map
	  character*4	version			! Header version
	  character*4	head_rep		! Header representation
	  character*4	data_rep		! Data representation
	  integer*4	detached		! detached header
	  integer*4	protected		! protected from overwrite
	  integer*4	pipe			! pipe mode

	  integer*4	ext_start		! extended header start block
	  integer*4	ext_size		! extended header size bytes
	  real*8	data_start		! data start in bytes
	  real*8	data_size		! data size in bytes

	  integer*4	type			! file type code
	  character*2	format			! data format code
	  integer*2	flagmask		! 16 bit flag mask (1=flagbit)
	  real*8	timecode		! time code field

	  integer*2	inlet			! inlet owner
	  integer*2	outlets			! number outlets
	  integer*4	outmask			! outlet async mask
	  integer*4	pipeloc			! pipe location
	  integer*4	pipesize		! pipe size in bytes
	  real*8	in_byte			! next input byte
	  real*8	out_byte		! next out byte (cumulative)
	  real*8	out_bytes(L_outlets)	! next out byte (each outlet)

	  integer*4	keylength		! length of keyword string
	  character*(L_keywords) keywords	! user defined keyword string
!
! The following should be at offset 0x100
!
	 end map

	 map					! type 1000/2000 adjunct
	  character*256 FILL1
	  real*8	xstart			! abscissa start
	  real*8	xdelta			! abscissa delta
	  integer*4	xunits			! abscissa units
	  integer*4	subsize			! element size (type 2000)
	  real*8	ystart			! secondary start
	  real*8	ydelta			! secondary delta
	  integer*4	yunits			! secondary units
	  integer*4     bid			! avl dr# 5349 TMS buffer id
	 end map

	 map					! type 3000 adjunct
	  character*256 FILL2
	  real*8	rstart			! abscissa start
	  real*8	rdelta			! abscissa delta
	  integer*4	runits			! abscissa units
	  integer*4	subrecords		! # subrecords per record
	  real*8	r2start			! abscissa 2 start
	  real*8	r2delta			! abscissa 2 delta
	  integer*4	r2units			! abscissa 2 units
	  integer*4	record_length		! record length in bytes
	  record /SUBRECSTRUCT/ subr(L_subrecords)
	 end map

	 map					! type 4000 adjunct
	  character*256 FILL3
	  real*8	vrstart			! abscissa start
	  real*8	vrdelta			! abscissa delta
	  integer*4	vrunits			! abscissa units
	  integer*4	nrecords		! # records in file
	  real*8	vr2start		! abscissa 2 start
	  real*8	vr2delta		! abscissa 2 delta
	  integer*4	vr2units		! abscissa 2 units
	  integer*4	vrecord_length		! record length in bytes
	 end map

	 map					! type 5000 adjunct
	  character*256 FILL4
	  real*8	tstart			! abscissa start
	  real*8	tdelta			! abscissa delta
	  integer*4	tunits			! abscissa units
	  integer*4	components		! # components per record
	  real*8	t2start			! abscissa 2 start
	  real*8	t2delta			! abscissa 2 delta
	  integer*4	t2units			! abscissa 2 units
	  integer*4	record_len		! record length in bytes
	  record /COMPSTRUCT/ comp(L_components)
	  real*8        quadwords(L_quadwords)	! systems modelling frame info
	 end map

	 map					! 5001 quadword specifics
	  character*416 FILL5
	  character*8	frame_of_ref		! coordinate reference frame
	  real*8	altitude		! altitude (meters) for topocent
	  real*8	latitude		! latitude (deg) for topocentric
	  real*8	longitude		! latitude (deg) for topocentric
	  real*8	azimuth			! azimuth (deg) for topocentric
	  real*8	elevation		! elevation (deg) for topocentric
	  real*8	roll			! roll (deg) for topocentric
	  real*8	notused1		!
	  real*8	notused2		!
	  real*8	epoch_year		! year of epoch for ECI
	  real*8	epoch_seconds		! seconds of epoch for ECI
	  real*8	hour_angle		! GW angle (rad) at epoch f ECI
	 end map
	 map					! 5001 quadword specifics
	  character*416 FILL6
	  character*96	quads			!
	 end map
	 map					! 5001 quadword specifics
	  character*416 FILL7
	  real*8	quad(L_quadwords)	!
	 end map
	end union

	union					! internals header
	 map
	  character*1024 internals
	 end map

	 map
!
! The following should be at offset 0x200
!
	  character*80	file_name		! file name to/from M$ALLOCATE
	  character*80	detach_name		! file name for .DET file
	  character*40	type_allow		! string of allowable type codes
	  character*40	format_allow		! string of allow format codes
	  character*40	qualifier		! qualifier string
	  character*20	label			! filename result label
	  character*12	rtype			! file record type
	  character*8	switch			! switch string
	  character*4	extension		! filename extension field
	  integer*4	qualifierlen		! qualifier string length
	  integer*4	file_lu			! file logical unit
	  integer*4	head_lu			! file logical unit
	  logical*4	exist			! file existence flag
	  logical*4	open			! file open flag
	  logical*4	output			! open for output/input
	  integer*4	class			! type code class (type/1000)

	  real*8	dbpe			! DP bytes per element
	  real*8	size			! file size in log elements
	  real*8	offset			! file read offset in log elem
	  real*8	save_offset		! byte offset of saved
	  real*8	first_byte		! data read first byte
	  real*8	last_byte		! data write last byte

	  integer*4	bps			! bytes per scalar
	  integer*4	spa			! scalars per atom
	  integer*4	bpa			! bytes per atom
	  integer*4	bpe			! bytes per element
	  integer*4	ape			! atoms per element
	  integer*4	save_size		! # elements in saved
	  integer*4	savebuf_loc		! save buffer location (GRABX) TODO: XM44: H_PTR
	  integer*4	savebuf_len		! save buffer size in bytes
	  integer*4	buffer_block		! block # of data in header buf
	  integer*4	xfer_len		! transfer length
	  integer*4	cons_len		! consume length

	  character*2	scr_format		! scratch format
	  character*1	buf_mode		! buffer format mode
	  character*1	buf_type		! buffer format type

	  logical*1	readonly		! open for read only ?
	  logical*1	overwrite		! open for over write ?
	  logical*1	contiguous		! open contiguous disk file ?
	  logical*1	flush			! keep output flushed to disk ?

	  integer*4	outlet			! pipe outlet number

	  integer*4	disk_blk		! disk address of file data
	  integer*4	disk_blk_hdr		! disk address of file header
	  integer*4	disk_blk_alloc		! disk allocation of data
	  integer*4	disk_address		! disk addr data in bytes

	  integer*4	ext_loc			! ext header mem location
	  integer*4	ext_alloc		! ext header mem allocated
	  integer*4	ext_scope_start		! ext header scope start
	  integer*4	ext_scope_end		! ext header scope end
	  integer*4	ext_next		! ext next

	  integer*2	last_bits		! bits in save_bits byte
	  byte		save_bits		! holds packed offset byte
	  byte		aux			! aux disk

	  integer*4	lph			! location of pipe header (ID mapping to DataFile object)
	  integer*4	blk			! DS disk block size

	  integer*4	status			! DS status flag
	  byte		mode			! DS IO mode
	  byte		engine			! DS engine #
	  byte		active			! DS active
	  byte		dsfill			! DS filler
c     512 bytes so far

	  record /H_PTR/ t6subr_loc		! T6 subrecords mem location
	  integer*4	t6subr_alloc		! T6 subrecords mem allocated
	  integer*4 filler			! filler to 8 byte boundary (alignment)

	  byte	ihdrfill(496)			! fill internal header to 1024

	 end map
	end union

	byte		buffer(512)		! primitive data buffer

	end structure


c************
c  Description:	Structure Def for Virtual recod buffer utilities
c************

	integer*4 L_vrbhead,L_keyhead
	parameter (L_vrbhead=24)	! size of vrb header
	parameter (L_keyhead=8)		! size of key header

	integer*4  MK_normal
	parameter  (MK_normal=0)
	integer*4  MK_main,MK_ext,MK_noscope,MK_next
	parameter  (MK_main=1,MK_ext=2,MK_noscope=4,MK_next=8)
	integer*4  MK_typeasis,MK_typetag,MK_replace
	parameter  (MK_typeasis=16,MK_typetag=32,MK_replace=64)
	integer*4  MK_index,MK_nextmatch,MK_last
	parameter  (MK_index=128,MK_nextmatch=256,MK_last=512)

	integer*4  BK_main,BK_ext,BK_noscope,BK_next
	parameter  (BK_main=0,BK_ext=1,BK_noscope=2,BK_next=3)
	integer*4  BK_typeasis,BK_typetag,BK_replace
	parameter  (BK_typeasis=4,BK_typetag=5,BK_replace=6)
	integer*4  BK_index,BK_nextmatch,BK_last
	parameter  (BK_index=7,BK_nextmatch=8,BK_last=9)

	integer*4  VRB_INIT,VRB_OFFSET,VRB_READ,VRB_WRITE
	parameter  (VRB_INIT=1,VRB_OFFSET=2,VRB_READ=3,VRB_WRITE=4)

	structure /KEYSTRUCT/
	  integer*4	lkey
	  integer*2	lext
	  byte		ltag
	  character	type
	end structure

	structure /VRB/
	  integer*4	fsize
	  integer*4	size
	  integer*4	buf_loc
	  integer*4	buf_len
	  integer*4	record_len
	  integer*4	next
	end structure

c************
c  Description:	Definitions for HCB function modes
c************

	integer*4 HCBF_CTG,HCBF_AUX,HCBF_SAM,HCBF_FLUSH
	integer*4 HCBF_APPEND,HCBF_REOPEN,HCBF_LASTBYTE
	integer*4 HCBF_DETACH,HCBF_INPUT,HCBF_OUTPUT
	integer*4 HCBF_HCBMOD,HCBF_OPTIONAL,HCBF_NOABORT
	integer*4 HCBF_PRESERVE,HCBF_WAIT,HCBF_EOF
	integer*4 HCBF_AVAIL,HCBF_SKIP,HCBF_PS

	parameter (HCBF_CTG=1,HCBF_AUX=2,HCBF_SAM=4,HCBF_FLUSH=8)
	parameter (HCBF_APPEND=16,HCBF_REOPEN=32,HCBF_LASTBYTE=64)
	parameter (HCBF_DETACH=128,HCBF_INPUT=256,HCBF_OUTPUT=512)
	parameter (HCBF_HCBMOD=1024,HCBF_OPTIONAL=2048,HCBF_NOABORT=4096)
	parameter (HCBF_PRESERVE=8192,HCBF_WAIT=16384,HCBF_EOF=32768)
	parameter (HCBF_AVAIL=65536,HCBF_SKIP=131072,HCBF_PS=262144)
