package nxm.xbc.lib;

import nxm.sys.inc.*;
import nxm.sys.lib.*;

/**
  A class to interface to an X-Midas session.
  This class also has ALL of the native methods for results/pipes/messages/environment.
  
  @author Jeff Schoen
  @version $Id: XmClient.java,v 1.1 2005/04/20 14:42:39 jgs Exp $

*/
public class XmClient implements ShellParent { 

  public static boolean useNative;
  public static XEnvironment XE;
  public static XResults XR;

  /** get the next command to run */
  public static native int getCommand (byte[] buf, int maxlen);

  /** acknowledge parsing of command to run */
  public static native void ackCommand (int mode);

  /** sets an X-Midas environment parameter */
  public static native int setEnvironment (String name, byte[] type, byte[] buf, int len);

  /** gets an X-Midas environment parameter */
  public static native int getEnvironment (String name, byte[] type, byte[] buf, int maxlen);

  /** sets an X-Midas results parameter */
  public static native int setResult (String name, byte[] type, byte[] buf, int len);

  /** gets an X-Midas results parameter */
  public static native int getResult (String name, byte[] type, byte[] buf, int maxlen);

  /** gets an X-Midas results parameter */
  public static native int getResultNames (String match, byte[] buf, int maxlen);


  public static void runShell() {
    XmClient xmc = new XmClient();
    xmc.run();
  }
  public void run() {
    int stat;
    byte[] type = new byte[1];
    byte[] buf = new byte[1024];
    System.out.println("OK Entering Shell");
    Shell shell = new Shell((ShellParent)this);
    shell.setPriv(Privileges.PRIV_ALL);
    shell.open();
    loadLib();
    XR = new XResults();
    XE = new XEnvironment();
    shell.M.results.put("XR",XR);
    shell.M.results.put("XE",XE);
    for (int i=0; i<100; i++) {
      String key = "AUX."+i;
      String name = (String)XE.getKey(key);
      if (name==null) continue;
      if (!name.endsWith("/")) name += "/";
      shell.M.results.put(key,name);
    }
    ackCommand(-1);
    while (shell.M.status != Commandable.EXIT) {
      stat = getCommand(buf,buf.length);
      if (stat>0) {
        String cmdline = new String(buf,0,stat);
        Args args = shell.parseCommand(shell.M,cmdline);
	ackCommand(-2);
        Command command = shell.loadCommand(shell.M,args);
        if (command==null) { System.out.println ("couldn't find command : " + cmdline); }
        else shell.execCommand(shell.M,command);
	ackCommand(-1);
        System.gc();  // be tidy
        if (command.name.equals("EXIT")) break;
      }
      else Time.sleep(0.1);
    }
    shell.close();
    ackCommand(0);
    System.out.println("OK Exiting Shell");
  }

  public static void loadLib() {
    useNative = Shell.loadLibrary ("xbc","lib","XmClient");
  }

  /* ShellParent methods */

  /** {@inheritDoc} */
  public String getAux() {
    final String defAux = "HOME,HOME|RAM|DAT";
    String aux = System.getProperty("AUX");
    if (aux==null) aux = defAux;
    return aux;
  }

  /** {@inheritDoc} */
  public String getNmRoot () {
    return System.getProperty("NMROOT");
  }

  /** {@inheritDoc} */
  public String getOpts() {
    return System.getProperty("OPTS");
  }

  /** {@inheritDoc} */
  public String getCWD() {
    return null;
  }

  /** {@inheritDoc} */
  public String getNMS() {
    return null;
  }

  /** {@inheritDoc} */
  public String getHome() { 
    return null;
  }

  /** {@inheritDoc} */
  public String getServer() {
    return null;
  }

  /** {@inheritDoc} */
  public String getTheme() {
    return null;
  }

  /** {@inheritDoc} */
  public String getStyle() {
    return null;
  }

}
