@echo %DEBUG% off
rem ############################################################################
rem NeXtMidas shell instantiation script
rem convention is to user all lower case for local variables
rem
rem @author Jeff Schoen / Neon Ngo
rem @version $Id: nm.bat,v 1.39 2012/08/13 15:35:00 jwb Exp $
rem ############################################################################

rem save all command line arguments into args (this does not include %0)
set args=%*

rem set source and target Java version (1.5, 1.6, etc.) for bootstrap build ONLY
rem if your option tree needs needs to target Java 6 or higher, update build.props
set jvm_target_ver=-source 1.6 -target 1.6

if not defined NM_BOOT_CP_APPEND (
  set nm_boot_classpath=%NMROOT%/tools
) else (
  set nm_boot_classpath=%NMROOT%/tools;%NM_BOOT_CP_APPEND%
)

set nm_java_cmd="%NM_JAVA%" %NM_JVM_FLAGS% -classpath "%nm_boot_classpath%;%JAVAHOME%\lib\tools.jar"
set nm_javac_cmd="%NM_JAVAC%" %jvm_target_ver% -classpath "%nm_boot_classpath%;%NMROOT%;%JAVAHOME%\lib\tools.jar"

if [%1] == [make] (
  rem # BUG 1607: only run the boot-strap build when building ALL or SYS option tree(s)
  rem #           or when making zip archive for distribution
  rem #           or when cleaning under $NMROOT
  rem #           if option tree (3rd argument) is empty, it defaults to ALL
  rem # /I switch is for case insensitive string compares
  if /I [%2] == [zip]   goto MAKE
  if /I [%2] == [clean] goto MAKE
  if    [%3] == []      goto MAKE
  if /I [%3] == [all]   goto MAKE
  if /I [%3] == [sys]   goto MAKE
)
if [%1] == [end]  goto END
if [%1] == [edit] goto EDIT
if not exist "%NMROOT%\nxm\sys\lib\Shell.class"   goto NOBUILD
if not exist "%NMROOT%\tools\NmClassLoader.class" goto BUILDNMCLASSLOADER
goto SHELL

rem ############################################################################
rem # NOTE:       The Java flags are used for boot strapping NeXtMidas and should
rem #             not be modified. To configure your source / target build, see the
rem #             NMROOT\nxm\sys\cfg\build.props and override with one in your option
rem #             tree.
rem # NeXtMidas (boot-strap) Build
:MAKE
  echo Compiling Tools
  %nm_javac_cmd% %NMROOT%\tools\*.java

  %nm_java_cmd% -DSHELL=tools.NmMake NmClassLoader %args%

  rem Need to run the bat file to rename the DLL's
  if not exist "%NM_SF%.bat" goto DONE
  call "%NM_SF%.bat"
  del  "%NM_SF%.bat"

  rem reload the nmstartup.mm file if shell is built now
  if not exist "%NMROOT%\nxm\sys\lib\Shell.class" goto DONE
  %nm_java_cmd% -DSHELL=nxm.sys.lib.Shell -DCONFIGURE=TRUE NmClassLoader
  goto DONE

rem ############################################################################
rem # Note: Unconfigured IDEs (Netbeans/Eclipse) may delete just NmClassLoader.class
rem #       and nothing else (as it is not in the "tools" package, so just compile
rem #       this one file and continue to run the shell
:BUILDNMCLASSLOADER
  echo Compiling NmClassLoader
  %nm_javac_cmd% "%NMROOT%\tools\NmClassLoader.java"
  goto SHELL

rem ############################################################################
:END
  call "%NM_OS%\nm%*"
  goto DONE

:EDIT
  call "%NM_OS%\nm%*"
  goto DONE

:NOBUILD
  echo NeXtMidas not built yet - see %NMROOT%\readme.txt  (or nm make all)
  goto DONE

rem ############################################################################
:SHELL
  if exist "%NM_SF%.bat" del "%NM_SF%.bat"

  %nm_java_cmd% -DSHELL=nxm.sys.lib.Shell NmClassLoader %args%

  if not exist "%NM_SF%.bat" goto DONE

  call "%NM_SF%.bat"
  rem if NM_SF var is no longer set, then goto DONE to prevent error messages
  if ["%NM_SF%"] == [""] goto DONE

  if exist "%NM_SF%.bat" del /Q "%NM_SF%.bat"

  @rem BUG 1956, only re-enter NeXtMidas shell when nm is run without any arguments
  if ["%args%"] == [""] goto SHELL

rem ############################################################################
:DONE
rem cleanup local variables
set args=
set jvm_target_ver=
set nm_boot_classpath=
set nm_java_cmd=
set nm_javac_cmd=
